// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       Error
//  TYPE:       c++ header
//  PROJECT:    VEGA Base Library
//  CONTENT:    simple error handler
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    hl      Helwig L"offelmann
//              rft     Robert F. Tobler
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 13:40:13  rft     last modification
//  08-mar-94 12:45:31  rft     changed naming conventions,
//                              introduced multiple handler facility
//  29-jan-94 16:50:00  hl      changed {...} to do {...} while (0)
//  15-jan-94 05:15:00  hl      no flag, ok_ind used any more
//                              Error(code, message) added
//  06-jan-94 16:35:00  hl      exitCode now -1 per default
//  04-jan-94 22:25:00  hl      created
// ===========================================================================
//  MANUAL:
//
//  DESCRIPTION
//      This module defines the ErrorHandler class, which can be used to
//      selectively modify the behaviour of the Library with respect to
//      certain errors that may occur.
//
//      For each module of the Library an ErrorHandler is defined in module
//      Library, that controls the behaviour of all Errors occuring in
//      methods and function in that module.
//
//      The default behaviour of an Errorhandler is to print its message and
//      terminate the program. Optionally ErrorNumbers can be used, so that
//      errors will not terminate the program but record the error. The
//      recorded error can be caught using macros defined in this module.
//
//      To catch an error the IF_EH_ERROR_IN macro can be used, which resets
//      a variable that contains the ErrorHandler with the last error, runs
//      the piece of code that is supplied and then tests if an error
//      occured:
//
//              IF_EH_ERROR_IN(<code>)
//              {
//                  // do something about the error
//              }
//
//      Alternatively the following sequence can be used:
//
//              EH_RESET();
//              <code>
//              IF_EH_ERROR()
//              {
//                  // do something about the error
//              }
//
//  EXAMPLE
//      Within the Crd3 module the 'inverseCopy()' function of Mat3 checks
//      for singularity and triggers an error using following piece of code:
//
//              if (det == 0.0)
//              {
//                  EH_ERROR(Crd3_errors, ERROR_CRD3_SINGULAR_MATRIX,
//                        "singular matrix");
//                  return mat;
//              }
//
//      To catch that error one could use following code:
//
//              Mat3 mat(1,2,3,4,5,6,7,8,9);
//              Mat3 inv;
//
//              Crd3_errors.setState(ErrorHanler::useErrorNumber);
//
//              IF_EH_ERROR_IN(inv = inverseCopy(mat))
//              {
//                  IF_EH_SWITCH(Crd3_errors)
//                  {
//                      case ERROR_CRD3_SINGULAR_MATRIX:
//                          // do something about the error
//                          break;
//                  }
//              }
//
//      Note, that it is easy to test for different errors handled by
//      different ErrorHandlers.
//
//  SEE ALSO
//      VEGA, BaseLibrary
// ===========================================================================
#ifndef _BASE_ERROR_H_
#define _BASE_ERROR_H_

extern "C"
{
#include <stdlib.h>                     //  _exit(), exit(), ...

extern void exit(int code);             // These declarations are here to
extern void _exit(int code);            // suppress a stupid gcc warning.
}

#include <iostream.h>

#include "VEGA/Base/defines.h"
#include "VEGA/Base/checks.h"
#include "VEGA/Base/system.h"

#ifdef __GNUC__
#pragma interface
#endif

extern "C++"
{

#define EH_NO_ERROR                0
        //
        //  If an ErrorHandler hasn't recorded any errors its error number
        //  has this value.
        //

// ======================================================== class ErrorHandler

class ErrorHandler
{
public:

    enum State
    {
        errorAndExit,                   // this is the default state
        useErrorNumber,                 
        errorAndContinue
    };

    enum Warnings
    {
	disabled,
	enabled
    };
    
    enum ExitFunction
    {
	exit,
	_exit
    };

    // ----------------------------------------------------------- constructor

    ErrorHandler(
        const char * newName
        );

    // ------------------------------------------------------ member functions

    //
    //  Change the behavior of an ErrorHandler. The following four functions
    //  return the old behaviour of the ErrorHandler.
    //
    ErrorHandler::State setState(
	State newState
	);
    ErrorHandler::Warnings setWarnings(
	Warnings newWarnings
	);
    ErrorHandler::ExitFunction setExitFunction(
	ExitFunction newExitFunction
	);
    int setExitCode(
	int newExitCode
	);

    //
    //  Normally one common stream is used for all errors, but these two
    //  methods make it possible to redirect all errors occuring in a
    //  specific ErrorHandler.
    //
    void enableLocalStream(
	ostream & s
	);
    void disableLocalStream(
	);

    static void setCommonStream(
	ostream & s
	);
        //
        //  The default stream for all error messages and warnings is cerr,
        //  but this function allows to change that.
        //

    // ----------------------------------------------------- special functions
    //
    // Allthough public, none of the following methods should be called
    // explicitly. They are implicitly called when the macros are used.
    //
    void doExit() const;
    void condExit() const;
    int printErrors() const;
    int printWarnings() const;
    int getErrorNumber() const;
    ostream & getStream() const;
    void setErrorNumber(int newErrorNumber);
    friend ostream & operator<<(ostream & s, const ErrorHandler & handler);
    static void setLastErrorHandlerPtr(ErrorHandler * errorHandlerPtr);
    static ErrorHandler * getLastErrorHandlerPtr();

private:
    State state;
    Warnings warnings;
    ExitFunction exitFunction;

    const char * name;
    int errorNumber;
    int exitCode;
    ostream * localStream;

    static ErrorHandler * lastErrorHandlerPtr;
    static ostream * commonStream;
};

//
//  Print warning using ErrorHandler:
//

#define EH_WARNING(handler,message)                                     \
        do                                                              \
        {                                                               \
            if (handler.printWarnings())                                \
                handler.getStream() << "W:" << handler << ":" << __FILE__ \
                    << ":" << __LINE__ << ":" << message << "\n";       \
        }                                                               \
        while (0)

//
//  An error can be generated using the following function. The behaviour of
//  the macro depends on the state of the ErrorHandler.
//

#define EH_ERROR(handler,number,message)                                \
        do                                                              \
        {                                                               \
            (handler).setErrorNumber(number);                           \
            if (handler.printErrors())                                  \
                handler.getStream() << "E:" << handler << ":" << __FILE__ \
                    << ":" << __LINE__ << ":" << message << "\n";       \
            handler.condExit();                                         \
        }                                                               \
        while (0)

//
//  Print a message and exit.
//

#define EH_FATAL(handler,message)                                       \
        do                                                              \
        {                                                               \
            (handler).getStream() << "F:" << handler << ":" << __FILE__ \
                << ":" << __LINE__ << ":" << message << "\n";           \
            (handler).doExit();                                         \
        }                                                               \
        while (0)

//
//  Test a condition and exit if the condition is true. Prints the condition
//  as error message.
//

#define EH_FATAL_CONDITION(handler,condition)                           \
        do                                                              \
        {                                                               \
            if (condition)                                              \
                EH_FATAL(handler,#condition);                           \
        }                                                               \
        while (0)


#define EH_RESET()	ErrorHandler::setLastErrorHandlerPtr(0)

#define IF_EH_ERROR()	if (ErrorHandler::getLastErrorHandlerPtr())

#define IF_EH_ERROR_IN(call)                                            \
        if (ErrorHandler::setLastErrorHandlerPtr(0), (call),            \
            ErrorHandler::getLastErrorHandlerPtr())

#define IF_EH_SWITCH(handler)                                           \
        if (ErrorHandler::getLastErrorHandlerPtr() == &(handler))       \
            switch ((handler).getErrorNumber())                         \

extern ErrorHandler mainErrors;

} // extern "C++"

#ifndef ERROR_NO_INLINE

#define INLINE_GLOBAL COMPILER_INLINE_GLOBAL
#define INLINE_MEMBER COMPILER_INLINE_MEMBER

#include "VEGA/Base/error.ipp"

#undef INLINE_GLOBAL
#undef INLINE_MEMBER

#endif

#endif // _BASE_ERROR_H_
// ===========================================================================
