// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       Option
//  TYPE:       c++ header
//  PROJECT:    VEGA Base Library
//  CONTENT:    commandline option parser
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rft     Robert F. Tobler
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 13:40:13  rft     last modification
//  23-mar-94 19:49:58  rft     created
// ===========================================================================
//  MANUAL:
//
//  DESCRIPTION
//      This is a command line parser that allows simple specification of all
//      options that may be put on the commandline.
//
//  EXAMPLE
//      Suppose we want to have a commandline utility that has the following
//      syntax:
//
//              tool [-h] [-n <int>] [-r <double>] [-s <string>] parameters
//
//      The options can also be used with their long names if they are
//      preceeded with '--' (e.g. --help, --number, --ratio, and --string),
//      the parameters can be intersperesed with the options, and the number
//      and ratio should get default values of 1 and 1.0.
//
//      To program that tool we would write the following code:
//
//              int main(int argc, char ** argv)
//              {
//
//                  OptionFlag          helpOpt  ("help",   "h");
//                  OptionLong          numberOpt("number", "n", 1);
//                  OptionDouble        ratioOpt ("ratio",  "r", 1.0);
//                  OptionString        stringOpt("string", "s");
//
//                  Option_errors.setState(ErrorHandler::useErrorNumber);
//                  IF_EH_ERROR_IN(Option::parse(argc,argv))
//                  {
//                      // print usage
//                  }
//
//                  if (helpOpt.isSpecified()) { ... }
//
//                  int number = numberOpt.getValue();
//                  double ratio = ratioOpt.getValue();
//
//                  if (stringOpt.isSpecified())
//                  {
//                      const char * string = stringOpt.getValue();
//                      ...
//                  }
//
//                  //
//                  // After the call to Option::parse all options and
//                  // parameters of options are stripped from argv.
//                  // argc is adjusted accordingly.
//                  //
//
//                  for (int i = 1; i < argc; i++)
//                  {
//                      // use argv[i]
//                      ...
//                  }
//              }
//
//      Note, that for the number and ratio options it is not necessary to
//      check if they were specified, since they both have default values.
//
//  BUGS
//      No type checking on the specified parameter is done. Incompletely
//      parsed numbers are not reported.
//
//      If multiple Options with the same name are initialized, only the last
//      one will be used. (No checks to guard against that are performed.)
//
//  SEE ALSO
//      VEGA, BaseLibrary
// ===========================================================================
#ifndef _BASE_OPTION_H_
#define _BASE_OPTION_H_

#include "VEGA/Base/globals.h"
#include "VEGA/Base/bool.h"
#include "VEGA/Base/system.h"

#ifdef __GNUC__
#pragma interface
#endif

extern "C++"
{

// =========================================== errors generated by this module

#define ERROR_OPTION_UNKNOWN            1
        //
        //  An unknown option was supplied on the commandline.
        //

#define ERROR_OPTION_SPECIFIED_TWICE    2
        //
        //  An option occured twice on the commandline.
        //

#define ERROR_OPTION_PARAMETER_MISSING  3
        //
        //  A parameter option was supplied without parameter.
        //

// ============================================================== class Option

class Option
{
public:

    int isSpecified(
	) const;

    static void parse(
	int & argc,
	char **  argv
	);
        //
        //  ERROR_OPTION_UNKNOWN
        //  ERROR_OPTION_SPECIFIED_TWICE
        //  ERROR_OPTION_PARAMETER_MISSING
        //

protected:
    Option(
        const char * newLongName,
        const char * newShortName,
        Bool newSpecified
        );

    virtual void setSpecified();
    virtual Bool missingParameter();
    virtual Bool takesParameter(const char * par);
    
    static Option * findByLongName(const char * name);
    static Option * findByShortName(const char * name);


    Bool specified;
    const char * longName;
    const char * shortName;
    Option * next;

    static Option * optionList;
};

// ========================================================== class OptionFlag

class OptionFlag :
    public Option
{
public:

    OptionFlag(
        const char * newLongName,
        const char * newShortName
        );

protected:
};

// ===================================================== class OptionParameter

class OptionParameter :
    public Option
{
protected:
    OptionParameter(
        const char * newLongName,
        const char * newShortName,
        Bool hasDefault
        );

    virtual void setSpecified();
    virtual Bool missingParameter();
    virtual void setParameter(const char * par);
    virtual Bool takesParameter(const char * par);

    Bool parameter;
};

// ======================================================== class OptionString

class OptionString :
    public OptionParameter
{
public:

    OptionString(
        const char * newLongName,
        const char * newShortName
        );
    OptionString(
        const char * newLongName,
        const char * newShortName,
        const char * defaultValue
        );

    const char * getValue() const;

protected:
    virtual void setParameter(const char * par);

    const char * value;
};

// ========================================================== class OptionLong

class OptionLong :
    public OptionParameter
{
public:

    OptionLong(
        const char * newLongName,
        const char * newShortName
        );
    OptionLong(
        const char * newLongName,
        const char * newShortName,
        int defaultValue
        );

    long getValue() const;

protected:
    virtual void setParameter(const char * par);

    long value;
};

// ======================================================== class OptionDouble

class OptionDouble :
    public OptionParameter
{
public:

    OptionDouble(
        const char * newLongName,
        const char * newShortName
        );
    OptionDouble(
        const char * newLongName,
        const char * newShortName,
        double defaultValue
        );

    double getValue() const;

protected:
    virtual void setParameter(const char * par);

    double value;
};

} // extern "C++"

#endif // _BASE_OPTION_H_
// ===========================================================================
