// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       Tracer
//  TYPE:       c++ header
//  PROJECT:    VEGA Base Library
//  CONTENT:    tracers which vanish if WITH_TRACERS isn't defined
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rft     Robert F. Tobler
//              hl      Helwig L"offelman
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 13:40:13  rft     last modification
//  15-Jul-93           rft     created
// ===========================================================================
//  MANUAL:
//
//  DESCRIPTION
//      Tracers are debugging aids that help managing debugging output. A
//      tracer allows to enable and disable all output routed through it.
//      Tracers can be defined for each group of trace messages to facilitate
//      detailed control of debugging output.
//
//      Tracers and trace message code is only compiled into the object if
//      the macro 'WITH_TRACERS' is defined. If it is not defined, NOTHING
//      remains in the resulting object file. This is made possible by
//      writing all trace code using the macros supplied in this module.
//      Therefore it is NOT necessary to remove the trace message statements
//      from the final program. If a number of tracers are defined to get
//      messages for all created and destoyed objects, to get messages for
//      all assignment operators etc. this can be a great debugging aid.
//
//  EXAMPLES
//	A tracer can be declared as follows:
//
//		#include <VEGA/Base/Tracer.h>
//
//		Tracer myTracer("myTracer",Tracer::on);
//
//      A simple trace message can then be produced as follows (here the
//	message is preeceded by the name of the tracer, and the source file,
//	and source line of the trace message):
//
//              TR_MESSAGE(myTracer,"entered section " << section);
//
//	To output additional information that is not prefixed by the tracer
//	name and outher information, the following statement can be used:
//
//		TR_OUT(myTracer,"additional information " << information);
//
//      An arbitrary piece of code can be made to compile only when macro
//      WITH_TRACERS is defined, if it is put into the TRACE macro like this:
//
//              TR_CODE(
//                  if (isTiny(squaredLength(vec)))
//                  {
//                      TR_MESSAGE(myTracer,"the length of vec is zero");
//                  }
//              )
//
//  SEE ALSO
//      VEGA, BaseLibrary
// ===========================================================================
#ifndef _BASE_TRACER_H_
#define _BASE_TRACER_H_

extern "C"
{
#include <stdlib.h>
}

#include <iostream.h>

#include "VEGA/Base/defines.h"
#include "VEGA/Base/system.h"

#ifdef __GNUC__
#pragma interface
#endif

extern "C++"
{

// ============================================================== class Tracer

class Tracer
{
public:

    enum State                          // per Tracer state
    {
        off = 0,                        // suppress all messages
        on = 1,                         // print all messages
    };

    // ----------------------------------------------------------- constructor

    Tracer(                             // create a Tracer
        const char * newName,
        State newState                  // initial state of the tracer
        );

    Tracer(                             // create a Tracer
        const char * newName,
        State newState,                 // initial state of the tracer
	Tracer & newParentTracer
        );

    // ------------------------------------------------------ member functions

    void turnOn();
    void turnOff();
    
    int isOn();
    int isOff();

    ostream & getStream();

    void enableLocalStream(ostream & s);
    void disableLocalStream();
    
    long getAndIncCounter();

    // ------------------------------------------------------ friend operators

    friend ostream & operator<<(ostream & s, const Tracer & tracer);

    // ------------------------------------------------------ static functions

    static void setCommonStream(ostream & s);    

private:
    static ostream * commonStream;      // the destination of all trace
                                        // messages

    Tracer *		parentTracer;
    ostream *		localStream;
    State		state;		// off | on
    const char *	name;		// the name of the tracer
    long		counter;
};

#ifdef WITH_TRACERS

#define TR_CODE(code)             code

#define TR_ON(tracer)                                   \
        do                                              \
        {                                               \
            if (tracer.isOff())                         \
            {                                           \
                tracer.turnOn();                        \
                TR_MESSAGE(tracer,"TURNED ON.");	\
            }                                           \
        }                                               \
        while(0)

#define TR_OFF(tracer)                                  \
        do                                              \
        {                                               \
            TR_MESSAGE(tracer,"TURNED OFF.");		\
            tracer.turnOff();                           \
        }                                               \
        while(0)

#define TR_ENABLE_LOCAL_STREAM(tracer,stream) \
        ((tracer).enableLocalStream(stream))
#define TR_DISABLE_LOCAL_STREAM(tracer) \
        ((tracer).disableLocalStream())

#define TR_SET_COMMON_STREAM(stream) \
        (Tracer::setCommonStream(stream))
        //
        //  The default stream for all trace messages is cerr, but this macro
        //  allows to change that.
        //

#define TR_MESSAGE(tracer,message)                                      \
        do                                                              \
        {                                                               \
            if (tracer.isOn())                                          \
            {                                                           \
                (tracer).getStream() << "T:" << tracer << ':'		\
		    << __FILE__ << ':' << __LINE__ << ":[#"		\
		    << (tracer).getAndIncCounter() << "]:"		\
		    << message << "\n";					\
            }                                                           \
        }                                                               \
        while (0)

#define TR_OUT(tracer,message)                                          \
        do                                                              \
        {                                                               \
            if (tracer.isOn())                                          \
            {                                                           \
                (tracer).getStream() << message;			\
            }                                                           \
        }                                                               \
        while (0)

#define TR_CONDITION(tracer,condition)                  \
        do                                              \
        {                                               \
            if (condition)                              \
                TR_MESSAGE(tracer,#condition);       \
        }                                               \
        while(0)
	//
	//  This macro will write a trace message if the condition is true.
	//

//
//  Following macros make it possible to easily output a function header to
//  a tracer by specifying the tracer, the function and all parameters in
//  succession.
//

#define TR_FUN0(tracer,fun) \
        TR_MESSAGE(tracer,#fun "()") 
#define TR_FUN1(tracer,fun,p1) \
        TR_MESSAGE(tracer,#fun "("<<p1<<")")
#define TR_FUN2(tracer,fun,p1,p2) \
        TR_MESSAGE(tracer,#fun "("<<p1<<","<<p2<<")")
#define TR_FUN3(tracer,fun,p1,p2,p3) \
        TR_MESSAGE(tracer,#fun "("<<p1<<","<<p2<<","<<p3<<")")
#define TR_FUN4(tracer,fun,p1,p2,p3,p4) \
        TR_MESSAGE(tracer,#fun "("<<p1<<","<<p2<<","<<p3<<","<<p4<<")")
#define TR_FUN5(tracer,fun,p1,p2,p3,p4,p5) \
        TR_MESSAGE(tracer,#fun "("<<p1<<","<<p2<<","<<p3<<","<<p4<<"," \
                <<p5<<")")
#define TR_FUN6(tracer,fun,p1,p2,p3,p4,p5,p6) \
        TR_MESSAGE(tracer,#fun "("<<p1<<","<<p2<<","<<p3<<","<<p4<<"," \
                <<p5<<","<<p6<<")")
#define TR_FUN7(tracer,fun,p1,p2,p3,p4,p5,p6,p7) \
        TR_MESSAGE(tracer,#fun "("<<p1<<","<<p2<<","<<p3<<","<<p4<<","\
                <<p5<<","<<p6<<","<<p7<<")")
#define TR_FUN8(tracer,fun,p1,p2,p3,p4,p5,p6,p7,p8) \
        TR_MESSAGE(tracer,#fun "("<<p1<<","<<p2<<","<<p3<<","<<p4<<","\
                <<p5<<","<<p6<<","<<p7<<","<<p8<<")")
#define TR_FUN9(tracer,fun,p1,p2,p3,p4,p5,p6,p7,p8,p9) \
        TR_MESSAGE(tracer,#fun "("<<p1<<","<<p2<<","<<p3<<","<<p4<<","\
                <<p5<<","<<p6<<","<<p7<<","<<p8<<","<<p9<<")")


#else // ! WITH_TRACERS

#define TR_CODE(code)
#define TR_ON(tracer)
#define TR_OFF(tracer)
#define TR_ENABLE_LOCAL_STREAM(tracer,stream)
#define TR_DISABLE_LOCAL_STREAM(tracer)
#define TR_SET_COMMON_STREAM(stream)
#define TR_MESSAGE(tracer,message)
#define TR_CONDITION(tracer,condition)

#define TR_FUN0(tracer,fun)
#define TR_FUN1(tracer,fun,p1)
#define TR_FUN2(tracer,fun,p1,p2)
#define TR_FUN3(tracer,fun,p1,p2,p3)
#define TR_FUN4(tracer,fun,p1,p2,p3,p4)
#define TR_FUN5(tracer,fun,p1,p2,p3,p4,p5)
#define TR_FUN6(tracer,fun,p1,p2,p3,p4,p5,p6)
#define TR_FUN7(tracer,fun,p1,p2,p3,p4,p5,p6,p7)
#define TR_FUN8(tracer,fun,p1,p2,p3,p4,p5,p6,p7,p8)
#define TR_FUN9(tracer,fun,p1,p2,p3,p4,p5,p6,p7,p8,p9)

#endif // ! WITH_TRACERS

extern Tracer mainTrace;

} // extern "C++"

#ifndef TRACER_NO_INLINE

#define INLINE_GLOBAL COMPILER_INLINE_GLOBAL
#define INLINE_MEMBER COMPILER_INLINE_MEMBER

#include "VEGA/Base/tracer.ipp"

#undef INLINE_GLOBAL
#undef INLINE_MEMBER

#endif

#endif // _BASE_TRACER_H_
// ===========================================================================
