// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       BiTrafo3D
//  TYPE:       c++ header
//  PROJECT:    VEGA Graphics Library
//  CONTENT:    bidirectional 3D transformation class
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rft     Robert F. Tobler
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 14:08:16  rft     last modification
//  30-mar-94 11:17:03  rft     created
// ===========================================================================
//  MANUAL:
//
//  DESCRIPTION
//      This module implements the class BiTrafo3D which allows bidirectional
//      transformation of Vectors between coordinate systems.
//
//      Whenever a BiTrafo3D is constructed from a Trafo3D or Lin3D either
//      by using the direct constructor or by using one of the mutliplication
//      operators, the inverse of the Trafo3D or Lin3D has to be calculated.
//      Since this is done by calculating determinants and sub-determinants
//      there is a possible danger of numerical instability. It is therefore
//      recommended to construct a BiTrafo3D from the simple transformations
//      (Scale3D, RotX3D, RotY3D, RotZ3D, Rot3D, ShearXY3D, ShearZX3D,
//      ShearYZ3D, Lin3D, and Trans3D) all of which use a simple and stable
//      way of calculating the inverse.
//
//  SEE ALSO
//      VEGA, GraphicsLibrary
// ===========================================================================
#ifndef _GRAPHICS_BITRAFO3D_H_
#define _GRAPHICS_BITRAFO3D_H_

#include "VEGA/Graphics/library.h"
#include "VEGA/Graphics/vec3d.h"

#ifdef __GNUC__
#pragma interface
#endif

extern "C++"
{

// =========================================================== class BiTrafo3D

class BiTrafo3D
{
public:

    // --------------------------------------------- constructors / destructor

    BiTrafo3D(				// Does NOT initialize the trafos.
	);
    BiTrafo3D(
	const BiTrafo3D & biTra
	);
    ~BiTrafo3D(
	);

    BiTrafo3D(
        const Trafo3D & newForward,
        const Trafo3D & newBackward
        );
    BiTrafo3D(
	const Scale3D & scale
	);
    BiTrafo3D(
	const RotX3D & rot
	);
    BiTrafo3D(
	const RotY3D & rot
	);
    BiTrafo3D(
	const RotZ3D & rot
	);
    BiTrafo3D(
	const Rot3D & rot
	);
    BiTrafo3D(
	const ShearXY3D & shear
	);
    BiTrafo3D(
	const ShearZX3D & shear
	);
    BiTrafo3D(
	const ShearYZ3D & shear
	);
    BiTrafo3D(
	const Lin3D & lin
	);
    BiTrafo3D(
	const Trans3D & trans
	);
    BiTrafo3D(
	const Trafo3D & trafo
	);
    // ------------------------------------------------------ member operators

    BiTrafo3D & operator*=(
	const BiTrafo3D & biTra
	);
    BiTrafo3D & operator*=(
	const Scale3D & scale
	);
    BiTrafo3D & operator*=(
	const RotX3D & rot
	);
    BiTrafo3D & operator*=(
	const RotY3D & rot
	);
    BiTrafo3D & operator*=(
	const RotZ3D & rot
	);
    BiTrafo3D & operator*=(
	const Rot3D & rot
	);
    BiTrafo3D & operator*=(
	const ShearXY3D & shear
	);
    BiTrafo3D & operator*=(
	const ShearZX3D & shear
	);
    BiTrafo3D & operator*=(
	const ShearYZ3D & shear
	);
    BiTrafo3D & operator*=(
	const Lin3D & lin
	);
    BiTrafo3D & operator*=(
	const Trans3D & trans
	);
    BiTrafo3D & operator*=(
	const Trafo3D & trafo
	);

    // ------------------------------------------------------ member functions

    Trafo3D getForward(
	) const;
    Trafo3D getBackward(
	) const;

    void invert(
	);

    // ------------------------------------------------------ friend operators

    friend BiTrafo3D operator*(
        const BiTrafo3D & biTrafo1,
        const BiTrafo3D & biTrafo2
        );

    friend BiTrafo3D operator*(
	const Scale3D & scale,
	const BiTrafo3D & biTra
	);
    friend BiTrafo3D operator*(
	const BiTrafo3D & biTra,
	const Scale3D & scale
	);

    friend BiTrafo3D operator*(
	const RotX3D & rot,
	const BiTrafo3D & biTra
	);
    friend BiTrafo3D operator*(
	const BiTrafo3D & biTra,
	const RotX3D & rot
	);

    friend BiTrafo3D operator*(
	const RotY3D & rot,
	const BiTrafo3D & biTra
	);
    friend BiTrafo3D operator*(
	const BiTrafo3D & biTra,
	const RotY3D & rot
	);

    friend BiTrafo3D operator*(
	const RotZ3D & rot,
	const BiTrafo3D & biTra
	);
    friend BiTrafo3D operator*(
	const BiTrafo3D & biTra,
	const RotZ3D & rot
	);

    friend BiTrafo3D operator*(
	const Rot3D & rot,
	const BiTrafo3D & biTra
	);
    friend BiTrafo3D operator*(
	const BiTrafo3D & biTra,
	const Rot3D & rot
	);

    friend BiTrafo3D operator*(
	const ShearXY3D & sh,
	const BiTrafo3D & biTra
	);
    friend BiTrafo3D operator*(
	const BiTrafo3D & biTra,
	const ShearXY3D & sh
	);

    friend BiTrafo3D operator*(
	const ShearZX3D & sh,
	const BiTrafo3D & biTra
	);
    friend BiTrafo3D operator*(
	const BiTrafo3D & biTra,
	const ShearZX3D & sh
	);

    friend BiTrafo3D operator*(
	const ShearYZ3D & sh,
	const BiTrafo3D & biTra
	);
    friend BiTrafo3D operator*(
	const BiTrafo3D & biTra,
	const ShearYZ3D & sh
	);

    friend BiTrafo3D operator*(
	const Trans3D & trans,
	const BiTrafo3D & biTra
	);
    friend BiTrafo3D operator*(
	const BiTrafo3D & biTra,
	const Trans3D & trans
	);

    friend BiTrafo3D operator*(
	const Lin3D & lin,
	const BiTrafo3D & biTra
	);
    friend BiTrafo3D operator*(
	const BiTrafo3D & biTra,
	const Lin3D & lin
	);

    friend BiTrafo3D operator*(
	const Trafo3D & trafo,
	const BiTrafo3D & biTra
	);
    friend BiTrafo3D operator*(
	const BiTrafo3D & biTra,
	const Trafo3D & trafo
	);

    // ------------------------------------------------------ friend functions

    friend Trafo3D getForward(
	const BiTrafo3D & biTrafo
	);
    friend Trafo3D getBackward(
	const BiTrafo3D & biTrafo
	);
    
    friend Vec3D forwardProduct(
	const Vec3D & vec,
	const BiTrafo3D & biTra
	);
    friend Pnt3D forwardProduct(
	const Pnt3D & pnt,
	const BiTrafo3D & biTra
	);

    friend Vec3D backwardProduct(
	const Vec3D & vec,
	const BiTrafo3D & biTra
	);
    friend Pnt3D backwardProduct(
	const Pnt3D & pnt,
	const BiTrafo3D & biTra
	);

    friend BiTrafo3D inverseCopy(
	const BiTrafo3D & biTra
	);

    // ----------------------------------------------- static member variables

    static const BiTrafo3D identity;

private:
    Trafo3D     forward;
    Trafo3D     backward;
};

} // extern "C++"

#ifndef BITRAFO3D_NO_INLINE

#define INLINE_GLOBAL COMPILER_INLINE_GLOBAL
#define INLINE_MEMBER COMPILER_INLINE_MEMBER

#include "VEGA/Graphics/bitra3d.ipp"

#undef INLINE_GLOBAL
#undef INLINE_MEMBER

#endif

#endif // _GRAPHICS_BITRAFO3D_H_
// ===========================================================================
