// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       BiTrafo3D
//  TYPE:       c++ inline
//  PROJECT:    VEGA Graphics Library
//  CONTENT:    bidirectional 3D transformation class
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rft     Robert F. Tobler
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 14:08:16  rft     last modification
//  30-mar-94 11:17:03  rft     created
// ===========================================================================

extern "C++"
{

INLINE_MEMBER BiTrafo3D::BiTrafo3D()
{
}

INLINE_MEMBER BiTrafo3D::BiTrafo3D(const BiTrafo3D& biTrafo)
:
    forward(biTrafo.forward),
    backward(biTrafo.backward)
{}

INLINE_MEMBER BiTrafo3D::~BiTrafo3D()
{
}

INLINE_MEMBER BiTrafo3D::BiTrafo3D(
	const Trafo3D& newForward,
	const Trafo3D& newBackward
	)
:
    forward(newForward),
    backward(newBackward)
{}

INLINE_GLOBAL BiTrafo3D inverseCopy(const BiTrafo3D& biTrafo)
{
    return BiTrafo3D(biTrafo.backward,biTrafo.forward);
}

INLINE_MEMBER BiTrafo3D::BiTrafo3D(
	const Scale3D & scale
	)
:
    forward(scale),
    backward(inverseCopy(scale))
{}

INLINE_MEMBER BiTrafo3D::BiTrafo3D(
	const RotX3D & rot
	)
:
    forward(rot),
    backward(inverseCopy(rot))
{}

INLINE_MEMBER BiTrafo3D::BiTrafo3D(
	const RotY3D & rot
	)
:
    forward(rot),
    backward(inverseCopy(rot))
{}

INLINE_MEMBER BiTrafo3D::BiTrafo3D(
	const RotZ3D & rot
	)
:
    forward(rot),
    backward(inverseCopy(rot))
{}

INLINE_MEMBER BiTrafo3D::BiTrafo3D(
	const Rot3D & rot
	)
:
    forward(rot),
    backward(inverseCopy(rot))
{}

INLINE_MEMBER BiTrafo3D::BiTrafo3D(
	const ShearXY3D & shear
	)
:
    forward(shear),
    backward(inverseCopy(shear))
{}

INLINE_MEMBER BiTrafo3D::BiTrafo3D(
	const ShearZX3D & shear
	)
:
    forward(shear),
    backward(inverseCopy(shear))
{}

INLINE_MEMBER BiTrafo3D::BiTrafo3D(
	const ShearYZ3D & shear
	)
:
    forward(shear),
    backward(inverseCopy(shear))
{}

INLINE_MEMBER BiTrafo3D::BiTrafo3D(
	const Lin3D & lin
	)
:
    forward(lin),
    backward(inverseCopy(lin))
{}

INLINE_MEMBER BiTrafo3D::BiTrafo3D(
	const Trans3D & trans
	)
:
    forward(trans),
    backward(inverseCopy(trans))
{}

INLINE_MEMBER BiTrafo3D::BiTrafo3D(
	const Trafo3D & trafo
	)
:
    forward(trafo),
    backward(inverseCopy(trafo))
{}

INLINE_MEMBER BiTrafo3D& BiTrafo3D::operator*=(const BiTrafo3D& biTrafo)
{
    forward *= biTrafo.forward;
    backward = biTrafo.backward * backward;
    return *this;
}

INLINE_MEMBER BiTrafo3D& BiTrafo3D::operator*=(const Scale3D& scale)
{
    forward *= scale;
    backward = inverseCopy(scale) * backward;
    return *this;
}

INLINE_MEMBER BiTrafo3D& BiTrafo3D::operator*=(const RotX3D& rot)
{
    forward *= rot;
    backward = inverseCopy(rot) * backward;
    return *this;
}

INLINE_MEMBER BiTrafo3D& BiTrafo3D::operator*=(const RotY3D& rot)
{
    forward *= rot;
    backward = inverseCopy(rot) * backward;
    return *this;
}

INLINE_MEMBER BiTrafo3D& BiTrafo3D::operator*=(const RotZ3D& rot)
{
    forward *= rot;
    backward = inverseCopy(rot) * backward;
    return *this;
}

INLINE_MEMBER BiTrafo3D& BiTrafo3D::operator*=(const Rot3D& rot)
{
    forward *= Lin3D(rot);
    backward = Lin3D(inverseCopy(rot)) * backward;
    return *this;
}

INLINE_MEMBER BiTrafo3D& BiTrafo3D::operator*=(const ShearXY3D & shear)
{
    forward *= shear;
    backward = inverseCopy(shear) * backward;
    return *this;
}

INLINE_MEMBER BiTrafo3D& BiTrafo3D::operator*=(const ShearZX3D & shear)
{
    forward *= shear;
    backward = inverseCopy(shear) * backward;
    return *this;
}

INLINE_MEMBER BiTrafo3D& BiTrafo3D::operator*=(const ShearYZ3D & shear)
{
    forward *= shear;
    backward = inverseCopy(shear) * backward;
    return *this;
}

INLINE_MEMBER BiTrafo3D& BiTrafo3D::operator*=(const Lin3D & lin)
{
    forward *= lin;
    backward = inverseCopy(lin) * backward;
    return *this;
}

INLINE_MEMBER BiTrafo3D& BiTrafo3D::operator*=(const Trans3D& trans)
{
    forward *= trans;
    backward = inverseCopy(trans) * backward;
    return *this;
}

INLINE_MEMBER BiTrafo3D& BiTrafo3D::operator*=(const Trafo3D& trafo)
{
    forward *= trafo;
    backward = inverseCopy(trafo) * backward;
    return *this;
}

INLINE_MEMBER Trafo3D BiTrafo3D::getForward() const
{
    return forward;
}

INLINE_MEMBER Trafo3D BiTrafo3D::getBackward() const
{
    return backward;
}

INLINE_MEMBER void BiTrafo3D::invert()
{
    *this = inverseCopy(*this);
}

INLINE_GLOBAL BiTrafo3D operator*(
	const Scale3D& scale, const BiTrafo3D& biTrafo
	)
{
    return BiTrafo3D(scale * biTrafo.forward,
                   biTrafo.backward * inverseCopy(scale));
}

INLINE_GLOBAL BiTrafo3D operator*(
	const BiTrafo3D& biTrafo, const Scale3D& scale
	)
{
    return BiTrafo3D(biTrafo.forward * scale,
                   inverseCopy(scale) * biTrafo.backward);
}

INLINE_GLOBAL BiTrafo3D operator*(
	const RotX3D& rot, const BiTrafo3D& biTrafo
	)
{
    return BiTrafo3D(rot * biTrafo.forward,
                   biTrafo.backward * inverseCopy(rot));
}

INLINE_GLOBAL BiTrafo3D operator*(
	const BiTrafo3D& biTrafo, const RotX3D& rot
	)
{
    return BiTrafo3D(biTrafo.forward * rot,
                   inverseCopy(rot) * biTrafo.backward);
}

INLINE_GLOBAL BiTrafo3D operator*(
	const RotY3D& rot, const BiTrafo3D& biTrafo
	)
{
    return BiTrafo3D(rot * biTrafo.forward,
                   biTrafo.backward * inverseCopy(rot));
}

INLINE_GLOBAL BiTrafo3D operator*(
	const BiTrafo3D& biTrafo, const RotY3D& rot
	)
{
    return BiTrafo3D(biTrafo.forward * rot,
                   inverseCopy(rot) * biTrafo.backward);
}

INLINE_GLOBAL BiTrafo3D operator*(
	const RotZ3D& rot, const BiTrafo3D& biTrafo
	)
{
    return BiTrafo3D(rot * biTrafo.forward,
                   biTrafo.backward * inverseCopy(rot));
}

INLINE_GLOBAL BiTrafo3D operator*(
	const BiTrafo3D& biTrafo, const RotZ3D& rot
	)
{
    return BiTrafo3D(biTrafo.forward * rot,
                   inverseCopy(rot) * biTrafo.backward);
}

INLINE_GLOBAL BiTrafo3D operator*(
	const Rot3D& rot, const BiTrafo3D& biTrafo
	)
{
    return BiTrafo3D(Lin3D(rot) * biTrafo.forward,
                   biTrafo.backward * Lin3D(inverseCopy(rot)));
}

INLINE_GLOBAL BiTrafo3D operator*(
	const BiTrafo3D& biTrafo, const Rot3D& rot
	)
{
    return BiTrafo3D(biTrafo.forward * Lin3D(rot),
                   Lin3D(inverseCopy(rot)) * biTrafo.backward);
}

INLINE_GLOBAL BiTrafo3D operator*(
	const ShearXY3D& sh, const BiTrafo3D& biTrafo
	)
{
    return BiTrafo3D(sh * biTrafo.forward,
                   biTrafo.backward * inverseCopy(sh));
}

INLINE_GLOBAL BiTrafo3D operator*(
	const BiTrafo3D& biTrafo, const ShearXY3D& sh
	)
{
    return BiTrafo3D(biTrafo.forward * sh,
                   inverseCopy(sh) * biTrafo.backward);
}

INLINE_GLOBAL BiTrafo3D operator*(
	const ShearZX3D& sh, const BiTrafo3D& biTrafo
	)
{
    return BiTrafo3D(sh * biTrafo.forward,
                   biTrafo.backward * inverseCopy(sh));
}

INLINE_GLOBAL BiTrafo3D operator*(
	const BiTrafo3D& biTrafo, const ShearZX3D& sh
	)
{
    return BiTrafo3D(biTrafo.forward * sh,
                   inverseCopy(sh) * biTrafo.backward);
}

INLINE_GLOBAL BiTrafo3D operator*(
	const ShearYZ3D& sh, const BiTrafo3D& biTrafo
	)
{
    return BiTrafo3D(sh * biTrafo.forward,
                   biTrafo.backward * inverseCopy(sh));
}

INLINE_GLOBAL BiTrafo3D operator*(
	const BiTrafo3D& biTrafo, const ShearYZ3D& sh
	)
{
    return BiTrafo3D(biTrafo.forward * sh,
                   inverseCopy(sh) * biTrafo.backward);
}

INLINE_GLOBAL BiTrafo3D operator*(
	const Trans3D& trans, const BiTrafo3D& biTrafo
	)
{
    return BiTrafo3D(trans * biTrafo.forward,
                   biTrafo.backward * inverseCopy(trans));
}

INLINE_GLOBAL BiTrafo3D operator*(
	const BiTrafo3D& biTrafo, const Trans3D& trans
	)
{
    return BiTrafo3D(biTrafo.forward * trans,
                   inverseCopy(trans) * biTrafo.backward);
}

INLINE_GLOBAL BiTrafo3D operator*(
	const Lin3D& lin, const BiTrafo3D& biTrafo
	)
{
    return BiTrafo3D(lin * biTrafo.forward,
                   biTrafo.backward * inverseCopy(lin));
}

INLINE_GLOBAL BiTrafo3D operator*(
	const BiTrafo3D& biTrafo, const Lin3D& lin
	)
{
    return BiTrafo3D(biTrafo.forward * lin,
                   inverseCopy(lin) * biTrafo.backward);
}

INLINE_GLOBAL BiTrafo3D operator*(
	const Trafo3D& trafo, const BiTrafo3D& biTrafo
	)
{
    return BiTrafo3D(trafo * biTrafo.forward,
                   biTrafo.backward * inverseCopy(trafo));
}

INLINE_GLOBAL BiTrafo3D operator*(
	const BiTrafo3D& biTrafo, const Trafo3D& trafo
	)
{
    return BiTrafo3D(biTrafo.forward * trafo,
                   inverseCopy(trafo) * biTrafo.backward);
}

INLINE_GLOBAL BiTrafo3D operator*(
	const BiTrafo3D& biTrafo1,
	const BiTrafo3D& biTrafo2
	)
{
    return BiTrafo3D(biTrafo1.forward * biTrafo2.forward,
		     biTrafo2.backward * biTrafo1.backward);
}

INLINE_GLOBAL Trafo3D getForward(const BiTrafo3D & biTrafo)
{
    return biTrafo.forward;
}

INLINE_GLOBAL Trafo3D getBackward(const BiTrafo3D & biTrafo)
{
    return biTrafo.backward;
}

INLINE_GLOBAL Vec3D forwardProduct(
	const Vec3D & vec, const BiTrafo3D & biTrafo
	)
{
    return (vec * biTrafo.forward);
}

INLINE_GLOBAL Pnt3D forwardProduct(
	const Pnt3D & pnt, const BiTrafo3D & biTrafo
	)
{
    return (pnt * biTrafo.forward);
}

INLINE_GLOBAL Vec3D backwardProduct(
	const Vec3D & vec, const BiTrafo3D & biTrafo
	)
{
    return (vec * biTrafo.backward);
}

INLINE_GLOBAL Pnt3D backwardProduct(
	const Pnt3D & pnt, const BiTrafo3D & biTrafo
	)
{
    return (pnt * biTrafo.backward);
}

} // extern "C++"

// ===========================================================================
