// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       FVec2D
//  TYPE:       c++ header
//  PROJECT:    VEGA Graphics Library
//  CONTENT:    2D vectors, points, and transformations using floats
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rft     Robert F. Tobler
//              hl      Helwig L"offelman
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 14:08:16  rft     last modification
//  08-Aug-93           rft     created
// ===========================================================================
//  MANUAL:
//
//  DESCRIPTION
//      The classes FVec2D and FPnt2D are the basic types for 2D geometry.
//      They can be transformed with the corresponding transformation types
//      that are called FTrans2D, FScale2D, FRot2D, FShearX2D, FShearY2D,
//      FLin2D, and FTrafo2D.
//
//      The classes FVec2D and FPnt2D have been designed in a typesafe way,
//      so that all operations among them are geometrically meaningful.
//      Thus there are some combinations of operations that are not allowed
//      since there is no meaningful geometric result. An example for such
//      an operation is the addition of two points.
//
//      In terms of matrix theory, FVec2D and FPnt2D can be viewed as
//      homogenous 2D vectors and points (the trailing zeros and ones are
//      not stored):
//
//              FVec2D vec;
//              FPnt2D pnt;
//
//              (vec.x(), vec.y(), 0)
//              (pnt.x(), pnt.y(), 1)
//
//      The transformation type FTrafo2D can be viewed as a 2D homogeneous
//      matrix. The multiplication of FVec2D and FPnt2D with FTrafo2D can
//      then be represented as follows (the elements of the transformation
//      cannot be accessed in the way shown here):
//
//              FTrafo2D t;
//              FVec2D v;
//              FPnt2D m;
//
//              FVec2D vec = v * t;
//              FVec2D pnt = p * t;
//
//                                (t.m[0,0], t.m[0,1], 0)
//                                (t.m[1,0], t.m[1,1], 0)
//                                (  t.c[0],   t.c[1], 1)
//
//              (v.x(),v.y(),0)   ( vec.x(),  vec.y(), 0)
//
//              (p.x(),p.y(),1)   ( pnt.x(),  pnt.y(), 1)
//
//      The transformation types FTrans2D, FScale2D, FShearX2D, FShearY2D,
//      FLin2D, and FRot2D represent subsets of the transformations
//      represented by the FTrafo2D type.
//
//      Rotations are defined to be counter-clockwise in the x-y-plane.
//      Since a right-handed coordinate system is assumed, this is consistent
//      with 3D-rotations that are defined to be clockwise if viewed in
//      direction of the axis.
//
//      Other ways to think about a FTrafo2D:
//
//              FTrafo2D <=> (FVec2D,     // vecX
//                           FVec2D,     // vecY
//                           FPnt2D)     // pntO (O for Origin)
//
//              (This can be used to directly specify a transformation by
//               just specifying the origin and the base vectors of the
//               coordinate system !)
//
//              FTrafo2D <=> (FLin2D,     // lin
//                           FTrans2D)   // trans
//
//      If you are using these models, there are constructors that take these
//      elements as parameters, and methods and functions to get these
//      elements (getVecX, getVecY, getPntO, getLin, and getTrans).
//
//  CHECKS
//      Range checking and attempts to normalize null-vectors will only be
//      caught if WITH_FCRD2_CHECKS is defined.
//
//  ERRORHANDLERS
//      FVec2D_errors
//
//  TRACERS
//      FVec2D_constructors
//
//  SEE ALSO
//      VEGA, GraphicsLibrary
// ===========================================================================
#ifndef _GRAPHICS_FVEC2D_H_
#define _GRAPHICS_FVEC2D_H_

#include <iostream.h>

#include <VEGA/Math/math.h>
#include <VEGA/Math/fcrd2.h>
#include <VEGA/Math/fcrd3.h>

#include "VEGA/Graphics/globals.h"

#ifdef __GNUC__
#pragma interface
#endif

extern "C++"
{

// ============================================ classes defined in this module

class FVec2D;			// vector
class FPnt2D;			// point
class FScale2D;			// scaling by (sx,sy)
class FRot2D;			// rotation around origin
class FShearX2D;			// shear along x-Axis
class FShearY2D;			// shear along y-Axis
class FLin2D;			// linear transformation
class FTrans2D;			// translation by vector
class FTrafo2D;			// affine transformation

// =========================================== errors generated by this module

#define ERROR_FVEC2D_SINGULAR_SCALE      1
        //
        //  An attempt was made to invert a singular scale.
        //


// =============================================================== class FVec2D

class FVec2D
{
    friend class FPnt2D;
    friend class FLin2D;
    friend class FTrans2D;
    friend class FTrafo2D;

    friend class Viewer3D;		//  needs FAST access to members

public:

    // --------------------------------------------- constructors / destructor

    FVec2D(
        );
    FVec2D(
        float xNew,
        float yNew
        );
    FVec2D(
        const FCrd2 & crd
        );
    FVec2D(
        const FVec2D & vec
        );
    ~FVec2D(
        );

    // ------------------------------------------------------ member operators

    float operator[](
        int index
        ) const;
    float & operator[](
        int index
        );

    FVec2D & operator+=(
        const FVec2D & vec
        );
    FVec2D & operator-=(
        const FVec2D & vec
        );

    FVec2D & operator*=(
        float scalar
        );

    // ------------------------------------------------------ member functions

    float x(
        ) const;
    float y(
        ) const;

    void x(
        float xNew
        );
    void y(
        float yNew
        );

    void normalize(
        );

    FCrd2 getFCrd2(
        ) const;
    FCrd3 getFCrd3(			// Get homogeneous representation.
        ) const;
    FPnt2D getFPnt2D(			// Explicit conversion to FPnt2D.
        ) const;

    // ------------------------------------------------------ friend operators

    friend ostream & operator<<(
        ostream & s,
	const FVec2D & vec
        );
    friend FVec2D operator-(
        const FVec2D & vec
        );
    friend FVec2D operator+(
        const FVec2D & vec1,
	const FVec2D & vec2
        );
    friend FPnt2D operator+(
        const FPnt2D & pnt,
	const FVec2D & vec
        );
    friend FVec2D operator-(
        const FVec2D & vec1,
	const FVec2D & vec2
        );
    friend FPnt2D operator-(
        const FPnt2D & pnt,
	const FVec2D & vec
        );
    friend FVec2D operator*(
        const FVec2D & vec,
	float scalar
        );
    friend FVec2D operator*(
        float scalar,
	const FVec2D & vec
        );

    // ------------------------------------------------------ friend functions

    friend FCrd2 getFCrd2(
        const FVec2D & vec
        );
    friend FCrd3 getFCrd3(		// Get homogeneous representation.
        const FVec2D & vec
        );
    friend FPnt2D getFPnt2D(		// Explicit conversion.
        const FVec2D & vec
        );
    friend float squaredLength(
        const FVec2D & vec
        );
    friend float length(
        const FVec2D & vec
        );
    friend float dotProduct(
        const FVec2D & vec1,
	const FVec2D & vec2
        );
    friend FVec2D normalizedCopy(
        const FVec2D & vec
        );

    // -------------------------------------- special purpose friend functions

    friend FVec2D cwReciprocalCopy(      // Componentwise multiplicative
        const FVec2D & vec		//  inverse. Zero components are not
        );				//  inverted.

    friend FVec2D cwMin(                 // Componentwise minimum.
        const FVec2D & vec1,
        const FVec2D & vec2
        );

    friend FVec2D cwMax(                 // Componentwise maximum.
        const FVec2D & vec1,
        const FVec2D & vec2
        );

    // ---------------------------------------------- transformation functions

    //
    //  Multiply vector by special transformation.
    //
    FVec2D & operator*=(
        const FScale2D & scale
        );
    FVec2D & operator*=(
        const FRot2D & rot
        );
    FVec2D & operator*=(
        const FShearX2D & shear
        );
    FVec2D & operator*=(
        const FShearY2D & shear
        );
    FVec2D & operator*=(
        const FLin2D & lin
        );
    FVec2D & operator*=(
        const FTrafo2D & trafo
        );

    //
    //  Product of vector and special tranformation.
    //
    friend FVec2D operator*(
        const FVec2D & vec,
	const FScale2D & scale
        );
    friend FVec2D operator*(
        const FVec2D & vec,
	const FRot2D & rot
        );
    friend FVec2D operator*(
        const FVec2D & vec,
	const FShearX2D & shear
        );
    friend FVec2D operator*(
        const FVec2D & vec,
	const FShearY2D & shear
        );
    friend FVec2D operator*(
        const FVec2D & vec,
	const FLin2D & lin
        );
    friend FVec2D operator*(
        const FVec2D & vec,
	const FTrafo2D & trafo
        );

    // ----------------------------------------------- static member variables

    static const FVec2D baseX;
    static const FVec2D baseY;

protected:

    FCrd2 c;
};

// ===================================================================== FPnt2D

class FPnt2D
{
    friend class FVec2D;
    friend class FLin2D;
    friend class FTrans2D;
    friend class FTrafo2D;

    friend class Viewer3D;		//  needs FAST access to members

public:

    // ---------------------------------------------------------- constructors

    FPnt2D(
        );
    FPnt2D(
        float x,
        float y
        );
    FPnt2D(
        const FCrd2 & crd
        );
    FPnt2D(
        const FPnt2D & pnt
        );
    ~FPnt2D(
        );

    // ------------------------------------------------------ member operators

    float operator[](
        int index
        ) const;
    float & operator[](
        int index
        );

    FPnt2D & operator+=(
        const FVec2D & vec
        );
    FPnt2D & operator-=(
        const FVec2D & vec
        );

    // ------------------------------------------------------ member functions

    float x(
        ) const;
    float y(
        ) const;

    void x(
        float xNew
        );
    void y(
        float yNew
        );

    FCrd2 getFCrd2(
        ) const;
    FCrd3 getFCrd3(
        ) const;               // get homogeneous representation
    FVec2D getFVec2D(
        ) const;             // explicit conversion to FVec2D

    // ------------------------------------------------------ friend operators

    friend ostream & operator<<(
        ostream & s,
	const FPnt2D & vec
        );
    friend FPnt2D operator+(
        const FPnt2D & pnt,
	const FVec2D & vec
        );
    friend FPnt2D operator-(
        const FPnt2D & pnt,
	const FVec2D & vec
        );
    friend FVec2D operator-(
        const FPnt2D & pnt1,
	const FPnt2D & pnt2
        );

    // ------------------------------------------------------ friend functions

    friend FCrd2 getFCrd2(
        const FPnt2D & pnt
        );
    friend FCrd3 getFCrd3(
        const FPnt2D & pnt
        );	// get homogeneous rep.
    friend FVec2D getFVec2D(
        const FPnt2D & pnt
        );	// explicit conversion
    friend float squaredDistanceFromOrigin(
        const FPnt2D & pnt
        );
    friend float distanceFromOrigin(
        const FPnt2D & pnt
        );
    friend float squaredDistance(
        const FPnt2D & pnt1,
	const FPnt2D & pnt2
        );
    friend float distance(
        const FPnt2D & pnt1,
	const FPnt2D & pnt2
        );

    // ----------------------------------------------- componentwise functions

    friend FPnt2D cwMin(
        const FPnt2D & vec1,
	const FPnt2D & vec2
        );
    friend FPnt2D cwMax(
        const FPnt2D & vec1,
	const FPnt2D & vec2
        );

    // ---------------------------------------------- transformation functions

    FPnt2D & operator*=(
        const FScale2D & scale
        );
    FPnt2D & operator*=(
        const FRot2D & rot
        );
    FPnt2D & operator*=(
        const FShearX2D & shear
        );
    FPnt2D & operator*=(
        const FShearY2D & shear
        );
    FPnt2D & operator*=(
        const FLin2D & lin
        );
    FPnt2D & operator*=(
        const FTrans2D & trans
        );
    FPnt2D & operator*=(
        const FTrafo2D & trafo
        );

    friend FPnt2D operator*(
        const FPnt2D & pnt,
	const FScale2D & scale
        );
    friend FPnt2D operator*(
        const FPnt2D & pnt,
	const FRot2D & rot
        );
    friend FPnt2D operator*(
        const FPnt2D & pnt,
	const FShearX2D & shear
        );
    friend FPnt2D operator*(
        const FPnt2D & pnt,
	const FShearY2D & shear
        );
    friend FPnt2D operator*(
        const FPnt2D & pnt,
	const FLin2D & lin
        );
    friend FPnt2D operator*(
        const FPnt2D & pnt,
	const FTrans2D & trans
        );
    friend FPnt2D operator*(
        const FPnt2D & pnt,
	const FTrafo2D & trafo
        );

    // ----------------------------------------------- static member variables

    static const FPnt2D origin;

protected:
    FCrd2 c;
};

// ============================================================= class FScale2D

class FScale2D
{
    friend class FVec2D;
    friend class FPnt2D;
    friend class FLin2D;
    friend class FTrafo2D;

public:

    FScale2D(				// initialize with identity
    	);
    FScale2D(
        float scalar
        );
    FScale2D(
        float x,
        float y
        );
    FScale2D(
        const FCrd2 & scale
        );
    FScale2D(
        const FScale2D & scale
        );
    ~FScale2D(
        );

    FScale2D & operator*=(
        const FScale2D & scale
        );

    void invert(
        );

    friend ostream & operator<<(
        ostream & s,
	const FScale2D & scale
        );
    friend FScale2D operator*(
        const FScale2D & sc1,
	const FScale2D & sc2
        );

    friend FScale2D inverseCopy(
        const FScale2D & scale
        );

    static const FScale2D identity;

    // -------------------------------------------- functions that need access

    friend FVec2D operator*(
        const FVec2D & vec,
	const FScale2D & scale
        );
    friend FPnt2D operator*(
        const FPnt2D & pnt,
	const FScale2D & scale
        );
    friend FLin2D operator*(
        const FScale2D & scale,
	const FLin2D & lin
        );
    friend FLin2D operator*(
        const FLin2D & lin,
	const FScale2D & scale
        );
    friend FTrafo2D operator*(
        const FScale2D & scale,
	const FTrafo2D & trafo
        );
    friend FTrafo2D operator*(
        const FTrafo2D & trafo,
	const FScale2D & scale
        );

protected:

    FCrd2 c;
};

// =============================================================== class FRot2D

class FRot2D
{
    friend class FVec2D;
    friend class FPnt2D;
    friend class FLin2D;
    friend class FTrafo2D;

public:

    FRot2D(				// initialize with identity
    	);
    FRot2D(
        float newAlpha
        );
    FRot2D(
        float newAlpha,
        float newCosAlpha,
        float newSinAlpha
        );
    FRot2D(
        const FRot2D & rot
        );
    ~FRot2D(
        );

    FRot2D & operator*=(
        const FRot2D & rot
        );

    float getAlpha(
        ) const;
    void invert(
        );
    
    friend ostream & operator<<(
        ostream & s,
	const FRot2D & rot
        );
    friend FRot2D operator*(
        const FRot2D & rot1,
	const FRot2D & rot2
        );

    friend FRot2D inverseCopy(
        const FRot2D & rot
        );

    static const FRot2D identity;

    // -------------------------------------------- functions that need access

    friend FVec2D operator*(
        const FVec2D & vec,
	const FRot2D & rot
        );
    friend FPnt2D operator*(
        const FPnt2D & pnt,
	const FRot2D & rot
        );
    friend FLin2D operator*(
        const FLin2D & lin,
	const FRot2D & rot
        );
    friend FLin2D operator*(
        const FRot2D & rot,
	const FLin2D & lin
        );
    friend FTrafo2D operator*(
        const FTrafo2D & trafo,
	const FRot2D & rot
        );
    friend FTrafo2D operator*(
        const FRot2D & rot,
	const FTrafo2D & trafo
        );

protected:
    float alpha;
    float cosAlpha;
    float sinAlpha;
};

// ============================================================ class FShearX2D

class FShearX2D
{
    friend class FVec2D;
    friend class FPnt2D;
    friend class FLin2D;
    friend class FTrafo2D;

public:

    FShearX2D(				// initialize with identity
    	);
    FShearX2D(
        float cNew
        );
    FShearX2D(
        const FShearX2D & shear
        );
    ~FShearX2D(
	);

    FShearX2D & operator*=(
        const FShearX2D & shear
        );

    void invert(
        );
    
    friend ostream & operator<<(
        ostream & s,
	const FShearX2D & shear
        );
    friend FShearX2D operator*(
        const FShearX2D & s1,
	const FShearX2D & s2
        );

    friend FShearX2D inverseCopy(
        const FShearX2D & shear
        );

    static const FShearX2D identity;

    // -------------------------------------------- functions that need access

    friend FVec2D operator*(
        const FVec2D & vec,
	const FShearX2D & shear
        );
    friend FPnt2D operator*(
        const FPnt2D & pnt,
	const FShearX2D & shear
        );
    friend FLin2D operator*(
        const FLin2D & lin,
	const FShearX2D & shear
        );
    friend FLin2D operator*(
        const FShearX2D & shear,
	const FLin2D & lin
        );
    friend FTrafo2D operator*(
        const FTrafo2D & trafo,
	const FShearX2D & sh
        );
    friend FTrafo2D operator*(
        const FShearX2D & sh,
	const FTrafo2D & trafo
        );

protected:
    float c;
};

// ============================================================ class FShearY2D

class FShearY2D
{
    friend class FVec2D;
    friend class FPnt2D;
    friend class FLin2D;
    friend class FTrafo2D;

public:

    FShearY2D(				// initialize with identity
    	);
    FShearY2D(
        float cNew
        );
    FShearY2D(
        const FShearY2D & shear
        );
    ~FShearY2D(
	);

    FShearY2D & operator*=(
        const FShearY2D & shear
        );

    void invert(
        );
    
    friend ostream & operator<<(
        ostream & s,
	const FShearY2D & shear
        );
    friend FShearY2D operator*(
        const FShearY2D & s1,
	const FShearY2D & s2
        );

    friend FShearY2D inverseCopy(
        const FShearY2D & shear
        );

    static const FShearY2D identity;

    // -------------------------------------------- functions that need access

    friend FVec2D operator*(
        const FVec2D & vec,
	const FShearY2D & shear
        );
    friend FPnt2D operator*(
        const FPnt2D & pnt,
	const FShearY2D & shear
        );
    friend FLin2D operator*(
        const FLin2D & lin,
	const FShearY2D & shear
        );
    friend FLin2D operator*(
        const FShearY2D & shear,
	const FLin2D & lin
        );
    friend FTrafo2D operator*(
        const FTrafo2D & trafo,
	const FShearY2D & sh
        );
    friend FTrafo2D operator*(
        const FShearY2D & sh,
	const FTrafo2D & trafo
        );

protected:
    float c;
};

// ============================================================= class FLin2D

class FLin2D
{
    friend class FVec2D;
    friend class FPnt2D;
    friend class FTrafo2D;

public:

    FLin2D(				// initialize with identity
        );
    FLin2D(
        const FMat2 & m
        );
    FLin2D(
        const FVec2D & vecX,
        const FVec2D & vecY
        );
    FLin2D(
        float xx, float xy,
        float yx, float yy
        );
    FLin2D(
        const FScale2D & scale
        );
    FLin2D(
        const FRot2D & rot
        );
    FLin2D(
        const FShearX2D & shear
        );
    FLin2D(
        const FShearY2D & shear
        );
    FLin2D(
        const FLin2D & lin
        );
    ~FLin2D(
        );

    // ------------------------------------------------------ member functions

    FVec2D getVecX(
        ) const;
    FVec2D getVecY(
        ) const;

    FMat2 getFMat2(
        ) const;

    void invert(
        );
        //
        //  ERROR_CRD2_SINGULAR_MATRIX see module FCrd2
        //

    // ------------------------------------------------------ friend operators

    friend ostream & operator<<(
        ostream & s,
	const FLin2D & lin
        );

    // ------------------------------------------------------ friend functions

    friend FVec2D getVecX(
        const FLin2D & lin
        );
    friend FVec2D getVecY(
        const FLin2D & lin
        );

    friend FMat2 getFMat2(
        const FLin2D & lin
        );
    friend float determinant(
        const FLin2D & lin
        );
    friend FLin2D inverseCopy(
        const FLin2D & lin,
	float det
        );
        //
        //  det is NOT checked against zero!
        //

    friend FLin2D inverseCopy(
        const FLin2D & lin
        );
        //
        //  ERROR_CRD2_SINGULAR_MATRIX see module FCrd2
        //

    // ----------------------------------------------- tranformation functions

    //
    //  Multiply transformation by special transformation.
    //
    FLin2D & operator*=(
        const FScale2D & scale
        );
    FLin2D & operator*=(
        const FRot2D & rot
        );
    FLin2D & operator*=(
        const FShearX2D & shear
        );
    FLin2D & operator*=(
        const FShearY2D & shear
        );
    FLin2D & operator*=(
        const FLin2D & lin
        );

    //
    //  Product of transformation and special transformation.
    //
    friend FLin2D operator*(
        const FLin2D & lin,
	const FScale2D & scale
        );
    friend FLin2D operator*(
        const FLin2D & lin,
	const FRot2D & rot
        );
    friend FLin2D operator*(
        const FLin2D & lin,
	const FShearX2D & shear
        );
    friend FLin2D operator*(
        const FLin2D & lin,
	const FShearY2D & shear
        );

    friend FLin2D operator*(		// Product of two transformations.
        const FLin2D & lin1,
        const FLin2D & lin2
        );

    //
    //  Product of special transformation and transformation.
    //
    friend FLin2D operator*(
        const FScale2D & scale,
	const FLin2D & lin
        );
    friend FLin2D operator*(
        const FRot2D & rot,
	const FLin2D & lin
        );
    friend FLin2D operator*(
        const FShearX2D & shear,
	const FLin2D & lin
        );
    friend FLin2D operator*(
        const FShearY2D & shear,
	const FLin2D & lin
        );

    //
    //  Product of vector/point and transformation.
    //
    friend FVec2D operator*(
        const FVec2D & vec,
	const FLin2D & lin
        );
    friend FPnt2D operator*(
        const FPnt2D & pnt,
	const FLin2D & lin
        );

    // ----------------------------------------------- static member variables

    static const FLin2D identity;

    // -------------------------------------------- functions that need access

    friend FTrafo2D operator*(
        const FLin2D & lin,
	const FTrafo2D & trafo
        );
    friend FTrafo2D operator*(
        const FTrafo2D & trafo,
	const FLin2D & lin
        );

protected:

    FMat2 m;
};

// ============================================================= class FTrans2D

class FTrans2D
{
    friend class FVec2D;
    friend class FPnt2D;
    friend class FTrafo2D;

public:

    FTrans2D(				// initialize with identity
	);
    FTrans2D(
        const FCrd2 & crd
        );
    FTrans2D(
        float xNew,
        float yNew
        );
    FTrans2D(
	const FVec2D & vec
	);
    FTrans2D(
	const FPnt2D & pntO
	);
    FTrans2D(
        const FTrans2D & trans
        );
    ~FTrans2D(
	);

    FTrans2D & operator*=(
        const FTrans2D & trans
        );

    void invert(
        );

    friend ostream & operator<<(
        ostream & s,
	const FTrans2D & trans
        );
    friend FTrans2D operator*(
        const FTrans2D & tr1,
	const FTrans2D & tr2
        );

    friend FTrans2D inverseCopy(
        const FTrans2D & trans
        );

    // -------------------------------------------- functions that need access

    friend FVec2D operator*(
        const FVec2D & vec,
	const FTrans2D & trans
        );
    friend FPnt2D operator*(
        const FPnt2D & pnt,
	const FTrans2D & trans
        );
    friend FTrafo2D operator*(
        const FTrans2D & trans,
	const FTrafo2D & trafo
        );
    friend FTrafo2D operator*(
        const FTrafo2D & trafo,
	const FTrans2D & trans
        );

protected:

    FCrd2 c;
};

// ============================================================= class FTrafo2D

class FTrafo2D
{
    friend class FVec2D;
    friend class FPnt2D;

public:

    FTrafo2D(				// initialize with identity
        );
    FTrafo2D(
        const FTrafo2D & trafo
        );
    FTrafo2D(
	const FLin2D & lin,
	const FTrans2D & trans
	);
    FTrafo2D(
        const FMat2 & m
        );
    FTrafo2D(
        const FMat2 & m,
        const FCrd2 & c
        );
    FTrafo2D(
        const FVec2D & x,
        const FVec2D & y,
        const FPnt2D & d
        );
    FTrafo2D(
        float xx, float xy,
        float yx, float yy,
        float dx, float dy
        );
    FTrafo2D(
        const FScale2D & scale
        );
    FTrafo2D(
        const FRot2D & rot
        );
    FTrafo2D(
        const FShearX2D & shear
        );
    FTrafo2D(
        const FShearY2D & shear
        );
    FTrafo2D(
	const FLin2D & lin
	);
    FTrafo2D(
        const FTrans2D & trans
        );
    ~FTrafo2D(
        );

    // ------------------------------------------------------ member functions

    FVec2D getVecX(
        ) const;
    FVec2D getVecY(
        ) const;
    FPnt2D getPntO(
        ) const;

    FMat2 getFMat2(
        ) const;
    FMat3 getFMat3(
        ) const;               // get homogeneous representation

    void invert(
        );
        //
        //  ERROR_CRD2_SINGULAR_MATRIX see module FCrd2
        //

    // ------------------------------------------------------ friend operators

    friend ostream & operator<<(
        ostream & s,
	const FTrafo2D & trafo
        );

    // ------------------------------------------------------ friend functions

    friend FVec2D getVecX(
        const FTrafo2D & trafo
        );
    friend FVec2D getVecY(
        const FTrafo2D & trafo
        );
    friend FPnt2D getPntO(
        const FTrafo2D & trafo
        );

    friend FMat2 getFMat2(
        const FTrafo2D & trafo
        );
    friend FMat3 getFMat3(
        const FTrafo2D & trafo
        );	// get homogeneous rep.
    friend float determinant(
        const FTrafo2D & trafo
        );
    friend FTrafo2D inverseCopy(
        const FTrafo2D & trafo,
	float det
        );
        //
        //  det is NOT checked against zero!
        //

    friend FTrafo2D inverseCopy(
        const FTrafo2D & trafo
        );
        //
        //  ERROR_CRD2_SINGULAR_MATRIX see module FCrd2
        //

    // ----------------------------------------------- tranformation functions

    //
    //  Multiply transformation by special transformation.
    //
    FTrafo2D & operator*=(
        const FScale2D & scale
        );
    FTrafo2D & operator*=(
        const FRot2D & rot
        );
    FTrafo2D & operator*=(
        const FShearX2D & shear
        );
    FTrafo2D & operator*=(
        const FShearY2D & shear
        );
    FTrafo2D & operator*=(
        const FLin2D & lin
        );
    FTrafo2D & operator*=(
        const FTrans2D & trans
        );
    FTrafo2D & operator*=(
        const FTrafo2D & trafo
        );

    //
    //  Product of transformation and special transformation.
    //
    friend FTrafo2D operator*(
        const FTrafo2D & trafo,
	const FScale2D & scale
        );
    friend FTrafo2D operator*(
        const FTrafo2D & trafo,
	const FRot2D & rot
        );
    friend FTrafo2D operator*(
        const FTrafo2D & trafo,
	const FShearX2D & sh
        );
    friend FTrafo2D operator*(
        const FTrafo2D & trafo,
	const FShearY2D & sh
        );
    friend FTrafo2D operator*(
        const FTrafo2D & trafo,
	const FLin2D & lin
        );
    friend FTrafo2D operator*(
        const FTrafo2D & trafo,
	const FTrans2D & trans
        );

    friend FTrafo2D operator*(		// Product of two transformations.
        const FTrafo2D & trafo1,
        const FTrafo2D & trafo2
        );

    //
    //  Product of special transformation and transformation.
    //
    friend FTrafo2D operator*(
        const FScale2D & scale,
	const FTrafo2D & trafo
        );
    friend FTrafo2D operator*(
        const FRot2D & rot,
	const FTrafo2D & trafo
        );
    friend FTrafo2D operator*(
        const FShearX2D & sh,
	const FTrafo2D & trafo
        );
    friend FTrafo2D operator*(
        const FShearY2D & sh,
	const FTrafo2D & trafo
        );
    friend FTrafo2D operator*(
        const FLin2D & lin,
	const FTrafo2D & trafo
        );
    friend FTrafo2D operator*(
        const FTrans2D & trans,
	const FTrafo2D & trafo
        );

    //
    //  Product of vector/point and transformation.
    //
    friend FVec2D operator*(
        const FVec2D & vec,
	const FTrafo2D & trafo
        );
    friend FPnt2D operator*(
        const FPnt2D & pnt,
	const FTrafo2D & trafo
        );

    // ----------------------------------------------- static member variables

    static const FTrafo2D identity;

protected:

    FMat2 m;
    FCrd2 c;
};

} // extern "C++"

#ifndef FVEC2D_NO_INLINE

#define INLINE_GLOBAL COMPILER_INLINE_GLOBAL
#define INLINE_MEMBER COMPILER_INLINE_MEMBER

#include "VEGA/Graphics/fvec2d.ipp"

#undef INLINE_GLOBAL
#undef INLINE_MEMBER

#endif


#endif // _GRAPHICS_FVEC2D_H_
// ===========================================================================
