// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       FVec2D
//  TYPE:       c++ inline
//  PROJECT:    VEGA Graphics Library
//  CONTENT:    2D vectors, points, and transformations using floats
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rft     Robert F. Tobler
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 14:08:16  rft     last modification
//  08-Aug-93           rft     created
// ===========================================================================

extern "C++"
{

// =============================================================== class FVec2D

// -------------------------------------------------------------- constructors

INLINE_MEMBER FVec2D::FVec2D()
{
    TR_FUN0(FVec2D_constructors,FVec2D);
}

INLINE_MEMBER FVec2D::FVec2D(float xNew, float yNew)
:
    c(xNew,yNew)
{
    TR_FUN2(FVec2D_constructors,FVec2D,xNew,yNew);
}

INLINE_MEMBER FVec2D::FVec2D(const FCrd2 & crd)
:
    c(crd)
{
    TR_FUN1(FVec2D_constructors,FVec2D,crd);
}

INLINE_MEMBER FVec2D::FVec2D(const FVec2D & vec)
:
    c(vec.c)
{
    TR_FUN1(FVec2D_constructors,FVec2D,vec);
}

INLINE_MEMBER FVec2D::~FVec2D()
{
    TR_MESSAGE(FVec2D_constructors,"~" << *this);
}

// ---- methods --------------------------------------------------------------

INLINE_MEMBER float FVec2D::operator[](int index) const
{
    return c[index];
}

INLINE_MEMBER float & FVec2D::operator[](int index)
{
    return c[index];
}

INLINE_MEMBER FVec2D & FVec2D::operator+=(const FVec2D & vec)
{
    c += vec.c;
    return *this;
}

INLINE_MEMBER FVec2D & FVec2D::operator-=(const FVec2D & vec)
{
    c -= vec.c;
    return *this;
}

INLINE_MEMBER FVec2D & FVec2D::operator*=(float scalar)
{
    c *= scalar;
    return *this;
}

INLINE_MEMBER float FVec2D::x() const { return c.x0(); }
INLINE_MEMBER float FVec2D::y() const { return c.x1(); }

INLINE_MEMBER void FVec2D::x(float xNew) { c.x0(xNew); }
INLINE_MEMBER void FVec2D::y(float yNew) { c.x1(yNew); }

INLINE_MEMBER void FVec2D::normalize()
{
    c.normalize();
}

INLINE_MEMBER FCrd2 FVec2D::getFCrd2() const
{
    return c;
}

INLINE_MEMBER FCrd3 FVec2D::getFCrd3() const
{
    return FCrd3(c.x0(),c.x1(),0.0);
}

INLINE_MEMBER FPnt2D::FPnt2D(const FCrd2 & crd)
:
    c(crd)
{
    TR_FUN1(FVec2D_constructors,FPnt2D,crd);
}

INLINE_MEMBER FPnt2D FVec2D::getFPnt2D() const
{
    return FPnt2D(c);
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const FVec2D & vec)
{
    s << "FVec2D(" << vec.x() << "," << vec.y() << ")";
    return s;
}

INLINE_GLOBAL FVec2D operator-(const FVec2D & vec)
{
    return FVec2D(-vec.c);
}

INLINE_GLOBAL FVec2D operator+(const FVec2D & vec1, const FVec2D & vec2)
{
    return FVec2D(vec1.c + vec2.c);
}

INLINE_GLOBAL FPnt2D operator+(const FPnt2D & pnt, const FVec2D & vec)
{
    return FPnt2D(pnt.c + vec.c);
}

INLINE_GLOBAL FVec2D operator-(const FVec2D & vec1, const FVec2D & vec2)
{
    return FVec2D(vec1.c - vec2.c);
}

INLINE_GLOBAL FPnt2D operator-(const FPnt2D & pnt, const FVec2D & vec)
{
    return FPnt2D(pnt.c - vec.c);
}

INLINE_GLOBAL FVec2D operator*(const FVec2D & vec, float scalar)
{
    return FVec2D(vec.c * scalar);
}

INLINE_GLOBAL FVec2D operator*(float scalar, const FVec2D & vec)
{
    return FVec2D(scalar * vec.c);
}

INLINE_GLOBAL FCrd2 getFCrd2(const FVec2D & vec)
{
    return vec.c;
}

INLINE_GLOBAL FCrd3 getFCrd3(const FVec2D & vec)
{
    return FCrd3(vec.c.x0(), vec.c.x1(),0.0);
}

INLINE_GLOBAL FPnt2D getFPnt2D(const FVec2D & vec)
{
    return FPnt2D(vec.c);
}

INLINE_GLOBAL float squaredLength(const FVec2D & vec)
{
    return squaredLength(vec.c);
}

INLINE_GLOBAL float length(const FVec2D & vec)
{
    return length(vec.c);
}

INLINE_GLOBAL float dotProduct(const FVec2D & vec1, const FVec2D & vec2)
{
    return dotProduct(vec1.c,vec2.c);
}

INLINE_GLOBAL FVec2D normalizedCopy(const FVec2D & vec)
{
    return FVec2D(normalizedCopy(vec.c));
}

INLINE_GLOBAL FVec2D cwReciprocalCopy(const FVec2D & vec)
{
    return FVec2D(cwReciprocalCopy(vec.c));
}

INLINE_GLOBAL FVec2D cwMin(const FVec2D & vec1, const FVec2D & vec2)
{
    return FVec2D(cwMin(vec1.c,vec2.c));
}

INLINE_GLOBAL FVec2D cwMax(const FVec2D & vec1, const FVec2D & vec2)
{
    return FVec2D(cwMax(vec1.c,vec2.c));
}

INLINE_MEMBER FVec2D & FVec2D::operator*=(const FScale2D & scale)
{
    c.scale(scale.c);
    return *this;
}

INLINE_MEMBER FVec2D & FVec2D::operator*=(const FRot2D & rot)
{
    c.rot(rot.cosAlpha,rot.sinAlpha);
    return *this;
}

INLINE_MEMBER FVec2D & FVec2D::operator*=(const FShearX2D & shear)
{
    c.shear0(shear.c);
    return *this;
}

INLINE_MEMBER FVec2D & FVec2D::operator*=(const FShearY2D & shear)
{
    c.shear1(shear.c);
    return *this;
}

INLINE_MEMBER FVec2D & FVec2D::operator*=(const FLin2D & lin)
{
    c *= lin.m;
    return *this;
}

INLINE_MEMBER FVec2D & FVec2D::operator*=(const FTrafo2D & trafo)
{
    c *= trafo.m;
    return *this;
}

INLINE_GLOBAL FVec2D operator*(const FVec2D & vec, const FScale2D & scale)
{
    return FVec2D(prCrdScale(vec.c,scale.c));
}

INLINE_GLOBAL FVec2D operator*(const FVec2D & vec, const FRot2D & rot)
{
    return FVec2D(prCrdRot(vec.c,rot.cosAlpha,rot.sinAlpha));
}

INLINE_GLOBAL FVec2D operator*(const FVec2D & vec, const FShearX2D & shear)
{
    return FVec2D(prCrdShear0(vec.c,shear.c));
}

INLINE_GLOBAL FVec2D operator*(const FVec2D & vec, const FShearY2D & shear)
{
    return FVec2D(prCrdShear1(vec.c,shear.c));
}

INLINE_GLOBAL FVec2D operator*(const FVec2D & vec, const FLin2D & lin)
{
    return FVec2D(vec.c * lin.m);
}

INLINE_GLOBAL FVec2D operator*(const FVec2D & vec, const FTrafo2D & trafo)
{
    return FVec2D(vec.c * trafo.m);
}

// =============================================================== class FPnt2D

// -------------------------------------------------------------- constructors

INLINE_MEMBER FPnt2D::FPnt2D()
{
    TR_FUN0(FVec2D_constructors,FPnt2D);
}

INLINE_MEMBER FPnt2D::FPnt2D(float xNew, float yNew)
:
    c(xNew,yNew)
{
    TR_FUN2(FVec2D_constructors,FPnt2D,xNew,yNew);
}

INLINE_MEMBER FPnt2D::FPnt2D(const FPnt2D & pnt)
:
    c(pnt.c)
{
    TR_FUN1(FVec2D_constructors,FPnt2D,pnt);
}

INLINE_MEMBER FPnt2D::~FPnt2D()
{
    TR_MESSAGE(FVec2D_constructors,"~" << *this);
}

// ---------------------------------------------------------- member functions

INLINE_MEMBER float FPnt2D::operator[](int index) const
{
    return c[index];
}

INLINE_MEMBER float & FPnt2D::operator[](int index)
{
    return c[index];
}

INLINE_MEMBER FPnt2D & FPnt2D::operator+=(const FVec2D & vec)
{
    c += vec.c;
    return *this;
}

INLINE_MEMBER FPnt2D & FPnt2D::operator-=(const FVec2D & vec)
{
    c -= vec.c;
    return *this;
}

INLINE_MEMBER float FPnt2D::x() const { return c.x0(); }
INLINE_MEMBER float FPnt2D::y() const { return c.x1(); }

INLINE_MEMBER void FPnt2D::x(float xNew) { c.x0(xNew); }
INLINE_MEMBER void FPnt2D::y(float yNew) { c.x1(yNew); }

INLINE_MEMBER FCrd2 FPnt2D::getFCrd2() const
{
    return c;
}

INLINE_MEMBER FCrd3 FPnt2D::getFCrd3() const
{
    return FCrd3(c.x0(),c.x1(),0.0);
}

INLINE_MEMBER FVec2D FPnt2D::getFVec2D() const
{
    return FVec2D(c);
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const FPnt2D & vec)
{
    s << "(" << vec.x() << "," << vec.y() << ")";
    return s;
}

INLINE_GLOBAL FVec2D operator-(const FPnt2D & pnt1, const FPnt2D & pnt2)
{
    return FVec2D(pnt1.c - pnt2.c);
}

INLINE_GLOBAL FCrd2 getFCrd2(const FPnt2D & pnt)
{
    return pnt.c;
}

INLINE_GLOBAL FCrd3 getFCrd3(const FPnt2D & pnt)
{
    return FCrd3(pnt.c.x0(), pnt.c.x1(),0.0);
}

INLINE_GLOBAL FVec2D getFVec2D(const FPnt2D & pnt)
{
    return FVec2D(pnt.c);
}

INLINE_GLOBAL float squaredDistanceFromOrigin(const FPnt2D & pnt)
{
    return (squaredLength(pnt.c));
}

INLINE_GLOBAL float distanceFromOrigin(const FPnt2D & pnt)
{
    return (length(pnt.c));
}

INLINE_GLOBAL float squaredDistance(const FPnt2D & pnt1, const FPnt2D & pnt2)
{
    return (squaredLength(pnt1.c - pnt2.c));
}

INLINE_GLOBAL float distance(const FPnt2D & pnt1, const FPnt2D & pnt2)
{
    return (length(pnt1.c - pnt2.c));
}

INLINE_GLOBAL FPnt2D cwMin(const FPnt2D & pnt1, const FPnt2D & pnt2)
{
    return FPnt2D(cwMin(pnt1.c,pnt2.c));
}

INLINE_GLOBAL FPnt2D cwMax(const FPnt2D & pnt1, const FPnt2D & pnt2)
{
    return FPnt2D(cwMax(pnt1.c,pnt2.c));
}

INLINE_MEMBER FPnt2D & FPnt2D::operator*=(const FScale2D & scale)
{
    c.scale(scale.c);
    return *this;
}

INLINE_MEMBER FPnt2D & FPnt2D::operator*=(const FRot2D & rot)
{
    c.rot(rot.cosAlpha,rot.sinAlpha);
    return *this;
}

INLINE_MEMBER FPnt2D & FPnt2D::operator*=(const FShearX2D & shear)
{
    c.shear0(shear.c);
    return *this;
}

INLINE_MEMBER FPnt2D & FPnt2D::operator*=(const FShearY2D & shear)
{
    c.shear1(shear.c);
    return *this;
}

INLINE_MEMBER FPnt2D & FPnt2D::operator*=(const FLin2D & lin)
{
    c *= lin.m;
    return *this;
}

INLINE_MEMBER FPnt2D & FPnt2D::operator*=(const FTrans2D & trans)
{
    c += trans.c;
    return *this;
}

INLINE_MEMBER FPnt2D & FPnt2D::operator*=(const FTrafo2D & trafo)
{
    c *= trafo.m;
    c += trafo.c;
    return *this;
}

INLINE_GLOBAL FPnt2D operator*(const FPnt2D & pnt, const FScale2D & scale)
{
    return FPnt2D(prCrdScale(pnt.c,scale.c));
}

INLINE_GLOBAL FPnt2D operator*(const FPnt2D & pnt, const FRot2D & rot)
{
    return FPnt2D(prCrdRot(pnt.c,rot.cosAlpha,rot.sinAlpha));
}

INLINE_GLOBAL FPnt2D operator*(const FPnt2D & pnt, const FShearX2D & shear)
{
    return FPnt2D(prCrdShear0(pnt.c,shear.c));
}

INLINE_GLOBAL FPnt2D operator*(const FPnt2D & pnt, const FShearY2D & shear)
{
    return FPnt2D(prCrdShear1(pnt.c,shear.c));
}

INLINE_GLOBAL FPnt2D operator*(const FPnt2D & pnt, const FLin2D & lin)
{
    return FPnt2D(pnt.c * lin.m);
}

INLINE_GLOBAL FPnt2D operator*(const FPnt2D & pnt, const FTrans2D & trans)
{
    return FPnt2D(pnt.c + trans.c);
}

INLINE_GLOBAL FPnt2D operator*(const FPnt2D & pnt, const FTrafo2D & trafo)
{
    return FPnt2D(pnt.c * trafo.m + trafo.c);
}


// ============================================================== class FScale2D

INLINE_MEMBER FScale2D::FScale2D(
	)
:
    c(1,1)
{
    TR_FUN0(FVec2D_constructors, FScale2D);
}

INLINE_MEMBER FScale2D::FScale2D(float scalar)
:
    c(scalar,scalar)
{
    TR_FUN1(FVec2D_constructors,FScale2D,scalar);
}

INLINE_MEMBER FScale2D::FScale2D(float x, float y)
:
    c(x,y)
{
    TR_FUN2(FVec2D_constructors,FScale2D,x,y);
}

INLINE_MEMBER FScale2D::FScale2D(const FCrd2 & scale)
:
    c(scale)
{
    TR_FUN1(FVec2D_constructors,FScale2D,scale);
}

INLINE_MEMBER FScale2D::FScale2D(const FScale2D & scale)
:
    c(scale.c)
{
    TR_FUN1(FVec2D_constructors,FScale2D,scale);
}

INLINE_MEMBER FScale2D::~FScale2D()
{
    TR_MESSAGE(FVec2D_constructors,"~" << *this);
}

INLINE_MEMBER FScale2D & FScale2D::operator*=(const FScale2D & scale)
{
    c.cwMultiplyBy(scale.c);
    return *this;
}

INLINE_MEMBER void FScale2D::invert()
{
    if (c.x0() == 0.0 ||                // use explicit accessor function
        c.x1() == 0.0 )                 // to avoid range checks
    {
        EH_ERROR(FVec2D_errors, ERROR_FVEC2D_SINGULAR_SCALE, "singular scale");
        return;
    }

    c.x0(1.0/c.x0());
    c.x1(1.0/c.x1());
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const FScale2D & scale)
{
    s << "FScale2D(" << scale.c.x0() << "," << scale.c.x1() << ")";
    return s;
}

INLINE_GLOBAL FScale2D operator*(const FScale2D & scale1, const FScale2D & scale2)
{
    return FScale2D(cwProduct(scale1.c,scale2.c));
}


INLINE_GLOBAL FScale2D inverseCopy(const FScale2D & scale)
{
    if (scale.c.x0() == 0.0 ||
        scale.c.x1() == 0.0)
    {
        EH_ERROR(FVec2D_errors, ERROR_FVEC2D_SINGULAR_SCALE, "singular scale");
        return scale;
    }

    return FScale2D(1.0/scale.c.x0(), 1.0/scale.c.x1());
}


// =============================================================== class FRot2D

INLINE_MEMBER FRot2D::FRot2D(
	)
:
    alpha(0.0),
    cosAlpha(1.0),
    sinAlpha(0.0)
{
    TR_FUN0(FVec2D_constructors,FRot2D);
}

INLINE_MEMBER FRot2D::FRot2D(float newAlpha)
:
    alpha(newAlpha),
    cosAlpha(cos(newAlpha)),
    sinAlpha(sin(newAlpha))
{
    TR_FUN1(FVec2D_constructors,FRot2D,newAlpha);
}

INLINE_MEMBER FRot2D::FRot2D(
	float newAlpha, float newCosAlpha, float newSinAlpha
	)
:
    alpha(newAlpha),
    cosAlpha(newCosAlpha),
    sinAlpha(newSinAlpha)
{
    TR_FUN3(FVec2D_constructors,FRot2D,newAlpha,newCosAlpha,newSinAlpha);
}

INLINE_MEMBER FRot2D::FRot2D(const FRot2D & rot)
:
    alpha(rot.alpha),
    cosAlpha(rot.cosAlpha),
    sinAlpha(rot.sinAlpha)
{
    TR_FUN1(FVec2D_constructors,FRot2D,rot);
}

INLINE_MEMBER FRot2D::~FRot2D()
{
    TR_MESSAGE(FVec2D_constructors,"~" << *this);
}

INLINE_MEMBER FRot2D & FRot2D::operator*=(const FRot2D & rot)
{
    alpha += rot.alpha;
    cosAlpha = cos(alpha);
    sinAlpha = sin(alpha);
    
    return *this;
}

INLINE_MEMBER float FRot2D::getAlpha() const
{
    return alpha;
}

INLINE_MEMBER void FRot2D::invert()
{
    alpha = -alpha;
    sinAlpha = -sinAlpha;
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const FRot2D & rot)
{
    s << "FRot2D(" << rot.alpha << ")";
    return s;
}

INLINE_GLOBAL FRot2D inverseCopy(const FRot2D & rot)
{
    return FRot2D(-rot.alpha,rot.cosAlpha,-rot.sinAlpha);
}

INLINE_GLOBAL FRot2D operator*(const FRot2D & rot1, const FRot2D & rot2)
{
    return FRot2D(rot1.alpha + rot2.alpha);
}

// ============================================================ class FShearX2D

INLINE_MEMBER FShearX2D::FShearX2D(
	)
:
    c(0.0)
{
    TR_FUN0(FVec2D_constructors,FShearX2D);
}

INLINE_MEMBER FShearX2D::FShearX2D(float cNew)
:
    c(cNew)
{
    TR_FUN1(FVec2D_constructors,FShearX2D,cNew);
}

INLINE_MEMBER FShearX2D::FShearX2D(const FShearX2D & shear)
:
    c(shear.c)
{
    TR_FUN1(FVec2D_constructors,FShearX2D,shear);
}

INLINE_MEMBER FShearX2D::~FShearX2D()
{
    TR_MESSAGE(FVec2D_constructors,"~" << *this);
}

INLINE_MEMBER FShearX2D & FShearX2D::operator*=(const FShearX2D & shear)
{
    c += shear.c;
    return *this;
}

INLINE_MEMBER void FShearX2D::invert()
{
    c = -c;
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const FShearX2D & shear)
{
    s << "FShearX2D(" << shear.c << ")";
    return s;
}

INLINE_GLOBAL FShearX2D inverseCopy(const FShearX2D & shear)
{
    return FShearX2D(-shear.c);
}

INLINE_GLOBAL FShearX2D operator*(const FShearX2D & shear1, const FShearX2D & shear2)
{
    return FShearX2D(shear1.c + shear2.c);
}

// ============================================================ class FShearY2D

INLINE_MEMBER FShearY2D::FShearY2D(
	)
:
    c(0.0)
{
    TR_FUN0(FVec2D_constructors,FShearY2D);
}

INLINE_MEMBER FShearY2D::FShearY2D(float cNew)
:
    c(cNew)
{
    TR_FUN1(FVec2D_constructors,FShearY2D,cNew);
}

INLINE_MEMBER FShearY2D::FShearY2D(const FShearY2D & shear)
:
    c(shear.c)
{
    TR_FUN1(FVec2D_constructors,FShearY2D,shear);
}

INLINE_MEMBER FShearY2D::~FShearY2D()
{
    TR_MESSAGE(FVec2D_constructors,"~" << *this);
}

INLINE_MEMBER FShearY2D & FShearY2D::operator*=(const FShearY2D & shear)
{
    c += shear.c;
    return *this;
}

INLINE_MEMBER void FShearY2D::invert()
{
    c = -c;
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const FShearY2D & shear)
{
    s << "FShearY2D(" << shear.c << ")";
    return s;
}

INLINE_GLOBAL FShearY2D inverseCopy(const FShearY2D & shear)
{
    return FShearY2D(-shear.c);
}

INLINE_GLOBAL FShearY2D operator*(const FShearY2D & shear1, const FShearY2D & shear2)
{
    return FShearY2D(shear1.c + shear2.c);
}

// ============================================================= class FLin2D

INLINE_MEMBER FLin2D::FLin2D()
:
    m(1.0,0.0, 0.0,1.0)
{
    TR_FUN0(FVec2D_constructors,FLin2D);
}

INLINE_MEMBER FLin2D::FLin2D(const FScale2D & scale)
:
    m(  scale.c.x0(),   0.0,
        0.0,            scale.c.x1()	)
{
    TR_FUN1(FVec2D_constructors,FLin2D,scale);
}

INLINE_MEMBER FLin2D::FLin2D(const FRot2D & rot)
:
    m(  rot.cosAlpha,   rot.sinAlpha,
        -rot.sinAlpha,  rot.cosAlpha	)
{
    TR_FUN1(FVec2D_constructors,FLin2D,rot);
}

INLINE_MEMBER FLin2D::FLin2D(const FShearX2D & shear)
:
    m(  1.0,	 0.0,
        shear.c, 1.0	)
{
    TR_FUN1(FVec2D_constructors,FLin2D,shear);
}

INLINE_MEMBER FLin2D::FLin2D(const FShearY2D & shear)
:
    m(  1.0,	 shear.c,
        0.0,	 1.0	)
{
    TR_FUN1(FVec2D_constructors,FLin2D,shear);
}

INLINE_MEMBER FLin2D::FLin2D(const FMat2 & mat)
:
    m(mat)
{
    TR_FUN1(FVec2D_constructors,FLin2D,mat);
}

INLINE_MEMBER FLin2D::FLin2D(const FLin2D & lin)
:
    m(lin.m)
{
    TR_FUN1(FVec2D_constructors,FLin2D,lin);
}

INLINE_MEMBER FLin2D::FLin2D(const FVec2D & vecX, const FVec2D & vecY)
:
    m(vecX.c.x0(), vecX.c.x1(),
      vecY.c.x0(), vecY.c.x1())
{
    TR_FUN2(FVec2D_constructors,FLin2D,vecX,vecY);
}

INLINE_MEMBER FLin2D::FLin2D(
        float xx, float xy,
        float yx, float yy
        )
:
    m(xx,xy, yx,yy)
{
    TR_FUN3(FVec2D_constructors,FLin2D,xx,xy,"...");
}

INLINE_MEMBER FLin2D::~FLin2D()
{
    TR_MESSAGE(FVec2D_constructors,"~" << *this);
}

INLINE_MEMBER FVec2D FLin2D::getVecX() const
{
    return FVec2D(m(0,0),m(0,1));
}

INLINE_MEMBER FVec2D FLin2D::getVecY() const
{
    return FVec2D(m(1,0),m(1,1));
}

INLINE_MEMBER FMat2 FLin2D::getFMat2() const
{
    return m;
}

INLINE_MEMBER FLin2D & FLin2D::operator*=(const FScale2D & scale)
{
    m.scale(scale.c);
    return *this;
}

INLINE_MEMBER FLin2D & FLin2D::operator*=(const FRot2D & rot)
{
    m.rot(rot.cosAlpha,rot.sinAlpha);
    return *this;
}

INLINE_MEMBER FLin2D & FLin2D::operator*=(const FShearX2D & shear)
{
    m.shear0(shear.c);
    return *this;
}

INLINE_MEMBER FLin2D & FLin2D::operator*=(const FShearY2D & shear)
{
    m.shear1(shear.c);
    return *this;
}

INLINE_MEMBER FLin2D & FLin2D::operator*=(const FLin2D & lin)
{
    m *= lin.m;
    return *this;
}

INLINE_GLOBAL FVec2D getVecX(const FLin2D & lin) { return lin.getVecX(); }
INLINE_GLOBAL FVec2D getVecY(const FLin2D & lin) { return lin.getVecY(); }
INLINE_GLOBAL FMat2 getFMat2(const FLin2D & lin) { return lin.m; }

INLINE_GLOBAL float determinant(const FLin2D & lin)
{
    return determinant(lin);
}

INLINE_GLOBAL FLin2D inverseCopy(const FLin2D & lin, float det)
{
    return FLin2D(inverseCopy(lin.m,det));
}

INLINE_MEMBER void FLin2D::invert(
        )
{
    *this = inverseCopy(*this,determinant(*this));
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const FLin2D & lin)
{
    s << "FLin2D(" << lin.m << ')';
    return s;
}

INLINE_GLOBAL FLin2D inverseCopy(const FLin2D & lin)
{
    return inverseCopy(lin,determinant(lin));
}

INLINE_GLOBAL FLin2D operator*(const FLin2D & lin, const FScale2D & scale)
{
    return FLin2D(prMatScale(lin.m,scale.c));
}

INLINE_GLOBAL FLin2D operator*(const FLin2D & lin, const FRot2D & rot)
{
    return FLin2D(prMatRot(lin.m,rot.cosAlpha,rot.sinAlpha));
}

INLINE_GLOBAL FLin2D operator*(const FLin2D & lin, const FShearX2D & shear)
{
    return FLin2D(prMatShear0(lin.m,shear.c));
}

INLINE_GLOBAL FLin2D operator*(const FLin2D & lin, const FShearY2D & shear)
{
    return FLin2D(prMatShear1(lin.m,shear.c));
}

INLINE_GLOBAL FLin2D operator*(const FLin2D & lin1, const FLin2D & lin2)
{
    return FLin2D(lin1.m * lin2.m);
}

INLINE_GLOBAL FLin2D operator*(const FScale2D & scale, const FLin2D & lin)
{
    return FLin2D(prScaleMat(scale.c,lin.m));
}

INLINE_GLOBAL FLin2D operator*(const FRot2D & rot, const FLin2D & lin)
{
    return FLin2D(prRotMat(rot.cosAlpha,rot.sinAlpha,lin.m));
}

INLINE_GLOBAL FLin2D operator*(const FShearX2D & shear, const FLin2D & lin)
{
    return FLin2D(prShear0Mat(shear.c, lin.m));
}

INLINE_GLOBAL FLin2D operator*(const FShearY2D & shear, const FLin2D & lin)
{
    return FLin2D(prShear1Mat(shear.c, lin.m));
}

// ============================================================= class FTrans2D

INLINE_MEMBER FTrans2D::FTrans2D(
	)
:
    c(0.0,0.0)
{
    TR_FUN0(FVec2D_constructors,FTrans2D);
}

INLINE_MEMBER FTrans2D::FTrans2D(const FCrd2 & crd)
:
    c(crd)
{
    TR_FUN1(FVec2D_constructors,FTrans2D,crd);
}

INLINE_MEMBER FTrans2D::FTrans2D(float xNew, float yNew)
:
    c(xNew, yNew)
{
    TR_FUN2(FVec2D_constructors,FTrans2D,xNew,yNew);
}

INLINE_MEMBER FTrans2D::FTrans2D(
	const FVec2D & vec
	)
:
    c(vec.c)
{
    TR_FUN1(FVec2D_constructors,FTrans2D,vec);
}

INLINE_MEMBER FTrans2D::FTrans2D(
	const FPnt2D & pntO
	)
:
    c(pntO.c)
{
    TR_FUN1(FVec2D_constructors,FTrans2D, pntO);
}

INLINE_MEMBER FTrans2D::FTrans2D(const FTrans2D & trans)
:
    c(trans.c)
{
    TR_FUN1(FVec2D_constructors,FTrans2D,"FTrans2D");
}

INLINE_MEMBER FTrans2D::~FTrans2D()
{
    TR_MESSAGE(FVec2D_constructors, "~" << *this);
}

INLINE_MEMBER FTrans2D & FTrans2D::operator*=(const FTrans2D & trans)
{
    c += trans.c;
    return *this;
}

INLINE_MEMBER void FTrans2D::invert(
        )
{
    c = -c;
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const FTrans2D & trans)
{
    s << "FTrans2D(" << trans.c.x0() << "," << trans.c.x1() << ")";
    return s;
}

INLINE_GLOBAL FTrans2D operator*(const FTrans2D & trans1, const FTrans2D & trans2)
{
    return FTrans2D(trans1.c + trans2.c);
}

INLINE_GLOBAL FTrans2D inverseCopy(const FTrans2D & trans)
{
    return FTrans2D(-trans.c);
}

// ============================================================= class FTrafo2D

INLINE_MEMBER FTrafo2D::FTrafo2D()
:
    m(1.0,0.0, 0.0,1.0), c(0.0,0.0)
{
    TR_FUN0(FVec2D_constructors,FTrafo2D);
}

INLINE_MEMBER FTrafo2D::FTrafo2D(const FScale2D & scale)
:
    m(  scale.c.x0(),   0.0,
        0.0,            scale.c.x1()	),
    c(  0.0,            0.0		)
{
    TR_FUN1(FVec2D_constructors,FTrafo2D,scale);
}

INLINE_MEMBER FTrafo2D::FTrafo2D(const FRot2D & rot)
:
    m(  rot.cosAlpha,   rot.sinAlpha,
        -rot.sinAlpha,  rot.cosAlpha	),
    c(  0.0,            0.0		)
{
    TR_FUN1(FVec2D_constructors,FTrafo2D,rot);
}

INLINE_MEMBER FTrafo2D::FTrafo2D(const FShearX2D & shear)
:
    m(  1.0,	 0.0,
        shear.c, 1.0	),
    c(  0.0,	 0.0	)
{
    TR_FUN1(FVec2D_constructors,FTrafo2D,shear);
}

INLINE_MEMBER FTrafo2D::FTrafo2D(const FShearY2D & shear)
:
    m(  1.0,	 shear.c,
        0.0,	 1.0	),
    c(  0.0,	 0.0	)
{
    TR_FUN1(FVec2D_constructors,FTrafo2D,shear);
}

INLINE_MEMBER FTrafo2D::FTrafo2D(const FLin2D & lin)
:
    m(lin.m), c(0.0,0.0)
{
    TR_FUN1(FVec2D_constructors,FTrafo2D,lin);
}

INLINE_MEMBER FTrafo2D::FTrafo2D(const FLin2D & lin, const FTrans2D & trans)
:
    m(lin.m), c(trans.c)
{
    TR_FUN2(FVec2D_constructors,FTrafo2D,lin,trans);
}

INLINE_MEMBER FTrafo2D::FTrafo2D(const FMat2 & mat)
:
    m(mat), c(0.0,0.0)
{
    TR_FUN1(FVec2D_constructors,FTrafo2D,mat);
}

INLINE_MEMBER FTrafo2D::FTrafo2D(const FMat2 & mat, const FCrd2 & crd)
:
    m(mat), c(crd)
{
    TR_FUN2(FVec2D_constructors,FTrafo2D,mat,crd);
}

INLINE_MEMBER FTrafo2D::FTrafo2D(const FTrans2D & trans)
:
    m(1.0,0.0, 0.0,1.0), c(trans.c)
{
    TR_FUN1(FVec2D_constructors,FTrafo2D,trans);
}

INLINE_MEMBER FTrafo2D::FTrafo2D(const FTrafo2D & trafo)
:
    m(trafo.m), c(trafo.c)
{
    TR_FUN1(FVec2D_constructors,FTrafo2D,trafo);
}

INLINE_MEMBER FTrafo2D::FTrafo2D(const FVec2D & vx, const FVec2D & vy,
                    const FPnt2D & pnt)
:
    m(vx.c.x0(),vx.c.x1(),
      vy.c.x0(),vy.c.x1()),
    c(pnt.c)
{
    TR_FUN3(FVec2D_constructors,FTrafo2D,vx,vy,pnt);
}

INLINE_MEMBER FTrafo2D::FTrafo2D(
        float xx, float xy,
        float yx, float yy,
        float dx, float dy
        )
:
    m(xx,xy, yx,yy), c(dx,dy)
{
    TR_FUN3(FVec2D_constructors,FTrafo2D,xx,xy,"...");
}

INLINE_MEMBER FTrafo2D::~FTrafo2D()
{
    TR_MESSAGE(FVec2D_constructors,"~" << *this);
}

INLINE_MEMBER FVec2D FTrafo2D::getVecX() const
{
    return FVec2D(m(0,0),m(0,1));
}

INLINE_MEMBER FVec2D FTrafo2D::getVecY() const
{
    return FVec2D(m(1,0),m(1,1));
}

INLINE_MEMBER FPnt2D FTrafo2D::getPntO() const
{
    return FPnt2D(c);
}

INLINE_MEMBER FMat2 FTrafo2D::getFMat2() const
{
    return m;
}

INLINE_MEMBER FMat3 FTrafo2D::getFMat3() const
{
    return FMat3(m(0,0), m(0,1), 0.0,
		m(1,0), m(1,1), 0.0,
		c.x0(), c.x1(), 1.0);
}

INLINE_MEMBER FTrafo2D & FTrafo2D::operator*=(const FScale2D & scale)
{
    m.scale(scale.c);
    c.scale(scale.c);
    return *this;
}

INLINE_MEMBER FTrafo2D & FTrafo2D::operator*=(const FRot2D & rot)
{
    m.rot(rot.cosAlpha,rot.sinAlpha);
    c.rot(rot.cosAlpha,rot.sinAlpha);
    return *this;
}

INLINE_MEMBER FTrafo2D & FTrafo2D::operator*=(const FShearX2D & shear)
{
    m.shear0(shear.c);
    c.shear0(shear.c);
    return *this;
}

INLINE_MEMBER FTrafo2D & FTrafo2D::operator*=(const FShearY2D & shear)
{
    m.shear1(shear.c);
    c.shear1(shear.c);
    return *this;
}

INLINE_MEMBER FTrafo2D & FTrafo2D::operator*=(const FLin2D & lin)
{
    m *= lin.m;
    c *= lin.m;
    return *this;
}

INLINE_MEMBER FTrafo2D & FTrafo2D::operator*=(const FTrans2D & trans)
{
    c += trans.c;
    return *this;
}

INLINE_MEMBER FTrafo2D & FTrafo2D::operator*=(const FTrafo2D & trafo)
{
    m *= trafo.m;
    c *= trafo.m;
    c += trafo.c;
    return *this;
}

INLINE_GLOBAL FVec2D getVecX(const FTrafo2D & trafo) { return trafo.getVecX(); }
INLINE_GLOBAL FVec2D getVecY(const FTrafo2D & trafo) { return trafo.getVecY(); }
INLINE_GLOBAL FPnt2D getPntO(const FTrafo2D & trafo) { return trafo.getPntO(); }
INLINE_GLOBAL FMat2 getFMat2(const FTrafo2D & trafo) { return trafo.m; }
INLINE_GLOBAL FMat3 getFMat3(const FTrafo2D & trafo) { return trafo.getFMat3(); }

INLINE_GLOBAL float determinant(const FTrafo2D & trafo)
{
    return determinant(trafo);
}

INLINE_GLOBAL FTrafo2D inverseCopy(const FTrafo2D & trafo, float det)
{
    FTrafo2D inv;

    inv.m = inverseCopy(trafo.m,det);

    inv.c = - (trafo.c * inv.m);
                 
    return inv;
}

INLINE_MEMBER void FTrafo2D::invert(
        )
{
    float det = determinant(*this);
    *this = inverseCopy(*this,det);
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const FTrafo2D & trafo)
{
    s << "FTrafo2D(" << trafo.m << ',' << trafo.c << ')';
    return s;
}

INLINE_GLOBAL FTrafo2D inverseCopy(const FTrafo2D & trafo)
{
    float det = determinant(trafo);
    return inverseCopy(trafo,det);
}

INLINE_GLOBAL FTrafo2D operator*(const FTrafo2D & trafo, const FScale2D & scale)
{
    return FTrafo2D(prMatScale(trafo.m,scale.c),
                   prCrdScale(trafo.c,scale.c));
}

INLINE_GLOBAL FTrafo2D operator*(const FTrafo2D & trafo, const FRot2D & rot)
{
    return FTrafo2D(prMatRot(trafo.m,rot.cosAlpha,rot.sinAlpha),
                   prCrdRot(trafo.c,rot.cosAlpha,rot.sinAlpha));
}

INLINE_GLOBAL FTrafo2D operator*(const FTrafo2D & trafo, const FShearX2D & shear)
{
    return FTrafo2D(prMatShear0(trafo.m,shear.c),
                   prCrdShear0(trafo.c,shear.c));
}

INLINE_GLOBAL FTrafo2D operator*(const FTrafo2D & trafo, const FShearY2D & shear)
{
    return FTrafo2D(prMatShear1(trafo.m,shear.c),
                   prCrdShear1(trafo.c,shear.c));
}

INLINE_GLOBAL FTrafo2D operator*(const FTrafo2D & trafo, const FLin2D & lin)
{
    return FTrafo2D(trafo.m * lin.m, trafo.c * lin.m);
}

INLINE_GLOBAL FTrafo2D operator*(const FTrafo2D & trafo, const FTrans2D & trans)
{
    return FTrafo2D(trafo.m, trafo.c + trans.c);
}

INLINE_GLOBAL FTrafo2D operator*(const FTrafo2D & trafo1, const FTrafo2D & trafo2)
{
    return FTrafo2D(trafo1.m * trafo2.m, trafo1.c * trafo2.m + trafo2.c);
}

INLINE_GLOBAL FTrafo2D operator*(const FScale2D & scale, const FTrafo2D & trafo)
{
    return FTrafo2D(prScaleMat(scale.c,trafo.m),trafo.c);
}

INLINE_GLOBAL FTrafo2D operator*(const FRot2D & rot, const FTrafo2D & trafo)
{
    return FTrafo2D(prRotMat(rot.cosAlpha,rot.sinAlpha,trafo.m),trafo.c);
}

INLINE_GLOBAL FTrafo2D operator*(const FShearX2D & shear, const FTrafo2D & trafo)
{
    return FTrafo2D(prShear0Mat(shear.c,trafo.m),trafo.c);
}

INLINE_GLOBAL FTrafo2D operator*(const FShearY2D & shear, const FTrafo2D & trafo)
{
    return FTrafo2D(prShear1Mat(shear.c,trafo.m),trafo.c);
}

INLINE_GLOBAL FTrafo2D operator*(const FLin2D & lin, const FTrafo2D & trafo)
{
    return FTrafo2D(lin.m * trafo.m,trafo.c);
}

INLINE_GLOBAL FTrafo2D operator*(const FTrans2D & trans, const FTrafo2D & trafo)
{
    return FTrafo2D(trafo.m,trans.c * trafo.m + trafo.c);
}

} // extern "C++"

// ===========================================================================

