// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       Vec2D
//  TYPE:       c++ header
//  PROJECT:    VEGA Graphics Library
//  CONTENT:    2D vectors, points, and transformations using doubles
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rft     Robert F. Tobler
//              hl      Helwig L"offelman
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 14:08:16  rft     last modification
//  08-Aug-93           rft     created
// ===========================================================================
//  MANUAL:
//
//  DESCRIPTION
//      The classes Vec2D and Pnt2D are the basic types for 2D geometry.
//      They can be transformed with the corresponding transformation types
//      that are called Trans2D, Scale2D, Rot2D, ShearX2D, ShearY2D,
//      Lin2D, and Trafo2D.
//
//      The classes Vec2D and Pnt2D have been designed in a typesafe way,
//      so that all operations among them are geometrically meaningful.
//      Thus there are some combinations of operations that are not allowed
//      since there is no meaningful geometric result. An example for such
//      an operation is the addition of two points.
//
//      In terms of matrix theory, Vec2D and Pnt2D can be viewed as
//      homogenous 2D vectors and points (the trailing zeros and ones are
//      not stored):
//
//              Vec2D vec;
//              Pnt2D pnt;
//
//              (vec.x(), vec.y(), 0)
//              (pnt.x(), pnt.y(), 1)
//
//      The transformation type Trafo2D can be viewed as a 2D homogeneous
//      matrix. The multiplication of Vec2D and Pnt2D with Trafo2D can
//      then be represented as follows (the elements of the transformation
//      cannot be accessed in the way shown here):
//
//              Trafo2D t;
//              Vec2D v;
//              Pnt2D m;
//
//              Vec2D vec = v * t;
//              Vec2D pnt = p * t;
//
//                                (t.m[0,0], t.m[0,1], 0)
//                                (t.m[1,0], t.m[1,1], 0)
//                                (  t.c[0],   t.c[1], 1)
//
//              (v.x(),v.y(),0)   ( vec.x(),  vec.y(), 0)
//
//              (p.x(),p.y(),1)   ( pnt.x(),  pnt.y(), 1)
//
//      The transformation types Trans2D, Scale2D, ShearX2D, ShearY2D,
//      Lin2D, and Rot2D represent subsets of the transformations
//      represented by the Trafo2D type.
//
//      Rotations are defined to be counter-clockwise in the x-y-plane.
//      Since a right-handed coordinate system is assumed, this is consistent
//      with 3D-rotations that are defined to be clockwise if viewed in
//      direction of the axis.
//
//      Other ways to think about a Trafo2D:
//
//              Trafo2D <=> (Vec2D,     // vecX
//                           Vec2D,     // vecY
//                           Pnt2D)     // pntO (O for Origin)
//
//              (This can be used to directly specify a transformation by
//               just specifying the origin and the base vectors of the
//               coordinate system !)
//
//              Trafo2D <=> (Lin2D,     // lin
//                           Trans2D)   // trans
//
//      If you are using these models, there are constructors that take these
//      elements as parameters, and methods and functions to get these
//      elements (getVecX, getVecY, getPntO, getLin, and getTrans).
//
//  CHECKS
//      Range checking and attempts to normalize null-vectors will only be
//      caught if WITH_CRD2_CHECKS is defined.
//
//  ERRORHANDLERS
//      Vec2D_errors
//
//  TRACERS
//      Vec2D_constructors
//
//  SEE ALSO
//      VEGA, GraphicsLibrary
// ===========================================================================
#ifndef _GRAPHICS_VEC2D_H_
#define _GRAPHICS_VEC2D_H_

#include <iostream.h>

#include <VEGA/Math/math.h>
#include <VEGA/Math/crd2.h>
#include <VEGA/Math/crd3.h>

#include "VEGA/Graphics/globals.h"

#ifdef __GNUC__
#pragma interface
#endif

extern "C++"
{

// ============================================ classes defined in this module

class Vec2D;			// vector
class Pnt2D;			// point
class Scale2D;			// scaling by (sx,sy)
class Rot2D;			// rotation around origin
class ShearX2D;			// shear along x-Axis
class ShearY2D;			// shear along y-Axis
class Lin2D;			// linear transformation
class Trans2D;			// translation by vector
class Trafo2D;			// affine transformation

// =========================================== errors generated by this module

#define ERROR_VEC2D_SINGULAR_SCALE      1
        //
        //  An attempt was made to invert a singular scale.
        //


// =============================================================== class Vec2D

class Vec2D
{
    friend class Pnt2D;
    friend class Lin2D;
    friend class Trans2D;
    friend class Trafo2D;

    friend class Viewer3D;		//  needs FAST access to members

public:

    // --------------------------------------------- constructors / destructor

    Vec2D(
        );
    Vec2D(
        double xNew,
        double yNew
        );
    Vec2D(
        const Crd2 & crd
        );
    Vec2D(
        const Vec2D & vec
        );
    ~Vec2D(
        );

    // ------------------------------------------------------ member operators

    double operator[](
        int index
        ) const;
    double & operator[](
        int index
        );

    Vec2D & operator+=(
        const Vec2D & vec
        );
    Vec2D & operator-=(
        const Vec2D & vec
        );

    Vec2D & operator*=(
        double scalar
        );

    // ------------------------------------------------------ member functions

    double x(
        ) const;
    double y(
        ) const;

    void x(
        double xNew
        );
    void y(
        double yNew
        );

    void normalize(
        );

    Crd2 getCrd2(
        ) const;
    Crd3 getCrd3(			// Get homogeneous representation.
        ) const;
    Pnt2D getPnt2D(			// Explicit conversion to Pnt2D.
        ) const;

    // ------------------------------------------------------ friend operators

    friend ostream & operator<<(
        ostream & s,
	const Vec2D & vec
        );
    friend Vec2D operator-(
        const Vec2D & vec
        );
    friend Vec2D operator+(
        const Vec2D & vec1,
	const Vec2D & vec2
        );
    friend Pnt2D operator+(
        const Pnt2D & pnt,
	const Vec2D & vec
        );
    friend Vec2D operator-(
        const Vec2D & vec1,
	const Vec2D & vec2
        );
    friend Pnt2D operator-(
        const Pnt2D & pnt,
	const Vec2D & vec
        );
    friend Vec2D operator*(
        const Vec2D & vec,
	double scalar
        );
    friend Vec2D operator*(
        double scalar,
	const Vec2D & vec
        );

    // ------------------------------------------------------ friend functions

    friend Crd2 getCrd2(
        const Vec2D & vec
        );
    friend Crd3 getCrd3(		// Get homogeneous representation.
        const Vec2D & vec
        );
    friend Pnt2D getPnt2D(		// Explicit conversion.
        const Vec2D & vec
        );
    friend double squaredLength(
        const Vec2D & vec
        );
    friend double length(
        const Vec2D & vec
        );
    friend double dotProduct(
        const Vec2D & vec1,
	const Vec2D & vec2
        );
    friend Vec2D normalizedCopy(
        const Vec2D & vec
        );

    // -------------------------------------- special purpose friend functions

    friend Vec2D cwReciprocalCopy(      // Componentwise multiplicative
        const Vec2D & vec		//  inverse. Zero components are not
        );				//  inverted.

    friend Vec2D cwMin(                 // Componentwise minimum.
        const Vec2D & vec1,
        const Vec2D & vec2
        );

    friend Vec2D cwMax(                 // Componentwise maximum.
        const Vec2D & vec1,
        const Vec2D & vec2
        );

    // ---------------------------------------------- transformation functions

    //
    //  Multiply vector by special transformation.
    //
    Vec2D & operator*=(
        const Scale2D & scale
        );
    Vec2D & operator*=(
        const Rot2D & rot
        );
    Vec2D & operator*=(
        const ShearX2D & shear
        );
    Vec2D & operator*=(
        const ShearY2D & shear
        );
    Vec2D & operator*=(
        const Lin2D & lin
        );
    Vec2D & operator*=(
        const Trafo2D & trafo
        );

    //
    //  Product of vector and special tranformation.
    //
    friend Vec2D operator*(
        const Vec2D & vec,
	const Scale2D & scale
        );
    friend Vec2D operator*(
        const Vec2D & vec,
	const Rot2D & rot
        );
    friend Vec2D operator*(
        const Vec2D & vec,
	const ShearX2D & shear
        );
    friend Vec2D operator*(
        const Vec2D & vec,
	const ShearY2D & shear
        );
    friend Vec2D operator*(
        const Vec2D & vec,
	const Lin2D & lin
        );
    friend Vec2D operator*(
        const Vec2D & vec,
	const Trafo2D & trafo
        );

    // ----------------------------------------------- static member variables

    static const Vec2D baseX;
    static const Vec2D baseY;

protected:

    Crd2 c;
};

// ===================================================================== Pnt2D

class Pnt2D
{
    friend class Vec2D;
    friend class Lin2D;
    friend class Trans2D;
    friend class Trafo2D;

    friend class Viewer3D;		//  needs FAST access to members

public:

    // ---------------------------------------------------------- constructors

    Pnt2D(
        );
    Pnt2D(
        double x,
        double y
        );
    Pnt2D(
        const Crd2 & crd
        );
    Pnt2D(
        const Pnt2D & pnt
        );
    ~Pnt2D(
        );

    // ------------------------------------------------------ member operators

    double operator[](
        int index
        ) const;
    double & operator[](
        int index
        );

    Pnt2D & operator+=(
        const Vec2D & vec
        );
    Pnt2D & operator-=(
        const Vec2D & vec
        );

    // ------------------------------------------------------ member functions

    double x(
        ) const;
    double y(
        ) const;

    void x(
        double xNew
        );
    void y(
        double yNew
        );

    Crd2 getCrd2(
        ) const;
    Crd3 getCrd3(
        ) const;               // get homogeneous representation
    Vec2D getVec2D(
        ) const;             // explicit conversion to Vec2D

    // ------------------------------------------------------ friend operators

    friend ostream & operator<<(
        ostream & s,
	const Pnt2D & vec
        );
    friend Pnt2D operator+(
        const Pnt2D & pnt,
	const Vec2D & vec
        );
    friend Pnt2D operator-(
        const Pnt2D & pnt,
	const Vec2D & vec
        );
    friend Vec2D operator-(
        const Pnt2D & pnt1,
	const Pnt2D & pnt2
        );

    // ------------------------------------------------------ friend functions

    friend Crd2 getCrd2(
        const Pnt2D & pnt
        );
    friend Crd3 getCrd3(
        const Pnt2D & pnt
        );	// get homogeneous rep.
    friend Vec2D getVec2D(
        const Pnt2D & pnt
        );	// explicit conversion
    friend double squaredDistanceFromOrigin(
        const Pnt2D & pnt
        );
    friend double distanceFromOrigin(
        const Pnt2D & pnt
        );
    friend double squaredDistance(
        const Pnt2D & pnt1,
	const Pnt2D & pnt2
        );
    friend double distance(
        const Pnt2D & pnt1,
	const Pnt2D & pnt2
        );

    // ----------------------------------------------- componentwise functions

    friend Pnt2D cwMin(
        const Pnt2D & vec1,
	const Pnt2D & vec2
        );
    friend Pnt2D cwMax(
        const Pnt2D & vec1,
	const Pnt2D & vec2
        );

    // ---------------------------------------------- transformation functions

    Pnt2D & operator*=(
        const Scale2D & scale
        );
    Pnt2D & operator*=(
        const Rot2D & rot
        );
    Pnt2D & operator*=(
        const ShearX2D & shear
        );
    Pnt2D & operator*=(
        const ShearY2D & shear
        );
    Pnt2D & operator*=(
        const Lin2D & lin
        );
    Pnt2D & operator*=(
        const Trans2D & trans
        );
    Pnt2D & operator*=(
        const Trafo2D & trafo
        );

    friend Pnt2D operator*(
        const Pnt2D & pnt,
	const Scale2D & scale
        );
    friend Pnt2D operator*(
        const Pnt2D & pnt,
	const Rot2D & rot
        );
    friend Pnt2D operator*(
        const Pnt2D & pnt,
	const ShearX2D & shear
        );
    friend Pnt2D operator*(
        const Pnt2D & pnt,
	const ShearY2D & shear
        );
    friend Pnt2D operator*(
        const Pnt2D & pnt,
	const Lin2D & lin
        );
    friend Pnt2D operator*(
        const Pnt2D & pnt,
	const Trans2D & trans
        );
    friend Pnt2D operator*(
        const Pnt2D & pnt,
	const Trafo2D & trafo
        );

    // ----------------------------------------------- static member variables

    static const Pnt2D origin;

protected:
    Crd2 c;
};

// ============================================================= class Scale2D

class Scale2D
{
    friend class Vec2D;
    friend class Pnt2D;
    friend class Lin2D;
    friend class Trafo2D;

public:

    Scale2D(				// initialize with identity
    	);
    Scale2D(
        double scalar
        );
    Scale2D(
        double x,
        double y
        );
    Scale2D(
        const Crd2 & scale
        );
    Scale2D(
        const Scale2D & scale
        );
    ~Scale2D(
        );

    Scale2D & operator*=(
        const Scale2D & scale
        );

    void invert(
        );

    friend ostream & operator<<(
        ostream & s,
	const Scale2D & scale
        );
    friend Scale2D operator*(
        const Scale2D & sc1,
	const Scale2D & sc2
        );

    friend Scale2D inverseCopy(
        const Scale2D & scale
        );

    static const Scale2D identity;

    // -------------------------------------------- functions that need access

    friend Vec2D operator*(
        const Vec2D & vec,
	const Scale2D & scale
        );
    friend Pnt2D operator*(
        const Pnt2D & pnt,
	const Scale2D & scale
        );
    friend Lin2D operator*(
        const Scale2D & scale,
	const Lin2D & lin
        );
    friend Lin2D operator*(
        const Lin2D & lin,
	const Scale2D & scale
        );
    friend Trafo2D operator*(
        const Scale2D & scale,
	const Trafo2D & trafo
        );
    friend Trafo2D operator*(
        const Trafo2D & trafo,
	const Scale2D & scale
        );

protected:

    Crd2 c;
};

// =============================================================== class Rot2D

class Rot2D
{
    friend class Vec2D;
    friend class Pnt2D;
    friend class Lin2D;
    friend class Trafo2D;

public:

    Rot2D(				// initialize with identity
    	);
    Rot2D(
        double newAlpha
        );
    Rot2D(
        double newAlpha,
        double newCosAlpha,
        double newSinAlpha
        );
    Rot2D(
        const Rot2D & rot
        );
    ~Rot2D(
        );

    Rot2D & operator*=(
        const Rot2D & rot
        );

    double getAlpha(
        ) const;
    void invert(
        );
    
    friend ostream & operator<<(
        ostream & s,
	const Rot2D & rot
        );
    friend Rot2D operator*(
        const Rot2D & rot1,
	const Rot2D & rot2
        );

    friend Rot2D inverseCopy(
        const Rot2D & rot
        );

    static const Rot2D identity;

    // -------------------------------------------- functions that need access

    friend Vec2D operator*(
        const Vec2D & vec,
	const Rot2D & rot
        );
    friend Pnt2D operator*(
        const Pnt2D & pnt,
	const Rot2D & rot
        );
    friend Lin2D operator*(
        const Lin2D & lin,
	const Rot2D & rot
        );
    friend Lin2D operator*(
        const Rot2D & rot,
	const Lin2D & lin
        );
    friend Trafo2D operator*(
        const Trafo2D & trafo,
	const Rot2D & rot
        );
    friend Trafo2D operator*(
        const Rot2D & rot,
	const Trafo2D & trafo
        );

protected:
    double alpha;
    double cosAlpha;
    double sinAlpha;
};

// ============================================================ class ShearX2D

class ShearX2D
{
    friend class Vec2D;
    friend class Pnt2D;
    friend class Lin2D;
    friend class Trafo2D;

public:

    ShearX2D(				// initialize with identity
    	);
    ShearX2D(
        double cNew
        );
    ShearX2D(
        const ShearX2D & shear
        );
    ~ShearX2D(
	);

    ShearX2D & operator*=(
        const ShearX2D & shear
        );

    void invert(
        );
    
    friend ostream & operator<<(
        ostream & s,
	const ShearX2D & shear
        );
    friend ShearX2D operator*(
        const ShearX2D & s1,
	const ShearX2D & s2
        );

    friend ShearX2D inverseCopy(
        const ShearX2D & shear
        );

    static const ShearX2D identity;

    // -------------------------------------------- functions that need access

    friend Vec2D operator*(
        const Vec2D & vec,
	const ShearX2D & shear
        );
    friend Pnt2D operator*(
        const Pnt2D & pnt,
	const ShearX2D & shear
        );
    friend Lin2D operator*(
        const Lin2D & lin,
	const ShearX2D & shear
        );
    friend Lin2D operator*(
        const ShearX2D & shear,
	const Lin2D & lin
        );
    friend Trafo2D operator*(
        const Trafo2D & trafo,
	const ShearX2D & sh
        );
    friend Trafo2D operator*(
        const ShearX2D & sh,
	const Trafo2D & trafo
        );

protected:
    double c;
};

// ============================================================ class ShearY2D

class ShearY2D
{
    friend class Vec2D;
    friend class Pnt2D;
    friend class Lin2D;
    friend class Trafo2D;

public:

    ShearY2D(				// initialize with identity
    	);
    ShearY2D(
        double cNew
        );
    ShearY2D(
        const ShearY2D & shear
        );
    ~ShearY2D(
	);

    ShearY2D & operator*=(
        const ShearY2D & shear
        );

    void invert(
        );
    
    friend ostream & operator<<(
        ostream & s,
	const ShearY2D & shear
        );
    friend ShearY2D operator*(
        const ShearY2D & s1,
	const ShearY2D & s2
        );

    friend ShearY2D inverseCopy(
        const ShearY2D & shear
        );

    static const ShearY2D identity;

    // -------------------------------------------- functions that need access

    friend Vec2D operator*(
        const Vec2D & vec,
	const ShearY2D & shear
        );
    friend Pnt2D operator*(
        const Pnt2D & pnt,
	const ShearY2D & shear
        );
    friend Lin2D operator*(
        const Lin2D & lin,
	const ShearY2D & shear
        );
    friend Lin2D operator*(
        const ShearY2D & shear,
	const Lin2D & lin
        );
    friend Trafo2D operator*(
        const Trafo2D & trafo,
	const ShearY2D & sh
        );
    friend Trafo2D operator*(
        const ShearY2D & sh,
	const Trafo2D & trafo
        );

protected:
    double c;
};

// ============================================================= class Lin2D

class Lin2D
{
    friend class Vec2D;
    friend class Pnt2D;
    friend class Trafo2D;

public:

    Lin2D(				// initialize with identity
        );
    Lin2D(
        const Mat2 & m
        );
    Lin2D(
        const Vec2D & vecX,
        const Vec2D & vecY
        );
    Lin2D(
        double xx, double xy,
        double yx, double yy
        );
    Lin2D(
        const Scale2D & scale
        );
    Lin2D(
        const Rot2D & rot
        );
    Lin2D(
        const ShearX2D & shear
        );
    Lin2D(
        const ShearY2D & shear
        );
    Lin2D(
        const Lin2D & lin
        );
    ~Lin2D(
        );

    // ------------------------------------------------------ member functions

    Vec2D getVecX(
        ) const;
    Vec2D getVecY(
        ) const;

    Mat2 getMat2(
        ) const;

    void invert(
        );
        //
        //  ERROR_CRD2_SINGULAR_MATRIX see module Crd2
        //

    // ------------------------------------------------------ friend operators

    friend ostream & operator<<(
        ostream & s,
	const Lin2D & lin
        );

    // ------------------------------------------------------ friend functions

    friend Vec2D getVecX(
        const Lin2D & lin
        );
    friend Vec2D getVecY(
        const Lin2D & lin
        );

    friend Mat2 getMat2(
        const Lin2D & lin
        );
    friend double determinant(
        const Lin2D & lin
        );
    friend Lin2D inverseCopy(
        const Lin2D & lin,
	double det
        );
        //
        //  det is NOT checked against zero!
        //

    friend Lin2D inverseCopy(
        const Lin2D & lin
        );
        //
        //  ERROR_CRD2_SINGULAR_MATRIX see module Crd2
        //

    // ----------------------------------------------- tranformation functions

    //
    //  Multiply transformation by special transformation.
    //
    Lin2D & operator*=(
        const Scale2D & scale
        );
    Lin2D & operator*=(
        const Rot2D & rot
        );
    Lin2D & operator*=(
        const ShearX2D & shear
        );
    Lin2D & operator*=(
        const ShearY2D & shear
        );
    Lin2D & operator*=(
        const Lin2D & lin
        );

    //
    //  Product of transformation and special transformation.
    //
    friend Lin2D operator*(
        const Lin2D & lin,
	const Scale2D & scale
        );
    friend Lin2D operator*(
        const Lin2D & lin,
	const Rot2D & rot
        );
    friend Lin2D operator*(
        const Lin2D & lin,
	const ShearX2D & shear
        );
    friend Lin2D operator*(
        const Lin2D & lin,
	const ShearY2D & shear
        );

    friend Lin2D operator*(		// Product of two transformations.
        const Lin2D & lin1,
        const Lin2D & lin2
        );

    //
    //  Product of special transformation and transformation.
    //
    friend Lin2D operator*(
        const Scale2D & scale,
	const Lin2D & lin
        );
    friend Lin2D operator*(
        const Rot2D & rot,
	const Lin2D & lin
        );
    friend Lin2D operator*(
        const ShearX2D & shear,
	const Lin2D & lin
        );
    friend Lin2D operator*(
        const ShearY2D & shear,
	const Lin2D & lin
        );

    //
    //  Product of vector/point and transformation.
    //
    friend Vec2D operator*(
        const Vec2D & vec,
	const Lin2D & lin
        );
    friend Pnt2D operator*(
        const Pnt2D & pnt,
	const Lin2D & lin
        );

    // ----------------------------------------------- static member variables

    static const Lin2D identity;

    // -------------------------------------------- functions that need access

    friend Trafo2D operator*(
        const Lin2D & lin,
	const Trafo2D & trafo
        );
    friend Trafo2D operator*(
        const Trafo2D & trafo,
	const Lin2D & lin
        );

protected:

    Mat2 m;
};

// ============================================================= class Trans2D

class Trans2D
{
    friend class Vec2D;
    friend class Pnt2D;
    friend class Trafo2D;

public:

    Trans2D(				// initialize with identity
	);
    Trans2D(
        const Crd2 & crd
        );
    Trans2D(
        double xNew,
        double yNew
        );
    Trans2D(
	const Vec2D & vec
	);
    Trans2D(
	const Pnt2D & pntO
	);
    Trans2D(
        const Trans2D & trans
        );
    ~Trans2D(
	);

    Trans2D & operator*=(
        const Trans2D & trans
        );

    void invert(
        );

    friend ostream & operator<<(
        ostream & s,
	const Trans2D & trans
        );
    friend Trans2D operator*(
        const Trans2D & tr1,
	const Trans2D & tr2
        );

    friend Trans2D inverseCopy(
        const Trans2D & trans
        );

    // -------------------------------------------- functions that need access

    friend Vec2D operator*(
        const Vec2D & vec,
	const Trans2D & trans
        );
    friend Pnt2D operator*(
        const Pnt2D & pnt,
	const Trans2D & trans
        );
    friend Trafo2D operator*(
        const Trans2D & trans,
	const Trafo2D & trafo
        );
    friend Trafo2D operator*(
        const Trafo2D & trafo,
	const Trans2D & trans
        );

protected:

    Crd2 c;
};

// ============================================================= class Trafo2D

class Trafo2D
{
    friend class Vec2D;
    friend class Pnt2D;

public:

    Trafo2D(				// initialize with identity
        );
    Trafo2D(
        const Trafo2D & trafo
        );
    Trafo2D(
	const Lin2D & lin,
	const Trans2D & trans
	);
    Trafo2D(
        const Mat2 & m
        );
    Trafo2D(
        const Mat2 & m,
        const Crd2 & c
        );
    Trafo2D(
        const Vec2D & x,
        const Vec2D & y,
        const Pnt2D & d
        );
    Trafo2D(
        double xx, double xy,
        double yx, double yy,
        double dx, double dy
        );
    Trafo2D(
        const Scale2D & scale
        );
    Trafo2D(
        const Rot2D & rot
        );
    Trafo2D(
        const ShearX2D & shear
        );
    Trafo2D(
        const ShearY2D & shear
        );
    Trafo2D(
	const Lin2D & lin
	);
    Trafo2D(
        const Trans2D & trans
        );
    ~Trafo2D(
        );

    // ------------------------------------------------------ member functions

    Vec2D getVecX(
        ) const;
    Vec2D getVecY(
        ) const;
    Pnt2D getPntO(
        ) const;

    Mat2 getMat2(
        ) const;
    Mat3 getMat3(
        ) const;               // get homogeneous representation

    void invert(
        );
        //
        //  ERROR_CRD2_SINGULAR_MATRIX see module Crd2
        //

    // ------------------------------------------------------ friend operators

    friend ostream & operator<<(
        ostream & s,
	const Trafo2D & trafo
        );

    // ------------------------------------------------------ friend functions

    friend Vec2D getVecX(
        const Trafo2D & trafo
        );
    friend Vec2D getVecY(
        const Trafo2D & trafo
        );
    friend Pnt2D getPntO(
        const Trafo2D & trafo
        );

    friend Mat2 getMat2(
        const Trafo2D & trafo
        );
    friend Mat3 getMat3(
        const Trafo2D & trafo
        );	// get homogeneous rep.
    friend double determinant(
        const Trafo2D & trafo
        );
    friend Trafo2D inverseCopy(
        const Trafo2D & trafo,
	double det
        );
        //
        //  det is NOT checked against zero!
        //

    friend Trafo2D inverseCopy(
        const Trafo2D & trafo
        );
        //
        //  ERROR_CRD2_SINGULAR_MATRIX see module Crd2
        //

    // ----------------------------------------------- tranformation functions

    //
    //  Multiply transformation by special transformation.
    //
    Trafo2D & operator*=(
        const Scale2D & scale
        );
    Trafo2D & operator*=(
        const Rot2D & rot
        );
    Trafo2D & operator*=(
        const ShearX2D & shear
        );
    Trafo2D & operator*=(
        const ShearY2D & shear
        );
    Trafo2D & operator*=(
        const Lin2D & lin
        );
    Trafo2D & operator*=(
        const Trans2D & trans
        );
    Trafo2D & operator*=(
        const Trafo2D & trafo
        );

    //
    //  Product of transformation and special transformation.
    //
    friend Trafo2D operator*(
        const Trafo2D & trafo,
	const Scale2D & scale
        );
    friend Trafo2D operator*(
        const Trafo2D & trafo,
	const Rot2D & rot
        );
    friend Trafo2D operator*(
        const Trafo2D & trafo,
	const ShearX2D & sh
        );
    friend Trafo2D operator*(
        const Trafo2D & trafo,
	const ShearY2D & sh
        );
    friend Trafo2D operator*(
        const Trafo2D & trafo,
	const Lin2D & lin
        );
    friend Trafo2D operator*(
        const Trafo2D & trafo,
	const Trans2D & trans
        );

    friend Trafo2D operator*(		// Product of two transformations.
        const Trafo2D & trafo1,
        const Trafo2D & trafo2
        );

    //
    //  Product of special transformation and transformation.
    //
    friend Trafo2D operator*(
        const Scale2D & scale,
	const Trafo2D & trafo
        );
    friend Trafo2D operator*(
        const Rot2D & rot,
	const Trafo2D & trafo
        );
    friend Trafo2D operator*(
        const ShearX2D & sh,
	const Trafo2D & trafo
        );
    friend Trafo2D operator*(
        const ShearY2D & sh,
	const Trafo2D & trafo
        );
    friend Trafo2D operator*(
        const Lin2D & lin,
	const Trafo2D & trafo
        );
    friend Trafo2D operator*(
        const Trans2D & trans,
	const Trafo2D & trafo
        );

    //
    //  Product of vector/point and transformation.
    //
    friend Vec2D operator*(
        const Vec2D & vec,
	const Trafo2D & trafo
        );
    friend Pnt2D operator*(
        const Pnt2D & pnt,
	const Trafo2D & trafo
        );

    // ----------------------------------------------- static member variables

    static const Trafo2D identity;

protected:

    Mat2 m;
    Crd2 c;
};

} // extern "C++"

#ifndef VEC2D_NO_INLINE

#define INLINE_GLOBAL COMPILER_INLINE_GLOBAL
#define INLINE_MEMBER COMPILER_INLINE_MEMBER

#include "VEGA/Graphics/vec2d.ipp"

#undef INLINE_GLOBAL
#undef INLINE_MEMBER

#endif


#endif // _GRAPHICS_VEC2D_H_
// ===========================================================================
