// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       Bool
//  TYPE:       c++ header
//  PROJECT:    VEGA Base Library
//  CONTENT:    Boolean type
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    hl      Helwig L"offelmann
//              rft     Robert F. Tobler
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 13:40:13  rft     last modification
//  13-mar-94 10:12:00  hl      enum instead of defines
//  12-mar-94 08:40:00  hl      created
// ===========================================================================
//  MANUAL:
//
//  DESCRIPTION
//      This module defines class Bool a boolean type that cannot be
//      implicitly converted to int. Since it has to be explicitly converted
//      to int using the functions 'int isTrue(Bool)' and 'int isFalse(Bool)'
//      it allows strict type safety when using boolean variables.
//
//	Due to a conflict with the proposed ANSI constants 'true' and 'false'
//	the constants had to be named: 'Bool::yes' and 'Bool::true'.
//
//      One of the main advantages of using a Bool as opposed to an int, is,
//      that it cannot be declared without an initializer, thus preventing
//      undefined status.
//
//  SEE ALSO
//      VEGA, BaseLibrary
// ===========================================================================
#ifndef _BASE_BOOL_H_
#define _BASE_BOOL_H_

#include <iostream.h>

#include "VEGA/Base/Globals.h"
#include "VEGA/Base/System.h"

#ifdef __GNUC__
#pragma interface
#endif

extern "C++"
{

// ================================================================ class Bool

class Bool
{
public:
    
    enum State
    {
        no = 0,
        yes  = 1
    };

    // ---------------------------------------------------------- constructors

    //
    //  The default constructor for Bool has been made private, in order to
    //	prevent creation of a Bool without explicitly setting its state.
    //

    Bool(
        State newState
	);
        //
        //  This constructor makes it possible to use 'Bool b = Bool::yes;'
        //  and 'Bool b = Bool::no;' as initializers.
        //

    // ------------------------------------------------------ member functions

    int isTrue(
	) const;
    int isFalse(
	) const;
    int isNotTrue(
	) const;
    int isNotFalse(
	) const;

    // ------------------------------------------------------ friend functions

    friend int isTrue(
	Bool b
	);
    friend int isFalse(
	Bool b
	);
    friend int isNotTrue(
	Bool b
	);
    friend int isNotFalse(
	Bool b
	);

    // ------------------------------------------------------ friend operators

    friend Bool operator!(
	Bool b
	);
    
    friend Bool operator||(
	Bool b1,
	Bool b2
	);
    friend Bool operator&&(
	Bool b1,
	Bool b2
	);

    friend Bool operator==(
	Bool b1,
	Bool b2
	);
    friend Bool operator!=(
	Bool b1,
	Bool b2
	);
    friend ostream & operator<<(
	ostream & s,
	Bool b
	);

private:
    Bool(
	);

    int state;                          // internally we use an int so that
                                        // we can use fast builtin operators
};

} // extern "C++"

#ifndef BOOL_NO_INLINE

#define INLINE_GLOBAL COMPILER_INLINE_GLOBAL
#define INLINE_MEMBER COMPILER_INLINE_MEMBER

#include "VEGA/Base/Bool.I"

#undef INLINE_GLOBAL
#undef INLINE_MEMBER

#endif


#endif // _BOOL_H_
// ===========================================================================

