// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       Option
//  TYPE:       c++ code
//  PROJECT:    VEGA Base Library
//  CONTENT:    commandline option parser
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rft     Robert F. Tobler
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 13:40:13  rft     last modification
//  23-mar-94 19:49:58  rft     created
// ===========================================================================

extern "C"
{
#include <string.h>
}

#include "VEGA/Base/Option.h"

Option * Option::optionList = 0;

Option::Option(
        const char * newLongName,
        const char * newShortName,
        Bool newSpecified
        )
:
    longName(newLongName),
    shortName(newShortName),
    specified(newSpecified),
    next(optionList)
{
    optionList = this;
}

int Option::isSpecified() const
{
    return isTrue(specified);
}

Option * Option::findByLongName(const char * name)
{
    for (Option * optionPtr = optionList;
         optionPtr;
         optionPtr = optionPtr->next)
    {
        if (strcmp(name,optionPtr->longName) == 0)
            return optionPtr;
    }
    return (0);
}

Option * Option::findByShortName(const char * name)
{
    for (Option * optionPtr = optionList;
         optionPtr;
         optionPtr = optionPtr->next)
    {
        if (strcmp(name,optionPtr->shortName) == 0)
            return optionPtr;
    }
    return (0);
}

void Option::setSpecified()
{
    specified = Bool::yes;
}

Bool Option::missingParameter()
{
    return Bool::no;
}

Bool Option::takesParameter(const char *)
{
    return Bool::no;
}

void Option::parse(
        int & argc,
        char **  argv
        )
{
    int i;
    Option * optionPtr = 0;
    int numberOfOptionsAndParameters = 0;

    TR_CODE
    (
        TR_MESSAGE(Option_tracer,"BEFORE PARSING:");
        TR_MESSAGE(Option_tracer,"  argc = " << argc);
        for (i = 1; i < argc; i++)
            TR_MESSAGE(Option_tracer,"  argv[" << i << "] = " << argv[i]);
    );

    for (i = 1; i < argc; i++)
    {
        if (numberOfOptionsAndParameters > 0)
            argv[i - numberOfOptionsAndParameters] = argv[i];

        if (argv[i][0] == '-')
        {
            if (optionPtr && isTrue(optionPtr->missingParameter()))
                return;

            if (argv[i][1] == '-')
            {
                optionPtr = findByLongName(argv[i] + 2);
            }
            else
            {
                optionPtr = findByShortName(argv[i] + 1);
            }
            if (! optionPtr)
            {
                EH_ERROR(Option_errors, ERROR_OPTION_UNKNOWN,
                      "unknown option '" << argv[i] << "'");
                return;
            }
            if (isTrue(optionPtr->specified))
            {
                EH_ERROR(Option_errors,ERROR_OPTION_SPECIFIED_TWICE,
                      "option '" << argv[i] << "' specified twice");
                return;
            }
            optionPtr->setSpecified();
            numberOfOptionsAndParameters++;
        }
        else
        {
            if (optionPtr &&
                isTrue(optionPtr->takesParameter(argv[i])))
                numberOfOptionsAndParameters++;
        }
    }
    if (optionPtr && isTrue(optionPtr->missingParameter()))
        return;

    argc -= numberOfOptionsAndParameters;

    TR_CODE
    (
        TR_MESSAGE(Option_tracer,"AFTER PARSING:");
        TR_MESSAGE(Option_tracer,"  argc = " << argc);
        for (i = 1; i < argc; i++)
            TR_MESSAGE(Option_tracer,"  argv[" << i << "] = " << argv[i]);
        TR_MESSAGE(Option_tracer,"OPTIONS:");
        for (optionPtr = optionList;
             optionPtr;
             optionPtr = optionPtr->next)
        {
            if (optionPtr->isSpecified())
                TR_MESSAGE(Option_tracer, "  " << optionPtr->longName);
        }
        
    );
}

OptionFlag::OptionFlag(
        const char * newLongName,
        const char * newShortName
        )
:
    Option(newLongName,newShortName,Bool::no)
{
    TR_FUN2(Option_tracer,OptionFlag,newLongName,newShortName);
}


OptionParameter::OptionParameter(
        const char * newLongName,
        const char * newShortName,
        Bool hasDefault
        )
:
    Option(newLongName,newShortName,Bool::no),
    parameter(hasDefault)
{}

void OptionParameter::setSpecified()
{
    Option::setSpecified();
    parameter = Bool::no;
}

Bool OptionParameter::takesParameter(const char * par)
{
    if (isFalse(parameter))
    {
        setParameter(par);      
        parameter = Bool::yes;
        return Bool::yes;
    }
    return Bool::no;
}

void OptionParameter::setParameter(const char *)
{
    EH_FATAL(Option_errors,"should be pure virtual");
}

Bool OptionParameter::missingParameter()
{
    if (isFalse(parameter))
    {
        EH_ERROR(Option_errors,ERROR_OPTION_PARAMETER_MISSING,
              "missing option parameter");
        return Bool::yes;
    }
    return Bool::no;
}

OptionString::OptionString(
        const char * newLongName,
        const char * newShortName
        )
:
    OptionParameter(newLongName,newShortName, Bool::no), value(0)
{
    TR_FUN2(Option_tracer,OptionString,newLongName,newShortName);
}

OptionString::OptionString(
        const char * newLongName,
        const char * newShortName,
        const char * defaultValue
        )
:
    OptionParameter(newLongName,newShortName, Bool::yes), value(defaultValue)
{
    TR_FUN3(Option_tracer,OptionString,newLongName,newShortName,
               defaultValue);
}

const char * OptionString::getValue() const { return value; }

void OptionString::setParameter(const char * par) { value = par; }

OptionLong::OptionLong(
        const char * newLongName,
        const char * newShortName
        )
:
    OptionParameter(newLongName,newShortName, Bool::no), value(0)
{
    TR_FUN2(Option_tracer,OptionLong,newLongName,newShortName);
}

OptionLong::OptionLong(
        const char * newLongName,
        const char * newShortName,
        int defaultValue
        )
:
    OptionParameter(newLongName,newShortName, Bool::yes), value(defaultValue)
{
    TR_FUN3(Option_tracer,OptionLong,newLongName,newShortName,
               defaultValue);
}

long OptionLong::getValue() const { return value; }

void OptionLong::setParameter(const char * par) { value = atol(par); }

OptionDouble::OptionDouble(
        const char * newLongName,
        const char * newShortName
        )
:
    OptionParameter(newLongName,newShortName, Bool::no), value(0)
{
    TR_FUN2(Option_tracer,OptionDouble,newLongName,newShortName);
}

OptionDouble::OptionDouble(
        const char * newLongName,
        const char * newShortName,
        double defaultValue
        )
:
    OptionParameter(newLongName,newShortName, Bool::yes), value(defaultValue)
{
    TR_FUN3(Option_tracer,OptionDouble,newLongName,newShortName,
               defaultValue);
}

double OptionDouble::getValue() const { return value; }

void OptionDouble::setParameter(const char * par) { value = atof(par); }

// ===========================================================================

