// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       Color
//  TYPE:       c++ header
//  PROJECT:    VEGA Graphics Library
//  CONTENT:    color classes
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rft     Robert F. Tobler
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 14:08:16  rft     last modification
//  05-apr-94 14:30:43  rft     created
// ===========================================================================
//  MANUAL:
//
//  DESCRIPTION
//      Color Classes
//
//  SEE ALSO
//      VEGA, GraphicsLibrary
// ===========================================================================
#ifndef _GRAPHICS_COLOR_H_
#define _GRAPHICS_COLOR_H_

#include <VEGA/Base/System.h>

#include <VEGA/Math/Math.h>
#include <VEGA/Math/Crd3.h>
#include <VEGA/Math/FCrd3.h>
#include <VEGA/Math/FCrd4.h>

#include "VEGA/Graphics/Globals.h"

#ifdef __GNUC__
#pragma interface
#endif

extern "C++"
{

class Color32;		// UInt8 (red,green,blue,alpha)
class RGB;		// float (red,green,blue)
class DRGB;		// double (red,green,blue)
class RGBA;		// float (red,green,blue,alpha)

// ============================================================= class Color32

class Color32
{
    friend class RGB;
    friend class DRGB;
    friend class RGBA;

    friend class Viewer3D;		//  needs FAST access to members

public:

    Color32(
	);
    Color32(
	UInt8 red,
	UInt8 green,
	UInt8 blue,
	UInt8 alpha
	);
    Color32(
	const RGB & rgb
	);
    Color32(
	const DRGB & drgb
	);
    Color32(
	const RGBA & rgba
	);

    UInt8 r() const;
    UInt8 b() const;
    UInt8 g() const;
    UInt8 a() const;

    void r(UInt8 rNew);
    void g(UInt8 gNew);
    void b(UInt8 bNew);
    void a(UInt8 aNew);

    UInt8 & operator[](int index);
    UInt8 operator[](int index) const;

    //
    //  The following two operations compare red, green, blue, AND alpha
    //  components of the two colors.
    //

    friend int operator==(const Color32 & col1, const Color32 & col2);
    friend int operator!=(const Color32 & col1, const Color32 & col2);

protected:
    UInt8 c[4];

private:
};

// ================================================================= class RGB

class RGB
{
    friend class Color32;
    friend class DRGB;
    friend class RGBA;

    friend class Viewer3D;		//  needs FAST access to members

public:

    RGB(
	);
    RGB(
    	float rNew,
	float gNew,
	float bNew
	);
    RGB(
	const FCrd3 & crd
	);
    RGB(
	const Color32 & col
	);
    RGB(
	const DRGB & drgb
	);
    RGB(
	const RGBA & rgba
	);

    float r() const;
    float g() const;
    float b() const;

    void r(float rNew);
    void g(float gNew);
    void b(float bNew);

    float & operator[](int index);
    float operator[](int index) const;

    //
    //  The following operators will not guard against over- or underflow of
    //  the range [0, 1]. The user has to do these checks himself.
    //

    RGB & operator+=(const RGB & rgb);
    RGB & operator-=(const RGB & rgb);
    RGB & operator*=(float scalar);

    friend RGB operator+(const RGB & rgb1, const RGB & rgb2);
    friend RGB operator-(const RGB & rgb1, const RGB & rgb2);
    friend RGB operator*(const RGB & rgb, float scalar);
    friend RGB operator*(float scalar, const RGB & rgb);

    static const RGB black;
    static const RGB red;
    static const RGB green;
    static const RGB blue;
    static const RGB cyan;
    static const RGB magenta;
    static const RGB yellow;
    static const RGB white;

protected:
/*

    friend RGB pow(const RGB & rgb, float p);

*/

    FCrd3 c;

private:
};

// ================================================================ class DRGB

class DRGB
{
    friend class Color32;
    friend class RGB;
    friend class RGBA;

    friend class Viewer3D;		//  needs FAST access to members

public:

    DRGB(
	);
    DRGB(
    	double rNew,
	double gNew,
	double bNew
	);
    DRGB(
	const Crd3 & crd
	);
    DRGB(
	const Color32 & col
	);
    DRGB(
	const RGB & rgb
	);
    DRGB(
	const RGBA & rgba
	);

    double r() const;
    double g() const;
    double b() const;

    void r(double rNew);
    void g(double gNew);
    void b(double bNew);

    double & operator[](int index);
    double operator[](int index) const;

    //
    //  The following operators will not guard against over- or underflow of
    //  the range [0, 1]. The user has to do these checks himself.
    //

    DRGB & operator+=(const DRGB & drgb);
    DRGB & operator-=(const DRGB & drgb);
    DRGB & operator*=(double scalar);

    friend DRGB operator+(const DRGB & drgb1, const DRGB & drgb2);
    friend DRGB operator-(const DRGB & drgb1, const DRGB & drgb2);
    friend DRGB operator*(const DRGB & drgb, double scalar);
    friend DRGB operator*(double scalar, const DRGB & drgb);

    static const DRGB black;
    static const DRGB red;
    static const DRGB green;
    static const DRGB blue;
    static const DRGB cyan;
    static const DRGB magenta;
    static const DRGB yellow;
    static const DRGB white;

protected:
/*

    friend DRGB pow(const DRGB & drgb, double p);

*/

    Crd3 c;

private:
};

// ================================================================ class RGBA

class RGBA
{
    friend class Color32;
    friend class RGB;
    friend class DRGB;

    friend class Viewer3D;		//  needs FAST access to members

public:

    RGBA(
	);
    RGBA(
    	float rNew,
	float gNew,
	float bNew,
	float aNew
	);
    RGBA(
	const FCrd4 & crd
	);
    RGBA(
	const Color32 & col
	);
    RGBA(
	const RGB & rgb, float alpha
	);
    RGBA(
	const DRGB & drgb, double alpha
	);

    float r() const;
    float g() const;
    float b() const;
    float a() const;

    void r(float rNew);
    void g(float gNew);
    void b(float bNew);
    void a(float aNew);

    float & operator[](int index);
    float operator[](int index) const;

    //
    //  The following operators will not guard against over- or underflow of
    //  the range [0, 1]. The user has to do these checks himself.
    //

    RGBA & operator+=(const RGBA & rgba);
    RGBA & operator-=(const RGBA & rgba);
    RGBA & operator*=(float scalar);

    friend RGBA operator+(const RGBA & rgba1, const RGBA & rgba2);
    friend RGBA operator-(const RGBA & rgba1, const RGBA & rgba2);
    friend RGBA operator*(const RGBA & rgba, float scalar);
    friend RGBA operator*(float scalar, const RGBA & rgba);

protected:
    FCrd4 c;

private:
};

} // extern "C++"

#ifndef COLOR_NO_INLINE

#define INLINE_GLOBAL COMPILER_INLINE_GLOBAL
#define INLINE_MEMBER COMPILER_INLINE_MEMBER

#include "VEGA/Graphics/Color.I"

#undef INLINE_GLOBAL
#undef INLINE_MEMBER

#endif

#endif // _GRAPHICS_COLOR_H_
// ===========================================================================
