// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       FVec3D
//  TYPE:       c++ header
//  PROJECT:    VEGA Graphics Library
//  CONTENT:    3D vectors, points, and transformations using floats
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rft     Robert F. Tobler
//              hl      Helwig L"offelman
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 14:08:16  rft     last modification
//  08-Aug-93           rft     created
// ===========================================================================
//  MANUAL:
//
//  DESCRIPTION
//      The classes FVec3D and FPnt3D are the basic types for 3D geometry.
//      They can be transformed with the corresponding transformation types
//      that are called FScale3D, FRotX3D, FRotY3D, FRotZ3D, FRot3D,
//      FShearXZ3D, FShearZX3D, FShearYZ3D, FLin3D, FTrans3D and FTrafo3D.
//
//      The classes FVec3D and FPnt3D have been designed in a typesafe way,
//      so that all operations among them are geometrically meaningful.
//      Thus there are some combinations of operations that are not allowed
//      since there is no meaningful geometric result. An example for such
//      an operation is the addition of two points.
//
//      In terms of matrix theory, FVec3D and FPnt3D can be viewed as
//      homogenous 3D vectors and points (the trailing zeros and ones are
//      not stored):
//
//              FVec3D vec;
//              FPnt3D pnt;
//
//              (vec.x(), vec.y(), vec.z(), 0)
//              (pnt.x(), pnt.y(), pnt.z(), 1)
//
//      The transformation type FTrafo3D can be viewed as a 3D homogeneous
//      matrix. The multiplication of FVec3D and FPnt3D with FTrafo3D can
//      then be represented as follows (the elements of the transformation
//      cannot be accessed in the way shown here):
//
//              FTrafo3D t;
//              FVec3D v;
//              FPnt3D m;
//
//              FVec3D vec = v * t;
//              FVec3D pnt = p * t;
//
//                                      (t.m[0,0], t.m[0,1], t.m[0,2], 0)
//                                      (t.m[1,0], t.m[1,1], t.m[0,1], 0)
//                                      (t.m[2,0], t.m[2,1], t.m[0,2], 0)
//                                      (t.c[0],   t.c[1],   t.c[2],   1)
//
//              (v.x(),v.y(),v.z(),0)   ( vec.x(),  vec.y(),  vec.z(), 0)
//
//              (p.x(),p.y(),p.z(),1)   ( pnt.x(),  pnt.y(),  pnt.z(), 1)
//
//      The transformation types FTrans3D, FScale3D, FRotX3D, FRotY3D,
//      FRotZ3D, FRot3D, FShearXZ3D, FShearZX3D, FShearYZ3D, and FLin3D
//      represent subsets of the transformations represented by the FTrafo3D
//      type.
//
//      Rotations always rotate clockwise if you look in the direction of the
//      axis. (For example FRotZ3D(90 DEGREES) transforms a point that lies on
//      the positive x-axis into a point on the positive y-axis.)
//
//      Other ways to think about FLin3D and FTrafo3D:
//
//              FLin3D <=>   (FVec3D,     // vecX
//                           FVec3D,     // vecY
//                           FVec3D)     // vecZ
//
//              FTrafo3D <=> (FVec3D,     // vecX
//                           FVec3D,     // vecY
//                           FVec3D,     // vecZ
//                           FPnt3D)     // pntO (O for Origin)
//
//              (This can be used to directly specify a transformation by
//               just specifying the origin and the base vectors of the
//               coordinate system !)
//
//              FTrafo3D <=> FLin3D * FTrans3D
//
//              FTrafo3D <=> (FLin3D,     // lin
//                           FVec3D)     // vec
//
//      If you are using these models, there are constructors that take these
//      elements as parameters, and methods and functions to get these
//      elements (getVecX, getVecY, getVecZ, getPntO, getLin, and getTrans).
//
//  CHECKS
//      Range checking and attempts to normalize null-vectors will only be
//      caught if WITH_FCRD3_CHECKS is defined.
//
//  ERRORHANDLERS
//      FVec3D_errors
//
//  TRACERS
//      FVec3D_constructors
//
//  SEE ALSO
//      VEGA, GraphicsLibrary
// ===========================================================================
#ifndef _GRAPHICS_FVEC3D_H_
#define _GRAPHICS_FVEC3D_H_

#include <iostream.h>

#include <VEGA/Math/Math.h>
#include <VEGA/Math/Crd2.h>
#include <VEGA/Math/FCrd3.h>
#include <VEGA/Math/FCrd4.h>

#include "VEGA/Graphics/Globals.h"

#ifdef __GNUC__
#pragma interface
#endif

extern "C++"
{

// ============================================ classes defined in this module

class FVec3D;				// vector
class FPnt3D;				// point
class FScale3D;				// scaling by (sx,sy,sz)
class FRotX3D;				// rotation around x-axis
class FRotY3D;				// rotation around y-axis
class FRotZ3D;				// rotation around z-axis
class FRot3D;				// rotation around axis through origin
class FShearXZ3D;			// shear along x-y-plane
class FShearZX3D;			// shear along z-x-plane
class FShearYZ3D;			// shear along y-z-plane
class FLin3D;				// linear transformation
class FTrans3D;				// translation by vector
class FTrafo3D;				// affine transformation

// =========================================== errors generated by this module

#define ERROR_FVEC3D_SINGULAR_SCALE      1
        //
        //  An attempt was made to invert a singular scale.
        //


// =============================================================== class FVec3D

class FVec3D
{
    friend class FPnt3D;
    friend class FLin3D;
    friend class FTrans3D;
    friend class FTrafo3D;

    friend class Viewer3D;		//  needs FAST access to members

public:

    // --------------------------------------------- constructors / destructor

    FVec3D();
    FVec3D(
        float xNew,
        float yNew,
        float zNew
        );
    FVec3D(
        const FCrd3 & crd
        );
    FVec3D(
        const FVec3D & vec
        );
    ~FVec3D();

    // ------------------------------------------------------ member operators

    float operator[](
	int index
	) const;
    float & operator[](
	int index
	);

    FVec3D & operator+=(
	const FVec3D & vec
	);
    FVec3D & operator-=(
	const FVec3D & vec
	);

    FVec3D & operator*=(
	float scalar
	);

    // ------------------------------------------------------ member functions

    float x(
	) const;
    float y(
	) const;
    float z(
	) const;

    void x(
	float xNew
	);
    void y(
	float yNew
	);
    void z(
	float zNew
	);

    void normalize(
	);

    FCrd3 getFCrd3(
	) const;

    FCrd4 getFCrd4(			// Get homogeneous representation.
	) const;

    FPnt3D getFPnt3D(			// Explicit conversion to FPnt3D.
	) const;

    // ------------------------------------------------------ friend operators

    friend ostream & operator<<(
	ostream & s,
	const FVec3D & vec
	);
    friend FVec3D operator-(
	const FVec3D & vec
	);
    friend FVec3D operator+(
	const FVec3D & vec1,
	const FVec3D & vec2
	);
    friend FPnt3D operator+(
	const FPnt3D & pnt,
	const FVec3D & vec
	);
    friend FVec3D operator-(
	const FVec3D & vec1,
	const FVec3D & vec2
	);
    friend FPnt3D operator-(
	const FPnt3D & pnt,
	const FVec3D & vec)
	;
    friend FVec3D operator*(
	const FVec3D & vec,
	float scalar
	);
    friend FVec3D operator*(
	float scalar,
	const FVec3D & vec
	);

    // ------------------------------------------------------ friend functions

    friend FCrd3 getFCrd3(
	const FVec3D & vec
	);
    friend FCrd4 getFCrd4(		// Get homogeneous representation.
	const FVec3D & vec
	);
    friend FPnt3D getFPnt3D(		// Explicit conversion.
	const FVec3D & vec
	);
    friend float squaredLength(
	const FVec3D & vec
	);
    friend float length(
	const FVec3D & vec
	);
    friend float dotProduct(
	const FVec3D & vec1,
	const FVec3D & vec2
	);
    friend FVec3D crossProduct(
	const FVec3D & vec1,		// (vec1, vec2, result) define the
	const FVec3D & vec2		//  base of a right-handed coordinate
	);				//  system

    friend FVec3D normalizedCopy(
	const FVec3D & vec
	);

    // -------------------------------------- special purpose friend functions

    friend FVec3D cwReciprocalCopy(      // Componentwise multiplicative
        const FVec3D & vec		//  inverse. Zero components are not
        );				//  inverted!

    friend FVec3D cwMin(                 // Componentwise minimum.
        const FVec3D & vec1,
        const FVec3D & vec2
        );

    friend FVec3D cwMax(                 // Componentwise maximum.
        const FVec3D & vec1,
        const FVec3D & vec2
        );

    // ---------------------------------------------- transformation functions

    //
    //  Multiply vector by special transformation.
    //
    FVec3D & operator*=(
	const FScale3D & scale
	);
    FVec3D & operator*=(
	const FRotX3D & rot
	);
    FVec3D & operator*=(
	const FRotY3D & rot
	);
    FVec3D & operator*=(
	const FRotZ3D & rot
	);
    FVec3D & operator*=(
	const FShearXZ3D & shear
	);
    FVec3D & operator*=(
	const FShearZX3D & shear
	);
    FVec3D & operator*=(
	const FShearYZ3D & shear
	);
    FVec3D & operator*=(
	const FLin3D & lin
	);
    FVec3D & operator*=(
	const FTrafo3D & trafo
	);

    //
    //  Product of vector and special tranformation.
    //
    friend FVec3D operator*(
	const FVec3D & vec,
	const FScale3D & scale
	);
    friend FVec3D operator*(
	const FVec3D & vec,
	const FRotX3D & rot
	);
    friend FVec3D operator*(
	const FVec3D & vec,
	const FRotY3D & rot
	);
    friend FVec3D operator*(
	const FVec3D & vec,
	const FRotZ3D & rot
	);
    friend FVec3D operator*(
	const FVec3D & vec,
	const FShearXZ3D & shear
	);
    friend FVec3D operator*(
	const FVec3D & vec,
	const FShearZX3D & shear
	);
    friend FVec3D operator*(
	const FVec3D & vec,
	const FShearYZ3D & shear
	);
    friend FVec3D operator*(
	const FVec3D & vec,
	const FLin3D & lin
	);
    friend FVec3D operator*(
	const FVec3D & vec,
	const FTrafo3D & trafo
	);

    // ----------------------------------------------- static member variables

    static const FVec3D baseX;
    static const FVec3D baseY;
    static const FVec3D baseZ;

protected:

    FCrd3 c;
};

// ===================================================================== FPnt3D

class FPnt3D
{
    friend class FVec3D;
    friend class FLin3D;
    friend class FTrans3D;
    friend class FTrafo3D;

    friend class Viewer3D;		//  needs FAST access to members

public:

    // ---------------------------------------------------------- constructors

    FPnt3D(
        );
    FPnt3D(
        float x,
        float y,
        float z
        );
    FPnt3D(
        const FCrd3 & crd
        );
    FPnt3D(
        const FPnt3D & pnt
        );
    ~FPnt3D(
        );

    // ------------------------------------------------------ member operators

    float operator[](
	int index
	) const;
    float & operator[](
	int index
	);

    FPnt3D & operator+=(
	const FVec3D & vec
	);
    FPnt3D & operator-=(
	const FVec3D & vec
	);

    // ------------------------------------------------------ member functions

    float x(
	) const;
    float y(
	) const;
    float z(
	) const;

    void x(
	float xNew
	);
    void y(
	float yNew
	);
    void z(
	float zNew
	);

    FCrd3 getFCrd3(
	) const;

    FCrd4 getFCrd4(			// Get homogeneous representation.
	) const;

    FVec3D getFVec3D(			// Explicit conversion to FVec3D.
	) const;

    // ------------------------------------------------------ friend operators

    friend ostream & operator<<(
	ostream & s,
	const FPnt3D & vec
	);
    friend FPnt3D operator+(
	const FPnt3D & pnt,
	const FVec3D & vec
	);
    friend FPnt3D operator-(
	const FPnt3D & pnt,
	const FVec3D & vec
	);
    friend FVec3D operator-(
	const FPnt3D & pnt1,
	const FPnt3D & pnt2
	);

    // ------------------------------------------------------ friend functions

    friend FCrd3 getFCrd3(
	const FPnt3D & pnt
	);
    friend FCrd4 getFCrd4(		// Get homogeneous representation.
	const FPnt3D & pnt
	);
    friend FVec3D getFVec3D(		// Explicit conversion.
	const FPnt3D & pnt
	);
    friend float squaredDistanceFromOrigin(
	const FPnt3D & pnt
	);
    friend float distanceFromOrigin(
	const FPnt3D & pnt
	);
    friend float squaredDistance(
	const FPnt3D & pnt1,
	const FPnt3D & pnt2
	);
    friend float distance(
	const FPnt3D & pnt1,
	const FPnt3D & pnt2
	);

    // ----------------------------------------------- componentwise functions

    friend FPnt3D cwMin(
	const FPnt3D & vec1,
	const FPnt3D & vec2
	);
    friend FPnt3D cwMax(
	const FPnt3D & vec1,
	const FPnt3D & vec2
	);

    // ---------------------------------------------- transformation functions

    FPnt3D & operator*=(
	const FScale3D & scale
	);
    FPnt3D & operator*=(
	const FRotX3D & rot
	);
    FPnt3D & operator*=(
	const FRotY3D & rot
	);
    FPnt3D & operator*=(
	const FRotZ3D & rot
	);
    FPnt3D & operator*=(
	const FShearXZ3D & shear
	);
    FPnt3D & operator*=(
	const FShearZX3D & shear
	);
    FPnt3D & operator*=(
	const FShearYZ3D & shear
	);
    FPnt3D & operator*=(
	const FTrans3D & trans
	);
    FPnt3D & operator*=(
	const FLin3D & lin
	);
    FPnt3D & operator*=(
	const FTrafo3D & trafo
	);

    friend FPnt3D operator*(
	const FPnt3D & pnt,
	const FScale3D & scale
	);
    friend FPnt3D operator*(
	const FPnt3D & pnt,
	const FRotX3D & rot
	);
    friend FPnt3D operator*(
	const FPnt3D & pnt,
	const FRotY3D & rot
	);
    friend FPnt3D operator*(
	const FPnt3D & pnt,
	const FRotZ3D & rot
	);
    friend FPnt3D operator*(
	const FPnt3D & pnt,
	const FShearXZ3D & shear
	);
    friend FPnt3D operator*(
	const FPnt3D & pnt,
	const FShearZX3D & shear
	);
    friend FPnt3D operator*(
	const FPnt3D & pnt,
	const FShearYZ3D & shear
	);
    friend FPnt3D operator*(
	const FPnt3D & pnt,
	const FTrans3D & trans
	);
    friend FPnt3D operator*(
	const FPnt3D & pnt,
	const FLin3D & lin
	);
    friend FPnt3D operator*(
	const FPnt3D & pnt,
	const FTrafo3D & trafo
	);

    // ----------------------------------------------- static member variables

    static const FPnt3D origin;

protected:
    FCrd3 c;
};

// ============================================================= class FScale3D

class FScale3D
{
    friend class FVec3D;
    friend class FPnt3D;
    friend class FLin3D;
    friend class FTrafo3D;

public:

    FScale3D(				// Create undefined object.
    	);
    FScale3D(
        float scalar
        );
    FScale3D(
        float x,
        float y,
        float z
        );
    FScale3D(
        const FCrd3 & scale
        );
    FScale3D(
        const FScale3D & scale
        );
    ~FScale3D();

    FScale3D & operator*=(
	const FScale3D & scale
	);

    void invert(
	);

    friend ostream & operator<<(
	ostream & s,
	const FScale3D & scale
	);
    friend FScale3D operator*(
	const FScale3D & scale1,
	const FScale3D & scale2
	);

    friend FScale3D inverseCopy(
	const FScale3D & scale
	);

    static const FScale3D identity;

    // -------------------------------------------- functions that need access

    friend FVec3D operator*(const FVec3D & vec, const FScale3D & scale);
    friend FPnt3D operator*(const FPnt3D & pnt, const FScale3D & scale);
    friend FLin3D operator*(const FScale3D & scale, const FLin3D & lin);
    friend FLin3D operator*(const FLin3D & lin, const FScale3D & scale);
    friend FTrafo3D operator*(const FScale3D & scale, const FTrafo3D & trafo);
    friend FTrafo3D operator*(const FTrafo3D & trafo, const FScale3D & scale);

protected:

    FCrd3 c;
};

// ============================================================== class FRotX3D

class FRotX3D
{
    friend class FVec3D;
    friend class FPnt3D;
    friend class FRot3D;
    friend class FLin3D;
    friend class FTrafo3D;

public:

    FRotX3D(				// Create undefined object.
    	);
    FRotX3D(
        float newAlpha
        );
    FRotX3D(				// WARNING:
        float newAlpha,		//  These parameters are not checked
        float newCosAlpha,		//  for consistency.
        float newSinAlpha
        );
    FRotX3D(
        const FRotX3D & rot
        );
    ~FRotX3D(
	);

    FRotX3D & operator*=(
	const FRotX3D & rot
	);

    float getAlpha(
	) const;
    void invert(
	);
    
    friend ostream & operator<<(
	ostream & s,
	const FRotX3D & rot
	);
    friend FRotX3D operator*(
	const FRotX3D & rot1,
	const FRotX3D & rot2
	);

    friend FRotX3D inverseCopy(
	const FRotX3D & rot
	);

    static const FRotX3D identity;

    // -------------------------------------------- functions that need access

    friend FVec3D operator*(const FVec3D & vec, const FRotX3D & rot);
    friend FPnt3D operator*(const FPnt3D & pnt, const FRotX3D & rot);
    friend FLin3D operator*(const FLin3D & lin, const FRotX3D & rot);
    friend FLin3D operator*(const FRotX3D & rot, const FLin3D & lin);
    friend FTrafo3D operator*(const FTrafo3D & trafo, const FRotX3D & rot);
    friend FTrafo3D operator*(const FRotX3D & rot, const FTrafo3D & trafo);

protected:
    float alpha;
    float cosAlpha;
    float sinAlpha;
};

// ============================================================== class FRotY3D

class FRotY3D
{
    friend class FVec3D;
    friend class FPnt3D;
    friend class FRot3D;
    friend class FLin3D;
    friend class FTrafo3D;

public:

    FRotY3D(				// Create undefined object.
    	);
    FRotY3D(
        float newAlpha
        );
    FRotY3D(
        float newAlpha,		// WARNING:
        float newCosAlpha,		//  These parameters are not checked
        float newSinAlpha		//  for consistency.
        );
    FRotY3D(
        const FRotY3D & rot
        );
    ~FRotY3D();

    FRotY3D & operator*=(
	const FRotY3D & rot
	);

    float getAlpha(
	) const;
    void invert(
	);

    friend ostream & operator<<(
	ostream & s,
	const FRotY3D & rot
	);
    friend FRotY3D operator*(
	const FRotY3D & rot1,
	const FRotY3D & rot2
	);

    friend FRotY3D inverseCopy(
	const FRotY3D & rot
	);

    static const FRotY3D identity;

    // -------------------------------------------- functions that need access

    friend FVec3D operator*(const FVec3D & vec, const FRotY3D & rot);
    friend FPnt3D operator*(const FPnt3D & pnt, const FRotY3D & rot);
    friend FLin3D operator*(const FLin3D & lin, const FRotY3D & rot);
    friend FLin3D operator*(const FRotY3D & rot, const FLin3D & lin);
    friend FTrafo3D operator*(const FTrafo3D & trafo, const FRotY3D & rot);
    friend FTrafo3D operator*(const FRotY3D & rot, const FTrafo3D & trafo);

protected:
    float alpha;
    float cosAlpha;
    float sinAlpha;
};

// ============================================================== class FRotZ3D

class FRotZ3D
{
    friend class FVec3D;
    friend class FPnt3D;
    friend class FRot3D;
    friend class FLin3D;
    friend class FTrafo3D;

public:

    FRotZ3D(				// Create undefined object.
    	);
    FRotZ3D(
        float newAlpha
        );
    FRotZ3D(
        float newAlpha,		// WARNING:
        float newCosAlpha,		// These parameters are not checked
        float newSinAlpha		// for consistency.
        );
    FRotZ3D(
        const FRotZ3D & rot
        );
    ~FRotZ3D();

    FRotZ3D & operator*=(
	const FRotZ3D & rot
	);

    float getAlpha(
	) const;
    void invert(
	);

    friend ostream & operator<<(
	ostream & s,
	const FRotZ3D & rot
	);
    friend FRotZ3D operator*(
	const FRotZ3D & rot1,
	const FRotZ3D & rot2
	);

    friend FRotZ3D inverseCopy(
	const FRotZ3D & rot
	);

    static const FRotZ3D identity;

    // -------------------------------------------- functions that need access

    friend FVec3D operator*(const FVec3D & vec, const FRotZ3D & rot);
    friend FPnt3D operator*(const FPnt3D & pnt, const FRotZ3D & rot);
    friend FLin3D operator*(const FLin3D & lin, const FRotZ3D & rot);
    friend FLin3D operator*(const FRotZ3D & rot, const FLin3D & lin);
    friend FTrafo3D operator*(const FTrafo3D & trafo, const FRotZ3D & rot);
    friend FTrafo3D operator*(const FRotZ3D & rot, const FTrafo3D & trafo);

protected:
    float alpha;
    float cosAlpha;
    float sinAlpha;
};

// =============================================================== class FRot3D

class FRot3D
{
    friend class FVec3D;
    friend class FPnt3D;
    friend class FLin3D;
    friend class FTrafo3D;

public:

    //
    //  This class represents general rotations around axes through the
    //	origin. They are implemented using quaternions.
    //

    FRot3D(				// Create undefined object.
        );
    FRot3D(
        const FRot3D & rot
        );
    FRot3D(				// WARNING:
    	float aNew,			//  The tuple is asumed to represent
	float bNew,			//  a unit-norm quaternion !
	float cNew,
	float dNew
	);
    FRot3D(				// WARNING: The FCrd4 is assumed to
        const FCrd4 & rot		//  represent a unit-norm quaternion !
        );
    FRot3D(				// Init with vector and angle:
        const FVec3D & vec,		//  A unit-norm quaternion is built
        float alpha			//  from the given vector and angle.
        );
    FRot3D(				// Init with vector and angle 0,
    	const FVec3D & vec		//  no unit-norm conversion is done.
	);
    FRot3D(
        const FRotX3D & rot
        );
    FRot3D(
        const FRotY3D & rot
        );
    FRot3D(
        const FRotZ3D & rot
        );
    ~FRot3D(
	);

    FRot3D & operator*=(
	const FRot3D & rot
	);

    float getAlpha(
	) const;			// Return value in [0, MATH_2_MUL_PI).

    FVec3D getFVec3D(
	) const;
    void invert(
	);

    friend ostream & operator<<(
	ostream & s,
	const FRot3D & rot
	);
    friend FRot3D operator*(
	const FRot3D & rot1,
	const FRot3D & rot2
	);

    friend FRot3D inverseCopy(
	const FRot3D & rot
	);

    static const FRot3D identity;

protected:
    FCrd4 c;
};

// =========================================================== class FShearXZ3D

class FShearXZ3D
{
    friend class FVec3D;
    friend class FPnt3D;
    friend class FLin3D;
    friend class FTrafo3D;
public:

    FShearXZ3D(				// Create undefined object.
    	);
    FShearXZ3D(
        const FShearXZ3D & shear
        );
    FShearXZ3D(
        float xNew,
	float yNew
        );
    FShearXZ3D(
	const Crd2 & crd
        );
    ~FShearXZ3D(
	);

    FShearXZ3D & operator*=(
	const FShearXZ3D & shear
	);

    void invert(
	);

    friend ostream & operator<<(
	ostream & s,
	const FShearXZ3D & shear
	);
    friend FShearXZ3D operator*(
	const FShearXZ3D & s1,
	const FShearXZ3D & s2
	);

    friend FShearXZ3D inverseCopy(
	const FShearXZ3D & shear
	);

    static const FShearXZ3D identity;
    
    // -------------------------------------------- functions that need access

    friend FVec3D operator*(const FVec3D & vec, const FShearXZ3D & shear);
    friend FPnt3D operator*(const FPnt3D & pnt, const FShearXZ3D & shear);
    friend FLin3D operator*(const FLin3D & lin, const FShearXZ3D & sh);
    friend FLin3D operator*(const FShearXZ3D & sh, const FLin3D & lin);
    friend FTrafo3D operator*(const FTrafo3D & trafo, const FShearXZ3D & sh);
    friend FTrafo3D operator*(const FShearXZ3D & sh, const FTrafo3D & trafo);

protected:
    Crd2 c;
};

// =========================================================== class FShearZX3D

class FShearZX3D
{
    friend class FVec3D;
    friend class FPnt3D;
    friend class FLin3D;
    friend class FTrafo3D;
public:

    FShearZX3D(				// Create undefined object.
    	);
    FShearZX3D(
        const FShearZX3D & shear
        );
    FShearZX3D(
        float zNew,
	float xNew
        );
    FShearZX3D(
	const Crd2 & crd
        );
    ~FShearZX3D(
	);

    FShearZX3D & operator*=(
	const FShearZX3D & shear
	);

    void invert();

    friend ostream & operator<<(
	ostream & s,
	const FShearZX3D & shear
	);
    friend FShearZX3D operator*(
	const FShearZX3D & s1,
	const FShearZX3D & s2
	);

    friend FShearZX3D inverseCopy(
	const FShearZX3D & shear
	);

    static const FShearZX3D identity;
    
    // -------------------------------------------- functions that need access

    friend FVec3D operator*(const FVec3D & vec, const FShearZX3D & shear);
    friend FPnt3D operator*(const FPnt3D & pnt, const FShearZX3D & shear);
    friend FLin3D operator*(const FLin3D & lin, const FShearZX3D & sh);
    friend FLin3D operator*(const FShearZX3D & sh, const FLin3D & lin);
    friend FTrafo3D operator*(const FTrafo3D & trafo, const FShearZX3D & sh);
    friend FTrafo3D operator*(const FShearZX3D & sh, const FTrafo3D & trafo);

protected:
    Crd2 c;
};

// =========================================================== class FShearYZ3D

class FShearYZ3D
{
    friend class FVec3D;
    friend class FPnt3D;
    friend class FLin3D;
    friend class FTrafo3D;
public:

    FShearYZ3D(				// Create undefined object.
    	);
    FShearYZ3D(
        const FShearYZ3D & shear
        );
    FShearYZ3D(
        float yNew,
	float zNew
        );
    FShearYZ3D(
	const Crd2 & crd
        );
    ~FShearYZ3D(
	);

    FShearYZ3D & operator*=(
	const FShearYZ3D & shear
	);

    void invert(
	);

    friend ostream & operator<<(
	ostream & s,
	const FShearYZ3D & shear
	);
    friend FShearYZ3D operator*(
	const FShearYZ3D & s1,
	const FShearYZ3D & s2
	);

    friend FShearYZ3D inverseCopy(
	const FShearYZ3D & shear
	);

    static const FShearYZ3D identity;
    
    // -------------------------------------------- functions that need access

    friend FVec3D operator*(const FVec3D & vec, const FShearYZ3D & shear);
    friend FPnt3D operator*(const FPnt3D & pnt, const FShearYZ3D & shear);
    friend FLin3D operator*(const FLin3D & lin, const FShearYZ3D & sh);
    friend FLin3D operator*(const FShearYZ3D & sh, const FLin3D & lin);
    friend FTrafo3D operator*(const FTrafo3D & trafo, const FShearYZ3D & sh);
    friend FTrafo3D operator*(const FShearYZ3D & sh, const FTrafo3D & trafo);

protected:
    Crd2 c;
};

// =============================================================== class FLin3D

class FLin3D
{
    friend class FVec3D;
    friend class FPnt3D;
    friend class FTrafo3D;
public:

    FLin3D(				// Create undefined object.
	);
    FLin3D(
	const FLin3D & lin
	);
    FLin3D(
	const FMat3 & m
	);
    FLin3D(
        const FVec3D & newVecX,
        const FVec3D & newVecY,
        const FVec3D & newVecZ
	);
    FLin3D(
        float xx, float xy, float xz,
        float yx, float yy, float yz,
        float zx, float zy, float zz
        );
    FLin3D(
        const FScale3D & scale
        );
    FLin3D(
        const FRotX3D & rot
        );
    FLin3D(
        const FRotY3D & rot
        );
    FLin3D(
        const FRotZ3D & rot
        );
    FLin3D(
        const FShearXZ3D & scale
        );
    FLin3D(
        const FShearZX3D & scale
        );
    FLin3D(
        const FShearYZ3D & scale
        );
    FLin3D(
        const FRot3D & rot
        );
    ~FLin3D(
	);

    // ------------------------------------------------------ member functions

    FVec3D getVecX(
	) const;
    FVec3D getVecY(
	) const;
    FVec3D getVecZ(
	) const;

    FMat3 getFMat3(
	) const;

    void invert(
	);
        //
        //  ERROR_CRD3_SINGULAR_MATRIX see module FCrd3
        //

    // ------------------------------------------------------ friend operators

    friend ostream & operator<<(
	ostream & s,
	const FLin3D & lin
	);

    // ------------------------------------------------------ friend functions

    friend FVec3D getVecX(
	const FLin3D & lin
	);
    friend FVec3D getVecY(
	const FLin3D & lin
	);
    friend FVec3D getVecZ(
	const FLin3D & lin
	);

    friend FMat3 getFMat3(
	const FLin3D & lin
	);
    friend float determinant(
	const FLin3D & lin
	);
    friend FLin3D inverseCopy(
	const FLin3D & lin,
	float det
	);
        //
        //  det is NOT checked against zero!
        //

    friend FLin3D inverseCopy(
	const FLin3D & lin
	);
        //
        //  ERROR_CRD3_SINGULAR_MATRIX see module FCrd3
        //

    // ----------------------------------------------- tranformation functions

    //
    //  Multiply transformation by special transformation.
    //
    FLin3D & operator*=(
	const FScale3D & scale
	);
    FLin3D & operator*=(
	const FRotX3D & rot
	);
    FLin3D & operator*=(
	const FRotY3D & rot
	);
    FLin3D & operator*=(
	const FRotZ3D & rot
	);
    FLin3D & operator*=(
	const FShearXZ3D & shear
	);
    FLin3D & operator*=(
	const FShearZX3D & shear
	);
    FLin3D & operator*=(
	const FShearYZ3D & shear
	);
    FLin3D & operator*=(
	const FLin3D & lin
	);

    //
    //  Product of transformation and special transformation.
    //
    friend FLin3D operator*(
	const FLin3D & lin,
	const FScale3D & scale
	);
    friend FLin3D operator*(
	const FLin3D & lin,
	const FRotX3D & rot
	);
    friend FLin3D operator*(
	const FLin3D & lin,
	const FRotY3D & rot
	);
    friend FLin3D operator*(
	const FLin3D & lin,
	const FRotZ3D & rot
	);
    friend FLin3D operator*(
	const FLin3D & lin,
	const FShearXZ3D & shear
	);
    friend FLin3D operator*(
	const FLin3D & lin,
	const FShearZX3D & shear
	);
    friend FLin3D operator*(
	const FLin3D & lin,
	const FShearYZ3D & shear
	);

    //
    //  Product of special transformation and transformation.
    //
    friend FLin3D operator*(
	const FScale3D & scale,
	const FLin3D & lin
	);
    friend FLin3D operator*(
	const FRotX3D & rot,
	const FLin3D & lin
	);
    friend FLin3D operator*(
	const FRotY3D & rot,
	const FLin3D & lin
	);
    friend FLin3D operator*(
	const FRotZ3D & rot,
	const FLin3D & lin
	);
    friend FLin3D operator*(
	const FShearXZ3D & shear,
	const FLin3D & lin
	);
    friend FLin3D operator*(
	const FShearZX3D & shear,
	const FLin3D & lin
	);
    friend FLin3D operator*(
	const FShearYZ3D & shear,
	const FLin3D & lin
	);

    friend FLin3D operator*(           // Product of two transformations.
        const FLin3D & lin1,
        const FLin3D & lin2
        );

    //
    //  Product of vector/point and transformation.
    //
    friend FVec3D operator*(
	const FVec3D & vec,
	const FLin3D & lin
	);
    friend FPnt3D operator*(
	const FPnt3D & pnt,
	const FLin3D & lin
	);

    // ----------------------------------------------- static member variables

    static const FLin3D identity;

    // -------------------------------------------- functions that need access

    friend FTrafo3D operator*(const FLin3D & lin, const FTrafo3D & trafo);
    friend FTrafo3D operator*(const FTrafo3D & trafo, const FLin3D & lin);

protected:
    FMat3 m;
};

// ============================================================= class FTrans3D

class FTrans3D
{
    friend class FVec3D;
    friend class FPnt3D;
    friend class FTrafo3D;

public:

    FTrans3D(				// Create undefined object.
	);
    FTrans3D(
        const FCrd3 & crd
        );
    FTrans3D(
        float xNew,
        float yNew,
        float zNew
        );
    FTrans3D(
	const FVec3D & vec
	);
    FTrans3D(
	const FPnt3D & pntO
	);
    FTrans3D(
        const FTrans3D & trans
        );
    ~FTrans3D(
	);

    FTrans3D & operator*=(
	const FTrans3D & trans
	);

    void invert(
	);

    friend ostream & operator<<(
	ostream & s,
	const FTrans3D & trans
	);
    friend FTrans3D operator*(
	const FTrans3D & tr1,
	const FTrans3D & tr2
	);
    friend FVec3D operator*(
	const FVec3D & vec,
	const FTrans3D & trans
	);
    friend FPnt3D operator*(
	const FPnt3D & pnt,
	const FTrans3D & trans
	);
    friend FTrafo3D operator*(
	const FTrans3D & trans,
	const FTrafo3D & trafo
	);
    friend FTrafo3D operator*(
	const FTrafo3D & trafo,
	const FTrans3D & trans
	);

    friend FTrans3D inverseCopy(
	const FTrans3D & trans
	);

    static const FTrans3D identity;

protected:

    FCrd3 c;
};

// ============================================================= class FTrafo3D

class FTrafo3D
{
    friend class FVec3D;
    friend class FPnt3D;

public:

    FTrafo3D(				// Create undefined object.
        );
    FTrafo3D(
        const FTrafo3D & trafo
        );
    FTrafo3D(
        const FMat3 & m
        );
    FTrafo3D(
        const FMat3 & m,
        const FCrd3 & c
        );
    FTrafo3D(
        float xx, float xy, float xz,
        float yx, float yy, float yz,
        float zx, float zy, float zz,
        float dx, float dy, float dz
        );
    FTrafo3D(
        const FVec3D & newVecX,
        const FVec3D & newVecY,
        const FVec3D & newVecZ,
        const FPnt3D & newPntO
        );
    FTrafo3D(
        const FLin3D & lin,
	const FTrans3D & trans
        );
    FTrafo3D(
        const FScale3D & scale
        );
    FTrafo3D(
        const FRotX3D & rot
        );
    FTrafo3D(
        const FRotY3D & rot
        );
    FTrafo3D(
        const FRotZ3D & rot
        );
    FTrafo3D(
        const FRot3D & rot
        );
    FTrafo3D(
        const FShearXZ3D & scale
        );
    FTrafo3D(
        const FShearZX3D & scale
        );
    FTrafo3D(
        const FShearYZ3D & scale
        );
    FTrafo3D(
        const FLin3D & lin
        );
    FTrafo3D(
        const FTrans3D & trans
        );
    ~FTrafo3D(
        );

    // ------------------------------------------------------ member functions

    FVec3D getVecX(
	) const;
    FVec3D getVecY(
	) const;
    FVec3D getVecZ(
	) const;
    FPnt3D getPntO(
	) const;

    FLin3D getLin(
	) const;
    FTrans3D getTrans(
	) const;

    FMat3 getFMat3(
	) const;
    FMat4 getFMat4(			// Get homogeneous representation.
	) const;

    void invert(
	);
        //
        //  ERROR_CRD3_SINGULAR_MATRIX see module FCrd3
        //

    // ------------------------------------------------------ friend operators

    friend ostream & operator<<(
	ostream & s,
	const FTrafo3D & trafo
	);

    // ------------------------------------------------------ friend functions

    friend FVec3D getVecX(
	const FTrafo3D & trafo
	);
    friend FVec3D getVecY(
	const FTrafo3D & trafo
	);
    friend FVec3D getVecZ(
	const FTrafo3D & trafo
	);
    friend FPnt3D getPntO(
	const FTrafo3D & trafo
	);

    friend FLin3D getLin(
	const FTrafo3D & trafo
	);
    friend FTrans3D getTrans(
	const FTrafo3D & trafo
	);

    friend FMat3 getFMat3(
	const FTrafo3D & trafo
	);
    friend FMat4 getFMat4(		// Get homogeneous representation
	const FTrafo3D & trafo
	);

    friend float determinant(
	const FTrafo3D & trafo
	);
    friend FTrafo3D inverseCopy(
	const FTrafo3D & trafo,
	float det			// det is NOT checked against zero!
	);

    friend FTrafo3D inverseCopy(
	const FTrafo3D & trafo
	);
        //
        //  ERROR_CRD3_SINGULAR_MATRIX see module FCrd3
        //

    // ----------------------------------------------- tranformation functions

    //
    //  Multiply transformation by special transformation.
    //
    FTrafo3D & operator*=(
	const FScale3D & scale
	);
    FTrafo3D & operator*=(
	const FRotX3D & rot
	);
    FTrafo3D & operator*=(
	const FRotY3D & rot
	);
    FTrafo3D & operator*=(
	const FRotZ3D & rot
	);
    FTrafo3D & operator*=(
	const FShearXZ3D & shear
	);
    FTrafo3D & operator*=(
	const FShearZX3D & shear
	);
    FTrafo3D & operator*=(
	const FShearYZ3D & shear
	);
    FTrafo3D & operator*=(
	const FTrans3D & trans
	);
    FTrafo3D & operator*=(
	const FLin3D & lin
	);
    FTrafo3D & operator*=(
	const FTrafo3D & trafo
	);

    //
    //  Product of transformation and special transformation.
    //
    friend FTrafo3D operator*(
	const FTrafo3D & trafo,
	const FScale3D & scale
	);
    friend FTrafo3D operator*(
	const FTrafo3D & trafo,
	const FRotX3D & rot
	);
    friend FTrafo3D operator*(
	const FTrafo3D & trafo,
	const FRotY3D & rot
	);
    friend FTrafo3D operator*(
	const FTrafo3D & trafo,
	const FRotZ3D & rot
	);
    friend FTrafo3D operator*(
	const FTrafo3D & trafo,
	const FShearXZ3D & sh
	);
    friend FTrafo3D operator*(
	const FTrafo3D & trafo,
	const FShearZX3D & sh
	);
    friend FTrafo3D operator*(
	const FTrafo3D & trafo,
	const FShearYZ3D & sh
	);
    friend FTrafo3D operator*(
	const FTrafo3D & trafo,
	const FTrans3D & trans
	);
    friend FTrafo3D operator*(
	const FTrafo3D & trafo,
	const FLin3D & lin
	);

    //
    //  Product of special transformation and transformation.
    //
    friend FTrafo3D operator*(
	const FScale3D & scale,
	const FTrafo3D & trafo
	);
    friend FTrafo3D operator*(
	const FRotX3D & rot,
	const FTrafo3D & trafo
	);
    friend FTrafo3D operator*(
	const FRotY3D & rot,
	const FTrafo3D & trafo
	);
    friend FTrafo3D operator*(
	const FRotZ3D & rot,
	const FTrafo3D & trafo
	);
    friend FTrafo3D operator*(
	const FShearXZ3D & sh,
	const FTrafo3D & trafo
	);
    friend FTrafo3D operator*(
	const FShearZX3D & sh,
	const FTrafo3D & trafo
	);
    friend FTrafo3D operator*(
	const FShearYZ3D & sh,
	const FTrafo3D & trafo
	);
    friend FTrafo3D operator*(
	const FTrans3D & trans,
	const FTrafo3D & trafo
	);
    friend FTrafo3D operator*(
	const FLin3D & lin,
	const FTrafo3D & trafo
	);

    friend FTrafo3D operator*(           // Product of two transformations.
        const FTrafo3D & trafo1,
        const FTrafo3D & trafo2
        );

    //
    //  Product of vector/point and transformation.
    //
    friend FVec3D operator*(
	const FVec3D & vec,
	const FTrafo3D & trafo
	);
    friend FPnt3D operator*(
	const FPnt3D & pnt,
	const FTrafo3D & trafo
	);

    // ----------------------------------------------- static member variables

    static const FTrafo3D identity;

protected:

    FMat3 m;
    FCrd3 c;
};

} // extern "C++"

#ifndef FVEC3D_NO_INLINE

#define INLINE_GLOBAL COMPILER_INLINE_GLOBAL
#define INLINE_MEMBER COMPILER_INLINE_MEMBER

#include "VEGA/Graphics/FVec3D.I"

#undef INLINE_GLOBAL
#undef INLINE_MEMBER

#endif


#endif // _GRAPHICS_FVEC3D_H_
// ===========================================================================
