// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       Vec2D
//  TYPE:       c++ inline
//  PROJECT:    VEGA Graphics Library
//  CONTENT:    2D vectors, points, and transformations using doubles
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rft     Robert F. Tobler
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 14:08:16  rft     last modification
//  08-Aug-93           rft     created
// ===========================================================================

extern "C++"
{

// =============================================================== class Vec2D

// -------------------------------------------------------------- constructors

INLINE_MEMBER Vec2D::Vec2D()
{
    TR_FUN0(Vec2D_constructors,Vec2D);
}

INLINE_MEMBER Vec2D::Vec2D(double xNew, double yNew)
:
    c(xNew,yNew)
{
    TR_FUN2(Vec2D_constructors,Vec2D,xNew,yNew);
}

INLINE_MEMBER Vec2D::Vec2D(const Crd2 & crd)
:
    c(crd)
{
    TR_FUN1(Vec2D_constructors,Vec2D,crd);
}

INLINE_MEMBER Vec2D::Vec2D(const Vec2D & vec)
:
    c(vec.c)
{
    TR_FUN1(Vec2D_constructors,Vec2D,vec);
}

INLINE_MEMBER Vec2D::~Vec2D()
{
    TR_MESSAGE(Vec2D_constructors,"~" << *this);
}

// ---- methods --------------------------------------------------------------

INLINE_MEMBER double Vec2D::operator[](int index) const
{
    return c[index];
}

INLINE_MEMBER double & Vec2D::operator[](int index)
{
    return c[index];
}

INLINE_MEMBER Vec2D & Vec2D::operator+=(const Vec2D & vec)
{
    c += vec.c;
    return *this;
}

INLINE_MEMBER Vec2D & Vec2D::operator-=(const Vec2D & vec)
{
    c -= vec.c;
    return *this;
}

INLINE_MEMBER Vec2D & Vec2D::operator*=(double scalar)
{
    c *= scalar;
    return *this;
}

INLINE_MEMBER double Vec2D::x() const { return c.x0(); }
INLINE_MEMBER double Vec2D::y() const { return c.x1(); }

INLINE_MEMBER void Vec2D::x(double xNew) { c.x0(xNew); }
INLINE_MEMBER void Vec2D::y(double yNew) { c.x1(yNew); }

INLINE_MEMBER void Vec2D::normalize()
{
    c.normalize();
}

INLINE_MEMBER Crd2 Vec2D::getCrd2() const
{
    return c;
}

INLINE_MEMBER Crd3 Vec2D::getCrd3() const
{
    return Crd3(c.x0(),c.x1(),0.0);
}

INLINE_MEMBER Pnt2D::Pnt2D(const Crd2 & crd)
:
    c(crd)
{
    TR_FUN1(Vec2D_constructors,Pnt2D,crd);
}

INLINE_MEMBER Pnt2D Vec2D::getPnt2D() const
{
    return Pnt2D(c);
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const Vec2D & vec)
{
    s << "Vec2D(" << vec.x() << "," << vec.y() << ")";
    return s;
}

INLINE_GLOBAL Vec2D operator-(const Vec2D & vec)
{
    return Vec2D(-vec.c);
}

INLINE_GLOBAL Vec2D operator+(const Vec2D & vec1, const Vec2D & vec2)
{
    return Vec2D(vec1.c + vec2.c);
}

INLINE_GLOBAL Pnt2D operator+(const Pnt2D & pnt, const Vec2D & vec)
{
    return Pnt2D(pnt.c + vec.c);
}

INLINE_GLOBAL Vec2D operator-(const Vec2D & vec1, const Vec2D & vec2)
{
    return Vec2D(vec1.c - vec2.c);
}

INLINE_GLOBAL Pnt2D operator-(const Pnt2D & pnt, const Vec2D & vec)
{
    return Pnt2D(pnt.c - vec.c);
}

INLINE_GLOBAL Vec2D operator*(const Vec2D & vec, double scalar)
{
    return Vec2D(vec.c * scalar);
}

INLINE_GLOBAL Vec2D operator*(double scalar, const Vec2D & vec)
{
    return Vec2D(scalar * vec.c);
}

INLINE_GLOBAL Crd2 getCrd2(const Vec2D & vec)
{
    return vec.c;
}

INLINE_GLOBAL Crd3 getCrd3(const Vec2D & vec)
{
    return Crd3(vec.c.x0(), vec.c.x1(),0.0);
}

INLINE_GLOBAL Pnt2D getPnt2D(const Vec2D & vec)
{
    return Pnt2D(vec.c);
}

INLINE_GLOBAL double squaredLength(const Vec2D & vec)
{
    return squaredLength(vec.c);
}

INLINE_GLOBAL double length(const Vec2D & vec)
{
    return length(vec.c);
}

INLINE_GLOBAL double dotProduct(const Vec2D & vec1, const Vec2D & vec2)
{
    return dotProduct(vec1.c,vec2.c);
}

INLINE_GLOBAL Vec2D normalizedCopy(const Vec2D & vec)
{
    return Vec2D(normalizedCopy(vec.c));
}

INLINE_GLOBAL Vec2D cwReciprocalCopy(const Vec2D & vec)
{
    return Vec2D(cwReciprocalCopy(vec.c));
}

INLINE_GLOBAL Vec2D cwMin(const Vec2D & vec1, const Vec2D & vec2)
{
    return Vec2D(cwMin(vec1.c,vec2.c));
}

INLINE_GLOBAL Vec2D cwMax(const Vec2D & vec1, const Vec2D & vec2)
{
    return Vec2D(cwMax(vec1.c,vec2.c));
}

INLINE_MEMBER Vec2D & Vec2D::operator*=(const Scale2D & scale)
{
    c.scale(scale.c);
    return *this;
}

INLINE_MEMBER Vec2D & Vec2D::operator*=(const Rot2D & rot)
{
    c.rot(rot.cosAlpha,rot.sinAlpha);
    return *this;
}

INLINE_MEMBER Vec2D & Vec2D::operator*=(const ShearX2D & shear)
{
    c.shear0(shear.c);
    return *this;
}

INLINE_MEMBER Vec2D & Vec2D::operator*=(const ShearY2D & shear)
{
    c.shear1(shear.c);
    return *this;
}

INLINE_MEMBER Vec2D & Vec2D::operator*=(const Lin2D & lin)
{
    c *= lin.m;
    return *this;
}

INLINE_MEMBER Vec2D & Vec2D::operator*=(const Trafo2D & trafo)
{
    c *= trafo.m;
    return *this;
}

INLINE_GLOBAL Vec2D operator*(const Vec2D & vec, const Scale2D & scale)
{
    return Vec2D(prCrdScale(vec.c,scale.c));
}

INLINE_GLOBAL Vec2D operator*(const Vec2D & vec, const Rot2D & rot)
{
    return Vec2D(prCrdRot(vec.c,rot.cosAlpha,rot.sinAlpha));
}

INLINE_GLOBAL Vec2D operator*(const Vec2D & vec, const ShearX2D & shear)
{
    return Vec2D(prCrdShear0(vec.c,shear.c));
}

INLINE_GLOBAL Vec2D operator*(const Vec2D & vec, const ShearY2D & shear)
{
    return Vec2D(prCrdShear1(vec.c,shear.c));
}

INLINE_GLOBAL Vec2D operator*(const Vec2D & vec, const Lin2D & lin)
{
    return Vec2D(vec.c * lin.m);
}

INLINE_GLOBAL Vec2D operator*(const Vec2D & vec, const Trafo2D & trafo)
{
    return Vec2D(vec.c * trafo.m);
}

// =============================================================== class Pnt2D

// -------------------------------------------------------------- constructors

INLINE_MEMBER Pnt2D::Pnt2D()
{
    TR_FUN0(Vec2D_constructors,Pnt2D);
}

INLINE_MEMBER Pnt2D::Pnt2D(double xNew, double yNew)
:
    c(xNew,yNew)
{
    TR_FUN2(Vec2D_constructors,Pnt2D,xNew,yNew);
}

INLINE_MEMBER Pnt2D::Pnt2D(const Pnt2D & pnt)
:
    c(pnt.c)
{
    TR_FUN1(Vec2D_constructors,Pnt2D,pnt);
}

INLINE_MEMBER Pnt2D::~Pnt2D()
{
    TR_MESSAGE(Vec2D_constructors,"~" << *this);
}

// ---------------------------------------------------------- member functions

INLINE_MEMBER double Pnt2D::operator[](int index) const
{
    return c[index];
}

INLINE_MEMBER double & Pnt2D::operator[](int index)
{
    return c[index];
}

INLINE_MEMBER Pnt2D & Pnt2D::operator+=(const Vec2D & vec)
{
    c += vec.c;
    return *this;
}

INLINE_MEMBER Pnt2D & Pnt2D::operator-=(const Vec2D & vec)
{
    c -= vec.c;
    return *this;
}

INLINE_MEMBER double Pnt2D::x() const { return c.x0(); }
INLINE_MEMBER double Pnt2D::y() const { return c.x1(); }

INLINE_MEMBER void Pnt2D::x(double xNew) { c.x0(xNew); }
INLINE_MEMBER void Pnt2D::y(double yNew) { c.x1(yNew); }

INLINE_MEMBER Crd2 Pnt2D::getCrd2() const
{
    return c;
}

INLINE_MEMBER Crd3 Pnt2D::getCrd3() const
{
    return Crd3(c.x0(),c.x1(),0.0);
}

INLINE_MEMBER Vec2D Pnt2D::getVec2D() const
{
    return Vec2D(c);
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const Pnt2D & vec)
{
    s << "(" << vec.x() << "," << vec.y() << ")";
    return s;
}

INLINE_GLOBAL Vec2D operator-(const Pnt2D & pnt1, const Pnt2D & pnt2)
{
    return Vec2D(pnt1.c - pnt2.c);
}

INLINE_GLOBAL Crd2 getCrd2(const Pnt2D & pnt)
{
    return pnt.c;
}

INLINE_GLOBAL Crd3 getCrd3(const Pnt2D & pnt)
{
    return Crd3(pnt.c.x0(), pnt.c.x1(),0.0);
}

INLINE_GLOBAL Vec2D getVec2D(const Pnt2D & pnt)
{
    return Vec2D(pnt.c);
}

INLINE_GLOBAL double squaredDistanceFromOrigin(const Pnt2D & pnt)
{
    return (squaredLength(pnt.c));
}

INLINE_GLOBAL double distanceFromOrigin(const Pnt2D & pnt)
{
    return (length(pnt.c));
}

INLINE_GLOBAL double squaredDistance(const Pnt2D & pnt1, const Pnt2D & pnt2)
{
    return (squaredLength(pnt1.c - pnt2.c));
}

INLINE_GLOBAL double distance(const Pnt2D & pnt1, const Pnt2D & pnt2)
{
    return (length(pnt1.c - pnt2.c));
}

INLINE_GLOBAL Pnt2D cwMin(const Pnt2D & pnt1, const Pnt2D & pnt2)
{
    return Pnt2D(cwMin(pnt1.c,pnt2.c));
}

INLINE_GLOBAL Pnt2D cwMax(const Pnt2D & pnt1, const Pnt2D & pnt2)
{
    return Pnt2D(cwMax(pnt1.c,pnt2.c));
}

INLINE_MEMBER Pnt2D & Pnt2D::operator*=(const Scale2D & scale)
{
    c.scale(scale.c);
    return *this;
}

INLINE_MEMBER Pnt2D & Pnt2D::operator*=(const Rot2D & rot)
{
    c.rot(rot.cosAlpha,rot.sinAlpha);
    return *this;
}

INLINE_MEMBER Pnt2D & Pnt2D::operator*=(const ShearX2D & shear)
{
    c.shear0(shear.c);
    return *this;
}

INLINE_MEMBER Pnt2D & Pnt2D::operator*=(const ShearY2D & shear)
{
    c.shear1(shear.c);
    return *this;
}

INLINE_MEMBER Pnt2D & Pnt2D::operator*=(const Lin2D & lin)
{
    c *= lin.m;
    return *this;
}

INLINE_MEMBER Pnt2D & Pnt2D::operator*=(const Trans2D & trans)
{
    c += trans.c;
    return *this;
}

INLINE_MEMBER Pnt2D & Pnt2D::operator*=(const Trafo2D & trafo)
{
    c *= trafo.m;
    c += trafo.c;
    return *this;
}

INLINE_GLOBAL Pnt2D operator*(const Pnt2D & pnt, const Scale2D & scale)
{
    return Pnt2D(prCrdScale(pnt.c,scale.c));
}

INLINE_GLOBAL Pnt2D operator*(const Pnt2D & pnt, const Rot2D & rot)
{
    return Pnt2D(prCrdRot(pnt.c,rot.cosAlpha,rot.sinAlpha));
}

INLINE_GLOBAL Pnt2D operator*(const Pnt2D & pnt, const ShearX2D & shear)
{
    return Pnt2D(prCrdShear0(pnt.c,shear.c));
}

INLINE_GLOBAL Pnt2D operator*(const Pnt2D & pnt, const ShearY2D & shear)
{
    return Pnt2D(prCrdShear1(pnt.c,shear.c));
}

INLINE_GLOBAL Pnt2D operator*(const Pnt2D & pnt, const Lin2D & lin)
{
    return Pnt2D(pnt.c * lin.m);
}

INLINE_GLOBAL Pnt2D operator*(const Pnt2D & pnt, const Trans2D & trans)
{
    return Pnt2D(pnt.c + trans.c);
}

INLINE_GLOBAL Pnt2D operator*(const Pnt2D & pnt, const Trafo2D & trafo)
{
    return Pnt2D(pnt.c * trafo.m + trafo.c);
}


// ============================================================== class Scale2D

INLINE_MEMBER Scale2D::Scale2D(
	)
:
    c(1,1)
{
    TR_FUN0(Vec2D_constructors, Scale2D);
}

INLINE_MEMBER Scale2D::Scale2D(double scalar)
:
    c(scalar,scalar)
{
    TR_FUN1(Vec2D_constructors,Scale2D,scalar);
}

INLINE_MEMBER Scale2D::Scale2D(double x, double y)
:
    c(x,y)
{
    TR_FUN2(Vec2D_constructors,Scale2D,x,y);
}

INLINE_MEMBER Scale2D::Scale2D(const Crd2 & scale)
:
    c(scale)
{
    TR_FUN1(Vec2D_constructors,Scale2D,scale);
}

INLINE_MEMBER Scale2D::Scale2D(const Scale2D & scale)
:
    c(scale.c)
{
    TR_FUN1(Vec2D_constructors,Scale2D,scale);
}

INLINE_MEMBER Scale2D::~Scale2D()
{
    TR_MESSAGE(Vec2D_constructors,"~" << *this);
}

INLINE_MEMBER Scale2D & Scale2D::operator*=(const Scale2D & scale)
{
    c.cwMultiplyBy(scale.c);
    return *this;
}

INLINE_MEMBER void Scale2D::invert()
{
    if (c.x0() == 0.0 ||                // use explicit accessor function
        c.x1() == 0.0 )                 // to avoid range checks
    {
        EH_ERROR(Vec2D_errors, ERROR_VEC2D_SINGULAR_SCALE, "singular scale");
        return;
    }

    c.x0(1.0/c.x0());
    c.x1(1.0/c.x1());
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const Scale2D & scale)
{
    s << "Scale2D(" << scale.c.x0() << "," << scale.c.x1() << ")";
    return s;
}

INLINE_GLOBAL Scale2D operator*(const Scale2D & scale1, const Scale2D & scale2)
{
    return Scale2D(cwProduct(scale1.c,scale2.c));
}


INLINE_GLOBAL Scale2D inverseCopy(const Scale2D & scale)
{
    if (scale.c.x0() == 0.0 ||
        scale.c.x1() == 0.0)
    {
        EH_ERROR(Vec2D_errors, ERROR_VEC2D_SINGULAR_SCALE, "singular scale");
        return scale;
    }

    return Scale2D(1.0/scale.c.x0(), 1.0/scale.c.x1());
}


// =============================================================== class Rot2D

INLINE_MEMBER Rot2D::Rot2D(
	)
:
    alpha(0.0),
    cosAlpha(1.0),
    sinAlpha(0.0)
{
    TR_FUN0(Vec2D_constructors,Rot2D);
}

INLINE_MEMBER Rot2D::Rot2D(double newAlpha)
:
    alpha(newAlpha),
    cosAlpha(cos(newAlpha)),
    sinAlpha(sin(newAlpha))
{
    TR_FUN1(Vec2D_constructors,Rot2D,newAlpha);
}

INLINE_MEMBER Rot2D::Rot2D(
	double newAlpha, double newCosAlpha, double newSinAlpha
	)
:
    alpha(newAlpha),
    cosAlpha(newCosAlpha),
    sinAlpha(newSinAlpha)
{
    TR_FUN3(Vec2D_constructors,Rot2D,newAlpha,newCosAlpha,newSinAlpha);
}

INLINE_MEMBER Rot2D::Rot2D(const Rot2D & rot)
:
    alpha(rot.alpha),
    cosAlpha(rot.cosAlpha),
    sinAlpha(rot.sinAlpha)
{
    TR_FUN1(Vec2D_constructors,Rot2D,rot);
}

INLINE_MEMBER Rot2D::~Rot2D()
{
    TR_MESSAGE(Vec2D_constructors,"~" << *this);
}

INLINE_MEMBER Rot2D & Rot2D::operator*=(const Rot2D & rot)
{
    alpha += rot.alpha;
    cosAlpha = cos(alpha);
    sinAlpha = sin(alpha);
    
    return *this;
}

INLINE_MEMBER double Rot2D::getAlpha() const
{
    return alpha;
}

INLINE_MEMBER void Rot2D::invert()
{
    alpha = -alpha;
    sinAlpha = -sinAlpha;
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const Rot2D & rot)
{
    s << "Rot2D(" << rot.alpha << ")";
    return s;
}

INLINE_GLOBAL Rot2D inverseCopy(const Rot2D & rot)
{
    return Rot2D(-rot.alpha,rot.cosAlpha,-rot.sinAlpha);
}

INLINE_GLOBAL Rot2D operator*(const Rot2D & rot1, const Rot2D & rot2)
{
    return Rot2D(rot1.alpha + rot2.alpha);
}

// ============================================================ class ShearX2D

INLINE_MEMBER ShearX2D::ShearX2D(
	)
:
    c(0.0)
{
    TR_FUN0(Vec2D_constructors,ShearX2D);
}

INLINE_MEMBER ShearX2D::ShearX2D(double cNew)
:
    c(cNew)
{
    TR_FUN1(Vec2D_constructors,ShearX2D,cNew);
}

INLINE_MEMBER ShearX2D::ShearX2D(const ShearX2D & shear)
:
    c(shear.c)
{
    TR_FUN1(Vec2D_constructors,ShearX2D,shear);
}

INLINE_MEMBER ShearX2D::~ShearX2D()
{
    TR_MESSAGE(Vec2D_constructors,"~" << *this);
}

INLINE_MEMBER ShearX2D & ShearX2D::operator*=(const ShearX2D & shear)
{
    c += shear.c;
    return *this;
}

INLINE_MEMBER void ShearX2D::invert()
{
    c = -c;
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const ShearX2D & shear)
{
    s << "ShearX2D(" << shear.c << ")";
    return s;
}

INLINE_GLOBAL ShearX2D inverseCopy(const ShearX2D & shear)
{
    return ShearX2D(-shear.c);
}

INLINE_GLOBAL ShearX2D operator*(const ShearX2D & shear1, const ShearX2D & shear2)
{
    return ShearX2D(shear1.c + shear2.c);
}

// ============================================================ class ShearY2D

INLINE_MEMBER ShearY2D::ShearY2D(
	)
:
    c(0.0)
{
    TR_FUN0(Vec2D_constructors,ShearY2D);
}

INLINE_MEMBER ShearY2D::ShearY2D(double cNew)
:
    c(cNew)
{
    TR_FUN1(Vec2D_constructors,ShearY2D,cNew);
}

INLINE_MEMBER ShearY2D::ShearY2D(const ShearY2D & shear)
:
    c(shear.c)
{
    TR_FUN1(Vec2D_constructors,ShearY2D,shear);
}

INLINE_MEMBER ShearY2D::~ShearY2D()
{
    TR_MESSAGE(Vec2D_constructors,"~" << *this);
}

INLINE_MEMBER ShearY2D & ShearY2D::operator*=(const ShearY2D & shear)
{
    c += shear.c;
    return *this;
}

INLINE_MEMBER void ShearY2D::invert()
{
    c = -c;
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const ShearY2D & shear)
{
    s << "ShearY2D(" << shear.c << ")";
    return s;
}

INLINE_GLOBAL ShearY2D inverseCopy(const ShearY2D & shear)
{
    return ShearY2D(-shear.c);
}

INLINE_GLOBAL ShearY2D operator*(const ShearY2D & shear1, const ShearY2D & shear2)
{
    return ShearY2D(shear1.c + shear2.c);
}

// ============================================================= class Lin2D

INLINE_MEMBER Lin2D::Lin2D()
:
    m(1.0,0.0, 0.0,1.0)
{
    TR_FUN0(Vec2D_constructors,Lin2D);
}

INLINE_MEMBER Lin2D::Lin2D(const Scale2D & scale)
:
    m(  scale.c.x0(),   0.0,
        0.0,            scale.c.x1()	)
{
    TR_FUN1(Vec2D_constructors,Lin2D,scale);
}

INLINE_MEMBER Lin2D::Lin2D(const Rot2D & rot)
:
    m(  rot.cosAlpha,   rot.sinAlpha,
        -rot.sinAlpha,  rot.cosAlpha	)
{
    TR_FUN1(Vec2D_constructors,Lin2D,rot);
}

INLINE_MEMBER Lin2D::Lin2D(const ShearX2D & shear)
:
    m(  1.0,	 0.0,
        shear.c, 1.0	)
{
    TR_FUN1(Vec2D_constructors,Lin2D,shear);
}

INLINE_MEMBER Lin2D::Lin2D(const ShearY2D & shear)
:
    m(  1.0,	 shear.c,
        0.0,	 1.0	)
{
    TR_FUN1(Vec2D_constructors,Lin2D,shear);
}

INLINE_MEMBER Lin2D::Lin2D(const Mat2 & mat)
:
    m(mat)
{
    TR_FUN1(Vec2D_constructors,Lin2D,mat);
}

INLINE_MEMBER Lin2D::Lin2D(const Lin2D & lin)
:
    m(lin.m)
{
    TR_FUN1(Vec2D_constructors,Lin2D,lin);
}

INLINE_MEMBER Lin2D::Lin2D(const Vec2D & vecX, const Vec2D & vecY)
:
    m(vecX.c.x0(), vecX.c.x1(),
      vecY.c.x0(), vecY.c.x1())
{
    TR_FUN2(Vec2D_constructors,Lin2D,vecX,vecY);
}

INLINE_MEMBER Lin2D::Lin2D(
        double xx, double xy,
        double yx, double yy
        )
:
    m(xx,xy, yx,yy)
{
    TR_FUN3(Vec2D_constructors,Lin2D,xx,xy,"...");
}

INLINE_MEMBER Lin2D::~Lin2D()
{
    TR_MESSAGE(Vec2D_constructors,"~" << *this);
}

INLINE_MEMBER Vec2D Lin2D::getVecX() const
{
    return Vec2D(m(0,0),m(0,1));
}

INLINE_MEMBER Vec2D Lin2D::getVecY() const
{
    return Vec2D(m(1,0),m(1,1));
}

INLINE_MEMBER Mat2 Lin2D::getMat2() const
{
    return m;
}

INLINE_MEMBER Lin2D & Lin2D::operator*=(const Scale2D & scale)
{
    m.scale(scale.c);
    return *this;
}

INLINE_MEMBER Lin2D & Lin2D::operator*=(const Rot2D & rot)
{
    m.rot(rot.cosAlpha,rot.sinAlpha);
    return *this;
}

INLINE_MEMBER Lin2D & Lin2D::operator*=(const ShearX2D & shear)
{
    m.shear0(shear.c);
    return *this;
}

INLINE_MEMBER Lin2D & Lin2D::operator*=(const ShearY2D & shear)
{
    m.shear1(shear.c);
    return *this;
}

INLINE_MEMBER Lin2D & Lin2D::operator*=(const Lin2D & lin)
{
    m *= lin.m;
    return *this;
}

INLINE_GLOBAL Vec2D getVecX(const Lin2D & lin) { return lin.getVecX(); }
INLINE_GLOBAL Vec2D getVecY(const Lin2D & lin) { return lin.getVecY(); }
INLINE_GLOBAL Mat2 getMat2(const Lin2D & lin) { return lin.m; }

INLINE_GLOBAL double determinant(const Lin2D & lin)
{
    return determinant(lin);
}

INLINE_GLOBAL Lin2D inverseCopy(const Lin2D & lin, double det)
{
    return Lin2D(inverseCopy(lin.m,det));
}

INLINE_MEMBER void Lin2D::invert(
        )
{
    *this = inverseCopy(*this,determinant(*this));
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const Lin2D & lin)
{
    s << "Lin2D(" << lin.m << ')';
    return s;
}

INLINE_GLOBAL Lin2D inverseCopy(const Lin2D & lin)
{
    return inverseCopy(lin,determinant(lin));
}

INLINE_GLOBAL Lin2D operator*(const Lin2D & lin, const Scale2D & scale)
{
    return Lin2D(prMatScale(lin.m,scale.c));
}

INLINE_GLOBAL Lin2D operator*(const Lin2D & lin, const Rot2D & rot)
{
    return Lin2D(prMatRot(lin.m,rot.cosAlpha,rot.sinAlpha));
}

INLINE_GLOBAL Lin2D operator*(const Lin2D & lin, const ShearX2D & shear)
{
    return Lin2D(prMatShear0(lin.m,shear.c));
}

INLINE_GLOBAL Lin2D operator*(const Lin2D & lin, const ShearY2D & shear)
{
    return Lin2D(prMatShear1(lin.m,shear.c));
}

INLINE_GLOBAL Lin2D operator*(const Lin2D & lin1, const Lin2D & lin2)
{
    return Lin2D(lin1.m * lin2.m);
}

INLINE_GLOBAL Lin2D operator*(const Scale2D & scale, const Lin2D & lin)
{
    return Lin2D(prScaleMat(scale.c,lin.m));
}

INLINE_GLOBAL Lin2D operator*(const Rot2D & rot, const Lin2D & lin)
{
    return Lin2D(prRotMat(rot.cosAlpha,rot.sinAlpha,lin.m));
}

INLINE_GLOBAL Lin2D operator*(const ShearX2D & shear, const Lin2D & lin)
{
    return Lin2D(prShear0Mat(shear.c, lin.m));
}

INLINE_GLOBAL Lin2D operator*(const ShearY2D & shear, const Lin2D & lin)
{
    return Lin2D(prShear1Mat(shear.c, lin.m));
}

// ============================================================= class Trans2D

INLINE_MEMBER Trans2D::Trans2D(
	)
:
    c(0.0,0.0)
{
    TR_FUN0(Vec2D_constructors,Trans2D);
}

INLINE_MEMBER Trans2D::Trans2D(const Crd2 & crd)
:
    c(crd)
{
    TR_FUN1(Vec2D_constructors,Trans2D,crd);
}

INLINE_MEMBER Trans2D::Trans2D(double xNew, double yNew)
:
    c(xNew, yNew)
{
    TR_FUN2(Vec2D_constructors,Trans2D,xNew,yNew);
}

INLINE_MEMBER Trans2D::Trans2D(
	const Vec2D & vec
	)
:
    c(vec.c)
{
    TR_FUN1(Vec2D_constructors,Trans2D,vec);
}

INLINE_MEMBER Trans2D::Trans2D(
	const Pnt2D & pntO
	)
:
    c(pntO.c)
{
    TR_FUN1(Vec2D_constructors,Trans2D, pntO);
}

INLINE_MEMBER Trans2D::Trans2D(const Trans2D & trans)
:
    c(trans.c)
{
    TR_FUN1(Vec2D_constructors,Trans2D,"Trans2D");
}

INLINE_MEMBER Trans2D::~Trans2D()
{
    TR_MESSAGE(Vec2D_constructors, "~" << *this);
}

INLINE_MEMBER Trans2D & Trans2D::operator*=(const Trans2D & trans)
{
    c += trans.c;
    return *this;
}

INLINE_MEMBER void Trans2D::invert(
        )
{
    c = -c;
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const Trans2D & trans)
{
    s << "Trans2D(" << trans.c.x0() << "," << trans.c.x1() << ")";
    return s;
}

INLINE_GLOBAL Trans2D operator*(const Trans2D & trans1, const Trans2D & trans2)
{
    return Trans2D(trans1.c + trans2.c);
}

INLINE_GLOBAL Trans2D inverseCopy(const Trans2D & trans)
{
    return Trans2D(-trans.c);
}

// ============================================================= class Trafo2D

INLINE_MEMBER Trafo2D::Trafo2D()
:
    m(1.0,0.0, 0.0,1.0), c(0.0,0.0)
{
    TR_FUN0(Vec2D_constructors,Trafo2D);
}

INLINE_MEMBER Trafo2D::Trafo2D(const Scale2D & scale)
:
    m(  scale.c.x0(),   0.0,
        0.0,            scale.c.x1()	),
    c(  0.0,            0.0		)
{
    TR_FUN1(Vec2D_constructors,Trafo2D,scale);
}

INLINE_MEMBER Trafo2D::Trafo2D(const Rot2D & rot)
:
    m(  rot.cosAlpha,   rot.sinAlpha,
        -rot.sinAlpha,  rot.cosAlpha	),
    c(  0.0,            0.0		)
{
    TR_FUN1(Vec2D_constructors,Trafo2D,rot);
}

INLINE_MEMBER Trafo2D::Trafo2D(const ShearX2D & shear)
:
    m(  1.0,	 0.0,
        shear.c, 1.0	),
    c(  0.0,	 0.0	)
{
    TR_FUN1(Vec2D_constructors,Trafo2D,shear);
}

INLINE_MEMBER Trafo2D::Trafo2D(const ShearY2D & shear)
:
    m(  1.0,	 shear.c,
        0.0,	 1.0	),
    c(  0.0,	 0.0	)
{
    TR_FUN1(Vec2D_constructors,Trafo2D,shear);
}

INLINE_MEMBER Trafo2D::Trafo2D(const Lin2D & lin)
:
    m(lin.m), c(0.0,0.0)
{
    TR_FUN1(Vec2D_constructors,Trafo2D,lin);
}

INLINE_MEMBER Trafo2D::Trafo2D(const Lin2D & lin, const Trans2D & trans)
:
    m(lin.m), c(trans.c)
{
    TR_FUN2(Vec2D_constructors,Trafo2D,lin,trans);
}

INLINE_MEMBER Trafo2D::Trafo2D(const Mat2 & mat)
:
    m(mat), c(0.0,0.0)
{
    TR_FUN1(Vec2D_constructors,Trafo2D,mat);
}

INLINE_MEMBER Trafo2D::Trafo2D(const Mat2 & mat, const Crd2 & crd)
:
    m(mat), c(crd)
{
    TR_FUN2(Vec2D_constructors,Trafo2D,mat,crd);
}

INLINE_MEMBER Trafo2D::Trafo2D(const Trans2D & trans)
:
    m(1.0,0.0, 0.0,1.0), c(trans.c)
{
    TR_FUN1(Vec2D_constructors,Trafo2D,trans);
}

INLINE_MEMBER Trafo2D::Trafo2D(const Trafo2D & trafo)
:
    m(trafo.m), c(trafo.c)
{
    TR_FUN1(Vec2D_constructors,Trafo2D,trafo);
}

INLINE_MEMBER Trafo2D::Trafo2D(const Vec2D & vx, const Vec2D & vy,
                    const Pnt2D & pnt)
:
    m(vx.c.x0(),vx.c.x1(),
      vy.c.x0(),vy.c.x1()),
    c(pnt.c)
{
    TR_FUN3(Vec2D_constructors,Trafo2D,vx,vy,pnt);
}

INLINE_MEMBER Trafo2D::Trafo2D(
        double xx, double xy,
        double yx, double yy,
        double dx, double dy
        )
:
    m(xx,xy, yx,yy), c(dx,dy)
{
    TR_FUN3(Vec2D_constructors,Trafo2D,xx,xy,"...");
}

INLINE_MEMBER Trafo2D::~Trafo2D()
{
    TR_MESSAGE(Vec2D_constructors,"~" << *this);
}

INLINE_MEMBER Vec2D Trafo2D::getVecX() const
{
    return Vec2D(m(0,0),m(0,1));
}

INLINE_MEMBER Vec2D Trafo2D::getVecY() const
{
    return Vec2D(m(1,0),m(1,1));
}

INLINE_MEMBER Pnt2D Trafo2D::getPntO() const
{
    return Pnt2D(c);
}

INLINE_MEMBER Mat2 Trafo2D::getMat2() const
{
    return m;
}

INLINE_MEMBER Mat3 Trafo2D::getMat3() const
{
    return Mat3(m(0,0), m(0,1), 0.0,
		m(1,0), m(1,1), 0.0,
		c.x0(), c.x1(), 1.0);
}

INLINE_MEMBER Trafo2D & Trafo2D::operator*=(const Scale2D & scale)
{
    m.scale(scale.c);
    c.scale(scale.c);
    return *this;
}

INLINE_MEMBER Trafo2D & Trafo2D::operator*=(const Rot2D & rot)
{
    m.rot(rot.cosAlpha,rot.sinAlpha);
    c.rot(rot.cosAlpha,rot.sinAlpha);
    return *this;
}

INLINE_MEMBER Trafo2D & Trafo2D::operator*=(const ShearX2D & shear)
{
    m.shear0(shear.c);
    c.shear0(shear.c);
    return *this;
}

INLINE_MEMBER Trafo2D & Trafo2D::operator*=(const ShearY2D & shear)
{
    m.shear1(shear.c);
    c.shear1(shear.c);
    return *this;
}

INLINE_MEMBER Trafo2D & Trafo2D::operator*=(const Lin2D & lin)
{
    m *= lin.m;
    c *= lin.m;
    return *this;
}

INLINE_MEMBER Trafo2D & Trafo2D::operator*=(const Trans2D & trans)
{
    c += trans.c;
    return *this;
}

INLINE_MEMBER Trafo2D & Trafo2D::operator*=(const Trafo2D & trafo)
{
    m *= trafo.m;
    c *= trafo.m;
    c += trafo.c;
    return *this;
}

INLINE_GLOBAL Vec2D getVecX(const Trafo2D & trafo) { return trafo.getVecX(); }
INLINE_GLOBAL Vec2D getVecY(const Trafo2D & trafo) { return trafo.getVecY(); }
INLINE_GLOBAL Pnt2D getPntO(const Trafo2D & trafo) { return trafo.getPntO(); }
INLINE_GLOBAL Mat2 getMat2(const Trafo2D & trafo) { return trafo.m; }
INLINE_GLOBAL Mat3 getMat3(const Trafo2D & trafo) { return trafo.getMat3(); }

INLINE_GLOBAL double determinant(const Trafo2D & trafo)
{
    return determinant(trafo);
}

INLINE_GLOBAL Trafo2D inverseCopy(const Trafo2D & trafo, double det)
{
    Trafo2D inv;

    inv.m = inverseCopy(trafo.m,det);

    inv.c = - (trafo.c * inv.m);
                 
    return inv;
}

INLINE_MEMBER void Trafo2D::invert(
        )
{
    double det = determinant(*this);
    *this = inverseCopy(*this,det);
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const Trafo2D & trafo)
{
    s << "Trafo2D(" << trafo.m << ',' << trafo.c << ')';
    return s;
}

INLINE_GLOBAL Trafo2D inverseCopy(const Trafo2D & trafo)
{
    double det = determinant(trafo);
    return inverseCopy(trafo,det);
}

INLINE_GLOBAL Trafo2D operator*(const Trafo2D & trafo, const Scale2D & scale)
{
    return Trafo2D(prMatScale(trafo.m,scale.c),
                   prCrdScale(trafo.c,scale.c));
}

INLINE_GLOBAL Trafo2D operator*(const Trafo2D & trafo, const Rot2D & rot)
{
    return Trafo2D(prMatRot(trafo.m,rot.cosAlpha,rot.sinAlpha),
                   prCrdRot(trafo.c,rot.cosAlpha,rot.sinAlpha));
}

INLINE_GLOBAL Trafo2D operator*(const Trafo2D & trafo, const ShearX2D & shear)
{
    return Trafo2D(prMatShear0(trafo.m,shear.c),
                   prCrdShear0(trafo.c,shear.c));
}

INLINE_GLOBAL Trafo2D operator*(const Trafo2D & trafo, const ShearY2D & shear)
{
    return Trafo2D(prMatShear1(trafo.m,shear.c),
                   prCrdShear1(trafo.c,shear.c));
}

INLINE_GLOBAL Trafo2D operator*(const Trafo2D & trafo, const Lin2D & lin)
{
    return Trafo2D(trafo.m * lin.m, trafo.c * lin.m);
}

INLINE_GLOBAL Trafo2D operator*(const Trafo2D & trafo, const Trans2D & trans)
{
    return Trafo2D(trafo.m, trafo.c + trans.c);
}

INLINE_GLOBAL Trafo2D operator*(const Trafo2D & trafo1, const Trafo2D & trafo2)
{
    return Trafo2D(trafo1.m * trafo2.m, trafo1.c * trafo2.m + trafo2.c);
}

INLINE_GLOBAL Trafo2D operator*(const Scale2D & scale, const Trafo2D & trafo)
{
    return Trafo2D(prScaleMat(scale.c,trafo.m),trafo.c);
}

INLINE_GLOBAL Trafo2D operator*(const Rot2D & rot, const Trafo2D & trafo)
{
    return Trafo2D(prRotMat(rot.cosAlpha,rot.sinAlpha,trafo.m),trafo.c);
}

INLINE_GLOBAL Trafo2D operator*(const ShearX2D & shear, const Trafo2D & trafo)
{
    return Trafo2D(prShear0Mat(shear.c,trafo.m),trafo.c);
}

INLINE_GLOBAL Trafo2D operator*(const ShearY2D & shear, const Trafo2D & trafo)
{
    return Trafo2D(prShear1Mat(shear.c,trafo.m),trafo.c);
}

INLINE_GLOBAL Trafo2D operator*(const Lin2D & lin, const Trafo2D & trafo)
{
    return Trafo2D(lin.m * trafo.m,trafo.c);
}

INLINE_GLOBAL Trafo2D operator*(const Trans2D & trans, const Trafo2D & trafo)
{
    return Trafo2D(trafo.m,trans.c * trafo.m + trafo.c);
}

} // extern "C++"

// ===========================================================================

