// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       Vec3D
//  TYPE:       c++ inline
//  PROJECT:    VEGA Graphics Library
//  CONTENT:    3D vectors, points, and transformations using doubles
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rft     Robert F. Tobler
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 14:08:16  rft     last modification
//  08-Aug-93           rft     created
// ===========================================================================

extern "C++"
{

// =============================================================== class Vec3D

// -------------------------------------------------------------- constructors

INLINE_MEMBER Vec3D::Vec3D()
{
    TR_FUN0(Vec3D_constructors,Vec3D);
}

INLINE_MEMBER Vec3D::Vec3D(double xNew, double yNew, double zNew)
:
    c(xNew,yNew,zNew)
{
    TR_FUN3(Vec3D_constructors,Vec3D,xNew,yNew,zNew);
}

INLINE_MEMBER Vec3D::Vec3D(const Crd3 & crd)
:
    c(crd)
{
    TR_FUN1(Vec3D_constructors,Vec3D,crd);
}

INLINE_MEMBER Vec3D::Vec3D(const Vec3D & vec)
:
    c(vec.c)
{
    TR_FUN1(Vec3D_constructors,Vec3D,vec);
}

INLINE_MEMBER Vec3D::~Vec3D()
{
    TR_MESSAGE(Vec3D_constructors,"~" << *this);
}

// ---- methods --------------------------------------------------------------

INLINE_MEMBER double Vec3D::operator[](int index) const
{
    return c[index];
}

INLINE_MEMBER double & Vec3D::operator[](int index)
{
    return c[index];
}

INLINE_MEMBER Vec3D & Vec3D::operator+=(const Vec3D & vec)
{
    c += vec.c;
    return *this;
}

INLINE_MEMBER Vec3D & Vec3D::operator-=(const Vec3D & vec)
{
    c -= vec.c;
    return *this;
}

INLINE_MEMBER Vec3D & Vec3D::operator*=(double scalar)
{
    c *= scalar;
    return *this;
}

INLINE_MEMBER double Vec3D::x() const { return c.x0(); }
INLINE_MEMBER double Vec3D::y() const { return c.x1(); }
INLINE_MEMBER double Vec3D::z() const { return c.x2(); }

INLINE_MEMBER void Vec3D::x(double xNew) { c.x0(xNew); }
INLINE_MEMBER void Vec3D::y(double yNew) { c.x1(yNew); }
INLINE_MEMBER void Vec3D::z(double zNew) { c.x2(zNew); }

INLINE_MEMBER void Vec3D::normalize()
{
    c.normalize();
}

INLINE_MEMBER Crd3 Vec3D::getCrd3() const
{
    return c;
}

INLINE_MEMBER Crd4 Vec3D::getCrd4() const
{
    return Crd4(c.x0(),c.x1(),c.x2(),0.0);
}

INLINE_MEMBER Pnt3D::Pnt3D(const Crd3 & crd)
:
    c(crd)
{
    TR_FUN1(Vec3D_constructors,Pnt3D,crd);
}

INLINE_MEMBER Pnt3D Vec3D::getPnt3D() const
{
    return Pnt3D(c);
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const Vec3D & vec)
{
    s << "Vec3D(" << vec.x() << "," << vec.y() << "," << vec.z() << ")";
    return s;
}

INLINE_GLOBAL Vec3D operator-(const Vec3D & vec)
{
    return Vec3D(-vec.c);
}

INLINE_GLOBAL Vec3D operator+(const Vec3D & vec1, const Vec3D & vec2)
{
    return Vec3D(vec1.c + vec2.c);
}

INLINE_GLOBAL Pnt3D operator+(const Pnt3D & pnt, const Vec3D & vec)
{
    return Pnt3D(pnt.c + vec.c);
}

INLINE_GLOBAL Vec3D operator-(const Vec3D & vec1, const Vec3D & vec2)
{
    return Vec3D(vec1.c - vec2.c);
}

INLINE_GLOBAL Pnt3D operator-(const Pnt3D & pnt, const Vec3D & vec)
{
    return Pnt3D(pnt.c - vec.c);
}

INLINE_GLOBAL Vec3D operator*(const Vec3D & vec, double scalar)
{
    return Vec3D(vec.c * scalar);
}

INLINE_GLOBAL Vec3D operator*(double scalar, const Vec3D & vec)
{
    return Vec3D(scalar * vec.c);
}

INLINE_GLOBAL Crd3 getCrd3(const Vec3D & vec)
{
    return vec.c;
}

INLINE_GLOBAL Crd4 getCrd4(const Vec3D & vec)
{
    return Crd4(vec.c.x0(), vec.c.x1(), vec.c.x2(),0.0);
}

INLINE_GLOBAL Pnt3D getPnt3D(const Vec3D & vec)
{
    return Pnt3D(vec.c);
}

INLINE_GLOBAL double squaredLength(const Vec3D & vec)
{
    return squaredLength(vec.c);
}

INLINE_GLOBAL double length(const Vec3D & vec)
{
    return length(vec.c);
}

INLINE_GLOBAL double dotProduct(const Vec3D & vec1, const Vec3D & vec2)
{
    return dotProduct(vec1.c,vec2.c);
}

INLINE_GLOBAL Vec3D crossProduct(const Vec3D & vec1, const Vec3D & vec2)
{
    return Vec3D(crossProduct(vec1.c,vec2.c));
}

INLINE_GLOBAL Vec3D normalizedCopy(const Vec3D & vec)
{
    return Vec3D(normalizedCopy(vec.c));
}

INLINE_GLOBAL Vec3D cwReciprocalCopy(const Vec3D & vec)
{
    return Vec3D(cwReciprocalCopy(vec.c));
}

INLINE_GLOBAL Vec3D cwMin(const Vec3D & vec1, const Vec3D & vec2)
{
    return Vec3D(cwMin(vec1.c,vec2.c));
}

INLINE_GLOBAL Vec3D cwMax(const Vec3D & vec1, const Vec3D & vec2)
{
    return Vec3D(cwMax(vec1.c,vec2.c));
}

INLINE_MEMBER Vec3D & Vec3D::operator*=(const Scale3D & scale)
{
    c.scale(scale.c);
    return *this;
}

INLINE_MEMBER Vec3D & Vec3D::operator*=(const RotX3D & rot)
{
    c.rot0(rot.cosAlpha,rot.sinAlpha);
    return *this;
}

INLINE_MEMBER Vec3D & Vec3D::operator*=(const RotY3D & rot)
{
    c.rot1(rot.cosAlpha,rot.sinAlpha);
    return *this;
}

INLINE_MEMBER Vec3D & Vec3D::operator*=(const RotZ3D & rot)
{
    c.rot2(rot.cosAlpha,rot.sinAlpha);
    return *this;
}

INLINE_MEMBER Vec3D & Vec3D::operator*=(const ShearXY3D & shear)
{
    c.shear01(shear.c.x0(),shear.c.x1());
    return *this;
}

INLINE_MEMBER Vec3D & Vec3D::operator*=(const ShearZX3D & shear)
{
    c.shear20(shear.c.x0(),shear.c.x1());
    return *this;
}

INLINE_MEMBER Vec3D & Vec3D::operator*=(const ShearYZ3D & shear)
{
    c.shear12(shear.c.x0(),shear.c.x1());
    return *this;
}

INLINE_MEMBER Vec3D & Vec3D::operator*=(const Lin3D & lin)
{
    c *= lin.m;
    return *this;
}

INLINE_MEMBER Vec3D & Vec3D::operator*=(const Trafo3D & trafo)
{
    c *= trafo.m;
    return *this;
}

INLINE_GLOBAL Vec3D operator*(const Vec3D & vec, const Scale3D & scale)
{
    return Vec3D(prCrdScale(vec.c,scale.c));
}

INLINE_GLOBAL Vec3D operator*(const Vec3D & vec, const RotX3D & rot)
{
    return Vec3D(prCrdRot0(vec.c,rot.cosAlpha,rot.sinAlpha));
}

INLINE_GLOBAL Vec3D operator*(const Vec3D & vec, const RotY3D & rot)
{
    return Vec3D(prCrdRot1(vec.c,rot.cosAlpha,rot.sinAlpha));
}

INLINE_GLOBAL Vec3D operator*(const Vec3D & vec, const RotZ3D & rot)
{
    return Vec3D(prCrdRot2(vec.c,rot.cosAlpha,rot.sinAlpha));
}

INLINE_GLOBAL Vec3D operator*(const Vec3D & vec, const ShearXY3D & shear)
{
    return Vec3D(prCrdShear01(vec.c,shear.c.x0(),shear.c.x1()));
}

INLINE_GLOBAL Vec3D operator*(const Vec3D & vec, const ShearZX3D & shear)
{
    return Vec3D(prCrdShear20(vec.c,shear.c.x0(),shear.c.x1()));
}

INLINE_GLOBAL Vec3D operator*(const Vec3D & vec, const ShearYZ3D & shear)
{
    return Vec3D(prCrdShear12(vec.c,shear.c.x0(),shear.c.x1()));
}

INLINE_GLOBAL Vec3D operator*(const Vec3D & vec, const Lin3D & lin)
{
    return Vec3D(vec.c * lin.m);
}

INLINE_GLOBAL Vec3D operator*(const Vec3D & vec, const Trafo3D & trafo)
{
    return Vec3D(vec.c * trafo.m);
}

// =============================================================== class Pnt3D

// -------------------------------------------------------------- constructors

INLINE_MEMBER Pnt3D::Pnt3D()
{
    TR_FUN0(Vec3D_constructors,Pnt3D);
}

INLINE_MEMBER Pnt3D::Pnt3D(double xNew, double yNew, double zNew)
:
    c(xNew,yNew,zNew)
{
    TR_FUN3(Vec3D_constructors,Pnt3D,xNew,yNew,zNew);
}

INLINE_MEMBER Pnt3D::Pnt3D(const Pnt3D & pnt)
:
    c(pnt.c)
{
    TR_FUN1(Vec3D_constructors,Pnt3D,pnt);
}

INLINE_MEMBER Pnt3D::~Pnt3D()
{
    TR_MESSAGE(Vec3D_constructors,"~" << *this);
}

// ---------------------------------------------------------- member functions

INLINE_MEMBER double Pnt3D::operator[](int index) const
{
    return c[index];
}

INLINE_MEMBER double & Pnt3D::operator[](int index)
{
    return c[index];
}

INLINE_MEMBER Pnt3D & Pnt3D::operator+=(const Vec3D & vec)
{
    c += vec.c;
    return *this;
}

INLINE_MEMBER Pnt3D & Pnt3D::operator-=(const Vec3D & vec)
{
    c -= vec.c;
    return *this;
}

INLINE_MEMBER double Pnt3D::x() const { return c.x0(); }
INLINE_MEMBER double Pnt3D::y() const { return c.x1(); }
INLINE_MEMBER double Pnt3D::z() const { return c.x2(); }

INLINE_MEMBER void Pnt3D::x(double xNew) { c.x0(xNew); }
INLINE_MEMBER void Pnt3D::y(double yNew) { c.x1(yNew); }
INLINE_MEMBER void Pnt3D::z(double zNew) { c.x2(zNew); }

INLINE_MEMBER Crd3 Pnt3D::getCrd3() const
{
    return c;
}

INLINE_MEMBER Crd4 Pnt3D::getCrd4() const
{
    return Crd4(c.x0(),c.x1(),c.x2(),0.0);
}

INLINE_MEMBER Vec3D Pnt3D::getVec3D() const
{
    return Vec3D(c);
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const Pnt3D & vec)
{
    s << "(" << vec.x() << "," << vec.y() << "," << vec.z() << ")";
    return s;
}

INLINE_GLOBAL Vec3D operator-(const Pnt3D & pnt1, const Pnt3D & pnt2)
{
    return Vec3D(pnt1.c - pnt2.c);
}

INLINE_GLOBAL Crd3 getCrd3(const Pnt3D & pnt)
{
    return pnt.c;
}

INLINE_GLOBAL Crd4 getCrd4(const Pnt3D & pnt)
{
    return Crd4(pnt.c.x0(), pnt.c.x1(), pnt.c.x2(),0.0);
}

INLINE_GLOBAL Vec3D getVec3D(const Pnt3D & pnt)
{
    return Vec3D(pnt.c);
}

INLINE_GLOBAL double squaredDistanceFromOrigin(const Pnt3D & pnt)
{
    return (squaredLength(pnt.c));
}

INLINE_GLOBAL double distanceFromOrigin(const Pnt3D & pnt)
{
    return (length(pnt.c));
}

INLINE_GLOBAL double squaredDistance(const Pnt3D & pnt1, const Pnt3D & pnt2)
{
    return (squaredLength(pnt1.c - pnt2.c));
}

INLINE_GLOBAL double distance(const Pnt3D & pnt1, const Pnt3D & pnt2)
{
    return (length(pnt1.c - pnt2.c));
}

INLINE_GLOBAL Pnt3D cwMin(const Pnt3D & pnt1, const Pnt3D & pnt2)
{
    return Pnt3D(cwMin(pnt1.c,pnt2.c));
}

INLINE_GLOBAL Pnt3D cwMax(const Pnt3D & pnt1, const Pnt3D & pnt2)
{
    return Pnt3D(cwMax(pnt1.c,pnt2.c));
}

INLINE_MEMBER Pnt3D & Pnt3D::operator*=(const Scale3D & scale)
{
    c.scale(scale.c);
    return *this;
}

INLINE_MEMBER Pnt3D & Pnt3D::operator*=(const RotX3D & rot)
{
    c.rot0(rot.cosAlpha,rot.sinAlpha);
    return *this;
}

INLINE_MEMBER Pnt3D & Pnt3D::operator*=(const RotY3D & rot)
{
    c.rot0(rot.cosAlpha,rot.sinAlpha);
    return *this;
}

INLINE_MEMBER Pnt3D & Pnt3D::operator*=(const RotZ3D & rot)
{
    c.rot0(rot.cosAlpha,rot.sinAlpha);
    return *this;
}

INLINE_MEMBER Pnt3D & Pnt3D::operator*=(const ShearXY3D & shear)
{
    c.shear01(shear.c.x0(),shear.c.x1());
    return *this;
}

INLINE_MEMBER Pnt3D & Pnt3D::operator*=(const ShearZX3D & shear)
{
    c.shear20(shear.c.x0(),shear.c.x1());
    return *this;
}

INLINE_MEMBER Pnt3D & Pnt3D::operator*=(const ShearYZ3D & shear)
{
    c.shear12(shear.c.x0(),shear.c.x1());
    return *this;
}

INLINE_MEMBER Pnt3D & Pnt3D::operator*=(const Trans3D & trans)
{
    c += trans.c;
    return *this;
}

INLINE_MEMBER Pnt3D & Pnt3D::operator*=(const Lin3D & lin)
{
    c *= lin.m;
    return *this;
}

INLINE_MEMBER Pnt3D & Pnt3D::operator*=(const Trafo3D & trafo)
{
    c *= trafo.m;
    c += trafo.c;
    return *this;
}

INLINE_GLOBAL Pnt3D operator*(const Pnt3D & pnt, const Scale3D & scale)
{
    return Pnt3D(prCrdScale(pnt.c,scale.c));
}

INLINE_GLOBAL Pnt3D operator*(const Pnt3D & pnt, const RotX3D & rot)
{
    return Pnt3D(prCrdRot0(pnt.c,rot.cosAlpha,rot.sinAlpha));
}

INLINE_GLOBAL Pnt3D operator*(const Pnt3D & pnt, const RotY3D & rot)
{
    return Pnt3D(prCrdRot1(pnt.c,rot.cosAlpha,rot.sinAlpha));
}

INLINE_GLOBAL Pnt3D operator*(const Pnt3D & pnt, const RotZ3D & rot)
{
    return Pnt3D(prCrdRot2(pnt.c,rot.cosAlpha,rot.sinAlpha));
}

INLINE_GLOBAL Pnt3D operator*(const Pnt3D & pnt, const ShearXY3D & shear)
{
    return Pnt3D(prCrdShear01(pnt.c,shear.c.x0(),shear.c.x1()));
}

INLINE_GLOBAL Pnt3D operator*(const Pnt3D & pnt, const ShearZX3D & shear)
{
    return Pnt3D(prCrdShear20(pnt.c,shear.c.x0(),shear.c.x1()));
}

INLINE_GLOBAL Pnt3D operator*(const Pnt3D & pnt, const ShearYZ3D & shear)
{
    return Pnt3D(prCrdShear12(pnt.c,shear.c.x0(),shear.c.x1()));
}

INLINE_GLOBAL Pnt3D operator*(const Pnt3D & pnt, const Trans3D & trans)
{
    return Pnt3D(pnt.c + trans.c);
}

INLINE_GLOBAL Pnt3D operator*(const Pnt3D & pnt, const Lin3D & lin)
{
    return Pnt3D(pnt.c * lin.m);
}

INLINE_GLOBAL Pnt3D operator*(const Pnt3D & pnt, const Trafo3D & trafo)
{
    return Pnt3D(pnt.c * trafo.m + trafo.c);
}

// ============================================================== class Scale3D

INLINE_MEMBER Scale3D::Scale3D(
	)
:
    c(1,1,1)
{
    TR_FUN0(Vec3D_constructors, Scale3D);
}

INLINE_MEMBER Scale3D::Scale3D(double scalar)
:
    c(scalar,scalar,scalar)
{
    TR_FUN1(Vec3D_constructors,Scale3D,scalar);
}

INLINE_MEMBER Scale3D::Scale3D(double x, double y, double z)
:
    c(x,y,z)
{
    TR_FUN3(Vec3D_constructors,Scale3D,x,y,z);
}

INLINE_MEMBER Scale3D::Scale3D(const Crd3 & scale)
:
    c(scale)
{
    TR_FUN1(Vec3D_constructors,Scale3D,scale);
}

INLINE_MEMBER Scale3D::Scale3D(const Scale3D & scale)
:
    c(scale.c)
{
    TR_FUN1(Vec3D_constructors,Scale3D,scale);
}

INLINE_MEMBER Scale3D::~Scale3D()
{
    TR_MESSAGE(Vec3D_constructors,"~" << *this);
}

INLINE_MEMBER Scale3D & Scale3D::operator*=(const Scale3D & scale)
{
    c.cwMultiplyBy(scale.c);
    return *this;
}

INLINE_MEMBER void Scale3D::invert()
{
    if (c.x0() == 0.0 ||                // use explicit accessor function
        c.x1() == 0.0 ||                // to avoid range checks
        c.x2() == 0.0)
    {
        EH_ERROR(Vec3D_errors, ERROR_VEC3D_SINGULAR_SCALE, "singular scale");
        return;
    }

    c.x0(1.0/c.x0());
    c.x1(1.0/c.x1());
    c.x2(1.0/c.x2());
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const Scale3D & scale)
{
    s << "Scale3D(" << scale.c.x0() << "," << scale.c.x1() << ","
      << scale.c.x2() << ")";
    return s;
}

INLINE_GLOBAL Scale3D operator*(const Scale3D & scale1, const Scale3D & scale2)
{
    return Scale3D(cwProduct(scale1.c,scale2.c));
}


INLINE_GLOBAL Scale3D inverseCopy(const Scale3D & scale)
{
    if (scale.c.x0() == 0.0 ||
        scale.c.x1() == 0.0 ||
        scale.c.x2() == 0.0)
    {
        EH_ERROR(Vec3D_errors, ERROR_VEC3D_SINGULAR_SCALE, "singular scale");
        return scale;
    }

    return Scale3D(1.0/scale.c.x0(), 1.0/scale.c.x1(), 1.0/scale.c.x2());
}


// =============================================================== class RotX3D

INLINE_MEMBER RotX3D::RotX3D(
	)
{
    TR_FUN0(Vec3D_constructors,RotX3D);
}

INLINE_MEMBER RotX3D::RotX3D(double newAlpha)
:
    alpha(newAlpha),
    cosAlpha(cos(newAlpha)),
    sinAlpha(sin(newAlpha))
{
    TR_FUN1(Vec3D_constructors,RotX3D,newAlpha);
}

INLINE_MEMBER RotX3D::RotX3D(
	double newAlpha, double newCosAlpha, double newSinAlpha
	)
:
    alpha(newAlpha),
    cosAlpha(newCosAlpha),
    sinAlpha(newSinAlpha)
{
    TR_FUN3(Vec3D_constructors,RotX3D,newAlpha,newCosAlpha,newSinAlpha);
}

INLINE_MEMBER RotX3D::RotX3D(const RotX3D & rot)
:
    alpha(rot.alpha),
    cosAlpha(rot.cosAlpha),
    sinAlpha(rot.sinAlpha)
{
    TR_FUN1(Vec3D_constructors,RotX3D,rot);
}

INLINE_MEMBER RotX3D::~RotX3D()
{
    TR_MESSAGE(Vec3D_constructors,"~" << *this);
}

INLINE_MEMBER RotX3D & RotX3D::operator*=(const RotX3D & rot)
{
    alpha += rot.alpha;
    cosAlpha = cos(alpha);
    sinAlpha = sin(alpha);
    
    return *this;
}

INLINE_MEMBER double RotX3D::getAlpha() const
{
    return alpha;
}

INLINE_MEMBER void RotX3D::invert()
{
    alpha = -alpha;
    sinAlpha = -sinAlpha;
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const RotX3D & rot)
{
    s << "RotX3D(" << rot.alpha << ")";
    return s;
}

INLINE_GLOBAL RotX3D inverseCopy(const RotX3D & rot)
{
    return RotX3D(-rot.alpha,rot.cosAlpha,-rot.sinAlpha);
}

INLINE_GLOBAL RotX3D operator*(const RotX3D & rot1, const RotX3D & rot2)
{
    return RotX3D(rot1.alpha + rot2.alpha);
}

// =============================================================== class RotY3D

INLINE_MEMBER RotY3D::RotY3D(
	)
{
    TR_FUN0(Vec3D_constructors,RotY3D);
}

INLINE_MEMBER RotY3D::RotY3D(double newAlpha)
:
    alpha(newAlpha),
    cosAlpha(cos(newAlpha)),
    sinAlpha(sin(newAlpha))
{
    TR_FUN1(Vec3D_constructors,RotY3D,newAlpha);
}

INLINE_MEMBER RotY3D::RotY3D(
	double newAlpha,double newCosAlpha, double newSinAlpha
	)
:
    alpha(newAlpha),
    cosAlpha(newCosAlpha),
    sinAlpha(newSinAlpha)
{
    TR_FUN3(Vec3D_constructors,RotY3D,newAlpha,newCosAlpha,newSinAlpha);
}

INLINE_MEMBER RotY3D::RotY3D(const RotY3D & rot)
:
    alpha(rot.alpha),
    cosAlpha(rot.cosAlpha),
    sinAlpha(rot.sinAlpha)
{
    TR_FUN1(Vec3D_constructors,RotY3D,rot);
}

INLINE_MEMBER RotY3D::~RotY3D()
{
    TR_MESSAGE(Vec3D_constructors,"~" << *this);
}

INLINE_MEMBER RotY3D & RotY3D::operator*=(const RotY3D & rot)
{
    alpha += rot.alpha;
    cosAlpha = cos(alpha);
    sinAlpha = sin(alpha);
    
    return *this;
}

INLINE_MEMBER double RotY3D::getAlpha() const
{
    return alpha;
}

INLINE_MEMBER void RotY3D::invert()
{
    alpha = -alpha;
    sinAlpha = -sinAlpha;
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const RotY3D & rot)
{
    s << "RotY3D(" << rot.alpha << ")";
    return s;
}

INLINE_GLOBAL RotY3D inverseCopy(const RotY3D & rot)
{
    return RotY3D(-rot.alpha,rot.cosAlpha,-rot.sinAlpha);
}

INLINE_GLOBAL RotY3D operator*(const RotY3D & rot1, const RotY3D & rot2)
{
    return RotY3D(rot1.alpha + rot2.alpha);
}

// =============================================================== class RotZ3D

INLINE_MEMBER RotZ3D::RotZ3D(
	)
{
    TR_FUN0(Vec3D_constructors,RotZ3D);
}

INLINE_MEMBER RotZ3D::RotZ3D(double newAlpha)
:
    alpha(newAlpha),
    cosAlpha(cos(newAlpha)),
    sinAlpha(sin(newAlpha))
{
    TR_FUN1(Vec3D_constructors,RotY3D,newAlpha);
}

INLINE_MEMBER RotZ3D::RotZ3D(
	double newAlpha, double newCosAlpha, double newSinAlpha
	)
:
    alpha(newAlpha),
    cosAlpha(newCosAlpha),
    sinAlpha(newSinAlpha)
{
    TR_FUN3(Vec3D_constructors,RotY3D,newAlpha,newCosAlpha,newSinAlpha);
}

INLINE_MEMBER RotZ3D::RotZ3D(const RotZ3D & rot)
:
    alpha(rot.alpha),
    cosAlpha(rot.cosAlpha),
    sinAlpha(rot.sinAlpha)
{
    TR_FUN1(Vec3D_constructors,RotY3D,rot);
}

INLINE_MEMBER RotZ3D::~RotZ3D()
{
    TR_MESSAGE(Vec3D_constructors,"~" << *this);
}

INLINE_MEMBER RotZ3D & RotZ3D::operator*=(const RotZ3D & rot)
{
    alpha += rot.alpha;
    cosAlpha = cos(alpha);
    sinAlpha = sin(alpha);

    return *this;
}

INLINE_MEMBER double RotZ3D::getAlpha() const
{
    return alpha;
}

INLINE_MEMBER void RotZ3D::invert()
{
    alpha = -alpha;
    sinAlpha = -sinAlpha;
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const RotZ3D & rot)
{
    s << "RotZ3D(" << rot.alpha << ")";
    return s;
}

INLINE_GLOBAL RotZ3D inverseCopy(const RotZ3D & rot)
{
    return RotZ3D(-rot.alpha,rot.cosAlpha,-rot.sinAlpha);
}

INLINE_GLOBAL RotZ3D operator*(const RotZ3D & rot1, const RotZ3D & rot2)
{
    return RotZ3D(rot1.alpha + rot2.alpha);
}

// =============================================================== class Rot3D

INLINE_MEMBER Rot3D::Rot3D(
        )
{
    TR_FUN0(Vec3D_constructors,Rot3D);
}

INLINE_MEMBER Rot3D::Rot3D(
        const Vec3D & vec,
        double alpha
        )
{
    alpha *= 0.5;			// quaternions always use half-angles

    double s = sin(alpha) / length(vec);
    c.x0(vec.x() * s);
    c.x1(vec.y() * s);
    c.x2(vec.z() * s);
    c.x3(- cos(alpha));
    TR_FUN2(Vec3D_constructors,Rot3D,vec,alpha);
}

INLINE_MEMBER Rot3D::Rot3D(
	const Vec3D & vec
	)
:
    c(vec.getCrd4())
{
    TR_FUN1(Vec3D_constructors,Rot3D,vec);
}

INLINE_MEMBER Rot3D::Rot3D(
    	double aNew,
	double bNew,
	double cNew,
	double dNew
        )
:
    c(aNew,bNew,cNew,dNew)
{
    TR_FUN4(Vec3D_constructors,Rot3D,aNew,bNew,cNew,dNew);
}

INLINE_MEMBER Rot3D::Rot3D(
        const Crd4 & rot
        )
:
    c(rot)
{
    TR_FUN1(Vec3D_constructors,Rot3D,rot);
}

INLINE_MEMBER Rot3D::Rot3D(const RotX3D & rot)
:
    c(sin(0.5 * rot.alpha), 0.0, 0.0, - cos(0.5 * rot.alpha))
{
    TR_FUN1(Vec3D_constructors,Rot3D,rot);
}

INLINE_MEMBER Rot3D::Rot3D(const RotY3D & rot)
:
    c(0.0, sin(0.5 * rot.alpha), 0.0, - cos(0.5 * rot.alpha))
{
    TR_FUN1(Vec3D_constructors,Rot3D,rot);
}

INLINE_MEMBER Rot3D::Rot3D(const RotZ3D & rot)
:
    c(0.0, 0.0, sin(0.5 * rot.alpha), - cos(0.5 * rot.alpha))
{
    TR_FUN1(Vec3D_constructors,Rot3D,rot);
}

INLINE_MEMBER Rot3D::Rot3D(const Rot3D & rot)
:
    c(rot.c)
{
    TR_FUN1(Vec3D_constructors,Rot3D,rot);
}

INLINE_MEMBER Rot3D::~Rot3D()
{
    TR_MESSAGE(Vec3D_constructors,"~" << *this);
}

INLINE_GLOBAL Rot3D operator*(const Rot3D & rot1, const Rot3D & rot2)
{
    return
        Rot3D
        (
	    Crd4
	    (
		rot1.c.x0() * rot2.c.x3() + rot1.c.x3() * rot2.c.x0() +
		rot1.c.x1() * rot2.c.x2() - rot1.c.x2() * rot2.c.x1(),
    
		rot1.c.x1() * rot2.c.x3() + rot1.c.x3() * rot2.c.x1() +
		rot1.c.x2() * rot2.c.x0() - rot1.c.x0() * rot2.c.x2(),
    
		rot1.c.x2() * rot2.c.x3() + rot1.c.x3() * rot2.c.x2() +
		rot1.c.x0() * rot2.c.x1() - rot1.c.x1() * rot2.c.x0(),
		
		rot1.c.x3() * rot2.c.x3() - rot1.c.x2() * rot2.c.x2() -
		rot1.c.x1() * rot2.c.x1() - rot1.c.x0() * rot2.c.x0()
	    )
        );
}

INLINE_MEMBER Rot3D & Rot3D::operator*=(const Rot3D & rot)
{
    *this = *this * rot;
    return *this;
}

INLINE_MEMBER double Rot3D::getAlpha() const
{
    return(2 * acos(-c.x3()));
}

INLINE_MEMBER Vec3D Rot3D::getVec3D() const
{
    return Vec3D(c.x0(),c.x1(),c.x2());
}

INLINE_MEMBER void Rot3D::invert()
{
    c.x3(-c.x3());
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const Rot3D & rot)
{
    s << "Rot3D(" << rot.c.x0() << "," << rot.c.x1() << "," << rot.c.x2()
      << "," << rot.c.x3() << ")";
    return s;
}

INLINE_GLOBAL Rot3D inverseCopy(const Rot3D & rot)
{
    return Rot3D(Crd4(rot.c.x0(),rot.c.x1(),rot.c.x2(),-rot.c.x3()));
}

// =========================================================== class ShearXY3D

INLINE_MEMBER ShearXY3D::ShearXY3D(
        )
{
    TR_FUN0(Vec3D_constructors,ShearXY3D);
}

INLINE_MEMBER ShearXY3D::ShearXY3D(const ShearXY3D & shear)
:
    c(shear.c)
{
    TR_FUN1(Vec3D_constructors,ShearXY3D,shear);
}

INLINE_MEMBER ShearXY3D::ShearXY3D(double xNew, double yNew)
:
    c(xNew,yNew)
{
    TR_FUN2(Vec3D_constructors,ShearXY3D,xNew,yNew);
}

INLINE_MEMBER ShearXY3D::ShearXY3D(const Crd2 & crd)
:
    c(crd)
{
    TR_FUN1(Vec3D_constructors,ShearXY3D,crd);
}

INLINE_MEMBER ShearXY3D::~ShearXY3D()
{
    TR_MESSAGE(Vec3D_constructors,"~" << *this);
}


INLINE_MEMBER ShearXY3D & ShearXY3D::operator*=(const ShearXY3D & shear)
{
    c += shear.c;
    return *this;
}

INLINE_MEMBER void ShearXY3D::invert()
{
    c = -c;
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const ShearXY3D & shear)
{
    s << "ShearXY3D(" << shear.c.x0() << ',' << shear.c.x1() << ')';
    return s;
}

INLINE_GLOBAL ShearXY3D operator*(const ShearXY3D & sh1, const ShearXY3D & sh2)
{
    return ShearXY3D(sh1.c + sh2.c);
}

INLINE_GLOBAL ShearXY3D inverseCopy(const ShearXY3D & shear)
{
    return ShearXY3D(-shear.c);
}

// =========================================================== class ShearZX3D

INLINE_MEMBER ShearZX3D::ShearZX3D(
        )
{
    TR_FUN0(Vec3D_constructors,ShearZX3D);
}

INLINE_MEMBER ShearZX3D::ShearZX3D(const ShearZX3D & shear)
:
    c(shear.c)
{
    TR_FUN1(Vec3D_constructors,ShearZX3D,shear);
}

INLINE_MEMBER ShearZX3D::ShearZX3D(double zNew, double xNew)
:
    c(zNew,xNew)
{
    TR_FUN2(Vec3D_constructors,ShearZX3D,zNew,xNew);
}

INLINE_MEMBER ShearZX3D::ShearZX3D(const Crd2 & crd)
:
    c(crd)
{
    TR_FUN1(Vec3D_constructors,ShearZX3D,crd);
}

INLINE_MEMBER ShearZX3D::~ShearZX3D()
{
    TR_MESSAGE(Vec3D_constructors,"~" << *this);
}


INLINE_MEMBER ShearZX3D & ShearZX3D::operator*=(const ShearZX3D & shear)
{
    c += shear.c;
    return *this;
}

INLINE_MEMBER void ShearZX3D::invert()
{
    c = -c;
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const ShearZX3D & shear)
{
    s << "ShearZX3D(" << shear.c.x0() << ',' << shear.c.x1() << ')';
    return s;
}

INLINE_GLOBAL ShearZX3D operator*(const ShearZX3D & sh1, const ShearZX3D & sh2)
{
    return ShearZX3D(sh1.c + sh2.c);
}

INLINE_GLOBAL ShearZX3D inverseCopy(const ShearZX3D & shear)
{
    return ShearZX3D(-shear.c);
}

// =========================================================== class ShearYZ3D

INLINE_MEMBER ShearYZ3D::ShearYZ3D(
        )
{
    TR_FUN0(Vec3D_constructors,ShearYZ3D);
}

INLINE_MEMBER ShearYZ3D::ShearYZ3D(const ShearYZ3D & shear)
:
    c(shear.c)
{
    TR_FUN1(Vec3D_constructors,ShearYZ3D,shear);
}

INLINE_MEMBER ShearYZ3D::ShearYZ3D(double yNew, double zNew)
:
    c(yNew,zNew)
{
    TR_FUN2(Vec3D_constructors,ShearYZ3D,yNew,zNew);
}

INLINE_MEMBER ShearYZ3D::ShearYZ3D(const Crd2 & crd)
:
    c(crd)
{
    TR_FUN1(Vec3D_constructors,ShearYZ3D,crd);
}

INLINE_MEMBER ShearYZ3D::~ShearYZ3D()
{
    TR_MESSAGE(Vec3D_constructors,"~" << *this);
}


INLINE_MEMBER ShearYZ3D & ShearYZ3D::operator*=(const ShearYZ3D & shear)
{
    c += shear.c;
    return *this;
}

INLINE_MEMBER void ShearYZ3D::invert()
{
    c = -c;
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const ShearYZ3D & shear)
{
    s << "ShearYZ3D(" << shear.c.x0() << ',' << shear.c.x1() << ')';
    return s;
}

INLINE_GLOBAL ShearYZ3D operator*(const ShearYZ3D & sh1, const ShearYZ3D & sh2)
{
    return ShearYZ3D(sh1.c + sh2.c);
}

INLINE_GLOBAL ShearYZ3D inverseCopy(const ShearYZ3D & shear)
{
    return ShearYZ3D(-shear.c);
}

// =============================================================== class Lin3D

INLINE_MEMBER Lin3D::Lin3D()
{
    TR_FUN0(Vec3D_constructors,Lin3D);
}

INLINE_MEMBER Lin3D::Lin3D(const Scale3D & scale)
:
    m(  scale.c.x0(),   0.0,            0.0,
        0.0,            scale.c.x1(),   0.0,
        0.0,            0.0,            scale.c.x2()    )
{
    TR_FUN1(Vec3D_constructors,Lin3D,scale);
}

INLINE_MEMBER Lin3D::Lin3D(const RotX3D & rot)
:
    m(  1.0,      0.0,            0.0,
        0.0,    rot.cosAlpha,   rot.sinAlpha,
        0.0,      -rot.sinAlpha,  rot.cosAlpha    )
{
    TR_FUN1(Vec3D_constructors,Lin3D,rot);
}

INLINE_MEMBER Lin3D::Lin3D(const RotY3D & rot)
:
    m(  rot.cosAlpha,   0.0,    -rot.sinAlpha,
        0.0,            1.0,    0.0,
        rot.sinAlpha,   0.0,    rot.cosAlpha    )
{
    TR_FUN1(Vec3D_constructors,Lin3D,rot);
}

INLINE_MEMBER Lin3D::Lin3D(const ShearXY3D & shear)
:
    m(  1.0,		0.0,		0.0,
        0.0,            1.0,		0.0,
        shear.c.x0(),	shear.c.x1(),	1.0    )
{
    TR_FUN1(Vec3D_constructors,Lin3D,shear);
}

INLINE_MEMBER Lin3D::Lin3D(const ShearZX3D & shear)
:
    m(  1.0,		0.0,		0.0,
        shear.c.x1(),	1.0,   shear.c.x0(),
        0.0,		0.0,		1.0    )
{
    TR_FUN1(Vec3D_constructors,Lin3D,shear);
}

INLINE_MEMBER Lin3D::Lin3D(const ShearYZ3D & shear)
:
    m(  1.0,   shear.c.x0(),   shear.c.x1(),
        0.0,            1.0,		0.0,
        0.0,		0.0,		1.0    )
{
    TR_FUN1(Vec3D_constructors,Lin3D,shear);
}

INLINE_MEMBER Lin3D::Lin3D(const RotZ3D & rot)
:
    m(  rot.cosAlpha,   rot.sinAlpha,   0.0,
        -rot.sinAlpha,  rot.cosAlpha,   0.0,
        0.0,            0.0,            1.0     )
{
    TR_FUN1(Vec3D_constructors,Lin3D,rot);
}

INLINE_MEMBER Lin3D::Lin3D(const Rot3D & rot)
{
    double aa = square(rot.c.x0());
    double ab2 = 2 * rot.c.x0() * rot.c.x1();
    double ac2 = 2 * rot.c.x0() * rot.c.x2();
    double ad2 = 2 * rot.c.x0() * rot.c.x3();
    double bb = square(rot.c.x1());
    double bc2 = 2 * rot.c.x1() * rot.c.x2();
    double bd2 = 2 * rot.c.x1() * rot.c.x3();
    double cc = square(rot.c.x2());
    double cd2 = 2 * rot.c.x2() * rot.c.x3();
    double dd = square(rot.c.x3());
    m = Mat3(aa - bb + dd - cc, ab2 - cd2, ac2 + bd2,
             ab2 + cd2, bb - aa + dd - cc, bc2 - ad2,
             ac2 - bd2, bc2 + ad2, cc - aa + dd - bb); 
    TR_FUN1(Vec3D_constructors,Lin3D,rot);
}

INLINE_MEMBER Lin3D::Lin3D(const Mat3 & mat)
:
    m(mat)
{
    TR_FUN1(Vec3D_constructors,Lin3D,mat);
}

INLINE_MEMBER Lin3D::Lin3D(const Lin3D & lin)
:
    m(lin.m)
{
    TR_FUN1(Vec3D_constructors,Lin3D,lin);
}

INLINE_MEMBER Lin3D::Lin3D(
	const Vec3D & vx, const Vec3D & vy, const Vec3D & vz
	)
:
    m(vx.c.x0(),vx.c.x1(),vx.c.x2(),
      vy.c.x0(),vy.c.x1(),vy.c.x2(),
      vz.c.x0(),vz.c.x1(),vz.c.x2())
{
    TR_FUN3(Vec3D_constructors,Lin3D,vx,vy,vz);
}

INLINE_MEMBER Lin3D::Lin3D(
        double xx, double xy, double xz,
        double yx, double yy, double yz,
        double zx, double zy, double zz
        )
:
    m(xx,xy,xz, yx,yy,yz, zx,zy,zz)
{
    TR_FUN4(Vec3D_constructors,Lin3D,xx,xy,xz,"...");
}

INLINE_MEMBER Lin3D::~Lin3D()
{
    TR_MESSAGE(Vec3D_constructors,"~" << *this);
}

INLINE_MEMBER Vec3D Lin3D::getVecX() const
{
    return Vec3D(m(0,0),m(0,1),m(0,2));
}

INLINE_MEMBER Vec3D Lin3D::getVecY() const
{
    return Vec3D(m(1,0),m(1,1),m(1,2));
}

INLINE_MEMBER Vec3D Lin3D::getVecZ() const
{
    return Vec3D(m(2,0),m(2,1),m(2,2));
}

INLINE_MEMBER Mat3 Lin3D::getMat3() const
{
    return m;
}

INLINE_MEMBER Lin3D & Lin3D::operator*=(const Scale3D & scale)
{
    m.scale(scale.c);
    return *this;
}

INLINE_MEMBER Lin3D & Lin3D::operator*=(const RotX3D & rot)
{
    m.rot0(rot.cosAlpha,rot.sinAlpha);
    return *this;
}

INLINE_MEMBER Lin3D & Lin3D::operator*=(const RotY3D & rot)
{
    m.rot1(rot.cosAlpha,rot.sinAlpha);
    return *this;
}

INLINE_MEMBER Lin3D & Lin3D::operator*=(const RotZ3D & rot)
{
    m.rot2(rot.cosAlpha,rot.sinAlpha);
    return *this;
}

INLINE_MEMBER Lin3D & Lin3D::operator*=(const ShearXY3D & shear)
{
    m.shear01(shear.c.x0(),shear.c.x1());
    return *this;
}

INLINE_MEMBER Lin3D & Lin3D::operator*=(const ShearZX3D & shear)
{
    m.shear20(shear.c.x0(),shear.c.x1());
    return *this;
}

INLINE_MEMBER Lin3D & Lin3D::operator*=(const ShearYZ3D & shear)
{
    m.shear12(shear.c.x0(),shear.c.x1());
    return *this;
}

INLINE_MEMBER Lin3D & Lin3D::operator*=(const Lin3D & lin)
{
    m *= lin.m;
    return *this;
}

INLINE_GLOBAL Vec3D getVecX(const Lin3D & lin) { return lin.getVecX(); }
INLINE_GLOBAL Vec3D getVecY(const Lin3D & lin) { return lin.getVecY(); }
INLINE_GLOBAL Vec3D getVecZ(const Lin3D & lin) { return lin.getVecZ(); }
INLINE_GLOBAL Mat3 getMat3(const Lin3D & lin) { return lin.m; }

INLINE_GLOBAL double determinant(const Lin3D & lin)
{
    return determinant(lin);
}

INLINE_GLOBAL Lin3D inverseCopy(const Lin3D & lin, double det)
{
    return Lin3D(inverseCopy(lin.m,det));
}

INLINE_MEMBER void Lin3D::invert(
        )
{
    *this = inverseCopy(*this,determinant(*this));
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const Lin3D & lin)
{
    s << "Lin3D(" << lin.m << ')';
    return s;
}

INLINE_GLOBAL Lin3D inverseCopy(const Lin3D & lin)
{
    return inverseCopy(lin,determinant(lin));
}

INLINE_GLOBAL Lin3D operator*(const Lin3D & lin, const Scale3D & scale)
{
    return Lin3D(prMatScale(lin.m,scale.c));
}

INLINE_GLOBAL Lin3D operator*(const Lin3D & lin, const RotX3D & rot)
{
    return Lin3D(prMatRot0(lin.m,rot.cosAlpha,rot.sinAlpha));
}

INLINE_GLOBAL Lin3D operator*(const Lin3D & lin, const RotY3D & rot)
{
    return Lin3D(prMatRot1(lin.m,rot.cosAlpha,rot.sinAlpha));
}

INLINE_GLOBAL Lin3D operator*(const Lin3D & lin, const RotZ3D & rot)
{
    return Lin3D(prMatRot2(lin.m,rot.cosAlpha,rot.sinAlpha));
}

INLINE_GLOBAL Lin3D operator*(const Lin3D & lin, const ShearXY3D & shear)
{
    return Lin3D(prMatShear01(lin.m,shear.c.x0(),shear.c.x1()));
}

INLINE_GLOBAL Lin3D operator*(const Lin3D & lin, const ShearZX3D & shear)
{
    return Lin3D(prMatShear20(lin.m,shear.c.x0(),shear.c.x1()));
}

INLINE_GLOBAL Lin3D operator*(const Lin3D & lin, const ShearYZ3D & shear)
{
    return Lin3D(prMatShear12(lin.m,shear.c.x0(),shear.c.x1()));
}

INLINE_GLOBAL Lin3D operator*(const Lin3D & lin1, const Lin3D & lin2)
{
    return Lin3D(lin1.m * lin2.m);
}

INLINE_GLOBAL Lin3D operator*(const Scale3D & scale, const Lin3D & lin)
{
    return Lin3D(prScaleMat(scale.c,lin.m));
}

INLINE_GLOBAL Lin3D operator*(const RotX3D & rot, const Lin3D & lin)
{
    return Lin3D(prRot0Mat(rot.cosAlpha,rot.sinAlpha,lin.m));
}

INLINE_GLOBAL Lin3D operator*(const RotY3D & rot, const Lin3D & lin)
{
    return Lin3D(prRot1Mat(rot.cosAlpha,rot.sinAlpha,lin.m));
}

INLINE_GLOBAL Lin3D operator*(const RotZ3D & rot, const Lin3D & lin)
{
    return Lin3D(prRot2Mat(rot.cosAlpha,rot.sinAlpha,lin.m));
}

INLINE_GLOBAL Lin3D operator*(const ShearXY3D & shear, const Lin3D & lin)
{
    return Lin3D(prShear01Mat(shear.c.x0(),shear.c.x1(),lin.m));
}

INLINE_GLOBAL Lin3D operator*(const ShearZX3D & shear, const Lin3D & lin)
{
    return Lin3D(prShear20Mat(shear.c.x0(),shear.c.x1(),lin.m));
}

INLINE_GLOBAL Lin3D operator*(const ShearYZ3D & shear, const Lin3D & lin)
{
    return Lin3D(prShear12Mat(shear.c.x0(),shear.c.x1(),lin.m));
}

// ============================================================= class Trans3D

INLINE_MEMBER Trans3D::Trans3D(
	)
{
    TR_FUN0(Vec3D_constructors,Trans3D);
}

INLINE_MEMBER Trans3D::Trans3D(const Crd3 & crd)
:
    c(crd)
{
    TR_FUN1(Vec3D_constructors,Trans3D,crd);
}

INLINE_MEMBER Trans3D::Trans3D(double xNew, double yNew, double zNew)
:
    c(xNew, yNew, zNew)
{
    TR_FUN3(Vec3D_constructors,Trans3D,xNew,yNew,zNew);
}

INLINE_MEMBER Trans3D::Trans3D(
	const Vec3D & vec
	)
:
    c(vec.c)
{
    TR_FUN1(Vec3D_constructors,Trans3D,vec);
}

INLINE_MEMBER Trans3D::Trans3D(
	const Pnt3D & pntO
	)
:
    c(pntO.c)
{
    TR_FUN1(Vec3D_constructors,Trans3D, pntO);
}

INLINE_MEMBER Trans3D::Trans3D(const Trans3D & trans)
:
    c(trans.c)
{
    TR_FUN1(Vec3D_constructors,Trans3D,"Trans3D");
}

INLINE_MEMBER Trans3D::~Trans3D()
{
    TR_MESSAGE(Vec3D_constructors, "~" << *this);
}

INLINE_MEMBER Trans3D & Trans3D::operator*=(const Trans3D & trans)
{
    c += trans.c;
    return *this;
}

INLINE_MEMBER void Trans3D::invert(
        )
{
    c = -c;
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const Trans3D & trans)
{
    s << "Trans3D(" << trans.c.x0() << "," << trans.c.x1() << ","
      << trans.c.x2() << ")";
    return s;
}

INLINE_GLOBAL Trans3D operator*(const Trans3D & trans1, const Trans3D & trans2)
{
    return Trans3D(trans1.c + trans2.c);
}

INLINE_GLOBAL Trans3D inverseCopy(const Trans3D & trans)
{
    return Trans3D(-trans.c);
}

// ============================================================= class Trafo3D

INLINE_MEMBER Trafo3D::Trafo3D()
{
    TR_FUN0(Vec3D_constructors,Trafo3D);
}

INLINE_MEMBER Trafo3D::Trafo3D(const Trafo3D & trafo)
:
    m(trafo.m), c(trafo.c)
{
    TR_FUN1(Vec3D_constructors,Trafo3D,trafo);
}

INLINE_MEMBER Trafo3D::Trafo3D(const Mat3 & mat)
:
    m(mat), c(0.0,0.0,0.0)
{
    TR_FUN1(Vec3D_constructors,Trafo3D,mat);
}

INLINE_MEMBER Trafo3D::Trafo3D(const Mat3 & mat, const Crd3 & crd)
:
    m(mat), c(crd)
{
    TR_FUN2(Vec3D_constructors,Trafo3D,mat,crd);
}

INLINE_MEMBER Trafo3D::Trafo3D(
        double xx, double xy, double xz,
        double yx, double yy, double yz,
        double zx, double zy, double zz,
        double dx, double dy, double dz
        )
:
    m(xx,xy,xz, yx,yy,yz, zx,zy,zz), c(dx,dy,dz)
{
    TR_FUN4(Vec3D_constructors,Trafo3D,xx,xy,xz,"...");
}

INLINE_MEMBER Trafo3D::Trafo3D(
	const Vec3D & vx, const Vec3D & vy, const Vec3D & vz,
	const Pnt3D & pnt
	)
:
    m(vx.c.x0(),vx.c.x1(),vx.c.x2(),
      vy.c.x0(),vy.c.x1(),vy.c.x2(),
      vz.c.x0(),vz.c.x1(),vz.c.x2()),
    c(pnt.c)
{
    TR_FUN4(Vec3D_constructors,Trafo3D,vx,vy,vz,pnt);
}

INLINE_MEMBER Trafo3D::Trafo3D(const Lin3D & lin, const Trans3D & trans)
:
    m(lin.m), c(trans.c)
{
    TR_FUN2(Vec3D_constructors,Trafo3D,lin,trans);
}


INLINE_MEMBER Trafo3D::Trafo3D(const Scale3D & scale)
:
    m(  scale.c.x0(),   0.0,            0.0,
        0.0,            scale.c.x1(),   0.0,
        0.0,            0.0,            scale.c.x2()    ),
    c(  0.0,            0.0,            0.0             )
{
    TR_FUN1(Vec3D_constructors,Trafo3D,scale);
}

INLINE_MEMBER Trafo3D::Trafo3D(const RotX3D & rot)
:
    m(  1.0,    0.0,            0.0,
        0.0,    rot.cosAlpha,   rot.sinAlpha,
        0.0,    -rot.sinAlpha,  rot.cosAlpha    ),
    c(  0.0,    0.0,            0.0             )
{
    TR_FUN1(Vec3D_constructors,Trafo3D,rot);
}

INLINE_MEMBER Trafo3D::Trafo3D(const RotY3D & rot)
:
    m(  rot.cosAlpha,   0.0,    -rot.sinAlpha,
        0.0,            1.0,    0.0,
        rot.sinAlpha,   0.0,    rot.cosAlpha    ),
    c(  0.0,            0.0,    0.0             )
{
    TR_FUN1(Vec3D_constructors,Trafo3D,rot);
}

INLINE_MEMBER Trafo3D::Trafo3D(const RotZ3D & rot)
:
    m(  rot.cosAlpha,   rot.sinAlpha,   0.0,
        -rot.sinAlpha,  rot.cosAlpha,   0.0,
        0.0,            0.0,            1.0     ),
    c(  0.0,            0.0,            0.0     )
{
    TR_FUN1(Vec3D_constructors,Trafo3D,rot);
}

INLINE_MEMBER Trafo3D::Trafo3D(const ShearXY3D & shear)
:
    m(  1.0,		0.0,		0.0,
        0.0,            1.0,		0.0,
        shear.c.x0(),	shear.c.x1(),	1.0	),
    c(  0.0,            0.0,            0.0	)
{
    TR_FUN1(Vec3D_constructors,Lin3D,shear);
}

INLINE_MEMBER Trafo3D::Trafo3D(const ShearZX3D & shear)
:
    m(  1.0,		0.0,		0.0,
        shear.c.x1(),	1.0,   shear.c.x0(),
        0.0,		0.0,		1.0	),
    c(  0.0,            0.0,            0.0	)
{
    TR_FUN1(Vec3D_constructors,Lin3D,shear);
}

INLINE_MEMBER Trafo3D::Trafo3D(const ShearYZ3D & shear)
:
    m(  1.0,   shear.c.x0(),   shear.c.x1(),
        0.0,            1.0,		0.0,
        0.0,		0.0,		1.0	),
    c(  0.0,            0.0,            0.0	)
{
    TR_FUN1(Vec3D_constructors,Lin3D,shear);
}

INLINE_MEMBER Trafo3D::Trafo3D(const Rot3D & rot)
:
    c(0.0,0.0,0.0)
{
    double aa = square(rot.c.x0());
    double ab2 = 2 * rot.c.x0() * rot.c.x1();
    double ac2 = 2 * rot.c.x0() * rot.c.x2();
    double ad2 = 2 * rot.c.x0() * rot.c.x3();
    double bb = square(rot.c.x1());
    double bc2 = 2 * rot.c.x1() * rot.c.x2();
    double bd2 = 2 * rot.c.x1() * rot.c.x3();
    double cc = square(rot.c.x2());
    double cd2 = 2 * rot.c.x2() * rot.c.x3();
    double dd = square(rot.c.x3());
    m = Mat3(aa - bb + dd - cc, ab2 - cd2, ac2 + bd2,
             ab2 + cd2, bb - aa + dd - cc, bc2 - ad2,
             ac2 - bd2, bc2 + ad2, cc - aa + dd - bb); 
    TR_FUN1(Vec3D_constructors,Trafo3D,rot);
}

INLINE_MEMBER Trafo3D::Trafo3D(const Lin3D & lin)
:
    m(lin.m), c(0.0,0.0,0.0)
{
    TR_FUN1(Vec3D_constructors,Trafo3D,lin);
}

INLINE_MEMBER Trafo3D::Trafo3D(const Trans3D & trans)
:
    m(1.0,0.0,0.0, 0.0,1.0,0.0, 0.0,0.0,1.0), c(trans.c)
{
    TR_FUN1(Vec3D_constructors,Trafo3D,trans);
}

INLINE_MEMBER Trafo3D::~Trafo3D()
{
    TR_MESSAGE(Vec3D_constructors,"~" << *this);
}

INLINE_MEMBER Vec3D Trafo3D::getVecX() const
{
    return Vec3D(m(0,0),m(0,1),m(0,2));
}

INLINE_MEMBER Vec3D Trafo3D::getVecY() const
{
    return Vec3D(m(1,0),m(1,1),m(1,2));
}

INLINE_MEMBER Vec3D Trafo3D::getVecZ() const
{
    return Vec3D(m(2,0),m(2,1),m(2,2));
}

INLINE_MEMBER Pnt3D Trafo3D::getPntO() const
{
    return Pnt3D(c);
}

INLINE_MEMBER Lin3D Trafo3D::getLin() const
{
    return Lin3D(m);
}

INLINE_MEMBER Trans3D Trafo3D::getTrans() const
{
    return Trans3D(c);
}

INLINE_MEMBER Mat3 Trafo3D::getMat3() const
{
    return m;
}

INLINE_MEMBER Mat4 Trafo3D::getMat4() const
{
    return Mat4(m(0,0), m(0,1), m(0,2), 0.0,
		m(1,0), m(1,1), m(1,2), 0.0,
		m(2,0), m(2,1), m(2,2), 0.0,
		c.x0(), c.x1(), c.x2(), 1.0);
}

INLINE_MEMBER Trafo3D & Trafo3D::operator*=(const Scale3D & scale)
{
    m.scale(scale.c);
    c.scale(scale.c);
    return *this;
}

INLINE_MEMBER Trafo3D & Trafo3D::operator*=(const RotX3D & rot)
{
    m.rot0(rot.cosAlpha,rot.sinAlpha);
    c.rot0(rot.cosAlpha,rot.sinAlpha);
    return *this;
}

INLINE_MEMBER Trafo3D & Trafo3D::operator*=(const RotY3D & rot)
{
    m.rot1(rot.cosAlpha,rot.sinAlpha);
    c.rot1(rot.cosAlpha,rot.sinAlpha);
    return *this;
}

INLINE_MEMBER Trafo3D & Trafo3D::operator*=(const RotZ3D & rot)
{
    m.rot2(rot.cosAlpha,rot.sinAlpha);
    c.rot2(rot.cosAlpha,rot.sinAlpha);
    return *this;
}

INLINE_MEMBER Trafo3D & Trafo3D::operator*=(const ShearXY3D & shear)
{
    m.shear01(shear.c.x0(),shear.c.x1());
    c.shear01(shear.c.x0(),shear.c.x1());
    return *this;
}

INLINE_MEMBER Trafo3D & Trafo3D::operator*=(const ShearZX3D & shear)
{
    m.shear20(shear.c.x0(),shear.c.x1());
    c.shear20(shear.c.x0(),shear.c.x1());
    return *this;
}

INLINE_MEMBER Trafo3D & Trafo3D::operator*=(const ShearYZ3D & shear)
{
    m.shear12(shear.c.x0(),shear.c.x1());
    c.shear12(shear.c.x0(),shear.c.x1());
    return *this;
}

INLINE_MEMBER Trafo3D & Trafo3D::operator*=(const Trans3D & trans)
{
    c += trans.c;
    return *this;
}

INLINE_MEMBER Trafo3D & Trafo3D::operator*=(const Lin3D & lin)
{
    m *= lin.m;
    c *= lin.m;
    return *this;
}

INLINE_MEMBER Trafo3D & Trafo3D::operator*=(const Trafo3D & trafo)
{
    m *= trafo.m;
    c *= trafo.m;
    c += trafo.c;
    return *this;
}

INLINE_GLOBAL Vec3D getVecX(const Trafo3D & trafo) { return trafo.getVecX(); }
INLINE_GLOBAL Vec3D getVecY(const Trafo3D & trafo) { return trafo.getVecY(); }
INLINE_GLOBAL Vec3D getVecZ(const Trafo3D & trafo) { return trafo.getVecZ(); }
INLINE_GLOBAL Pnt3D getPntO(const Trafo3D & trafo) { return trafo.getPntO(); }
INLINE_GLOBAL Lin3D getLin(const Trafo3D & trafo) { return trafo.getLin(); }
INLINE_GLOBAL Trans3D getTrans(const Trafo3D & trafo) { return trafo.getTrans(); }
INLINE_GLOBAL Mat3 getMat3(const Trafo3D & trafo) { return trafo.m; }
INLINE_GLOBAL Mat4 getMat4(const Trafo3D & trafo) { return trafo.getMat4(); }

INLINE_GLOBAL double determinant(const Trafo3D & trafo)
{
    return determinant(trafo);
}

INLINE_GLOBAL Trafo3D inverseCopy(const Trafo3D & trafo, double det)
{
    Trafo3D inv;

    inv.m = inverseCopy(trafo.m,det);

    inv.c = - (trafo.c * inv.m);
                 
    return inv;
}

INLINE_MEMBER void Trafo3D::invert(
        )
{
    double det = determinant(*this);
    *this = inverseCopy(*this,det);
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const Trafo3D & trafo)
{
    s << "Trafo3D(" << trafo.m << ',' << trafo.c << ')';
    return s;
}

INLINE_GLOBAL Trafo3D inverseCopy(const Trafo3D & trafo)
{
    double det = determinant(trafo);
    return inverseCopy(trafo,det);
}

INLINE_GLOBAL Trafo3D operator*(const Trafo3D & trafo, const Scale3D & scale)
{
    return Trafo3D(prMatScale(trafo.m,scale.c),
                   prCrdScale(trafo.c,scale.c));
}

INLINE_GLOBAL Trafo3D operator*(const Trafo3D & trafo, const RotX3D & rot)
{
    return Trafo3D(prMatRot0(trafo.m,rot.cosAlpha,rot.sinAlpha),
                   prCrdRot0(trafo.c,rot.cosAlpha,rot.sinAlpha));
}

INLINE_GLOBAL Trafo3D operator*(const Trafo3D & trafo, const RotY3D & rot)
{
    return Trafo3D(prMatRot1(trafo.m,rot.cosAlpha,rot.sinAlpha),
                   prCrdRot1(trafo.c,rot.cosAlpha,rot.sinAlpha));
}

INLINE_GLOBAL Trafo3D operator*(const Trafo3D & trafo, const RotZ3D & rot)
{
    return Trafo3D(prMatRot2(trafo.m,rot.cosAlpha,rot.sinAlpha),
                   prCrdRot2(trafo.c,rot.cosAlpha,rot.sinAlpha));
}

INLINE_GLOBAL Trafo3D operator*(const Trafo3D & trafo, const ShearXY3D & shear)
{
    return Trafo3D(prMatShear01(trafo.m,shear.c.x0(),shear.c.x1()),
		   prCrdShear01(trafo.c,shear.c.x0(),shear.c.x1()));
}

INLINE_GLOBAL Trafo3D operator*(const Trafo3D & trafo, const ShearZX3D & shear)
{
    return Trafo3D(prMatShear20(trafo.m,shear.c.x0(),shear.c.x1()),
    		   prCrdShear20(trafo.c,shear.c.x0(),shear.c.x1()));
}

INLINE_GLOBAL Trafo3D operator*(const Trafo3D & trafo, const ShearYZ3D & shear)
{
    return Trafo3D(prMatShear12(trafo.m,shear.c.x0(),shear.c.x1()),
		   prCrdShear12(trafo.c,shear.c.x0(),shear.c.x1()));
}

INLINE_GLOBAL Trafo3D operator*(const Trafo3D & trafo, const Trans3D & trans)
{
    return Trafo3D(trafo.m, trafo.c + trans.c);
}

INLINE_GLOBAL Trafo3D operator*(const Trafo3D & trafo, const Lin3D & lin)
{
    return Trafo3D(trafo.m * lin.m, trafo.c * lin.m);
}

INLINE_GLOBAL Trafo3D operator*(const Scale3D & scale, const Trafo3D & trafo)
{
    return Trafo3D(prScaleMat(scale.c,trafo.m),trafo.c);
}

INLINE_GLOBAL Trafo3D operator*(const RotX3D & rot, const Trafo3D & trafo)
{
    return Trafo3D(prRot0Mat(rot.cosAlpha,rot.sinAlpha,trafo.m),trafo.c);
}

INLINE_GLOBAL Trafo3D operator*(const RotY3D & rot, const Trafo3D & trafo)
{
    return Trafo3D(prRot1Mat(rot.cosAlpha,rot.sinAlpha,trafo.m),trafo.c);
}

INLINE_GLOBAL Trafo3D operator*(const RotZ3D & rot, const Trafo3D & trafo)
{
    return Trafo3D(prRot2Mat(rot.cosAlpha,rot.sinAlpha,trafo.m),trafo.c);
}

INLINE_GLOBAL Trafo3D operator*(const ShearXY3D & shear, const Trafo3D & trafo)
{
    return Trafo3D(prShear01Mat(shear.c.x0(),shear.c.x1(),trafo.m),
		   trafo.c);
}

INLINE_GLOBAL Trafo3D operator*(const ShearZX3D & shear, const Trafo3D & trafo)
{
    return Trafo3D(prShear20Mat(shear.c.x0(),shear.c.x1(),trafo.m),
		   trafo.c);
}

INLINE_GLOBAL Trafo3D operator*(const ShearYZ3D & shear, const Trafo3D & trafo)
{
    return Trafo3D(prShear12Mat(shear.c.x0(),shear.c.x1(), trafo.m),
		   trafo.c);
}

INLINE_GLOBAL Trafo3D operator*(const Trans3D & trans, const Trafo3D & trafo)
{
    return Trafo3D(trafo.m,trans.c * trafo.m + trafo.c);
}

INLINE_GLOBAL Trafo3D operator*(const Lin3D & lin, const Trafo3D & trafo)
{
    return Trafo3D(lin.m * trafo.m,trafo.c);
}

INLINE_GLOBAL Trafo3D operator*(const Trafo3D & trafo1, const Trafo3D & trafo2)
{
    return Trafo3D(trafo1.m * trafo2.m, trafo1.c * trafo2.m + trafo2.c);
}

} // extern "C++"

// ===========================================================================

