// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       Vec3D
//  TYPE:       c++ header
//  PROJECT:    VEGA Graphics Library
//  CONTENT:    3D vectors, points, and transformations using doubles
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rft     Robert F. Tobler
//              hl      Helwig L"offelman
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 14:08:16  rft     last modification
//  08-Aug-93           rft     created
// ===========================================================================
//  MANUAL:
//
//  DESCRIPTION
//      The classes Vec3D and Pnt3D are the basic types for 3D geometry.
//      They can be transformed with the corresponding transformation types
//      that are called Scale3D, RotX3D, RotY3D, RotZ3D, Rot3D,
//      ShearXY3D, ShearZX3D, ShearYZ3D, Lin3D, Trans3D and Trafo3D.
//
//      The classes Vec3D and Pnt3D have been designed in a typesafe way,
//      so that all operations among them are geometrically meaningful.
//      Thus there are some combinations of operations that are not allowed
//      since there is no meaningful geometric result. An example for such
//      an operation is the addition of two points.
//
//      In terms of matrix theory, Vec3D and Pnt3D can be viewed as
//      homogenous 3D vectors and points (the trailing zeros and ones are
//      not stored):
//
//              Vec3D vec;
//              Pnt3D pnt;
//
//              (vec.x(), vec.y(), vec.z(), 0)
//              (pnt.x(), pnt.y(), pnt.z(), 1)
//
//      The transformation type Trafo3D can be viewed as a 3D homogeneous
//      matrix. The multiplication of Vec3D and Pnt3D with Trafo3D can
//      then be represented as follows (the elements of the transformation
//      cannot be accessed in the way shown here):
//
//              Trafo3D t;
//              Vec3D v;
//              Pnt3D m;
//
//              Vec3D vec = v * t;
//              Vec3D pnt = p * t;
//
//                                      (t.m[0,0], t.m[0,1], t.m[0,2], 0)
//                                      (t.m[1,0], t.m[1,1], t.m[0,1], 0)
//                                      (t.m[2,0], t.m[2,1], t.m[0,2], 0)
//                                      (t.c[0],   t.c[1],   t.c[2],   1)
//
//              (v.x(),v.y(),v.z(),0)   ( vec.x(),  vec.y(),  vec.z(), 0)
//
//              (p.x(),p.y(),p.z(),1)   ( pnt.x(),  pnt.y(),  pnt.z(), 1)
//
//      The transformation types Trans3D, Scale3D, RotX3D, RotY3D,
//      RotZ3D, Rot3D, ShearXY3D, ShearZX3D, ShearYZ3D, and Lin3D
//      represent subsets of the transformations represented by the Trafo3D
//      type.
//
//      Rotations always rotate clockwise if you look in the direction of the
//      axis. (For example RotZ3D(90 DEGREES) transforms a point that lies on
//      the positive x-axis into a point on the positive y-axis.)
//
//      Other ways to think about Lin3D and Trafo3D:
//
//              Lin3D <=>   (Vec3D,     // vecX
//                           Vec3D,     // vecY
//                           Vec3D)     // vecZ
//
//              Trafo3D <=> (Vec3D,     // vecX
//                           Vec3D,     // vecY
//                           Vec3D,     // vecZ
//                           Pnt3D)     // pntO (O for Origin)
//
//              (This can be used to directly specify a transformation by
//               just specifying the origin and the base vectors of the
//               coordinate system !)
//
//              Trafo3D <=> Lin3D * Trans3D
//
//              Trafo3D <=> (Lin3D,     // lin
//                           Vec3D)     // vec
//
//      If you are using these models, there are constructors that take these
//      elements as parameters, and methods and functions to get these
//      elements (getVecX, getVecY, getVecZ, getPntO, getLin, and getTrans).
//
//  CHECKS
//      Range checking and attempts to normalize null-vectors will only be
//      caught if WITH_CRD3_CHECKS is defined.
//
//  ERRORHANDLERS
//      Vec3D_errors
//
//  TRACERS
//      Vec3D_constructors
//
//  SEE ALSO
//      VEGA, GraphicsLibrary
// ===========================================================================
#ifndef _GRAPHICS_VEC3D_H_
#define _GRAPHICS_VEC3D_H_

#include <iostream.h>

#include <VEGA/Math/Math.h>
#include <VEGA/Math/Crd2.h>
#include <VEGA/Math/Crd3.h>
#include <VEGA/Math/Crd4.h>

#include "VEGA/Graphics/Globals.h"

#ifdef __GNUC__
#pragma interface
#endif

extern "C++"
{

// ============================================ classes defined in this module

class Vec3D;				// vector
class Pnt3D;				// point
class Scale3D;				// scaling by (sx,sy,sz)
class RotX3D;				// rotation around x-axis
class RotY3D;				// rotation around y-axis
class RotZ3D;				// rotation around z-axis
class Rot3D;				// rotation around axis through origin
class ShearXY3D;			// shear along x-y-plane
class ShearZX3D;			// shear along z-x-plane
class ShearYZ3D;			// shear along y-z-plane
class Lin3D;				// linear transformation
class Trans3D;				// translation by vector
class Trafo3D;				// affine transformation

// =========================================== errors generated by this module

#define ERROR_VEC3D_SINGULAR_SCALE      1
        //
        //  An attempt was made to invert a singular scale.
        //


// =============================================================== class Vec3D

class Vec3D
{
    friend class Pnt3D;
    friend class Lin3D;
    friend class Trans3D;
    friend class Trafo3D;

    friend class Viewer3D;		//  needs FAST access to members

public:

    // --------------------------------------------- constructors / destructor

    Vec3D();
    Vec3D(
        double xNew,
        double yNew,
        double zNew
        );
    Vec3D(
        const Crd3 & crd
        );
    Vec3D(
        const Vec3D & vec
        );
    ~Vec3D();

    // ------------------------------------------------------ member operators

    double operator[](
	int index
	) const;
    double & operator[](
	int index
	);

    Vec3D & operator+=(
	const Vec3D & vec
	);
    Vec3D & operator-=(
	const Vec3D & vec
	);

    Vec3D & operator*=(
	double scalar
	);

    // ------------------------------------------------------ member functions

    double x(
	) const;
    double y(
	) const;
    double z(
	) const;

    void x(
	double xNew
	);
    void y(
	double yNew
	);
    void z(
	double zNew
	);

    void normalize(
	);

    Crd3 getCrd3(
	) const;

    Crd4 getCrd4(			// Get homogeneous representation.
	) const;

    Pnt3D getPnt3D(			// Explicit conversion to Pnt3D.
	) const;

    // ------------------------------------------------------ friend operators

    friend ostream & operator<<(
	ostream & s,
	const Vec3D & vec
	);
    friend Vec3D operator-(
	const Vec3D & vec
	);
    friend Vec3D operator+(
	const Vec3D & vec1,
	const Vec3D & vec2
	);
    friend Pnt3D operator+(
	const Pnt3D & pnt,
	const Vec3D & vec
	);
    friend Vec3D operator-(
	const Vec3D & vec1,
	const Vec3D & vec2
	);
    friend Pnt3D operator-(
	const Pnt3D & pnt,
	const Vec3D & vec)
	;
    friend Vec3D operator*(
	const Vec3D & vec,
	double scalar
	);
    friend Vec3D operator*(
	double scalar,
	const Vec3D & vec
	);

    // ------------------------------------------------------ friend functions

    friend Crd3 getCrd3(
	const Vec3D & vec
	);
    friend Crd4 getCrd4(		// Get homogeneous representation.
	const Vec3D & vec
	);
    friend Pnt3D getPnt3D(		// Explicit conversion.
	const Vec3D & vec
	);
    friend double squaredLength(
	const Vec3D & vec
	);
    friend double length(
	const Vec3D & vec
	);
    friend double dotProduct(
	const Vec3D & vec1,
	const Vec3D & vec2
	);
    friend Vec3D crossProduct(
	const Vec3D & vec1,		// (vec1, vec2, result) define the
	const Vec3D & vec2		//  base of a right-handed coordinate
	);				//  system

    friend Vec3D normalizedCopy(
	const Vec3D & vec
	);

    // -------------------------------------- special purpose friend functions

    friend Vec3D cwReciprocalCopy(      // Componentwise multiplicative
        const Vec3D & vec		//  inverse. Zero components are not
        );				//  inverted!

    friend Vec3D cwMin(                 // Componentwise minimum.
        const Vec3D & vec1,
        const Vec3D & vec2
        );

    friend Vec3D cwMax(                 // Componentwise maximum.
        const Vec3D & vec1,
        const Vec3D & vec2
        );

    // ---------------------------------------------- transformation functions

    //
    //  Multiply vector by special transformation.
    //
    Vec3D & operator*=(
	const Scale3D & scale
	);
    Vec3D & operator*=(
	const RotX3D & rot
	);
    Vec3D & operator*=(
	const RotY3D & rot
	);
    Vec3D & operator*=(
	const RotZ3D & rot
	);
    Vec3D & operator*=(
	const ShearXY3D & shear
	);
    Vec3D & operator*=(
	const ShearZX3D & shear
	);
    Vec3D & operator*=(
	const ShearYZ3D & shear
	);
    Vec3D & operator*=(
	const Lin3D & lin
	);
    Vec3D & operator*=(
	const Trafo3D & trafo
	);

    //
    //  Product of vector and special tranformation.
    //
    friend Vec3D operator*(
	const Vec3D & vec,
	const Scale3D & scale
	);
    friend Vec3D operator*(
	const Vec3D & vec,
	const RotX3D & rot
	);
    friend Vec3D operator*(
	const Vec3D & vec,
	const RotY3D & rot
	);
    friend Vec3D operator*(
	const Vec3D & vec,
	const RotZ3D & rot
	);
    friend Vec3D operator*(
	const Vec3D & vec,
	const ShearXY3D & shear
	);
    friend Vec3D operator*(
	const Vec3D & vec,
	const ShearZX3D & shear
	);
    friend Vec3D operator*(
	const Vec3D & vec,
	const ShearYZ3D & shear
	);
    friend Vec3D operator*(
	const Vec3D & vec,
	const Lin3D & lin
	);
    friend Vec3D operator*(
	const Vec3D & vec,
	const Trafo3D & trafo
	);

    // ----------------------------------------------- static member variables

    static const Vec3D baseX;
    static const Vec3D baseY;
    static const Vec3D baseZ;

protected:

    Crd3 c;
};

// ===================================================================== Pnt3D

class Pnt3D
{
    friend class Vec3D;
    friend class Lin3D;
    friend class Trans3D;
    friend class Trafo3D;

    friend class Viewer3D;		//  needs FAST access to members

public:

    // ---------------------------------------------------------- constructors

    Pnt3D(
        );
    Pnt3D(
        double x,
        double y,
        double z
        );
    Pnt3D(
        const Crd3 & crd
        );
    Pnt3D(
        const Pnt3D & pnt
        );
    ~Pnt3D(
        );

    // ------------------------------------------------------ member operators

    double operator[](
	int index
	) const;
    double & operator[](
	int index
	);

    Pnt3D & operator+=(
	const Vec3D & vec
	);
    Pnt3D & operator-=(
	const Vec3D & vec
	);

    // ------------------------------------------------------ member functions

    double x(
	) const;
    double y(
	) const;
    double z(
	) const;

    void x(
	double xNew
	);
    void y(
	double yNew
	);
    void z(
	double zNew
	);

    Crd3 getCrd3(
	) const;

    Crd4 getCrd4(			// Get homogeneous representation.
	) const;

    Vec3D getVec3D(			// Explicit conversion to Vec3D.
	) const;

    // ------------------------------------------------------ friend operators

    friend ostream & operator<<(
	ostream & s,
	const Pnt3D & vec
	);
    friend Pnt3D operator+(
	const Pnt3D & pnt,
	const Vec3D & vec
	);
    friend Pnt3D operator-(
	const Pnt3D & pnt,
	const Vec3D & vec
	);
    friend Vec3D operator-(
	const Pnt3D & pnt1,
	const Pnt3D & pnt2
	);

    // ------------------------------------------------------ friend functions

    friend Crd3 getCrd3(
	const Pnt3D & pnt
	);
    friend Crd4 getCrd4(		// Get homogeneous representation.
	const Pnt3D & pnt
	);
    friend Vec3D getVec3D(		// Explicit conversion.
	const Pnt3D & pnt
	);
    friend double squaredDistanceFromOrigin(
	const Pnt3D & pnt
	);
    friend double distanceFromOrigin(
	const Pnt3D & pnt
	);
    friend double squaredDistance(
	const Pnt3D & pnt1,
	const Pnt3D & pnt2
	);
    friend double distance(
	const Pnt3D & pnt1,
	const Pnt3D & pnt2
	);

    // ----------------------------------------------- componentwise functions

    friend Pnt3D cwMin(
	const Pnt3D & vec1,
	const Pnt3D & vec2
	);
    friend Pnt3D cwMax(
	const Pnt3D & vec1,
	const Pnt3D & vec2
	);

    // ---------------------------------------------- transformation functions

    Pnt3D & operator*=(
	const Scale3D & scale
	);
    Pnt3D & operator*=(
	const RotX3D & rot
	);
    Pnt3D & operator*=(
	const RotY3D & rot
	);
    Pnt3D & operator*=(
	const RotZ3D & rot
	);
    Pnt3D & operator*=(
	const ShearXY3D & shear
	);
    Pnt3D & operator*=(
	const ShearZX3D & shear
	);
    Pnt3D & operator*=(
	const ShearYZ3D & shear
	);
    Pnt3D & operator*=(
	const Trans3D & trans
	);
    Pnt3D & operator*=(
	const Lin3D & lin
	);
    Pnt3D & operator*=(
	const Trafo3D & trafo
	);

    friend Pnt3D operator*(
	const Pnt3D & pnt,
	const Scale3D & scale
	);
    friend Pnt3D operator*(
	const Pnt3D & pnt,
	const RotX3D & rot
	);
    friend Pnt3D operator*(
	const Pnt3D & pnt,
	const RotY3D & rot
	);
    friend Pnt3D operator*(
	const Pnt3D & pnt,
	const RotZ3D & rot
	);
    friend Pnt3D operator*(
	const Pnt3D & pnt,
	const ShearXY3D & shear
	);
    friend Pnt3D operator*(
	const Pnt3D & pnt,
	const ShearZX3D & shear
	);
    friend Pnt3D operator*(
	const Pnt3D & pnt,
	const ShearYZ3D & shear
	);
    friend Pnt3D operator*(
	const Pnt3D & pnt,
	const Trans3D & trans
	);
    friend Pnt3D operator*(
	const Pnt3D & pnt,
	const Lin3D & lin
	);
    friend Pnt3D operator*(
	const Pnt3D & pnt,
	const Trafo3D & trafo
	);

    // ----------------------------------------------- static member variables

    static const Pnt3D origin;

protected:
    Crd3 c;
};

// ============================================================= class Scale3D

class Scale3D
{
    friend class Vec3D;
    friend class Pnt3D;
    friend class Lin3D;
    friend class Trafo3D;

public:

    Scale3D(				// Create undefined object.
    	);
    Scale3D(
        double scalar
        );
    Scale3D(
        double x,
        double y,
        double z
        );
    Scale3D(
        const Crd3 & scale
        );
    Scale3D(
        const Scale3D & scale
        );
    ~Scale3D();

    Scale3D & operator*=(
	const Scale3D & scale
	);

    void invert(
	);

    friend ostream & operator<<(
	ostream & s,
	const Scale3D & scale
	);
    friend Scale3D operator*(
	const Scale3D & scale1,
	const Scale3D & scale2
	);

    friend Scale3D inverseCopy(
	const Scale3D & scale
	);

    static const Scale3D identity;

    // -------------------------------------------- functions that need access

    friend Vec3D operator*(const Vec3D & vec, const Scale3D & scale);
    friend Pnt3D operator*(const Pnt3D & pnt, const Scale3D & scale);
    friend Lin3D operator*(const Scale3D & scale, const Lin3D & lin);
    friend Lin3D operator*(const Lin3D & lin, const Scale3D & scale);
    friend Trafo3D operator*(const Scale3D & scale, const Trafo3D & trafo);
    friend Trafo3D operator*(const Trafo3D & trafo, const Scale3D & scale);

protected:

    Crd3 c;
};

// ============================================================== class RotX3D

class RotX3D
{
    friend class Vec3D;
    friend class Pnt3D;
    friend class Rot3D;
    friend class Lin3D;
    friend class Trafo3D;

public:

    RotX3D(				// Create undefined object.
    	);
    RotX3D(
        double newAlpha
        );
    RotX3D(				// WARNING:
        double newAlpha,		//  These parameters are not checked
        double newCosAlpha,		//  for consistency.
        double newSinAlpha
        );
    RotX3D(
        const RotX3D & rot
        );
    ~RotX3D(
	);

    RotX3D & operator*=(
	const RotX3D & rot
	);

    double getAlpha(
	) const;
    void invert(
	);
    
    friend ostream & operator<<(
	ostream & s,
	const RotX3D & rot
	);
    friend RotX3D operator*(
	const RotX3D & rot1,
	const RotX3D & rot2
	);

    friend RotX3D inverseCopy(
	const RotX3D & rot
	);

    static const RotX3D identity;

    // -------------------------------------------- functions that need access

    friend Vec3D operator*(const Vec3D & vec, const RotX3D & rot);
    friend Pnt3D operator*(const Pnt3D & pnt, const RotX3D & rot);
    friend Lin3D operator*(const Lin3D & lin, const RotX3D & rot);
    friend Lin3D operator*(const RotX3D & rot, const Lin3D & lin);
    friend Trafo3D operator*(const Trafo3D & trafo, const RotX3D & rot);
    friend Trafo3D operator*(const RotX3D & rot, const Trafo3D & trafo);

protected:
    double alpha;
    double cosAlpha;
    double sinAlpha;
};

// ============================================================== class RotY3D

class RotY3D
{
    friend class Vec3D;
    friend class Pnt3D;
    friend class Rot3D;
    friend class Lin3D;
    friend class Trafo3D;

public:

    RotY3D(				// Create undefined object.
    	);
    RotY3D(
        double newAlpha
        );
    RotY3D(
        double newAlpha,		// WARNING:
        double newCosAlpha,		//  These parameters are not checked
        double newSinAlpha		//  for consistency.
        );
    RotY3D(
        const RotY3D & rot
        );
    ~RotY3D();

    RotY3D & operator*=(
	const RotY3D & rot
	);

    double getAlpha(
	) const;
    void invert(
	);

    friend ostream & operator<<(
	ostream & s,
	const RotY3D & rot
	);
    friend RotY3D operator*(
	const RotY3D & rot1,
	const RotY3D & rot2
	);

    friend RotY3D inverseCopy(
	const RotY3D & rot
	);

    static const RotY3D identity;

    // -------------------------------------------- functions that need access

    friend Vec3D operator*(const Vec3D & vec, const RotY3D & rot);
    friend Pnt3D operator*(const Pnt3D & pnt, const RotY3D & rot);
    friend Lin3D operator*(const Lin3D & lin, const RotY3D & rot);
    friend Lin3D operator*(const RotY3D & rot, const Lin3D & lin);
    friend Trafo3D operator*(const Trafo3D & trafo, const RotY3D & rot);
    friend Trafo3D operator*(const RotY3D & rot, const Trafo3D & trafo);

protected:
    double alpha;
    double cosAlpha;
    double sinAlpha;
};

// ============================================================== class RotZ3D

class RotZ3D
{
    friend class Vec3D;
    friend class Pnt3D;
    friend class Rot3D;
    friend class Lin3D;
    friend class Trafo3D;

public:

    RotZ3D(				// Create undefined object.
    	);
    RotZ3D(
        double newAlpha
        );
    RotZ3D(
        double newAlpha,		// WARNING:
        double newCosAlpha,		// These parameters are not checked
        double newSinAlpha		// for consistency.
        );
    RotZ3D(
        const RotZ3D & rot
        );
    ~RotZ3D();

    RotZ3D & operator*=(
	const RotZ3D & rot
	);

    double getAlpha(
	) const;
    void invert(
	);

    friend ostream & operator<<(
	ostream & s,
	const RotZ3D & rot
	);
    friend RotZ3D operator*(
	const RotZ3D & rot1,
	const RotZ3D & rot2
	);

    friend RotZ3D inverseCopy(
	const RotZ3D & rot
	);

    static const RotZ3D identity;

    // -------------------------------------------- functions that need access

    friend Vec3D operator*(const Vec3D & vec, const RotZ3D & rot);
    friend Pnt3D operator*(const Pnt3D & pnt, const RotZ3D & rot);
    friend Lin3D operator*(const Lin3D & lin, const RotZ3D & rot);
    friend Lin3D operator*(const RotZ3D & rot, const Lin3D & lin);
    friend Trafo3D operator*(const Trafo3D & trafo, const RotZ3D & rot);
    friend Trafo3D operator*(const RotZ3D & rot, const Trafo3D & trafo);

protected:
    double alpha;
    double cosAlpha;
    double sinAlpha;
};

// =============================================================== class Rot3D

class Rot3D
{
    friend class Vec3D;
    friend class Pnt3D;
    friend class Lin3D;
    friend class Trafo3D;

public:

    //
    //  This class represents general rotations around axes through the
    //	origin. They are implemented using quaternions.
    //

    Rot3D(				// Create undefined object.
        );
    Rot3D(
        const Rot3D & rot
        );
    Rot3D(				// WARNING:
    	double aNew,			//  The tuple is asumed to represent
	double bNew,			//  a unit-norm quaternion !
	double cNew,
	double dNew
	);
    Rot3D(				// WARNING: The Crd4 is assumed to
        const Crd4 & rot		//  represent a unit-norm quaternion !
        );
    Rot3D(				// Init with vector and angle:
        const Vec3D & vec,		//  A unit-norm quaternion is built
        double alpha			//  from the given vector and angle.
        );
    Rot3D(				// Init with vector and angle 0,
    	const Vec3D & vec		//  no unit-norm conversion is done.
	);
    Rot3D(
        const RotX3D & rot
        );
    Rot3D(
        const RotY3D & rot
        );
    Rot3D(
        const RotZ3D & rot
        );
    ~Rot3D(
	);

    Rot3D & operator*=(
	const Rot3D & rot
	);

    double getAlpha(
	) const;			// Return value in [0, MATH_2_MUL_PI).

    Vec3D getVec3D(
	) const;
    void invert(
	);

    friend ostream & operator<<(
	ostream & s,
	const Rot3D & rot
	);
    friend Rot3D operator*(
	const Rot3D & rot1,
	const Rot3D & rot2
	);

    friend Rot3D inverseCopy(
	const Rot3D & rot
	);

    static const Rot3D identity;

protected:
    Crd4 c;
};

// =========================================================== class ShearXY3D

class ShearXY3D
{
    friend class Vec3D;
    friend class Pnt3D;
    friend class Lin3D;
    friend class Trafo3D;
public:

    ShearXY3D(				// Create undefined object.
    	);
    ShearXY3D(
        const ShearXY3D & shear
        );
    ShearXY3D(
        double xNew,
	double yNew
        );
    ShearXY3D(
	const Crd2 & crd
        );
    ~ShearXY3D(
	);

    ShearXY3D & operator*=(
	const ShearXY3D & shear
	);

    void invert(
	);

    friend ostream & operator<<(
	ostream & s,
	const ShearXY3D & shear
	);
    friend ShearXY3D operator*(
	const ShearXY3D & s1,
	const ShearXY3D & s2
	);

    friend ShearXY3D inverseCopy(
	const ShearXY3D & shear
	);

    static const ShearXY3D identity;
    
    // -------------------------------------------- functions that need access

    friend Vec3D operator*(const Vec3D & vec, const ShearXY3D & shear);
    friend Pnt3D operator*(const Pnt3D & pnt, const ShearXY3D & shear);
    friend Lin3D operator*(const Lin3D & lin, const ShearXY3D & sh);
    friend Lin3D operator*(const ShearXY3D & sh, const Lin3D & lin);
    friend Trafo3D operator*(const Trafo3D & trafo, const ShearXY3D & sh);
    friend Trafo3D operator*(const ShearXY3D & sh, const Trafo3D & trafo);

protected:
    Crd2 c;
};

// =========================================================== class ShearZX3D

class ShearZX3D
{
    friend class Vec3D;
    friend class Pnt3D;
    friend class Lin3D;
    friend class Trafo3D;
public:

    ShearZX3D(				// Create undefined object.
    	);
    ShearZX3D(
        const ShearZX3D & shear
        );
    ShearZX3D(
        double zNew,
	double xNew
        );
    ShearZX3D(
	const Crd2 & crd
        );
    ~ShearZX3D(
	);

    ShearZX3D & operator*=(
	const ShearZX3D & shear
	);

    void invert();

    friend ostream & operator<<(
	ostream & s,
	const ShearZX3D & shear
	);
    friend ShearZX3D operator*(
	const ShearZX3D & s1,
	const ShearZX3D & s2
	);

    friend ShearZX3D inverseCopy(
	const ShearZX3D & shear
	);

    static const ShearZX3D identity;
    
    // -------------------------------------------- functions that need access

    friend Vec3D operator*(const Vec3D & vec, const ShearZX3D & shear);
    friend Pnt3D operator*(const Pnt3D & pnt, const ShearZX3D & shear);
    friend Lin3D operator*(const Lin3D & lin, const ShearZX3D & sh);
    friend Lin3D operator*(const ShearZX3D & sh, const Lin3D & lin);
    friend Trafo3D operator*(const Trafo3D & trafo, const ShearZX3D & sh);
    friend Trafo3D operator*(const ShearZX3D & sh, const Trafo3D & trafo);

protected:
    Crd2 c;
};

// =========================================================== class ShearYZ3D

class ShearYZ3D
{
    friend class Vec3D;
    friend class Pnt3D;
    friend class Lin3D;
    friend class Trafo3D;
public:

    ShearYZ3D(				// Create undefined object.
    	);
    ShearYZ3D(
        const ShearYZ3D & shear
        );
    ShearYZ3D(
        double yNew,
	double zNew
        );
    ShearYZ3D(
	const Crd2 & crd
        );
    ~ShearYZ3D(
	);

    ShearYZ3D & operator*=(
	const ShearYZ3D & shear
	);

    void invert(
	);

    friend ostream & operator<<(
	ostream & s,
	const ShearYZ3D & shear
	);
    friend ShearYZ3D operator*(
	const ShearYZ3D & s1,
	const ShearYZ3D & s2
	);

    friend ShearYZ3D inverseCopy(
	const ShearYZ3D & shear
	);

    static const ShearYZ3D identity;
    
    // -------------------------------------------- functions that need access

    friend Vec3D operator*(const Vec3D & vec, const ShearYZ3D & shear);
    friend Pnt3D operator*(const Pnt3D & pnt, const ShearYZ3D & shear);
    friend Lin3D operator*(const Lin3D & lin, const ShearYZ3D & sh);
    friend Lin3D operator*(const ShearYZ3D & sh, const Lin3D & lin);
    friend Trafo3D operator*(const Trafo3D & trafo, const ShearYZ3D & sh);
    friend Trafo3D operator*(const ShearYZ3D & sh, const Trafo3D & trafo);

protected:
    Crd2 c;
};

// =============================================================== class Lin3D

class Lin3D
{
    friend class Vec3D;
    friend class Pnt3D;
    friend class Trafo3D;
public:

    Lin3D(				// Create undefined object.
	);
    Lin3D(
	const Lin3D & lin
	);
    Lin3D(
	const Mat3 & m
	);
    Lin3D(
        const Vec3D & newVecX,
        const Vec3D & newVecY,
        const Vec3D & newVecZ
	);
    Lin3D(
        double xx, double xy, double xz,
        double yx, double yy, double yz,
        double zx, double zy, double zz
        );
    Lin3D(
        const Scale3D & scale
        );
    Lin3D(
        const RotX3D & rot
        );
    Lin3D(
        const RotY3D & rot
        );
    Lin3D(
        const RotZ3D & rot
        );
    Lin3D(
        const ShearXY3D & scale
        );
    Lin3D(
        const ShearZX3D & scale
        );
    Lin3D(
        const ShearYZ3D & scale
        );
    Lin3D(
        const Rot3D & rot
        );
    ~Lin3D(
	);

    // ------------------------------------------------------ member functions

    Vec3D getVecX(
	) const;
    Vec3D getVecY(
	) const;
    Vec3D getVecZ(
	) const;

    Mat3 getMat3(
	) const;

    void invert(
	);
        //
        //  ERROR_CRD3_SINGULAR_MATRIX see module Crd3
        //

    // ------------------------------------------------------ friend operators

    friend ostream & operator<<(
	ostream & s,
	const Lin3D & lin
	);

    // ------------------------------------------------------ friend functions

    friend Vec3D getVecX(
	const Lin3D & lin
	);
    friend Vec3D getVecY(
	const Lin3D & lin
	);
    friend Vec3D getVecZ(
	const Lin3D & lin
	);

    friend Mat3 getMat3(
	const Lin3D & lin
	);
    friend double determinant(
	const Lin3D & lin
	);
    friend Lin3D inverseCopy(
	const Lin3D & lin,
	double det
	);
        //
        //  det is NOT checked against zero!
        //

    friend Lin3D inverseCopy(
	const Lin3D & lin
	);
        //
        //  ERROR_CRD3_SINGULAR_MATRIX see module Crd3
        //

    // ----------------------------------------------- tranformation functions

    //
    //  Multiply transformation by special transformation.
    //
    Lin3D & operator*=(
	const Scale3D & scale
	);
    Lin3D & operator*=(
	const RotX3D & rot
	);
    Lin3D & operator*=(
	const RotY3D & rot
	);
    Lin3D & operator*=(
	const RotZ3D & rot
	);
    Lin3D & operator*=(
	const ShearXY3D & shear
	);
    Lin3D & operator*=(
	const ShearZX3D & shear
	);
    Lin3D & operator*=(
	const ShearYZ3D & shear
	);
    Lin3D & operator*=(
	const Lin3D & lin
	);

    //
    //  Product of transformation and special transformation.
    //
    friend Lin3D operator*(
	const Lin3D & lin,
	const Scale3D & scale
	);
    friend Lin3D operator*(
	const Lin3D & lin,
	const RotX3D & rot
	);
    friend Lin3D operator*(
	const Lin3D & lin,
	const RotY3D & rot
	);
    friend Lin3D operator*(
	const Lin3D & lin,
	const RotZ3D & rot
	);
    friend Lin3D operator*(
	const Lin3D & lin,
	const ShearXY3D & shear
	);
    friend Lin3D operator*(
	const Lin3D & lin,
	const ShearZX3D & shear
	);
    friend Lin3D operator*(
	const Lin3D & lin,
	const ShearYZ3D & shear
	);

    //
    //  Product of special transformation and transformation.
    //
    friend Lin3D operator*(
	const Scale3D & scale,
	const Lin3D & lin
	);
    friend Lin3D operator*(
	const RotX3D & rot,
	const Lin3D & lin
	);
    friend Lin3D operator*(
	const RotY3D & rot,
	const Lin3D & lin
	);
    friend Lin3D operator*(
	const RotZ3D & rot,
	const Lin3D & lin
	);
    friend Lin3D operator*(
	const ShearXY3D & shear,
	const Lin3D & lin
	);
    friend Lin3D operator*(
	const ShearZX3D & shear,
	const Lin3D & lin
	);
    friend Lin3D operator*(
	const ShearYZ3D & shear,
	const Lin3D & lin
	);

    friend Lin3D operator*(           // Product of two transformations.
        const Lin3D & lin1,
        const Lin3D & lin2
        );

    //
    //  Product of vector/point and transformation.
    //
    friend Vec3D operator*(
	const Vec3D & vec,
	const Lin3D & lin
	);
    friend Pnt3D operator*(
	const Pnt3D & pnt,
	const Lin3D & lin
	);

    // ----------------------------------------------- static member variables

    static const Lin3D identity;

    // -------------------------------------------- functions that need access

    friend Trafo3D operator*(const Lin3D & lin, const Trafo3D & trafo);
    friend Trafo3D operator*(const Trafo3D & trafo, const Lin3D & lin);

protected:
    Mat3 m;
};

// ============================================================= class Trans3D

class Trans3D
{
    friend class Vec3D;
    friend class Pnt3D;
    friend class Trafo3D;

public:

    Trans3D(				// Create undefined object.
	);
    Trans3D(
        const Crd3 & crd
        );
    Trans3D(
        double xNew,
        double yNew,
        double zNew
        );
    Trans3D(
	const Vec3D & vec
	);
    Trans3D(
	const Pnt3D & pntO
	);
    Trans3D(
        const Trans3D & trans
        );
    ~Trans3D(
	);

    Trans3D & operator*=(
	const Trans3D & trans
	);

    void invert(
	);

    friend ostream & operator<<(
	ostream & s,
	const Trans3D & trans
	);
    friend Trans3D operator*(
	const Trans3D & tr1,
	const Trans3D & tr2
	);
    friend Vec3D operator*(
	const Vec3D & vec,
	const Trans3D & trans
	);
    friend Pnt3D operator*(
	const Pnt3D & pnt,
	const Trans3D & trans
	);
    friend Trafo3D operator*(
	const Trans3D & trans,
	const Trafo3D & trafo
	);
    friend Trafo3D operator*(
	const Trafo3D & trafo,
	const Trans3D & trans
	);

    friend Trans3D inverseCopy(
	const Trans3D & trans
	);

    static const Trans3D identity;

protected:

    Crd3 c;
};

// ============================================================= class Trafo3D

class Trafo3D
{
    friend class Vec3D;
    friend class Pnt3D;

public:

    Trafo3D(				// Create undefined object.
        );
    Trafo3D(
        const Trafo3D & trafo
        );
    Trafo3D(
        const Mat3 & m
        );
    Trafo3D(
        const Mat3 & m,
        const Crd3 & c
        );
    Trafo3D(
        double xx, double xy, double xz,
        double yx, double yy, double yz,
        double zx, double zy, double zz,
        double dx, double dy, double dz
        );
    Trafo3D(
        const Vec3D & newVecX,
        const Vec3D & newVecY,
        const Vec3D & newVecZ,
        const Pnt3D & newPntO
        );
    Trafo3D(
        const Lin3D & lin,
	const Trans3D & trans
        );
    Trafo3D(
        const Scale3D & scale
        );
    Trafo3D(
        const RotX3D & rot
        );
    Trafo3D(
        const RotY3D & rot
        );
    Trafo3D(
        const RotZ3D & rot
        );
    Trafo3D(
        const Rot3D & rot
        );
    Trafo3D(
        const ShearXY3D & scale
        );
    Trafo3D(
        const ShearZX3D & scale
        );
    Trafo3D(
        const ShearYZ3D & scale
        );
    Trafo3D(
        const Lin3D & lin
        );
    Trafo3D(
        const Trans3D & trans
        );
    ~Trafo3D(
        );

    // ------------------------------------------------------ member functions

    Vec3D getVecX(
	) const;
    Vec3D getVecY(
	) const;
    Vec3D getVecZ(
	) const;
    Pnt3D getPntO(
	) const;

    Lin3D getLin(
	) const;
    Trans3D getTrans(
	) const;

    Mat3 getMat3(
	) const;
    Mat4 getMat4(			// Get homogeneous representation.
	) const;

    void invert(
	);
        //
        //  ERROR_CRD3_SINGULAR_MATRIX see module Crd3
        //

    // ------------------------------------------------------ friend operators

    friend ostream & operator<<(
	ostream & s,
	const Trafo3D & trafo
	);

    // ------------------------------------------------------ friend functions

    friend Vec3D getVecX(
	const Trafo3D & trafo
	);
    friend Vec3D getVecY(
	const Trafo3D & trafo
	);
    friend Vec3D getVecZ(
	const Trafo3D & trafo
	);
    friend Pnt3D getPntO(
	const Trafo3D & trafo
	);

    friend Lin3D getLin(
	const Trafo3D & trafo
	);
    friend Trans3D getTrans(
	const Trafo3D & trafo
	);

    friend Mat3 getMat3(
	const Trafo3D & trafo
	);
    friend Mat4 getMat4(		// Get homogeneous representation
	const Trafo3D & trafo
	);

    friend double determinant(
	const Trafo3D & trafo
	);
    friend Trafo3D inverseCopy(
	const Trafo3D & trafo,
	double det			// det is NOT checked against zero!
	);

    friend Trafo3D inverseCopy(
	const Trafo3D & trafo
	);
        //
        //  ERROR_CRD3_SINGULAR_MATRIX see module Crd3
        //

    // ----------------------------------------------- tranformation functions

    //
    //  Multiply transformation by special transformation.
    //
    Trafo3D & operator*=(
	const Scale3D & scale
	);
    Trafo3D & operator*=(
	const RotX3D & rot
	);
    Trafo3D & operator*=(
	const RotY3D & rot
	);
    Trafo3D & operator*=(
	const RotZ3D & rot
	);
    Trafo3D & operator*=(
	const ShearXY3D & shear
	);
    Trafo3D & operator*=(
	const ShearZX3D & shear
	);
    Trafo3D & operator*=(
	const ShearYZ3D & shear
	);
    Trafo3D & operator*=(
	const Trans3D & trans
	);
    Trafo3D & operator*=(
	const Lin3D & lin
	);
    Trafo3D & operator*=(
	const Trafo3D & trafo
	);

    //
    //  Product of transformation and special transformation.
    //
    friend Trafo3D operator*(
	const Trafo3D & trafo,
	const Scale3D & scale
	);
    friend Trafo3D operator*(
	const Trafo3D & trafo,
	const RotX3D & rot
	);
    friend Trafo3D operator*(
	const Trafo3D & trafo,
	const RotY3D & rot
	);
    friend Trafo3D operator*(
	const Trafo3D & trafo,
	const RotZ3D & rot
	);
    friend Trafo3D operator*(
	const Trafo3D & trafo,
	const ShearXY3D & sh
	);
    friend Trafo3D operator*(
	const Trafo3D & trafo,
	const ShearZX3D & sh
	);
    friend Trafo3D operator*(
	const Trafo3D & trafo,
	const ShearYZ3D & sh
	);
    friend Trafo3D operator*(
	const Trafo3D & trafo,
	const Trans3D & trans
	);
    friend Trafo3D operator*(
	const Trafo3D & trafo,
	const Lin3D & lin
	);

    //
    //  Product of special transformation and transformation.
    //
    friend Trafo3D operator*(
	const Scale3D & scale,
	const Trafo3D & trafo
	);
    friend Trafo3D operator*(
	const RotX3D & rot,
	const Trafo3D & trafo
	);
    friend Trafo3D operator*(
	const RotY3D & rot,
	const Trafo3D & trafo
	);
    friend Trafo3D operator*(
	const RotZ3D & rot,
	const Trafo3D & trafo
	);
    friend Trafo3D operator*(
	const ShearXY3D & sh,
	const Trafo3D & trafo
	);
    friend Trafo3D operator*(
	const ShearZX3D & sh,
	const Trafo3D & trafo
	);
    friend Trafo3D operator*(
	const ShearYZ3D & sh,
	const Trafo3D & trafo
	);
    friend Trafo3D operator*(
	const Trans3D & trans,
	const Trafo3D & trafo
	);
    friend Trafo3D operator*(
	const Lin3D & lin,
	const Trafo3D & trafo
	);

    friend Trafo3D operator*(           // Product of two transformations.
        const Trafo3D & trafo1,
        const Trafo3D & trafo2
        );

    //
    //  Product of vector/point and transformation.
    //
    friend Vec3D operator*(
	const Vec3D & vec,
	const Trafo3D & trafo
	);
    friend Pnt3D operator*(
	const Pnt3D & pnt,
	const Trafo3D & trafo
	);

    // ----------------------------------------------- static member variables

    static const Trafo3D identity;

protected:

    Mat3 m;
    Crd3 c;
};

} // extern "C++"

#ifndef VEC3D_NO_INLINE

#define INLINE_GLOBAL COMPILER_INLINE_GLOBAL
#define INLINE_MEMBER COMPILER_INLINE_MEMBER

#include "VEGA/Graphics/Vec3D.I"

#undef INLINE_GLOBAL
#undef INLINE_MEMBER

#endif


#endif // _GRAPHICS_VEC3D_H_
// ===========================================================================
