// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       Crd2
//  TYPE:       c++ inline
//  PROJECT:    VEGA Math Library
//  CONTENT:    base types for 2D coordinate manipulation using doubles
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rft     Robert F. Tobler
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 14:06:28  rft     last modification
//  30-mar-94 10:30:36  rft     created
// ===========================================================================

extern "C++"
{

// ================================================================ class Crd2

INLINE_MEMBER Crd2::Crd2()
{
    TR_FUN0(Crd2_constructors,Crd2);
}

INLINE_MEMBER Crd2::Crd2(double x0new, double x1new)
{
    x[0] = x0new;
    x[1] = x1new;

    TR_FUN2(Crd2_constructors,Crd2,x0new,x1new);
}

INLINE_MEMBER Crd2::Crd2(const Crd2 & crd)
{
    x[0] = crd.x[0];
    x[1] = crd.x[1];

    TR_FUN1(Crd2_constructors,Crd2,crd);
}

INLINE_MEMBER Crd2::~Crd2()
{
    TR_MESSAGE(Crd2_constructors,"~" << *this);
}

INLINE_MEMBER double Crd2::operator[](int index) const
{
    CH_CRD2_CODE(
        if (index < 0 || index > 1)
            EH_FATAL(Crd2_errors,"index == " << index);
    )

    return x[index];
}

INLINE_MEMBER double & Crd2::operator[](int index)
{
    CH_CRD2_CODE(
        if (index < 0 || index > 1)
            EH_FATAL(Crd2_errors,"index == " << index);
    )

    return x[index];
}

INLINE_MEMBER Crd2 & Crd2::operator+=(const Crd2 & crd)
{
    x[0] += crd.x[0];
    x[1] += crd.x[1];

    return *this;
}

INLINE_MEMBER Crd2 & Crd2::operator-=(const Crd2 & crd)
{
    x[0] -= crd.x[0];
    x[1] -= crd.x[1];

    return *this;
}

INLINE_MEMBER Crd2 & Crd2::operator*=(double scalar)
{
    x[0] *= scalar;
    x[1] *= scalar;

    return *this;
}

INLINE_MEMBER Crd2 & Crd2::operator/=(double scalar)
{
    CH_CRD2_CODE(EH_FATAL_CONDITION(Crd2_errors,scalar == 0.0));

    x[0] /= scalar;
    x[1] /= scalar;

    return *this;
}

INLINE_MEMBER double Crd2::x0() const { return x[0]; }
INLINE_MEMBER double Crd2::x1() const { return x[1]; }

INLINE_MEMBER void Crd2::x0(double x0new) { x[0] = x0new; }
INLINE_MEMBER void Crd2::x1(double x1new) { x[1] = x1new; }

INLINE_GLOBAL double squaredLength(const Crd2 & crd)
{
    return (square(crd.x[0]) + square(crd.x[1]));
}

INLINE_GLOBAL double length(const Crd2 & crd)
{
    return sqrt(squaredLength(crd));
}

INLINE_MEMBER void Crd2::normalize()
{
    double crdLength = length(*this);

    *this /= crdLength;
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const Crd2 & crd)
{
    s << "Crd2(" << crd.x[0] << "," << crd.x[1] << ")";
    return s;
}

INLINE_GLOBAL Crd2 operator-(const Crd2 & crd)
{
    return Crd2(-crd.x[0], -crd.x[1]);
}

INLINE_GLOBAL Crd2 operator+(const Crd2 & crd1, const Crd2 & crd2)
{
    return Crd2(crd1.x[0] + crd2.x[0],
                crd1.x[1] + crd2.x[1]);
}

INLINE_GLOBAL Crd2 operator-(const Crd2 & crd1, const Crd2 & crd2)
{
    return Crd2(crd1.x[0] - crd2.x[0],
                crd1.x[1] - crd2.x[1]);
}

INLINE_GLOBAL Crd2 operator*(const Crd2 & crd, double scalar)
{
    return Crd2(crd.x[0] * scalar,
                crd.x[1] * scalar);
}

INLINE_GLOBAL Crd2 operator*(double scalar, const Crd2 & crd)
{
    return Crd2(crd.x[0] * scalar,
                crd.x[1] * scalar);
}

INLINE_GLOBAL Crd2 operator/(const Crd2 & crd, double scalar)
{
    CH_CRD2_CODE(EH_FATAL_CONDITION(Crd2_errors,scalar == 0.0));

    return Crd2(crd.x[0] / scalar,
                crd.x[1] / scalar);
}

INLINE_GLOBAL double dotProduct(const Crd2 & crd1, const Crd2 & crd2)
{
    return (crd1.x[0] * crd2.x[0] +
            crd1.x[1] * crd2.x[1]);
}

INLINE_GLOBAL Crd2 normalizedCopy(const Crd2 & crd)
{
    double crdLength = length(crd);

    return (crd / crdLength);
}

INLINE_MEMBER void Crd2::cwMultiplyBy(const Crd2 & crd)
{
    x[0] *= crd.x[0];
    x[1] *= crd.x[1];
}

INLINE_MEMBER void Crd2::cwDevideBy(const Crd2 & crd)
{
    CH_CRD2_CODE(EH_FATAL_CONDITION(Crd2_errors,crd[0] == 0.0));
    CH_CRD2_CODE(EH_FATAL_CONDITION(Crd2_errors,crd[1] == 0.0));

    x[0] /= crd.x[0];
    x[1] /= crd.x[1];
}

INLINE_GLOBAL Crd2 cwReciprocalCopy(const Crd2 & crd)
{
    Crd2 newVec;
    
    if (crd.x[0] != 0.0)
        newVec.x[0] = 1.0 / crd.x[0];
    else
        newVec.x[0] = 0.0;

    if (crd.x[1] != 0.0)
        newVec.x[1] = 1.0 / crd.x[1];
    else
        newVec.x[1] = 0.0;

    return (newVec);
}

INLINE_GLOBAL Crd2 cwProduct(const Crd2 & crd1, const Crd2 & crd2)
{
    return (Crd2(crd1.x[0] * crd2.x[0],
                 crd1.x[1] * crd2.x[1]));
}

INLINE_GLOBAL Crd2 cwQuotient(const Crd2 & crd1, const Crd2 & crd2)
{
    CH_CRD2_CODE(EH_FATAL_CONDITION(Crd2_errors,crd2[0] == 0.0));
    CH_CRD2_CODE(EH_FATAL_CONDITION(Crd2_errors,crd2[1] == 0.0));

    return Crd2(crd1.x[0] / crd2.x[0],
                crd1.x[1] / crd2.x[1]);
}

INLINE_GLOBAL Crd2 cwMin(const Crd2 & crd1, const Crd2 & crd2)
{
    return Crd2(min(crd1.x[0],crd2.x[0]),
                min(crd1.x[1],crd2.x[1]));
}

INLINE_GLOBAL Crd2 cwMax(const Crd2 & crd1, const Crd2 & crd2)
{
    return Crd2(max(crd1.x[0],crd2.x[0]),
                max(crd1.x[1],crd2.x[1]));
}

INLINE_MEMBER Crd2 & Crd2::scale(const Crd2 & s)
{
    x[0] *= s[0];
    x[1] *= s[1];

    return *this;
}

INLINE_MEMBER Crd2 & Crd2::scale(double s0, double s1)
{
    x[0] *= s0;
    x[1] *= s1;

    return *this;
}

INLINE_MEMBER Crd2 & Crd2::rot(double cosAlpha, double sinAlpha)
{
    double t = x[0];
    
    x[0] = x[0] * cosAlpha - x[1] * sinAlpha;
    x[1] = x[1] * cosAlpha +    t * sinAlpha;

    return *this;
}

INLINE_MEMBER Crd2 & Crd2::shear0(double x0s)
{
    x[0] += x[1] * x0s;

    return *this;
}

INLINE_MEMBER Crd2 & Crd2::shear1(double x1s)
{
    x[1] += x[0] * x1s;

    return *this;
}

INLINE_MEMBER Crd2 & Crd2::operator*=(const Mat2 & mat)
{
    double t = x[0];

    x[0] = x[0] * mat.x[0][0] + x[1] * mat.x[1][0];
    x[1] =    t * mat.x[0][1] + x[1] * mat.x[1][1];
    
    return *this;
}

INLINE_GLOBAL Crd2 prCrdScale(const Crd2 & crd, const Crd2 & s)
{
    return Crd2(crd.x[0] * s[0], crd.x[1] * s[1]);
}

INLINE_GLOBAL Crd2 prCrdScale(const Crd2 & crd, double s0, double s1)
{
    return Crd2(crd.x[0] * s0, crd.x[1] * s1);
}

INLINE_GLOBAL Crd2 prCrdRot(const Crd2 & crd, double cosAlpha, double sinAlpha)
{
    return Crd2(crd.x[0] * cosAlpha - crd.x[1] * sinAlpha,
                crd.x[1] * cosAlpha + crd.x[0] * sinAlpha);
}

INLINE_GLOBAL Crd2 prCrdShear0(const Crd2 & crd, double x0s)
{
    return Crd2(crd.x[0] + crd.x[1] * x0s, crd.x[1]);
}

INLINE_GLOBAL Crd2 prCrdShear1(const Crd2 & crd, double x1s)
{
    return Crd2(crd.x[0], crd.x[1] + crd.x[0] * x1s);
}

INLINE_GLOBAL Crd2 operator*(const Crd2 & crd, const Mat2 & mat)
{
    return
        Crd2
        (
            crd.x[0] * mat.x[0][0] + crd.x[1] * mat.x[1][0],
	    crd.x[0] * mat.x[0][1] + crd.x[1] * mat.x[1][1]
        );
}

// ================================================================ class Mat2

INLINE_MEMBER Mat2::Mat2()
{
    TR_FUN0(Crd2_constructors,Mat2);
}

INLINE_MEMBER Mat2::Mat2(double x00, double x01,
                  double x10, double x11)
{
    x[0][0] = x00; x[0][1] = x01;
    x[1][0] = x10; x[1][1] = x11;

    TR_FUN4(Crd2_constructors,Mat2,x00,x01,x10,x11);
}

INLINE_MEMBER Mat2::Mat2(const Mat2 & mat)
{
    x[0][0] = mat.x[0][0];
    x[0][1] = mat.x[0][1];
    x[1][0] = mat.x[1][0];
    x[1][1] = mat.x[1][1];
    TR_FUN1(Crd2_constructors,Mat2,mat);
}


INLINE_MEMBER Mat2::~Mat2()
{
    TR_MESSAGE(Crd2_constructors,"~" << *this);
}

INLINE_MEMBER double Mat2::operator()(int i, int j) const
{
    CH_CRD2_CODE(
        if (i < 0 || i > 1)
            EH_FATAL(Crd2_errors,"i == " << i);
        if (j < 0 || j > 1)
            EH_FATAL(Crd2_errors,"j == " << j);
    )
    return x[i][j];
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const Mat2 & mat)
{
    s << "Mat2(" << mat.x[0][0] << "," << mat.x[0][1] << ","
        << mat.x[1][0] << "," << mat.x[1][1] << ")";
    return s;
}


INLINE_GLOBAL double determinant(const Mat2 & mat)
{
    return
        det2x2(
            mat.x[0][0], mat.x[0][1],
            mat.x[1][0], mat.x[1][1]
        );
}

INLINE_GLOBAL Mat2 inverseCopy(const Mat2 & mat, double det)
{
    det = 1.0 / det;

    return
	Mat2
	(
	     det * mat.x[1][1], -det * mat.x[0][1],
	    -det * mat.x[1][0],  det * mat.x[0][0]
	);
}

INLINE_GLOBAL Mat2 inverseCopy(const Mat2 & mat)
{
    double det = determinant(mat);

    if (det == 0.0)
    {
        EH_ERROR(Crd2_errors, ERROR_CRD2_SINGULAR_MATRIX, "singular matrix");
        return mat;
    }

    return inverseCopy(mat,det);
}

INLINE_MEMBER void Mat2::invert()
{
    *this = inverseCopy(*this);
}

INLINE_GLOBAL Mat2 transposedCopy(const Mat2 & mat)
{
    return
        Mat2(
            mat.x[0][0], mat.x[1][0],
            mat.x[0][1], mat.x[1][1]
        );
}

INLINE_MEMBER void Mat2::transpose()
{
    *this = transposedCopy(*this);
}

INLINE_GLOBAL Mat2 operator-(const Mat2 & mat)
{
    return
        Mat2
        (
            - mat.x[0][0], - mat.x[0][1],
            - mat.x[1][0], - mat.x[1][1]
        );
}

INLINE_GLOBAL Mat2 operator+(const Mat2 & mat1, const Mat2 & mat2)
{
    return
        Mat2
        (
            mat1.x[0][0] + mat2.x[0][0], mat1.x[0][1] + mat2.x[0][1],
            mat1.x[1][0] + mat2.x[1][0], mat1.x[1][1] + mat2.x[1][1]
        );
}

INLINE_GLOBAL Mat2 operator-(const Mat2 & mat1, const Mat2 & mat2)
{
    return
        Mat2
        (
            mat1.x[0][0] - mat2.x[0][0], mat1.x[0][1] - mat2.x[0][1],
            mat1.x[1][0] - mat2.x[1][0], mat1.x[1][1] - mat2.x[1][1]
        );
}

INLINE_GLOBAL Mat2 operator*(const Mat2 & mat, double s)
{
    return
        Mat2
        (
            mat.x[0][0] * s, mat.x[0][1] * s,
            mat.x[1][0] * s, mat.x[1][1] * s
        );
}

INLINE_GLOBAL Mat2 operator*(double s, const Mat2 & mat)
{
    return
        Mat2
        (
            s * mat.x[0][0], s * mat.x[0][1],
            s * mat.x[1][0], s * mat.x[1][1]
        );
}

INLINE_MEMBER Mat2 & Mat2::operator*=(double s)
{
    x[0][0] *= s; x[0][1] *= s;
    x[1][0] *= s; x[1][1] *= s;
    return *this;
}

INLINE_MEMBER Mat2 & Mat2::scale(const Crd2 & s)
{
    x[0][0] *= s[0]; x[0][1] *= s[1];
    x[1][0] *= s[0]; x[1][1] *= s[1];
    return *this;
}

INLINE_MEMBER Mat2 & Mat2::scale(double s0, double s1)
{
    x[0][0] *= s0; x[0][1] *= s1;
    x[1][0] *= s0; x[1][1] *= s1;
    return *this;
}

INLINE_MEMBER Mat2 & Mat2::rot(double cosAlpha, double sinAlpha)
{
    double t;

    t = x[0][0];
    x[0][0] = t * cosAlpha - x[0][1] * sinAlpha;
    x[0][1] = t * sinAlpha + x[0][1] * cosAlpha;
    t = x[1][0];
    x[1][0] = t * cosAlpha - x[1][1] * sinAlpha;
    x[1][1] = t * sinAlpha + x[1][1] * cosAlpha;
    return *this;
}

INLINE_MEMBER Mat2 & Mat2::shear0(double x0s)
{
    x[0][0] += x[0][1] * x0s;
    x[1][0] += x[1][1] * x0s;

    return *this;
}

INLINE_MEMBER Mat2 & Mat2::shear1(double x1s)
{
    x[0][1] += x[0][0] * x1s;
    x[1][1] += x[1][0] * x1s;

    return *this;
}

INLINE_MEMBER Mat2 & Mat2::operator+=(const Mat2 & mat)
{
    x[0][0] += mat.x[0][0];
    x[0][1] += mat.x[0][1];
    x[1][0] += mat.x[1][0];
    x[1][1] += mat.x[1][1];

    return *this;
}

INLINE_GLOBAL Mat2 operator*(const Mat2 & mat1, const Mat2 & mat2)
{
    return
        Mat2
        (
            mat1.x[0][0] * mat2.x[0][0] + mat1.x[0][1] * mat2.x[1][0],
            mat1.x[0][0] * mat2.x[0][1] + mat1.x[0][1] * mat2.x[1][1],
            mat1.x[1][0] * mat2.x[0][0] + mat1.x[1][1] * mat2.x[1][0],
            mat1.x[1][0] * mat2.x[0][1] + mat1.x[1][1] * mat2.x[1][1]
        );
}

INLINE_MEMBER Mat2 & Mat2::operator*=(const Mat2 & mat)
{
    *this = *this * mat;

    return *this;
}

INLINE_GLOBAL Mat2 prMatScale(const Mat2 & mat, const Crd2 & s)
{
    return
        Mat2
        (
            mat.x[0][0] * s[0], mat.x[0][1] * s[1],
            mat.x[1][0] * s[0], mat.x[1][1] * s[1]
        );
}

INLINE_GLOBAL Mat2 prMatScale(const Mat2 & mat, double s0, double s1)
{
    return
        Mat2
        (
            mat.x[0][0] * s0, mat.x[0][1] * s1,
            mat.x[1][0] * s0, mat.x[1][1] * s1
        );
}

INLINE_GLOBAL Mat2 prMatRot(const Mat2 & mat, double cosAlpha, double sinAlpha)
{
    return
        Mat2
        (
            mat.x[0][0] * cosAlpha - mat.x[0][1] * sinAlpha,
            mat.x[0][1] * cosAlpha + mat.x[0][0] * sinAlpha,

            mat.x[1][0] * cosAlpha - mat.x[1][1] * sinAlpha,
            mat.x[1][1] * cosAlpha + mat.x[1][0] * sinAlpha
        );
}

INLINE_GLOBAL Mat2 prMatShear0(const Mat2 & mat, double x0s)
{
    return
	Mat2
	(
	    mat.x[0][0] + mat.x[0][1] * x0s, mat.x[0][1],
	    mat.x[1][0] + mat.x[1][1] * x0s, mat.x[1][1]
	);
}

INLINE_GLOBAL Mat2 prMatShear1(const Mat2 & mat, double x1s)
{
    return
	Mat2
	(
	    mat.x[0][0], mat.x[0][1] + mat.x[0][0] * x1s,
	    mat.x[1][0], mat.x[1][1] + mat.x[1][0] * x1s
	);
}

INLINE_GLOBAL Mat2 prScaleMat(const Crd2 & s, const Mat2 & mat)
{
    return
        Mat2
        (
            s[0] * mat.x[0][0], s[0] * mat.x[0][1],
            s[1] * mat.x[1][0], s[1] * mat.x[1][1]
        );
}

INLINE_GLOBAL Mat2 prScaleMat(double s0, double s1, const Mat2 & mat)
{
    return
        Mat2
        (
            s0 * mat.x[0][0], s0 * mat.x[0][1],
            s1 * mat.x[1][0], s1 * mat.x[1][1]
        );
}

INLINE_GLOBAL Mat2 prRotMat(double cosAlpha, double sinAlpha, const Mat2 & mat)
{
    return
        Mat2
        (
            cosAlpha * mat.x[0][0] + sinAlpha * mat.x[1][0],
            cosAlpha * mat.x[0][1] + sinAlpha * mat.x[1][1],
            cosAlpha * mat.x[1][0] - sinAlpha * mat.x[0][0],
            cosAlpha * mat.x[1][1] - sinAlpha * mat.x[0][1]
        );
}

INLINE_GLOBAL Mat2 prShear0Mat(double x0s, const Mat2 & mat)
{
    return
	Mat2
	(
	    mat.x[0][0],
	    mat.x[0][1],
	    x0s * mat.x[0][0] + mat.x[1][0],
	    x0s * mat.x[0][1] + mat.x[1][1]
	);
}

INLINE_GLOBAL Mat2 prShear1Mat(double x1s, const Mat2 & mat)
{
    return
	Mat2
	(
	    mat.x[0][0] + x1s * mat.x[1][0],
	    mat.x[0][1] + x1s * mat.x[1][1],
	    mat.x[1][0],
	    mat.x[1][1]
	);
}

} // extern "C++"

// ===========================================================================
