// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       Crd4
//  TYPE:       c++ inline
//  PROJECT:    VEGA Math Library
//  CONTENT:    base types for 4D coordinate manipulation using doubles
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rft     Robert F. Tobler
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 14:06:28  rft     last modification
//  25-mar-94 23:09:46  rft     created
// ===========================================================================

extern "C++"
{

// ================================================================ class Crd4

INLINE_MEMBER Crd4::Crd4()
{
    TR_FUN0(Crd4_constructors,Crd4);
}

INLINE_MEMBER Crd4::Crd4(
	double x0new, double x1new, double x2new, double x3new
	)
{
    x[0] = x0new;
    x[1] = x1new;
    x[2] = x2new;
    x[3] = x3new;

    TR_FUN4(Crd4_constructors,Crd4,x0new,x1new,x2new,x3new);
}

INLINE_MEMBER Crd4::Crd4(const Crd4 & crd)
{
    x[0] = crd.x[0];
    x[1] = crd.x[1];
    x[2] = crd.x[2];
    x[3] = crd.x[3];

    TR_FUN1(Crd4_constructors,Crd4,crd);
}

INLINE_MEMBER Crd4::~Crd4()
{
    TR_MESSAGE(Crd4_constructors,"~" << *this);
}

INLINE_MEMBER double Crd4::operator[](int index) const
{
    CH_CRD4_CODE(
        if (index < 0 || index > 3)
            EH_FATAL(Crd4_errors,"index == " << index);
    )

    return x[index];
}

INLINE_MEMBER double & Crd4::operator[](int index)
{
    CH_CRD4_CODE(
        if (index < 0 || index > 3)
            EH_FATAL(Crd4_errors,"index == " << index);
    )

    return x[index];
}

INLINE_MEMBER Crd4 & Crd4::operator+=(const Crd4 & crd)
{
    x[0] += crd.x[0];
    x[1] += crd.x[1];
    x[2] += crd.x[2];
    x[3] += crd.x[3];

    return *this;
}

INLINE_MEMBER Crd4 & Crd4::operator-=(const Crd4 & crd)
{
    x[0] -= crd.x[0];
    x[1] -= crd.x[1];
    x[2] -= crd.x[2];
    x[3] -= crd.x[3];

    return *this;
}

INLINE_MEMBER Crd4 & Crd4::operator*=(double scalar)
{
    x[0] *= scalar;
    x[1] *= scalar;
    x[2] *= scalar;
    x[3] *= scalar;

    return *this;
}

INLINE_MEMBER Crd4 & Crd4::operator/=(double scalar)
{
    CH_CRD4_CODE(EH_FATAL_CONDITION(Crd4_errors,scalar == 0.0));

    x[0] /= scalar;
    x[1] /= scalar;
    x[2] /= scalar;
    x[3] /= scalar;

    return *this;
}

INLINE_MEMBER double Crd4::x0() const { return x[0]; }
INLINE_MEMBER double Crd4::x1() const { return x[1]; }
INLINE_MEMBER double Crd4::x2() const { return x[2]; }
INLINE_MEMBER double Crd4::x3() const { return x[3]; }

INLINE_MEMBER void Crd4::x0(double x0new) { x[0] = x0new; }
INLINE_MEMBER void Crd4::x1(double x1new) { x[1] = x1new; }
INLINE_MEMBER void Crd4::x2(double x2new) { x[2] = x2new; }
INLINE_MEMBER void Crd4::x3(double x3new) { x[3] = x3new; }

INLINE_GLOBAL double squaredLength(const Crd4 & crd)
{
    return (square(crd.x[0]) + square(crd.x[1]) +
            square(crd.x[2]) + square(crd.x[3]));
}

INLINE_GLOBAL double length(const Crd4 & crd)
{
    return sqrt(squaredLength(crd));
}

INLINE_MEMBER void Crd4::normalize()
{
    double crdLength = length(*this);

    *this /= crdLength;
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const Crd4 & crd)
{
    s << "Crd4(" << crd.x[0] << "," << crd.x[1] << "," << crd.x[2]
      << "," << crd.x[3] << ")";
    return s;
}

INLINE_GLOBAL Crd4 operator-(const Crd4 & crd)
{
    return Crd4(-crd.x[0], -crd.x[1], -crd.x[2], -crd.x[3]);
}

INLINE_GLOBAL Crd4 operator+(const Crd4 & crd1, const Crd4 & crd2)
{
    return Crd4(crd1.x[0] + crd2.x[0],
                crd1.x[1] + crd2.x[1],
                crd1.x[2] + crd2.x[2],
                crd1.x[3] + crd2.x[3]);
}

INLINE_GLOBAL Crd4 operator-(const Crd4 & crd1, const Crd4 & crd2)
{
    return Crd4(crd1.x[0] - crd2.x[0],
                crd1.x[1] - crd2.x[1],
                crd1.x[2] - crd2.x[2],
                crd1.x[3] - crd2.x[3]);
}

INLINE_GLOBAL Crd4 operator*(const Crd4 & crd, double scalar)
{
    return Crd4(crd.x[0] * scalar,
                crd.x[1] * scalar,
                crd.x[2] * scalar,
                crd.x[3] * scalar);
}

INLINE_GLOBAL Crd4 operator*(double scalar, const Crd4 & crd)
{
    return Crd4(crd.x[0] * scalar,
                crd.x[1] * scalar,
                crd.x[2] * scalar,
                crd.x[3] * scalar);
}

INLINE_GLOBAL Crd4 operator/(const Crd4 & crd, double scalar)
{
    CH_CRD4_CODE(EH_FATAL_CONDITION(Crd4_errors,scalar == 0.0));

    return Crd4(crd.x[0] / scalar,
                crd.x[1] / scalar,
                crd.x[2] / scalar,
                crd.x[3] / scalar);
}

INLINE_GLOBAL double dotProduct(const Crd4 & crd1, const Crd4 & crd2)
{
    return (crd1.x[0] * crd2.x[0] +
            crd1.x[1] * crd2.x[1] +
            crd1.x[2] * crd2.x[2] +
            crd1.x[3] * crd2.x[3]);
}

INLINE_GLOBAL Crd4 normalizedCopy(const Crd4 & crd)
{
    double crdLength = length(crd);

    return (crd / crdLength);
}

INLINE_MEMBER void Crd4::cwMultiplyBy(const Crd4 & crd)
{
    x[0] *= crd.x[0];
    x[1] *= crd.x[1];
    x[2] *= crd.x[2];
    x[3] *= crd.x[3];
}

INLINE_MEMBER void Crd4::cwDevideBy(const Crd4 & crd)
{
    CH_CRD4_CODE(EH_FATAL_CONDITION(Crd4_errors,crd[0] == 0.0));
    CH_CRD4_CODE(EH_FATAL_CONDITION(Crd4_errors,crd[1] == 0.0));
    CH_CRD4_CODE(EH_FATAL_CONDITION(Crd4_errors,crd[2] == 0.0));
    CH_CRD4_CODE(EH_FATAL_CONDITION(Crd4_errors,crd[3] == 0.0));

    x[0] /= crd.x[0];
    x[1] /= crd.x[1];
    x[2] /= crd.x[2];
    x[3] /= crd.x[3];
}

INLINE_GLOBAL Crd4 cwReciprocalCopy(const Crd4 & crd)
{
    Crd4 newVec;
    
    if (crd.x[0] != 0.0)
        newVec.x[0] = 1.0 / crd.x[0];
    else
        newVec.x[0] = 0.0;

    if (crd.x[1] != 0.0)
        newVec.x[1] = 1.0 / crd.x[1];
    else
        newVec.x[1] = 0.0;

    if (crd.x[2] != 0.0)
        newVec.x[2] = 1.0 / crd.x[2];
    else
        newVec.x[2] = 0.0;

    if (crd.x[3] != 0.0)
        newVec.x[3] = 1.0 / crd.x[3];
    else
        newVec.x[3] = 0.0;

    return (newVec);
}

INLINE_GLOBAL Crd4 cwProduct(const Crd4 & crd1, const Crd4 & crd2)
{
    return (Crd4(crd1.x[0] * crd2.x[0],
                 crd1.x[1] * crd2.x[1],
                 crd1.x[2] * crd2.x[2],
                 crd1.x[3] * crd2.x[3]));
}

INLINE_GLOBAL Crd4 cwQuotient(const Crd4 & crd1, const Crd4 & crd2)
{
    CH_CRD4_CODE(EH_FATAL_CONDITION(Crd4_errors,crd2[0] == 0.0));
    CH_CRD4_CODE(EH_FATAL_CONDITION(Crd4_errors,crd2[1] == 0.0));
    CH_CRD4_CODE(EH_FATAL_CONDITION(Crd4_errors,crd2[2] == 0.0));

    return Crd4(crd1.x[0] / crd2.x[0],
                crd1.x[1] / crd2.x[1],
                crd1.x[2] / crd2.x[2],
                crd1.x[3] / crd2.x[3]);
}

INLINE_GLOBAL Crd4 cwMin(const Crd4 & crd1, const Crd4 & crd2)
{
    return Crd4(min(crd1.x[0],crd2.x[0]),
                min(crd1.x[1],crd2.x[1]),
                min(crd1.x[2],crd2.x[2]),
                min(crd1.x[3],crd2.x[3]));
}

INLINE_GLOBAL Crd4 cwMax(const Crd4 & crd1, const Crd4 & crd2)
{
    return Crd4(max(crd1.x[0],crd2.x[0]),
                max(crd1.x[1],crd2.x[1]),
                max(crd1.x[2],crd2.x[2]),
                max(crd1.x[3],crd2.x[3]));
}

INLINE_MEMBER Crd4 & Crd4::scale(const Crd4 & s)
{
    x[0] *= s[0];
    x[1] *= s[1];
    x[2] *= s[2];
    x[3] *= s[3];

    return *this;
}

INLINE_MEMBER Crd4 & Crd4::scale(double s0, double s1, double s2, double s3)
{
    x[0] *= s0;
    x[1] *= s1;
    x[2] *= s2;
    x[3] *= s3;

    return *this;
}

INLINE_MEMBER Crd4 & Crd4::operator*=(const Mat4 & mat)
{
    double t0 = x[0];
    double t1 = x[1];
    double t2 = x[2];

    x[0] = x[0] * mat.x[0][0] + x[1] * mat.x[1][0] +
           x[2] * mat.x[2][0] + x[3] * mat.x[3][0];
    x[1] = t0   * mat.x[0][1] + x[1] * mat.x[1][1] +
           x[2] * mat.x[2][1] + x[3] * mat.x[3][1];
    x[2] = t0   * mat.x[0][2] + t1   * mat.x[1][2] +
           x[2] * mat.x[2][2] + x[3] * mat.x[3][2];
    x[3] = t0   * mat.x[0][3] + t1   * mat.x[1][3] +
           t2   * mat.x[2][3] + x[3] * mat.x[3][3];
    
    return *this;
}

INLINE_GLOBAL Crd4 prCrdScale(const Crd4 & crd, const Crd4 & s)
{
    return Crd4(crd.x[0] * s[0], crd.x[1] * s[1],
                crd.x[2] * s[2], crd.x[3] * s[3]);
}

INLINE_GLOBAL Crd4 prCrdScale(const Crd4 & crd,
        double s0, double s1, double s2, double s3)
{
    return Crd4(crd.x[0] * s0, crd.x[1] * s1,
                crd.x[2] * s2, crd.x[3] * s3);
}

INLINE_GLOBAL Crd4 operator*(const Crd4 & crd, const Mat4 & mat)
{
    return
        Crd4
        (
            crd.x[0] * mat.x[0][0] +
            crd.x[1] * mat.x[1][0] +
            crd.x[2] * mat.x[2][0] +
            crd.x[3] * mat.x[3][0],

            crd.x[0] * mat.x[0][1] +
            crd.x[1] * mat.x[1][1] +
            crd.x[2] * mat.x[2][1] +
            crd.x[3] * mat.x[3][1],

            crd.x[0] * mat.x[0][2] +
            crd.x[1] * mat.x[1][2] +
            crd.x[2] * mat.x[2][2] +
            crd.x[3] * mat.x[3][2],

            crd.x[0] * mat.x[0][3] +
            crd.x[1] * mat.x[1][3] +
            crd.x[2] * mat.x[2][3] +
            crd.x[3] * mat.x[3][3]
        );
}

// ================================================================ class Mat4

INLINE_MEMBER Mat4::Mat4()
{
    TR_FUN0(Crd4_constructors,Mat4);
}

INLINE_MEMBER Mat4::Mat4(double x00, double x01, double x02, double x03,
                  double x10, double x11, double x12, double x13,
                  double x20, double x21, double x22, double x23,
                  double x30, double x31, double x32, double x33)
{
    x[0][0] = x00; x[0][1] = x01; x[0][2] = x02; x[0][3] = x03;
    x[1][0] = x10; x[1][1] = x11; x[1][2] = x12; x[1][3] = x13;
    x[2][0] = x20; x[2][1] = x21; x[2][2] = x22; x[2][3] = x23;
    x[3][0] = x30; x[3][1] = x31; x[3][2] = x32; x[3][3] = x33;

    TR_FUN4(Crd4_constructors,Mat4,x00,x01,x02,"...");
}

INLINE_MEMBER Mat4::Mat4(const Mat4 & mat)
{
    x[0][0] = mat.x[0][0];
    x[0][1] = mat.x[0][1];
    x[0][2] = mat.x[0][2];
    x[0][3] = mat.x[0][3];

    x[1][0] = mat.x[1][0];
    x[1][1] = mat.x[1][1];
    x[1][2] = mat.x[1][2];
    x[1][3] = mat.x[1][3];

    x[2][0] = mat.x[2][0];
    x[2][1] = mat.x[2][1];
    x[2][2] = mat.x[2][2];
    x[2][3] = mat.x[2][3];

    x[3][0] = mat.x[3][0];
    x[3][1] = mat.x[3][1];
    x[3][2] = mat.x[3][2];
    x[3][3] = mat.x[3][3];

    TR_FUN1(Crd4_constructors,Mat4,mat);
}


INLINE_MEMBER Mat4::~Mat4()
{
    TR_MESSAGE(Crd4_constructors,"~" << *this);
}

INLINE_MEMBER double Mat4::operator()(int i, int j) const
{
    CH_CRD4_CODE(
        if (i < 0 || i > 3)
            EH_FATAL(Crd3_errors,"i == " << i);
        if (j < 0 || j > 3)
            EH_FATAL(Crd3_errors,"j == " << j);
    )
    return x[i][j];
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const Mat4 & mat)
{
    s << "Mat4("
	<< mat.x[0][0] << ',' << mat.x[0][1] << ','
	<< mat.x[0][2] << ',' << mat.x[0][3] << ','
	<< mat.x[1][0] << ',' << mat.x[1][1] << ','
	<< mat.x[1][2] << ',' << mat.x[1][3] << ','
	<< mat.x[2][0] << ',' << mat.x[2][1] << ','
	<< mat.x[2][2] << ',' << mat.x[2][3] << ','
	<< mat.x[3][0] << ',' << mat.x[3][1] << ','
	<< mat.x[3][2] << ',' << mat.x[3][3] << ')';
    return s;
}

INLINE_GLOBAL double determinant(const Mat4 & m)
{
    return (  m.x[0][0] * det3x3(m.x[1][1], m.x[1][2], m.x[1][3],
                                 m.x[2][1], m.x[2][2], m.x[2][3],
                                 m.x[3][1], m.x[3][2], m.x[3][3])

            - m.x[0][1] * det3x3(m.x[1][0], m.x[1][2], m.x[1][3],
                                 m.x[2][0], m.x[2][2], m.x[2][3],
                                 m.x[3][0], m.x[3][2], m.x[3][3])

            + m.x[0][2] * det3x3(m.x[1][0], m.x[1][1], m.x[1][3],
                                 m.x[2][0], m.x[2][1], m.x[2][3],
                                 m.x[3][0], m.x[3][1], m.x[3][3])

            - m.x[0][3] * det3x3(m.x[1][0], m.x[1][1], m.x[1][2],
                                 m.x[2][0], m.x[2][1], m.x[2][2],
                                 m.x[3][0], m.x[3][1], m.x[3][2]));
}

INLINE_GLOBAL Mat4 adjointCopy(const Mat4 & m)
{
    Mat4 a;
    //
    //  Row column labeling reversed since we transpose rows & columns.
    //

    a.x[0][0] =   det3x3(m.x[1][1], m.x[1][2], m.x[1][3],
                         m.x[2][1], m.x[2][2], m.x[2][3],
                         m.x[3][1], m.x[3][2], m.x[3][3]);
    a.x[1][0] = - det3x3(m.x[1][0], m.x[1][2], m.x[1][3],
                         m.x[2][0], m.x[2][2], m.x[2][3],
                         m.x[3][0], m.x[3][2], m.x[3][3]);
    a.x[2][0] =   det3x3(m.x[1][0], m.x[1][1], m.x[1][3],
                         m.x[2][0], m.x[2][1], m.x[2][3],
                         m.x[3][0], m.x[3][1], m.x[3][3]);
    a.x[3][0] = - det3x3(m.x[1][0], m.x[1][1], m.x[1][2],
                         m.x[2][0], m.x[2][1], m.x[2][2],
                         m.x[3][0], m.x[3][1], m.x[3][2]);
        
    a.x[0][1] = - det3x3(m.x[0][1], m.x[0][2], m.x[0][3],
                         m.x[2][1], m.x[2][2], m.x[2][3],
                         m.x[3][1], m.x[3][2], m.x[3][3]);
    a.x[1][1] =   det3x3(m.x[0][0], m.x[0][2], m.x[0][3],
                         m.x[2][0], m.x[2][2], m.x[2][3],
                         m.x[3][0], m.x[3][2], m.x[3][3]);
    a.x[2][1] = - det3x3(m.x[0][0], m.x[0][1], m.x[0][3],
                         m.x[2][0], m.x[2][1], m.x[2][3],
                         m.x[3][0], m.x[3][1], m.x[3][3]);
    a.x[3][1] =   det3x3(m.x[0][0], m.x[0][1], m.x[0][2],
                         m.x[2][0], m.x[2][1], m.x[2][2],
                         m.x[3][0], m.x[3][1], m.x[3][2]);

    a.x[0][2] =   det3x3(m.x[0][1], m.x[0][2], m.x[0][3],
                         m.x[1][1], m.x[1][2], m.x[1][3],
                         m.x[3][1], m.x[3][2], m.x[3][3]);
    a.x[1][2] = - det3x3(m.x[0][0], m.x[0][2], m.x[0][3],
                         m.x[1][0], m.x[1][2], m.x[1][3],
                         m.x[3][0], m.x[3][2], m.x[3][3]);
    a.x[2][2] =   det3x3(m.x[0][0], m.x[0][1], m.x[0][3],
                         m.x[1][0], m.x[1][1], m.x[1][3],
                         m.x[3][0], m.x[3][1], m.x[3][3]);
    a.x[3][2] = - det3x3(m.x[0][0], m.x[0][1], m.x[0][2],
                         m.x[1][0], m.x[1][1], m.x[1][2],
                         m.x[3][0], m.x[3][1], m.x[3][2]);
        
    a.x[0][3] = - det3x3(m.x[0][1], m.x[0][2], m.x[0][3],
                         m.x[1][1], m.x[1][2], m.x[1][3],
                         m.x[2][1], m.x[2][2], m.x[2][3]);
    a.x[1][3] =   det3x3(m.x[0][0], m.x[0][2], m.x[0][3],
                         m.x[1][0], m.x[1][2], m.x[1][3],
                         m.x[2][0], m.x[2][2], m.x[2][3]);
    a.x[2][3] = - det3x3(m.x[0][0], m.x[0][1], m.x[0][3],
                         m.x[1][0], m.x[1][1], m.x[1][3],
                         m.x[2][0], m.x[2][1], m.x[2][3]);
    a.x[3][3] =   det3x3(m.x[0][0], m.x[0][1], m.x[0][2],
                         m.x[1][0], m.x[1][1], m.x[1][2],
                         m.x[2][0], m.x[2][1], m.x[2][2]);
        
    return a;
}

INLINE_GLOBAL Mat4 operator*(const Mat4 & mat, double s)
{
    return
        Mat4
        (
            mat.x[0][0] * s, mat.x[0][1] * s,
            mat.x[0][2] * s, mat.x[0][3] * s,
            mat.x[1][0] * s, mat.x[1][1] * s,
            mat.x[1][2] * s, mat.x[1][3] * s,
            mat.x[2][0] * s, mat.x[2][1] * s,
            mat.x[2][2] * s, mat.x[2][3] * s,
            mat.x[3][0] * s, mat.x[3][1] * s,
            mat.x[3][2] * s, mat.x[3][3] * s
        );
}

INLINE_GLOBAL Mat4 inverseCopy(const Mat4 & mat, double det)
{
    det = 1.0 / det;

    return
        adjointCopy(mat) * det;
}

INLINE_GLOBAL Mat4 inverseCopy(const Mat4 & mat)
{
    double det = determinant(mat);

    if (det == 0.0)
    {
        EH_ERROR(Crd4_errors, ERROR_CRD4_SINGULAR_MATRIX, "singular matrix");
        return mat;
    }

    return inverseCopy(mat,det);
}

INLINE_MEMBER void Mat4::invert()
{
    *this = inverseCopy(*this);
}

INLINE_GLOBAL Mat4 transposedCopy(const Mat4 & mat)
{
    return
        Mat4(
            mat.x[0][0], mat.x[1][0], mat.x[2][0], mat.x[3][0],
            mat.x[0][1], mat.x[1][1], mat.x[2][1], mat.x[3][1],
            mat.x[0][2], mat.x[1][2], mat.x[2][2], mat.x[3][2],
            mat.x[0][3], mat.x[1][3], mat.x[2][3], mat.x[3][3]
        );
}

INLINE_MEMBER void Mat4::transpose()
{
    *this = transposedCopy(*this);
}

INLINE_MEMBER Mat4 & Mat4::operator*=(double s)
{
    x[0][0] *= s; x[0][1] *= s; x[0][2] *= s; x[0][3] *= s;
    x[1][0] *= s; x[1][1] *= s; x[1][2] *= s; x[1][3] *= s;
    x[2][0] *= s; x[2][1] *= s; x[2][2] *= s; x[2][3] *= s;
    x[3][0] *= s; x[3][1] *= s; x[3][2] *= s; x[3][3] *= s;
    return *this;
}

INLINE_MEMBER Mat4 & Mat4::scale(const Crd4 & s)
{
    x[0][0] *= s[0]; x[0][1] *= s[1]; x[0][2] *= s[2]; x[0][3] *= s[3];
    x[1][0] *= s[0]; x[1][1] *= s[1]; x[1][2] *= s[2]; x[1][3] *= s[3];
    x[2][0] *= s[0]; x[2][1] *= s[1]; x[2][2] *= s[2]; x[2][3] *= s[3];
    x[3][0] *= s[0]; x[3][1] *= s[1]; x[3][2] *= s[2]; x[3][3] *= s[3];
    return *this;
}

INLINE_MEMBER Mat4 & Mat4::scale(double s0, double s1, double s2, double s3)
{
    x[0][0] *= s0; x[0][1] *= s1; x[0][2] *= s2; x[0][3] *= s3;
    x[1][0] *= s0; x[1][1] *= s1; x[1][2] *= s2; x[1][3] *= s3;
    x[2][0] *= s0; x[2][1] *= s1; x[2][2] *= s2; x[2][3] *= s3;
    x[3][0] *= s0; x[3][1] *= s1; x[3][2] *= s2; x[3][3] *= s3;
    return *this;
}

INLINE_MEMBER Mat4 & Mat4::operator+=(const Mat4 & mat)
{
    x[0][0] += mat.x[0][0];
    x[0][1] += mat.x[0][1];
    x[0][2] += mat.x[0][2];
    x[0][3] += mat.x[0][3];

    x[1][0] += mat.x[1][0];
    x[1][1] += mat.x[1][1];
    x[1][2] += mat.x[1][2];
    x[1][3] += mat.x[1][3];

    x[2][0] += mat.x[2][0];
    x[2][1] += mat.x[2][1];
    x[2][2] += mat.x[2][2];
    x[2][3] += mat.x[2][3];

    x[3][0] += mat.x[3][0];
    x[3][1] += mat.x[3][1];
    x[3][2] += mat.x[3][2];
    x[3][3] += mat.x[3][3];

    return *this;
}

INLINE_GLOBAL Mat4 operator*(const Mat4 & mat1, const Mat4 & mat2)
{
    return
        Mat4
        (
            mat1.x[0][0] * mat2.x[0][0] + mat1.x[0][1] * mat2.x[1][0] +
            mat1.x[0][2] * mat2.x[2][0] + mat1.x[0][3] * mat2.x[3][0],
            mat1.x[0][0] * mat2.x[0][1] + mat1.x[0][1] * mat2.x[1][1] +
            mat1.x[0][2] * mat2.x[2][1] + mat1.x[0][3] * mat2.x[3][1],
            mat1.x[0][0] * mat2.x[0][2] + mat1.x[0][1] * mat2.x[1][2] +
            mat1.x[0][2] * mat2.x[2][2] + mat1.x[0][3] * mat2.x[3][2],
            mat1.x[0][0] * mat2.x[0][3] + mat1.x[0][1] * mat2.x[1][3] +
            mat1.x[0][2] * mat2.x[2][3] + mat1.x[0][3] * mat2.x[3][3],

            mat1.x[1][0] * mat2.x[0][0] + mat1.x[1][1] * mat2.x[1][0] +
            mat1.x[1][2] * mat2.x[2][0] + mat1.x[1][3] * mat2.x[3][0],
            mat1.x[1][0] * mat2.x[0][1] + mat1.x[1][1] * mat2.x[1][1] +
            mat1.x[1][2] * mat2.x[2][1] + mat1.x[1][3] * mat2.x[3][1],
            mat1.x[1][0] * mat2.x[0][2] + mat1.x[1][1] * mat2.x[1][2] +
            mat1.x[1][2] * mat2.x[2][2] + mat1.x[1][3] * mat2.x[3][2],
            mat1.x[1][0] * mat2.x[0][3] + mat1.x[1][1] * mat2.x[1][3] +
            mat1.x[1][2] * mat2.x[2][3] + mat1.x[1][3] * mat2.x[3][3],

            mat1.x[2][0] * mat2.x[0][0] + mat1.x[2][1] * mat2.x[1][0] +
            mat1.x[2][2] * mat2.x[2][0] + mat1.x[2][3] * mat2.x[3][0],
            mat1.x[2][0] * mat2.x[0][1] + mat1.x[2][1] * mat2.x[1][1] +
            mat1.x[2][2] * mat2.x[2][1] + mat1.x[2][3] * mat2.x[3][1],
            mat1.x[2][0] * mat2.x[0][2] + mat1.x[2][1] * mat2.x[1][2] +
            mat1.x[2][2] * mat2.x[2][2] + mat1.x[2][3] * mat2.x[3][2],
            mat1.x[2][0] * mat2.x[0][3] + mat1.x[2][1] * mat2.x[1][3] +
            mat1.x[2][2] * mat2.x[2][3] + mat1.x[2][3] * mat2.x[3][3],

            mat1.x[3][0] * mat2.x[0][0] + mat1.x[3][1] * mat2.x[1][0] +
            mat1.x[3][2] * mat2.x[2][0] + mat1.x[3][3] * mat2.x[3][0],
            mat1.x[3][0] * mat2.x[0][1] + mat1.x[3][1] * mat2.x[1][1] +
            mat1.x[3][2] * mat2.x[2][1] + mat1.x[3][3] * mat2.x[3][1],
            mat1.x[3][0] * mat2.x[0][2] + mat1.x[3][1] * mat2.x[1][2] +
            mat1.x[3][2] * mat2.x[2][2] + mat1.x[3][3] * mat2.x[3][2],
            mat1.x[3][0] * mat2.x[0][3] + mat1.x[3][1] * mat2.x[1][3] +
            mat1.x[3][2] * mat2.x[2][3] + mat1.x[3][3] * mat2.x[3][3]
        );
}

INLINE_MEMBER Mat4 & Mat4::operator*=(const Mat4 & mat)
{
    *this = *this * mat;

    return *this;
}

INLINE_GLOBAL Mat4 prMatScale(const Mat4 & mat, const Crd4 & s)
{
    return
        Mat4
        (
            mat.x[0][0] * s[0], mat.x[0][1] * s[1],
            mat.x[0][2] * s[2], mat.x[0][3] * s[3],
            mat.x[1][0] * s[0], mat.x[1][1] * s[1],
            mat.x[1][2] * s[2], mat.x[1][3] * s[3],
            mat.x[2][0] * s[0], mat.x[2][1] * s[1],
            mat.x[2][2] * s[2], mat.x[2][3] * s[3],
            mat.x[3][0] * s[0], mat.x[3][1] * s[1],
            mat.x[3][2] * s[2], mat.x[3][3] * s[3]
        );
}

INLINE_GLOBAL Mat4 prMatScale(
        const Mat4 & mat,
        double s0, double s1, double s2, double s3)
{
    return
        Mat4
        (
            mat.x[0][0] * s0, mat.x[0][1] * s1,
            mat.x[0][2] * s2, mat.x[0][3] * s3,
            mat.x[1][0] * s0, mat.x[1][1] * s1,
            mat.x[1][2] * s2, mat.x[1][3] * s3,
            mat.x[2][0] * s0, mat.x[2][1] * s1,
            mat.x[2][2] * s2, mat.x[2][3] * s3,
            mat.x[3][0] * s0, mat.x[3][1] * s1,
            mat.x[3][2] * s2, mat.x[3][3] * s3
        );
}

INLINE_GLOBAL Mat4 operator*(double s, const Mat4 & mat)
{
    return
        Mat4
        (
            s * mat.x[0][0], s * mat.x[0][1],
            s * mat.x[0][2], s * mat.x[0][3],
            s * mat.x[1][0], s * mat.x[1][1],
            s * mat.x[1][2], s * mat.x[1][3],
            s * mat.x[2][0], s * mat.x[2][1],
            s * mat.x[2][2], s * mat.x[2][3],
            s * mat.x[3][0], s * mat.x[3][1],
            s * mat.x[3][2], s * mat.x[3][3]
        );
}

INLINE_GLOBAL Mat4 prScaleMat(const Crd4 & s, const Mat4 & mat)
{
    return
        Mat4
        (
            s[0] * mat.x[0][0], s[0] * mat.x[0][1],
            s[0] * mat.x[0][2], s[0] * mat.x[0][3],
            s[1] * mat.x[1][0], s[1] * mat.x[1][1],
            s[1] * mat.x[1][2], s[1] * mat.x[1][3],
            s[2] * mat.x[2][0], s[2] * mat.x[2][1],
            s[2] * mat.x[2][2], s[2] * mat.x[2][3],
            s[3] * mat.x[3][0], s[3] * mat.x[3][1],
            s[3] * mat.x[3][2], s[3] * mat.x[3][3]
        );
}

INLINE_GLOBAL Mat4 prScaleMat(
        double s0, double s1, double s2, double s3,
        const Mat4 & mat
        )
{
    return
        Mat4
        (
            s0 * mat.x[0][0], s0 * mat.x[0][1],
            s0 * mat.x[0][2], s0 * mat.x[0][3],
            s1 * mat.x[1][0], s1 * mat.x[1][1],
            s1 * mat.x[1][2], s1 * mat.x[1][3],
            s2 * mat.x[2][0], s2 * mat.x[2][1],
            s2 * mat.x[2][2], s2 * mat.x[2][3],
            s3 * mat.x[3][0], s3 * mat.x[3][1],
            s3 * mat.x[3][2], s3 * mat.x[3][3]
        );
}

INLINE_GLOBAL Mat4 operator-(const Mat4 & mat)
{
    return
        Mat4
        (
	    - mat.x[0][0], - mat.x[0][1], - mat.x[0][2], - mat.x[0][3],
	    - mat.x[1][0], - mat.x[1][1], - mat.x[1][2], - mat.x[1][3],
	    - mat.x[2][0], - mat.x[2][1], - mat.x[2][2], - mat.x[2][3],
	    - mat.x[3][0], - mat.x[3][1], - mat.x[3][2], - mat.x[3][3]
        );
}

INLINE_GLOBAL Mat4 operator+(const Mat4 & mat1, const Mat4 & mat2)
{
    return
        Mat4
        (
	    mat1.x[0][0] + mat2.x[0][0],
	    mat1.x[0][1] + mat2.x[0][1],
	    mat1.x[0][2] + mat2.x[0][2],
	    mat1.x[0][3] + mat2.x[0][3],
	
	    mat1.x[1][0] + mat2.x[1][0],
	    mat1.x[1][1] + mat2.x[1][1],
	    mat1.x[1][2] + mat2.x[1][2],
	    mat1.x[1][3] + mat2.x[1][3],
	
	    mat1.x[2][0] + mat2.x[2][0],
	    mat1.x[2][1] + mat2.x[2][1],
	    mat1.x[2][2] + mat2.x[2][2],
	    mat1.x[2][3] + mat2.x[2][3],
	
	    mat1.x[3][0] + mat2.x[3][0],
	    mat1.x[3][1] + mat2.x[3][1],
	    mat1.x[3][2] + mat2.x[3][2],
	    mat1.x[3][3] + mat2.x[3][3]
        );
}

INLINE_GLOBAL Mat4 operator-(const Mat4 & mat1, const Mat4 & mat2)
{
    return
        Mat4
        (
	    mat1.x[0][0] - mat2.x[0][0],
	    mat1.x[0][1] - mat2.x[0][1],
	    mat1.x[0][2] - mat2.x[0][2],
	    mat1.x[0][3] - mat2.x[0][3],
	
	    mat1.x[1][0] - mat2.x[1][0],
	    mat1.x[1][1] - mat2.x[1][1],
	    mat1.x[1][2] - mat2.x[1][2],
	    mat1.x[1][3] - mat2.x[1][3],
	
	    mat1.x[2][0] - mat2.x[2][0],
	    mat1.x[2][1] - mat2.x[2][1],
	    mat1.x[2][2] - mat2.x[2][2],
	    mat1.x[2][3] - mat2.x[2][3],
	
	    mat1.x[3][0] - mat2.x[3][0],
	    mat1.x[3][1] - mat2.x[3][1],
	    mat1.x[3][2] - mat2.x[3][2],
	    mat1.x[3][3] - mat2.x[3][3]
        );
}

} // extern "C++"

// ===========================================================================
