// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       FCrd2
//  TYPE:       c++ inline
//  PROJECT:    VEGA Math Library
//  CONTENT:    base types for 2D coordinate manipulation using floats
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rft     Robert F. Tobler
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 14:06:28  rft     last modification
//  30-mar-94 10:30:36  rft     created
// ===========================================================================

extern "C++"
{

// ================================================================ class FCrd2

INLINE_MEMBER FCrd2::FCrd2()
{
    TR_FUN0(FCrd2_constructors,FCrd2);
}

INLINE_MEMBER FCrd2::FCrd2(float x0new, float x1new)
{
    x[0] = x0new;
    x[1] = x1new;

    TR_FUN2(FCrd2_constructors,FCrd2,x0new,x1new);
}

INLINE_MEMBER FCrd2::FCrd2(const FCrd2 & crd)
{
    x[0] = crd.x[0];
    x[1] = crd.x[1];

    TR_FUN1(FCrd2_constructors,FCrd2,crd);
}

INLINE_MEMBER FCrd2::~FCrd2()
{
    TR_MESSAGE(FCrd2_constructors,"~" << *this);
}

INLINE_MEMBER float FCrd2::operator[](int index) const
{
    CH_FCRD2_CODE(
        if (index < 0 || index > 1)
            EH_FATAL(FCrd2_errors,"index == " << index);
    )

    return x[index];
}

INLINE_MEMBER float & FCrd2::operator[](int index)
{
    CH_FCRD2_CODE(
        if (index < 0 || index > 1)
            EH_FATAL(FCrd2_errors,"index == " << index);
    )

    return x[index];
}

INLINE_MEMBER FCrd2 & FCrd2::operator+=(const FCrd2 & crd)
{
    x[0] += crd.x[0];
    x[1] += crd.x[1];

    return *this;
}

INLINE_MEMBER FCrd2 & FCrd2::operator-=(const FCrd2 & crd)
{
    x[0] -= crd.x[0];
    x[1] -= crd.x[1];

    return *this;
}

INLINE_MEMBER FCrd2 & FCrd2::operator*=(float scalar)
{
    x[0] *= scalar;
    x[1] *= scalar;

    return *this;
}

INLINE_MEMBER FCrd2 & FCrd2::operator/=(float scalar)
{
    CH_FCRD2_CODE(EH_FATAL_CONDITION(FCrd2_errors,scalar == 0.0));

    x[0] /= scalar;
    x[1] /= scalar;

    return *this;
}

INLINE_MEMBER float FCrd2::x0() const { return x[0]; }
INLINE_MEMBER float FCrd2::x1() const { return x[1]; }

INLINE_MEMBER void FCrd2::x0(float x0new) { x[0] = x0new; }
INLINE_MEMBER void FCrd2::x1(float x1new) { x[1] = x1new; }

INLINE_GLOBAL float squaredLength(const FCrd2 & crd)
{
    return (square(crd.x[0]) + square(crd.x[1]));
}

INLINE_GLOBAL float length(const FCrd2 & crd)
{
    return sqrt(squaredLength(crd));
}

INLINE_MEMBER void FCrd2::normalize()
{
    float crdLength = length(*this);

    *this /= crdLength;
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const FCrd2 & crd)
{
    s << "FCrd2(" << crd.x[0] << "," << crd.x[1] << ")";
    return s;
}

INLINE_GLOBAL FCrd2 operator-(const FCrd2 & crd)
{
    return FCrd2(-crd.x[0], -crd.x[1]);
}

INLINE_GLOBAL FCrd2 operator+(const FCrd2 & crd1, const FCrd2 & crd2)
{
    return FCrd2(crd1.x[0] + crd2.x[0],
                crd1.x[1] + crd2.x[1]);
}

INLINE_GLOBAL FCrd2 operator-(const FCrd2 & crd1, const FCrd2 & crd2)
{
    return FCrd2(crd1.x[0] - crd2.x[0],
                crd1.x[1] - crd2.x[1]);
}

INLINE_GLOBAL FCrd2 operator*(const FCrd2 & crd, float scalar)
{
    return FCrd2(crd.x[0] * scalar,
                crd.x[1] * scalar);
}

INLINE_GLOBAL FCrd2 operator*(float scalar, const FCrd2 & crd)
{
    return FCrd2(crd.x[0] * scalar,
                crd.x[1] * scalar);
}

INLINE_GLOBAL FCrd2 operator/(const FCrd2 & crd, float scalar)
{
    CH_FCRD2_CODE(EH_FATAL_CONDITION(FCrd2_errors,scalar == 0.0));

    return FCrd2(crd.x[0] / scalar,
                crd.x[1] / scalar);
}

INLINE_GLOBAL float dotProduct(const FCrd2 & crd1, const FCrd2 & crd2)
{
    return (crd1.x[0] * crd2.x[0] +
            crd1.x[1] * crd2.x[1]);
}

INLINE_GLOBAL FCrd2 normalizedCopy(const FCrd2 & crd)
{
    float crdLength = length(crd);

    return (crd / crdLength);
}

INLINE_MEMBER void FCrd2::cwMultiplyBy(const FCrd2 & crd)
{
    x[0] *= crd.x[0];
    x[1] *= crd.x[1];
}

INLINE_MEMBER void FCrd2::cwDevideBy(const FCrd2 & crd)
{
    CH_FCRD2_CODE(EH_FATAL_CONDITION(FCrd2_errors,crd[0] == 0.0));
    CH_FCRD2_CODE(EH_FATAL_CONDITION(FCrd2_errors,crd[1] == 0.0));

    x[0] /= crd.x[0];
    x[1] /= crd.x[1];
}

INLINE_GLOBAL FCrd2 cwReciprocalCopy(const FCrd2 & crd)
{
    FCrd2 newVec;
    
    if (crd.x[0] != 0.0)
        newVec.x[0] = 1.0 / crd.x[0];
    else
        newVec.x[0] = 0.0;

    if (crd.x[1] != 0.0)
        newVec.x[1] = 1.0 / crd.x[1];
    else
        newVec.x[1] = 0.0;

    return (newVec);
}

INLINE_GLOBAL FCrd2 cwProduct(const FCrd2 & crd1, const FCrd2 & crd2)
{
    return (FCrd2(crd1.x[0] * crd2.x[0],
                 crd1.x[1] * crd2.x[1]));
}

INLINE_GLOBAL FCrd2 cwQuotient(const FCrd2 & crd1, const FCrd2 & crd2)
{
    CH_FCRD2_CODE(EH_FATAL_CONDITION(FCrd2_errors,crd2[0] == 0.0));
    CH_FCRD2_CODE(EH_FATAL_CONDITION(FCrd2_errors,crd2[1] == 0.0));

    return FCrd2(crd1.x[0] / crd2.x[0],
                crd1.x[1] / crd2.x[1]);
}

INLINE_GLOBAL FCrd2 cwMin(const FCrd2 & crd1, const FCrd2 & crd2)
{
    return FCrd2(min(crd1.x[0],crd2.x[0]),
                min(crd1.x[1],crd2.x[1]));
}

INLINE_GLOBAL FCrd2 cwMax(const FCrd2 & crd1, const FCrd2 & crd2)
{
    return FCrd2(max(crd1.x[0],crd2.x[0]),
                max(crd1.x[1],crd2.x[1]));
}

INLINE_MEMBER FCrd2 & FCrd2::scale(const FCrd2 & s)
{
    x[0] *= s[0];
    x[1] *= s[1];

    return *this;
}

INLINE_MEMBER FCrd2 & FCrd2::scale(float s0, float s1)
{
    x[0] *= s0;
    x[1] *= s1;

    return *this;
}

INLINE_MEMBER FCrd2 & FCrd2::rot(float cosAlpha, float sinAlpha)
{
    float t = x[0];
    
    x[0] = x[0] * cosAlpha - x[1] * sinAlpha;
    x[1] = x[1] * cosAlpha +    t * sinAlpha;

    return *this;
}

INLINE_MEMBER FCrd2 & FCrd2::shear0(float x0s)
{
    x[0] += x[1] * x0s;

    return *this;
}

INLINE_MEMBER FCrd2 & FCrd2::shear1(float x1s)
{
    x[1] += x[0] * x1s;

    return *this;
}

INLINE_MEMBER FCrd2 & FCrd2::operator*=(const FMat2 & mat)
{
    float t = x[0];

    x[0] = x[0] * mat.x[0][0] + x[1] * mat.x[1][0];
    x[1] =    t * mat.x[0][1] + x[1] * mat.x[1][1];
    
    return *this;
}

INLINE_GLOBAL FCrd2 prCrdScale(const FCrd2 & crd, const FCrd2 & s)
{
    return FCrd2(crd.x[0] * s[0], crd.x[1] * s[1]);
}

INLINE_GLOBAL FCrd2 prCrdScale(const FCrd2 & crd, float s0, float s1)
{
    return FCrd2(crd.x[0] * s0, crd.x[1] * s1);
}

INLINE_GLOBAL FCrd2 prCrdRot(const FCrd2 & crd, float cosAlpha, float sinAlpha)
{
    return FCrd2(crd.x[0] * cosAlpha - crd.x[1] * sinAlpha,
                crd.x[1] * cosAlpha + crd.x[0] * sinAlpha);
}

INLINE_GLOBAL FCrd2 prCrdShear0(const FCrd2 & crd, float x0s)
{
    return FCrd2(crd.x[0] + crd.x[1] * x0s, crd.x[1]);
}

INLINE_GLOBAL FCrd2 prCrdShear1(const FCrd2 & crd, float x1s)
{
    return FCrd2(crd.x[0], crd.x[1] + crd.x[0] * x1s);
}

INLINE_GLOBAL FCrd2 operator*(const FCrd2 & crd, const FMat2 & mat)
{
    return
        FCrd2
        (
            crd.x[0] * mat.x[0][0] + crd.x[1] * mat.x[1][0],
	    crd.x[0] * mat.x[0][1] + crd.x[1] * mat.x[1][1]
        );
}

// ================================================================ class FMat2

INLINE_MEMBER FMat2::FMat2()
{
    TR_FUN0(FCrd2_constructors,FMat2);
}

INLINE_MEMBER FMat2::FMat2(float x00, float x01,
                  float x10, float x11)
{
    x[0][0] = x00; x[0][1] = x01;
    x[1][0] = x10; x[1][1] = x11;

    TR_FUN4(FCrd2_constructors,FMat2,x00,x01,x10,x11);
}

INLINE_MEMBER FMat2::FMat2(const FMat2 & mat)
{
    x[0][0] = mat.x[0][0];
    x[0][1] = mat.x[0][1];
    x[1][0] = mat.x[1][0];
    x[1][1] = mat.x[1][1];
    TR_FUN1(FCrd2_constructors,FMat2,mat);
}


INLINE_MEMBER FMat2::~FMat2()
{
    TR_MESSAGE(FCrd2_constructors,"~" << *this);
}

INLINE_MEMBER float FMat2::operator()(int i, int j) const
{
    CH_FCRD2_CODE(
        if (i < 0 || i > 1)
            EH_FATAL(FCrd2_errors,"i == " << i);
        if (j < 0 || j > 1)
            EH_FATAL(FCrd2_errors,"j == " << j);
    )
    return x[i][j];
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const FMat2 & mat)
{
    s << "FMat2(" << mat.x[0][0] << "," << mat.x[0][1] << ","
        << mat.x[1][0] << "," << mat.x[1][1] << ")";
    return s;
}


INLINE_GLOBAL float determinant(const FMat2 & mat)
{
    return
        det2x2(
            mat.x[0][0], mat.x[0][1],
            mat.x[1][0], mat.x[1][1]
        );
}

INLINE_GLOBAL FMat2 inverseCopy(const FMat2 & mat, float det)
{
    det = 1.0 / det;

    return
	FMat2
	(
	     det * mat.x[1][1], -det * mat.x[0][1],
	    -det * mat.x[1][0],  det * mat.x[0][0]
	);
}

INLINE_GLOBAL FMat2 inverseCopy(const FMat2 & mat)
{
    float det = determinant(mat);

    if (det == 0.0)
    {
        EH_ERROR(FCrd2_errors, ERROR_FCRD2_SINGULAR_MATRIX, "singular matrix");
        return mat;
    }

    return inverseCopy(mat,det);
}

INLINE_MEMBER void FMat2::invert()
{
    *this = inverseCopy(*this);
}

INLINE_GLOBAL FMat2 transposedCopy(const FMat2 & mat)
{
    return
        FMat2(
            mat.x[0][0], mat.x[1][0],
            mat.x[0][1], mat.x[1][1]
        );
}

INLINE_MEMBER void FMat2::transpose()
{
    *this = transposedCopy(*this);
}

INLINE_GLOBAL FMat2 operator-(const FMat2 & mat)
{
    return
        FMat2
        (
            - mat.x[0][0], - mat.x[0][1],
            - mat.x[1][0], - mat.x[1][1]
        );
}

INLINE_GLOBAL FMat2 operator+(const FMat2 & mat1, const FMat2 & mat2)
{
    return
        FMat2
        (
            mat1.x[0][0] + mat2.x[0][0], mat1.x[0][1] + mat2.x[0][1],
            mat1.x[1][0] + mat2.x[1][0], mat1.x[1][1] + mat2.x[1][1]
        );
}

INLINE_GLOBAL FMat2 operator-(const FMat2 & mat1, const FMat2 & mat2)
{
    return
        FMat2
        (
            mat1.x[0][0] - mat2.x[0][0], mat1.x[0][1] - mat2.x[0][1],
            mat1.x[1][0] - mat2.x[1][0], mat1.x[1][1] - mat2.x[1][1]
        );
}

INLINE_GLOBAL FMat2 operator*(const FMat2 & mat, float s)
{
    return
        FMat2
        (
            mat.x[0][0] * s, mat.x[0][1] * s,
            mat.x[1][0] * s, mat.x[1][1] * s
        );
}

INLINE_GLOBAL FMat2 operator*(float s, const FMat2 & mat)
{
    return
        FMat2
        (
            s * mat.x[0][0], s * mat.x[0][1],
            s * mat.x[1][0], s * mat.x[1][1]
        );
}

INLINE_MEMBER FMat2 & FMat2::operator*=(float s)
{
    x[0][0] *= s; x[0][1] *= s;
    x[1][0] *= s; x[1][1] *= s;
    return *this;
}

INLINE_MEMBER FMat2 & FMat2::scale(const FCrd2 & s)
{
    x[0][0] *= s[0]; x[0][1] *= s[1];
    x[1][0] *= s[0]; x[1][1] *= s[1];
    return *this;
}

INLINE_MEMBER FMat2 & FMat2::scale(float s0, float s1)
{
    x[0][0] *= s0; x[0][1] *= s1;
    x[1][0] *= s0; x[1][1] *= s1;
    return *this;
}

INLINE_MEMBER FMat2 & FMat2::rot(float cosAlpha, float sinAlpha)
{
    float t;

    t = x[0][0];
    x[0][0] = t * cosAlpha - x[0][1] * sinAlpha;
    x[0][1] = t * sinAlpha + x[0][1] * cosAlpha;
    t = x[1][0];
    x[1][0] = t * cosAlpha - x[1][1] * sinAlpha;
    x[1][1] = t * sinAlpha + x[1][1] * cosAlpha;
    return *this;
}

INLINE_MEMBER FMat2 & FMat2::shear0(float x0s)
{
    x[0][0] += x[0][1] * x0s;
    x[1][0] += x[1][1] * x0s;

    return *this;
}

INLINE_MEMBER FMat2 & FMat2::shear1(float x1s)
{
    x[0][1] += x[0][0] * x1s;
    x[1][1] += x[1][0] * x1s;

    return *this;
}

INLINE_MEMBER FMat2 & FMat2::operator+=(const FMat2 & mat)
{
    x[0][0] += mat.x[0][0];
    x[0][1] += mat.x[0][1];
    x[1][0] += mat.x[1][0];
    x[1][1] += mat.x[1][1];

    return *this;
}

INLINE_GLOBAL FMat2 operator*(const FMat2 & mat1, const FMat2 & mat2)
{
    return
        FMat2
        (
            mat1.x[0][0] * mat2.x[0][0] + mat1.x[0][1] * mat2.x[1][0],
            mat1.x[0][0] * mat2.x[0][1] + mat1.x[0][1] * mat2.x[1][1],
            mat1.x[1][0] * mat2.x[0][0] + mat1.x[1][1] * mat2.x[1][0],
            mat1.x[1][0] * mat2.x[0][1] + mat1.x[1][1] * mat2.x[1][1]
        );
}

INLINE_MEMBER FMat2 & FMat2::operator*=(const FMat2 & mat)
{
    *this = *this * mat;

    return *this;
}

INLINE_GLOBAL FMat2 prMatScale(const FMat2 & mat, const FCrd2 & s)
{
    return
        FMat2
        (
            mat.x[0][0] * s[0], mat.x[0][1] * s[1],
            mat.x[1][0] * s[0], mat.x[1][1] * s[1]
        );
}

INLINE_GLOBAL FMat2 prMatScale(const FMat2 & mat, float s0, float s1)
{
    return
        FMat2
        (
            mat.x[0][0] * s0, mat.x[0][1] * s1,
            mat.x[1][0] * s0, mat.x[1][1] * s1
        );
}

INLINE_GLOBAL FMat2 prMatRot(const FMat2 & mat, float cosAlpha, float sinAlpha)
{
    return
        FMat2
        (
            mat.x[0][0] * cosAlpha - mat.x[0][1] * sinAlpha,
            mat.x[0][1] * cosAlpha + mat.x[0][0] * sinAlpha,

            mat.x[1][0] * cosAlpha - mat.x[1][1] * sinAlpha,
            mat.x[1][1] * cosAlpha + mat.x[1][0] * sinAlpha
        );
}

INLINE_GLOBAL FMat2 prMatShear0(const FMat2 & mat, float x0s)
{
    return
	FMat2
	(
	    mat.x[0][0] + mat.x[0][1] * x0s, mat.x[0][1],
	    mat.x[1][0] + mat.x[1][1] * x0s, mat.x[1][1]
	);
}

INLINE_GLOBAL FMat2 prMatShear1(const FMat2 & mat, float x1s)
{
    return
	FMat2
	(
	    mat.x[0][0], mat.x[0][1] + mat.x[0][0] * x1s,
	    mat.x[1][0], mat.x[1][1] + mat.x[1][0] * x1s
	);
}

INLINE_GLOBAL FMat2 prScaleMat(const FCrd2 & s, const FMat2 & mat)
{
    return
        FMat2
        (
            s[0] * mat.x[0][0], s[0] * mat.x[0][1],
            s[1] * mat.x[1][0], s[1] * mat.x[1][1]
        );
}

INLINE_GLOBAL FMat2 prScaleMat(float s0, float s1, const FMat2 & mat)
{
    return
        FMat2
        (
            s0 * mat.x[0][0], s0 * mat.x[0][1],
            s1 * mat.x[1][0], s1 * mat.x[1][1]
        );
}

INLINE_GLOBAL FMat2 prRotMat(float cosAlpha, float sinAlpha, const FMat2 & mat)
{
    return
        FMat2
        (
            cosAlpha * mat.x[0][0] + sinAlpha * mat.x[1][0],
            cosAlpha * mat.x[0][1] + sinAlpha * mat.x[1][1],
            cosAlpha * mat.x[1][0] - sinAlpha * mat.x[0][0],
            cosAlpha * mat.x[1][1] - sinAlpha * mat.x[0][1]
        );
}

INLINE_GLOBAL FMat2 prShear0Mat(float x0s, const FMat2 & mat)
{
    return
	FMat2
	(
	    mat.x[0][0],
	    mat.x[0][1],
	    x0s * mat.x[0][0] + mat.x[1][0],
	    x0s * mat.x[0][1] + mat.x[1][1]
	);
}

INLINE_GLOBAL FMat2 prShear1Mat(float x1s, const FMat2 & mat)
{
    return
	FMat2
	(
	    mat.x[0][0] + x1s * mat.x[1][0],
	    mat.x[0][1] + x1s * mat.x[1][1],
	    mat.x[1][0],
	    mat.x[1][1]
	);
}

} // extern "C++"

// ===========================================================================
