// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       FCrd3
//  TYPE:       c++ header
//  PROJECT:    VEGA Math Library
//  CONTENT:    base types for 3D coordinate manipulation using floats
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rft     Robert F. Tobler
//              hl      Helwig L"offelman
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 14:06:28  rft     last modification
//  08-Aug-93           rft     created
// ===========================================================================
//  MANUAL:
//
//  DESCRIPTION
//      FCrd3 is the base type for three dimensional coordinate manipulation.
//      It is a set of three coordinates that can be transformed with the
//      associated matrix type called FMat3.
//
//      The FCrd3 and FMat3 classes are based on floats internally.
//
//      In terms of matrix theory a FCrd3 is a 1 by 3 line vector and a FMat3
//      is a 3 by 3 Matrix. Thus the transformation of a FCrd3 by a FMat3 is
//      defined as follows.
//
//              FCrd3    c;
//              FMat3    mat;
//
//              FCrd3    crd = c * mat;
//
//                                  (mat(0,0), mat(0,1), mat(0,2))
//                                  (mat(1,0), mat(1,1), mat(1,2))
//                                  (mat(2,0), mat(2,1), mat(2,2))
//
//              (c[0],c[1],c[2])    (  crd[0],   crd[1],   crd[2])
//
//  CHECKS
//      Range checking and attempts to normalize null-vectors will only be
//      caught if WITH_FCRD3_CHECKS is defined.
//
//  ERRORHANDLERS
//      FCrd3_errors
//
//  TRACERS
//      FCrd3_constructors
//
//  SEE ALSO
//      VEGA, MathLibrary
// ===========================================================================
#ifndef _MATH_FCRD3_H_
#define _MATH_FCRD3_H_

#include <iostream.h>

#include "VEGA/Math/Globals.h"
#include "VEGA/Math/Math.h"

#ifdef __GNUC__
#pragma interface
#endif

extern "C++"
{

// ============================================ classes defined in this module

class FCrd3;
class FMat3;

// =========================================== errors generated by this module

#define ERROR_FCRD3_SINGULAR_MATRIX      1
        //
        //  An attempt was made to invert a singular matrix.
        //

// ================================================================ class FCrd3

class FCrd3
{
    friend class FMat3;

    friend class Viewer3D;		//  needs FAST access to members
public:

    // ---------------------------------------------------------- constructors

    FCrd3(
        );
    FCrd3(
        float x0new,
        float x1new,
        float x2new
        );
    FCrd3(
        const FCrd3 & crd
        );
    ~FCrd3(
        );

    // ------------------------------------------------------ member operators

    float operator[](
	int index
	) const;
    float & operator[](
	int index
	);
    FCrd3 & operator+=(
	const FCrd3 & crd
	);
    FCrd3 & operator-=(
	const FCrd3 & crd
	);
    FCrd3 & operator*=(
	float scalar
	);
    FCrd3 & operator/=(
	float scalar
	);
    FCrd3 & operator*=(
	const FMat3 & mat
	);

    // ------------------------------------------------------ member functions

    float x0(
	) const;
    float x1(
	) const;
    float x2(
	) const;
    void x0(
	float x0new
	);
    void x1(
	float x1new
	);
    void x2(
	float x2new
	);

    void normalize(
	);

    // ------------------------------------------------------ friend operators

    friend ostream & operator<<(
	ostream & s,
	const FCrd3 & crd
	);
    friend FCrd3 operator-(
	const FCrd3 & crd
	);
    friend FCrd3 operator+(
	const FCrd3 & crd1,
	const FCrd3 & crd2
	);
    friend FCrd3 operator-(
	const FCrd3 & crd1,
	const FCrd3 & crd2
	);
    friend FCrd3 operator*(
	const FCrd3 & crd,
	float scalar
	);
    friend FCrd3 operator*(
	float scalar,
	const FCrd3 & crd
	);
    friend FCrd3 operator/(
	const FCrd3 & crd,
	float scalar
	);
    friend FCrd3 operator*(
	const FCrd3 & crd,
	const FMat3 & mat
	);

    // ------------------------------------------------------ friend functions

    friend float squaredLength(
	const FCrd3 & crd
	);
    friend float length(
	const FCrd3 & crd
	);
    friend float dotProduct(
	const FCrd3 & crd1,
	const FCrd3 & crd2
	);
    friend FCrd3 crossProduct(
	const FCrd3 & crd1,		// (crd1, crd2, result) defines the
	const FCrd3 & crd2		//  base of a right-handed coordinate
	);				//  system.

    friend FCrd3 normalizedCopy(
	const FCrd3 & crd
	);

    // ----------------------------------------------- componentwise functions
    //
    //  The following functions apply each operation componentwise.
    //

    void cwMultiplyBy(
	const FCrd3 & crd
	);
    void cwDevideBy(
	const FCrd3 & crd
	);

    friend FCrd3 cwProduct(
	const FCrd3 & crd1,
	const FCrd3 & crd2
	);
    friend FCrd3 cwQuotient(
	const FCrd3 & crd1,
	const FCrd3 & crd2
	);
    friend FCrd3 cwMin(
	const FCrd3 & crd1,
	const FCrd3 & crd2
	);
    friend FCrd3 cwMax(
	const FCrd3 & crd1,
	const FCrd3 & crd2
	);

    friend FCrd3 cwReciprocalCopy(       // Componentwise multiplicative
        const FCrd3 & crd		//  inverse. Zero components are not
        );				//  inverted.

    // ---------------------------------------------- transformation functions
    //
    //  The following transformation functions and methods are provided to
    //  facilitate the easy construction of complex matrices. They are mainly
    //  used in the Vec3D module to implement the special purpose
    //  transformation classes.
    //

    FCrd3 & scale(
	const FCrd3 & s
	);
    FCrd3 & scale(
	float s0,
	float s1,
	float s2
	);
    FCrd3 & rot0(
	float cosAlpha,
	float sinAlpha
	);
    FCrd3 & rot1(
	float cosAlpha,
	float sinAlpha
	);
    FCrd3 & rot2(
	float cosAlpha,
	float sinAlpha
	);
    FCrd3 & shear01(
	float x0s,
	float x1s
	);
    FCrd3 & shear20(
	float x2s,
	float x0s
	);
    FCrd3 & shear12(
	float x1s,
	float x2s
	);
        
    friend FCrd3 prCrdScale(
	const FCrd3 & crd,
	const FCrd3 & s
	);
    friend FCrd3 prCrdScale(
	const FCrd3 & crd,
	float s0, float s1, float s2
	);
    friend FCrd3 prCrdRot0(
	const FCrd3 & crd,
	float cosAlpha, float sinAlpha
	);
    friend FCrd3 prCrdRot1(
	const FCrd3 & crd,
	float cosAlpha, float sinAlpha
	);
    friend FCrd3 prCrdRot2(
	const FCrd3 & crd,
	float cosAlpha,
	float sinAlpha
	);
    friend FCrd3 prCrdShear01(
	const FCrd3 & crd,
	float x0s, float x1s
	);
    friend FCrd3 prCrdShear20(
	const FCrd3 & crd,
	float x0s, float x1s
	);
    friend FCrd3 prCrdShear12(
	const FCrd3 & crd,
	float x0s, float x1s
	);

    // ----------------------------------------------- static member variables

    static const FCrd3 base0;
    static const FCrd3 base1;
    static const FCrd3 base2;

protected:

    float x[3];
};

// =============================================================== class FMat3

class FMat3
{
    friend class FCrd3;

    friend class Viewer3D;		//  needs FAST access to members
public:

    // --------------------------------------------- constructors / destructor

    FMat3(
        );
    FMat3(
        float x00, float x01, float x02,
        float x10, float x11, float x12,
        float x20, float x21, float x22
        );
    FMat3(
        const FMat3 & mat
        );
    ~FMat3(
        );

    // ------------------------------------------------------ member operators

    FMat3 & operator+=(
	const FMat3 & mat
	);
    FMat3 & operator*=(
	float scalar
	);
    FMat3 & operator*=(
	const FMat3 & mat
	);

    // ------------------------------------------------------ member functions

    float operator()(			// Access element.
	int i, int j
	) const;

    void invert(
	);
        //
        //  ERROR_FCRD3_SINGULAR_MATRIX
        //

    void transpose(
	);

    // ------------------------------------------------------ friend operators

    friend FMat3 operator-(
	const FMat3 & mat
	);
    friend FMat3 operator+(
	const FMat3 & mat1,
	const FMat3 & mat2
	);
    friend FMat3 operator-(
	const FMat3 & mat1,
	const FMat3 & mat2
	);
    friend FMat3 operator*(
	const FMat3 & mat,
	float scalar
	);
    friend FMat3 operator*(
	float scalar,
	const FMat3 & mat
	);
    friend FCrd3 operator*(
	const FCrd3 & crd,
	const FMat3 & mat
	);
    friend FMat3 operator*(
	const FMat3 & mat1,
	const FMat3 & mat2
	);
    friend ostream & operator<<(
	ostream & s,
	const FMat3 & mat
	);

    // ------------------------------------------------------ friend functions

    friend float determinant(
	const FMat3 & mat
	);
    friend FMat3 inverseCopy(
	const FMat3 & mat,
	float det			//  det is NOT checked against zero!
	);

    friend FMat3 inverseCopy(
	const FMat3 & mat
	);
        //
        //  ERROR_FCRD3_SINGULAR_MATRIX
        //

    friend FMat3 transposedCopy(
	const FMat3 & mat
	);

    // ---------------------------------------------- transformation functions
    //
    //  The following transformation functions and methods are provided to
    //  facilitate the easy construction of complex matrices. They are mainly
    //  used in the Vec3D module to implement the special purpose
    //  transformation classes.
    //

    FMat3 & scale(
	const FCrd3 & s
	);
    FMat3 & scale(
	float s0,
	float s1,
	float s2
	);
    FMat3 & rot0(
	float cosAlpha,
	float sinAlpha
	);
    FMat3 & rot1(
	float cosAlpha,
	float sinAlpha
	);
    FMat3 & rot2(
	float cosAlpha,
	float sinAlpha
	);
    FMat3 & shear01(
	float x0s,
	float x1s
	);
    FMat3 & shear20(
	float x2s,
	float x0s
	);
    FMat3 & shear12(
	float x1s,
	float x2s
	);
    
    friend FMat3 prMatScale(
	const FMat3 & mat,
	const FCrd3 & s
	);
    friend FMat3 prMatScale(
	const FMat3 & mat,
	float s0, float s1, float s2
	);
    friend FMat3 prMatRot0(
	const FMat3 & mat,
	float cosAlpha, float sinAlpha
	);
    friend FMat3 prMatRot1(
	const FMat3 & mat,
	float cosAlpha, float sinAlpha
	);
    friend FMat3 prMatRot2(
	const FMat3 & mat,
	float cosAlpha, float sinAlpha
	);
    friend FMat3 prMatShear01(
	const FMat3 & mat,
	float x0s, float x1s
	);
    friend FMat3 prMatShear20(
	const FMat3 & mat,
	float x2s, float x0s
	);
    friend FMat3 prMatShear12(
	const FMat3 & mat,
	float x1s, float x2s
	);

    friend FMat3 prScaleMat(
	const FCrd3 & s,
	const FMat3 & mat
	);
    friend FMat3 prScaleMat(
	float s0, float s1, float s2,
	const FMat3 & mat
	);
    friend FMat3 prRot0Mat(
	float cosAlpha, float sinAlpha,
	const FMat3 & mat
	);
    friend FMat3 prRot1Mat(
	float cosAlpha, float sinAlpha,
	const FMat3 & mat
	);
    friend FMat3 prRot2Mat(
	float cosAlpha, float sinAlpha,
	const FMat3 & mat
	);
    friend FMat3 prShear01Mat(
	float x0s, float x1s,
	const FMat3 & mat
	);
    friend FMat3 prShear20Mat(
	float x2s, float x0s,
	const FMat3 & mat
	);
    friend FMat3 prShear12Mat(
	float x1s, float x2s,
	const FMat3 & mat
	);

    // ----------------------------------------------- static member variables

    static const FMat3 identity;

protected:

    float x[3][3];
};

} // extern "C++"

#ifndef CRD3_NO_INLINE

#define INLINE_GLOBAL COMPILER_INLINE_GLOBAL
#define INLINE_MEMBER COMPILER_INLINE_MEMBER

#include "VEGA/Math/FCrd3.I"

#undef INLINE_GLOBAL
#undef INLINE_MEMBER

#endif

#endif // _MATH_FCRD3_H_
// ===========================================================================

