// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       FVectorN
//  TYPE:       c++ header
//  PROJECT:    VEGA Math Library
//  CONTENT:    vector and matrix of floats
//  VERSION:    0.6.0
// ===========================================================================
//  AUTHORS:    rft     Robert F. Tobler
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 14:06:28  rft     last modification
//  28-jul-94 15:24:07  rft     created
// ===========================================================================
//  MANUAL:
//
//  DESCRIPTION
//      This module implements float Vector and Matrix classes. Although it
//      is recommended to work with vectors (matrices) with first index 0,
//      both classes are implemented in such a way that vectors and matrices
//      with different first indices can be created.
//
//	Both FVectorN and FMatrixNM implement subclasses calles SubFVectorN and
//	SubFMatrixNM, that do NOT own the actual memory they are using. Thus
//	a change to a submatrix will be reflected in the original matrix with
//	which it was initialized.
//
//	WARNING: If a FVectorN (FMatrixNM) is deleted before a SubFVectorN
//	(SubFMatrixNM) that was initialized with it, the SubFVectorN
//	(SubFMatrixNM) will contain an illegal pointer.
//
//  SEE ALSO
//      VEGA, MathLibrary
// ===========================================================================
#ifndef _MATH_FVECTORN_H_
#define _MATH_FVECTORN_H_

#include <VEGA/Misc/TTools.h>

#include "VEGA/Math/Globals.h"

#ifdef __GNUC__
#pragma interface
#endif

extern "C++"
{

#ifdef SGI
#define VECTOR2_NO_INLINE
#endif

#define ERROR_VECTOR_MISMATCH		1
	//
	//  A vector operation is attempted with two vectors that do not match
	//  in length (or base offset).
	//

#define ERROR_VECTOR_NOT_POWER_OF_2	2
	//
	//  Some vector operation require the vector length to be a power of
	//  two. If any other vector is supplied, this error is raised.
	//

// ============================================================= class FVectorN

class FVectorN
{
    friend class SubFVectorN;

public:

    FVectorN(
	);
    FVectorN(
	int newLength			// implicit: newLower = 0
	);
    FVectorN(
	int newLower,			// lowest valid index
	int newLength
	);
    FVectorN(
	const float * values,		// implicit: newLower = 0
	int newLength
	);
    FVectorN(
	const float * values,		// values[0] is placed in
	int newLower,			//  vector[newLower]
	int newLength
	);
    FVectorN(
	const FVectorN & vector
	);

    virtual ~FVectorN(
	);

    float operator[](
	int index
	) const;
    float & operator[](
	int index
	);

    int getLength(
	) const;    
    int getLower(			// Get index of first valid element.
	) const;
    int getUpper(			// 1 + index of last valid element.
	) const;

    void setLower(			// Set index of first valid element
	int newLower			//  to newLower.
	);

    float get(
	int index
	) const;
    void set(
	int index,
	float value
	);

    FVectorN & fill(			// Fill all elements with the same
	float value			//  specified value.
	);

    FVectorN & operator=(
	const FVectorN & vector
	);

    FVectorN & operator*=(
	float scalar
	);
    FVectorN & operator+=(
	const FVectorN & vector
	);
    FVectorN & operator-=(
	const FVectorN & vector
	);


    friend FVectorN operator-(
	const FVectorN & vector
	);

    //
    //  Following operators only work on vectors of same length and lower
    //  index.
    //

    friend FVectorN operator+(
	const FVectorN & vector1,
	const FVectorN & vector2
	);
    friend FVectorN operator-(
	const FVectorN & vector1,
	const FVectorN & vector2
	);
    friend FVectorN operator*(
	float scalar,
	const FVectorN & vector
	);
    friend FVectorN operator*(
	const FVectorN & vector,
	float scalar
	);

    // sfmap: protect: double

    friend double dotProduct(
	const FVectorN & vector1,
	const FVectorN & vector2
	);

    FVectorN & bitReverseIndex(
	);
    FVectorN bitReverseIndexedCopy(
	) const;


    FVectorN & hartleyTransform(
	);
    FVectorN hartleyTransformedCopy(
	) const;


protected:
    void make();
    void kill();


    FVectorN & fastHartleyTransform();


    float * getValuePtr();
    float * getLowerPtr();
    float * getUpperPtr();
    const float * getValuePtr() const;
    const float * getLowerPtr() const;
    const float * getUpperPtr() const;

    int lower;
    int length;
    float * valuePtr;
};

// ========================================================== class SubFVectorN

class SubFVectorN : public FVectorN
{
public:
    SubFVectorN(
	FVectorN & vector
	);

    SubFVectorN(				// lower is inherited
	FVectorN & vector,
	int subLower,
	int subLength
	);

    SubFVectorN(
	FVectorN & vector,
	int subLower,
	int subLength,
	int newLower
	);

    SubFVectorN(
	float * values,
	int newLength
	);

    SubFVectorN(
	float * values,
	int newLower,
	int newLength
	);

    SubFVectorN & operator=(
	const FVectorN & vector
	);

    ~SubFVectorN(
	);
};

// ============================================================ class FMatrixNM

class FMatrixNM
{
    friend class SubFMatrixNM;

public:
    FMatrixNM(
	);
    FMatrixNM(
	const FMatrixNM & matrix
	);
    FMatrixNM(
	int newLengthN,
	int newLengthM
	);
    FMatrixNM(
	int newLowerN,
	int newLowerM,
	int newLengthN,
	int newLengthM
	);
    FMatrixNM(
	const float * values,
	int newLengthN,
	int newLengthM
	);
    FMatrixNM(
	const float * values,
	int newLowerN,
	int newLowerM,
	int newLengthN,
	int newLengthM
	);
    virtual ~FMatrixNM(
	);

    FMatrixNM & operator=(
	const FMatrixNM & matrix
	);

    const float * operator[](
	int index
	) const;
    float * operator[](
	int index
	);

    FMatrixNM & fill(
	float value
	);

    FMatrixNM & operator*=(
	float scalar
	);
    FMatrixNM & operator+=(
	const FMatrixNM & matrix
	);
    FMatrixNM & operator-=(
	const FMatrixNM & matrix
	);

    int getLengthN(
	) const;
    int getLengthM(
	) const;
    int getLowerN(
	) const;
    int getLowerM(
	) const;
    int getUpperN(
	) const;
    int getUpperM(
	) const;
    int getDeltaN(
	) const;
/*
    SubFVectorN getSubRow(
	int row
	);
    SubFVectorN getSubRow(
	int row,
	int colLower,
	int colLength
	);

    FVectorN getColumn(
	int col
	);

    FVectorN getColumn(
	int col,
	int rowLower,
	int rowLength
	);
*/

protected:
    void make();
    void kill();

    float * getValuePtr();
    float * getLowerNptr();
    float * getUpperNptr();
    float * getLowerMptr(float * value1ptr);
    float * getUpperMptr(float * value1ptr);

    const float * getValuePtr() const;
    const float * getLowerNptr() const;
    const float * getUpperNptr() const;
    const float * getLowerMptr(const float * value1ptr) const;
    const float * getUpperMptr(const float * value1ptr) const;

    int lowerN;				// lower index in major axis
    int lowerM;				// lower index in minor axis
    int lengthN;			// length in major axis
    int lengthM;			// length in minor axis
    int deltaN;				// difference of two
    					// matrix[i][j] and matrix[i+1][j]

    float * valuePtr;			// pointer to matrix[0][0]
};

// ========================================================= class SubFMatrixNM

class SubFMatrixNM : public FMatrixNM
{
public:
    SubFMatrixNM(
	FMatrixNM & matrix
	);

    SubFMatrixNM(			// lowerN and lowerM are inherited
	FMatrixNM & matrix,
	int subLowerN,
	int subLowerM,
	int subLengthN,
	int subLengthM
	);

    SubFMatrixNM(
	FMatrixNM & matrix,
	int subLowerN,
	int subLowerM,
	int subLengthN,
	int subLengthM,
	int newLowerN,
	int newLowerM
	);

    SubFMatrixNM(			// implicit: lowerN = lowerM = 0
	float * values,
	int newLengthN,
	int newLengthM
	);

    SubFMatrixNM(
	float * values,
	int newLowerN,
	int newLowerM,
	int newLengthN,
	int newLengthM
	);

    SubFMatrixNM & operator=(
	const FMatrixNM & matrix
	);

    ~SubFMatrixNM();
};

} // extern "C++"

#ifndef FVECTORN1_NO_INLINE

#define INLINE_GLOBAL COMPILER_INLINE_GLOBAL
#define INLINE_MEMBER COMPILER_INLINE_MEMBER

#include "VEGA/Math/FVectorN1.I"

#undef INLINE_GLOBAL
#undef INLINE_MEMBER

#endif

#ifndef FVECTORN2_NO_INLINE

#define INLINE_GLOBAL COMPILER_INLINE_GLOBAL
#define INLINE_MEMBER COMPILER_INLINE_MEMBER

#include "VEGA/Math/FVectorN2.I"

#undef INLINE_GLOBAL
#undef INLINE_MEMBER

#endif

#endif // _MATH_FVECTORN_H_
// ===========================================================================
