// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       FVectorN
//  TYPE:       c++ inline
//  PROJECT:    VEGA Math Library
//  CONTENT:    vector and matrix of floats
//  VERSION:    0.6.0
// ===========================================================================
//  AUTHORS:    rft     Robert F. Tobler
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 14:06:28  rft     last modification
//  28-jul-94 15:24:07  rft     created
// ===========================================================================

extern "C++"
{

INLINE_MEMBER FVectorN & FVectorN::fill(float value)
{
    float * pNow = getLowerPtr();
    float * pEnd = getUpperPtr();

    while (pNow < pEnd)
	*pNow++ = value;

    return *this;
}

INLINE_MEMBER FVectorN & FVectorN::operator*=(float scalar)
{
    float * pNow = getLowerPtr();
    float * pEnd = getUpperPtr();

    while (pNow < pEnd)
	*pNow++ *= scalar;

    return *this;
}

INLINE_MEMBER FVectorN & FVectorN::operator+=(const FVectorN & vector)
{
    CH_FVECTORN_CODE(
	EH_FATAL_CONDITION(FVectorN_errors,getLength() != vector.getLength());
    )

    float * pNow = getLowerPtr();
    float * pEnd = getUpperPtr();
    const float * pOne = vector.getLowerPtr();

    while (pNow < pEnd)
	*pNow++ += *pOne++;

    return *this;
}

INLINE_MEMBER FVectorN & FVectorN::operator-=(const FVectorN & vector)
{
    CH_FVECTORN_CODE(
	EH_FATAL_CONDITION(FVectorN_errors,getLength() != vector.getLength());
    )

    float * pNow = getLowerPtr();
    float * pEnd = getUpperPtr();
    const float * pOne = vector.getLowerPtr();

    while (pNow < pEnd)
	*pNow++ -= *pOne++;

    return *this;
}

// sfmap: protect: double

INLINE_GLOBAL double dotProduct(const FVectorN & vector1, const FVectorN & vector2)
{
    CH_FVECTORN_CODE(
	EH_FATAL_CONDITION(FVectorN_errors,
			   vector1.getLength() != vector2.getLength());
    )

    // sfmap: protect: double

    double result = 0;

    const float * pOne = vector1.getLowerPtr();
    const float * pEnd = vector1.getUpperPtr();
    const float * pTwo = vector2.getLowerPtr();


    // sfmap: protect: double double

    while (pOne < pEnd)
	result += (double)(*pOne++) * (double)(*pTwo++);

    return result;
}

INLINE_GLOBAL FVectorN operator-(const FVectorN & vector)
{
    FVectorN result = FVectorN(vector.getLower(), vector.getLength());

    float * pNow = result.getLowerPtr();
    float * pEnd = result.getUpperPtr();
    const float * pOne = vector.getLowerPtr();

    while (pNow < pEnd)
	*pNow++ = - *pOne++;

    return result;
}

INLINE_GLOBAL FVectorN operator+(const FVectorN & vector1, const FVectorN & vector2)
{
    CH_FVECTORN_CODE(
	EH_FATAL_CONDITION(FVectorN_errors,
			   vector1.getLength() != vector2.getLength());
	EH_FATAL_CONDITION(FVectorN_errors,
			   vector1.getLower() != vector2.getLower());
    )

    FVectorN result = FVectorN(vector1.getLower(),vector1.getLength());

    float * pNow = result.getLowerPtr();
    float * pEnd = result.getUpperPtr();
    const float * pOne = vector1.getLowerPtr();
    const float * pTwo = vector2.getLowerPtr();

    while (pNow < pEnd)
	*pNow++ = *pOne++ + *pTwo++;

    return result;
}

INLINE_GLOBAL FVectorN operator-(const FVectorN & vector1, const FVectorN & vector2)
{
    CH_FVECTORN_CODE(
	EH_FATAL_CONDITION(FVectorN_errors,
			   vector1.getLength() != vector2.getLength());
	EH_FATAL_CONDITION(FVectorN_errors,
			   vector1.getLower() != vector2.getLower());
    )

    FVectorN result = FVectorN(vector1.getLower(),vector1.getLength());

    float * pNow = result.getLowerPtr();
    float * pEnd = result.getUpperPtr();
    const float * pOne = vector1.getLowerPtr();
    const float * pTwo = vector2.getLowerPtr();

    while (pNow < pEnd)
	*pNow++ = *pOne++ - *pTwo++;

    return result;
}

INLINE_GLOBAL FVectorN operator*(float scalar, const FVectorN & vector)
{
    FVectorN result = FVectorN(vector.getLower(), vector.getLength());

    float * pNow = result.getLowerPtr();
    float * pEnd = result.getUpperPtr();
    const float * pOne = vector.getLowerPtr();

    while (pNow < pEnd)
	*pNow++ = scalar * *pOne++;

    return result;
}

INLINE_GLOBAL FVectorN operator*(const FVectorN & vector, float scalar)
{
    FVectorN result = FVectorN(vector.getLower(), vector.getLength());

    float * pNow = result.getLowerPtr();
    float * pEnd = result.getUpperPtr();
    const float * pOne = vector.getLowerPtr();

    while (pNow < pEnd)
	*pNow++ = *pOne++ * scalar;

    return result;
}

/*

INLINE_GLOBAL static void siftDownInHeap(
	float top, float * upperPtr, int t, int end
	)
{
    int i = t;
    int j = t+t;
    while (j >= end)
    {
	if (j > end && upperPtr[j] > upperPtr[j-1]) j--;
	if (top > upperPtr[j])
	{
	    upperPtr[i] = upperPtr[j];
	    i = j;
	    j <<= 1;
	}
	else
	{
	    j = end - 1;
	}
    }
    upperPtr[i] = top;
}

INLINE_GLOBAL FVectorN sort()
{
    float * upperPtr = getUpperPtr();
    
    int end = -getLength();
    
    for (int t = end/2; t < 0; t++)
    {
	siftDownInHeap(upperPtr[t],upperPtr,t,end);
    }

    float top;
    
    while (end < -1)
    {
	top = upperPtr[end];
	upperPtr[end] = upperPtr[-1];
	end++;
	siftDownInHeap(top,upperPtr,-1,end);
    }
}

*/

INLINE_MEMBER SubFVectorN & SubFVectorN::operator=(const FVectorN & vector)
{
    CH_FVECTORN_CODE(
	EH_FATAL_CONDITION(FVectorN_errors,getLength() != vector.getLength());
    )

#ifdef __GNUC__
    copy((float*)vector.getLowerPtr(),getLength(),getLowerPtr());
    //   ^^^^^^^^^^
    //  this was put there due to a bug in gcc 2.5.8
    //
#else
    copy(vector.getLowerPtr(),getLength(),getLowerPtr());
#endif
    return *this;
}

INLINE_MEMBER FMatrixNM::FMatrixNM(const FMatrixNM & matrix)
:
    lowerN(matrix.getLowerN()),
    lowerM(matrix.getLowerM()),
    lengthN(matrix.getLengthN()),
    lengthM(matrix.getLengthM())
{
    make();

    float * pNowN = getLowerNptr();
    float * pEndN = getUpperNptr();
    const float * pOneN = matrix.getLowerNptr();

    while (pNowN < pEndN)
    {
#ifdef __GNUC__
	copy((float*)matrix.getLowerMptr(pOneN),
	//   ^^^^^^^^^^
	//  this was put there due to a bug in gcc 2.5.8
	//
	     getLengthM(),
	     getLowerMptr(pNowN));
#else
	copy(matrix.getLowerMptr(pOneN),
	     getLengthM(),
	     getLowerMptr(pNowN));
#endif
	pNowN += getDeltaN();
	pOneN += matrix.getDeltaN();
    }
}

INLINE_MEMBER FMatrixNM & FMatrixNM::fill(float value)
{
    float * pNowN = getLowerNptr();
    float * pEndN = getUpperNptr();
    
    while (pNowN < pEndN)
    {
	float * pNowM = getLowerMptr(pNowN);
	float * pEndM = getUpperMptr(pNowN);

	while (pNowM < pEndM)
	    *pNowM++ = value;

	pNowN += getDeltaN();
    }
    
    return *this;
}

INLINE_MEMBER FMatrixNM & FMatrixNM::operator=(const FMatrixNM & matrix)
{
    if (this == &matrix)
	return *this;

    kill();
    lowerN = matrix.getLowerN();
    lowerM = matrix.getLowerM();
    lengthN = matrix.getLengthN();
    lengthM = matrix.getLengthM();
    make();

    float * pNowN = getLowerNptr();
    float * pEndN = getUpperNptr();
    const float * pOneN = matrix.getLowerNptr();

    while (pNowN < pEndN)
    {
#ifdef __GNUC__
	copy((float*)matrix.getLowerMptr(pOneN),
	//   ^^^^^^^^^^
	//  this was put there due to a bug in gcc 2.5.8
	//
	     getLengthM(),
	     getLowerMptr(pNowN));
#else
	copy(matrix.getLowerMptr(pOneN),
	     getLengthM(),
	     getLowerMptr(pNowN));
#endif
	pNowN += getDeltaN();
	pOneN += matrix.getDeltaN();
    }

    return *this;
}

INLINE_MEMBER FMatrixNM & FMatrixNM::operator*=(float scalar)
{
    float * pNowN = getLowerNptr();
    float * pEndN = getUpperNptr();
    
    while (pNowN < pEndN)
    {
	float * pNowM = getLowerMptr(pNowN);
	float * pEndM = getUpperMptr(pNowN);

	while (pNowM < pEndM)
	    *pNowM++ *= scalar;

	pNowN += getDeltaN();
    }
    
    return *this;
}

INLINE_MEMBER FMatrixNM & FMatrixNM::operator+=(const FMatrixNM & matrix)
{
    CH_FVECTORN_CODE(
	EH_FATAL_CONDITION(FVectorN_errors,getLengthN() != matrix.getLengthN());
	EH_FATAL_CONDITION(FVectorN_errors,getLengthM() != matrix.getLengthM());
    )

    float * pNowN = getLowerNptr();
    float * pEndN = getUpperNptr();
    const float * pOneN = matrix.getLowerNptr();
    
    while (pNowN < pEndN)
    {
	float * pNowM = getLowerMptr(pNowN);
	float * pEndM = getUpperMptr(pNowN);
	const float * pOneM = matrix.getLowerMptr(pOneN);

	while (pNowM < pEndM)
	    *pNowM++ += *pOneM++;

	pNowN += getDeltaN();
	pOneN += matrix.getDeltaN();
    }
    
    return *this;
}

INLINE_MEMBER FMatrixNM & FMatrixNM::operator-=(const FMatrixNM & matrix)
{
    CH_FVECTORN_CODE(
	EH_FATAL_CONDITION(FVectorN_errors,getLengthN() != matrix.getLengthN());
	EH_FATAL_CONDITION(FVectorN_errors,getLengthM() != matrix.getLengthM());
    )

    float * pNowN = getLowerNptr();
    float * pEndN = getUpperNptr();
    const float * pOneN = matrix.getLowerNptr();
    
    while (pNowN < pEndN)
    {
	float * pNowM = getLowerMptr(pNowN);
	float * pEndM = getUpperMptr(pNowN);
	const float * pOneM = matrix.getLowerMptr(pOneN);

	while (pNowM < pEndM)
	    *pNowM++ -= *pOneM++;

	pNowN += getDeltaN();
	pOneN += matrix.getDeltaN();
    }
    
    return *this;
}

INLINE_MEMBER SubFMatrixNM & SubFMatrixNM::operator=(const FMatrixNM & matrix)
{
    CH_FVECTORN_CODE(
	EH_FATAL_CONDITION(FVectorN_errors,getLengthN() != matrix.getLengthN());
	EH_FATAL_CONDITION(FVectorN_errors,getLengthM() != matrix.getLengthM());
    )

    float * pNowN = getLowerNptr();
    float * pEndN = getUpperNptr();
    const float * pOneN = matrix.getLowerNptr();

    while (pNowN < pEndN)
    {
#ifdef __GNUC__
	copy((float*)matrix.getLowerMptr(pOneN),
	//   ^^^^^^^^^^
	//  this was put there due to a bug in gcc 2.5.8
	//
	     getLengthM(),
	     getLowerMptr(pNowN));
#else
	copy(matrix.getLowerMptr(pOneN),
	     getLengthM(),
	     getLowerMptr(pNowN));
#endif
	pNowN += getDeltaN();
	pOneN += matrix.getDeltaN();
    }

    return *this;
}

} // extern "C++"

// ===========================================================================
