// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       Math
//  TYPE:       c++ code
//  PROJECT:    VEGA Math Library
//  CONTENT:    mathematical functions
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    jv      J.A.M. Vermaseren
//              rft     Robert F. Tobler
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 14:06:28  rft     last modification
//  28-Aug-93           rft     ported jv math routines to C++
// ===========================================================================

extern "C"
{
#include <stdio.h>
#include <stdlib.h>
}

#include <VEGA/Base/System.h>

#include "VEGA/Math/Globals.h"

#define MATH_NO_INLINE

#include "VEGA/Math/Math.h"

#define INLINE_GLOBAL
#define INLINE_MEMBER

#include "VEGA/Math/Math.I"

float float_eps;
float float_onePlusEps;
double double_eps;
double double_onePlusEps;

class MathInit
{
public:
    MathInit();
};

static MathInit mathInit;

MathInit::MathInit()
{
    volatile float f = 1.0;
    volatile float onePlusF = 2.0;

    while (onePlusF != 1.0)
    {
        float_eps = f;
        f /= 2.0;
	onePlusF = 1.0 + f;
    }
    float_onePlusEps = 1.0 + float_eps;

    TR_MESSAGE(Math_tracer,"sizeof(float) = " << sizeof(float));
    TR_MESSAGE(Math_tracer,"float_eps = " << float_eps);

    volatile double d = 1.0;
    volatile double onePlusD = 2.0;

    while (onePlusD != 1.0)
    {
        double_eps = d;
        d /= 2.0;
	onePlusD = 1.0 + d;
    }
    double_onePlusEps = 1.0 + double_eps;

    TR_MESSAGE(Math_tracer,"sizeof(double) = " << sizeof(double));
    TR_MESSAGE(Math_tracer,"double_eps = " << double_eps);
}

// ============================================================ class SoftMath

double SoftMath::asin(double x)
{
    double y,y2,sum;
    int sflag, yflag;
    if (x < 0.0)
        { sflag = 1; x = -x; }
    else
        sflag = 0;
    if (x > 0.5)
    {
        if (x > 1.0)
        {
            if (sflag)
                x = -x;
            fprintf(stderr,"Argument out of range in asin: %f\n",x);
            exit(-1);
        }
        y = sqrt(0.5 * (1.0 - x));
        yflag = 1;
    }
    else
    {
        y = x; yflag = 0;
    }
    y2 = y*y;
    sum = y * (1.00000000000000001365+y2*(
                0.16666666666664838839+y2*(
                0.07500000000404881646+y2*(
                0.04464285679119552341+y2*(
                0.03038196027237537659+y2*(
                0.02237173658111094020+y2*(
                0.01735997031086759961+y2*(
                0.01388302431756389624+y2*(
                0.01218170270685370993+y2*(
                0.00648075111416853345+y2*(
                0.01964268701248804731+y2*(
                -0.01638489913725409907+y2*(
                0.03201166198466493988)))))))))))));
    if (yflag)
        sum = MATH_PI_DIV_2 - 2.*sum;
    if (sflag)
        return (-sum);
    else
        return (sum);
}

double SoftMath::atan2(double y, double x)
{
    // -----------------------------------------------------------------------
    //  arctangent between -MATH_PI and MATH_PI.
    // -----------------------------------------------------------------------
    int sflag, piflag, pflag, qflag;
    double z,sum,z2;
    if (y < 0.0)
        { y = -y;  sflag = 1; }
    else 
        sflag = 0;
    if (x < 0.0)
        { x = -x; piflag = 1; }
    else
        piflag = 0;
    if (y > x)
        { pflag = 1; z = y; y = x; x = z; }
    else
        pflag = 0;
    // -----------------------------------------------------------------------
    //  We are now in the range 0-1
    // -----------------------------------------------------------------------
    if (y == 0.0 && x == 0.0)
    {
        return (0.0);
    }
    if (y > 0.2679491924311227065*x)
    {
        z = (y * MATH_SQRT_3)/(y + MATH_SQRT_3 * x);
        qflag = 1;
    }
    else
    {
        z = y/x;
        qflag = 0;
    }
    // -----------------------------------------------------------------------
    //  Now we are in the range -(2-sqrt(3))<z<2-sqrt(3)
    // -----------------------------------------------------------------------
    z2 = z*z;
    sum = z*(0.99999999999999998493+z2*(
            -0.33333333333329920123+z2*(
            0.19999999998726326765+z2*(
            -0.14285714102479574031+z2*(
            0.11111097879199028445+z2*(
            -0.09090370583742286501+z2*(
            0.07679359531984028197+z2*(
            -0.06483115238224471189+z2*(
            0.04438679771995954480)))))))));
    if (qflag) sum += 0.5235987755982988731;
    if (pflag) sum = MATH_PI_DIV_2 - sum;
    if (piflag) sum = MATH_PI - sum;
    if (sflag)
        return (-sum);
    else
        return (sum);
}

double SoftMath::cos(double x)
{
    double x2,sum;
    int sflag;
    if (x < 0.0) x = -x;
    if (x >= MATH_2_MUL_PI) x = fmod(x,MATH_2_MUL_PI);
    if (x > MATH_PI) x = MATH_2_MUL_PI - x;
    if (x > MATH_PI_DIV_2) { sflag = 1; x = MATH_PI - x; }
    else sflag = 0;
    // -----------------------------------------------------------------------
    //  Now we are in the first quadrant
    // -----------------------------------------------------------------------
    if (x < MATH_PI_DIV_4)
    {
        x2 = x*x;
        sum =  0.99999999999999995287+x2*(
        -0.49999999999999251135+x2*(
        0.04166666666647237368+x2*(
        -0.00138888888699833768+x2*(
        0.00002480157854004439+x2*(
        -0.00000027555234099211+x2*(
        0.00000000206304660116))))));
    }
    else
    {
        x = MATH_PI_DIV_2 - x;
        x2 = x*x;
        sum = x*(0.99999999999999999685+x2*(
            -0.16666666666666616691+x2*(
            0.00833333333332036780+x2*(
            -0.00019841269828654287+x2*(
            0.00000275573133777316+x2*(
            -0.00000002505071717794+x2*(
            0.00000000015894748309)))))));
    }

    if (sflag)
        return (-sum);
    else
        return (sum);
}

double SoftMath::sin(double x)
{
    double x2,sum;
    int sflag;

    if (x < 0.0)
        { sflag = 1; x = -x; }
    else
        sflag = 0;

    if (x >= MATH_2_MUL_PI)
        x = fmod(x,MATH_2_MUL_PI);

    if (x >= MATH_PI)
    {
        x = MATH_2_MUL_PI - x;
        sflag = 1-sflag;
    }

    if (x > MATH_PI_DIV_2)
        x = MATH_PI - x;

    // -----------------------------------------------------------------------
    //  Now we are in the first quadrant
    // -----------------------------------------------------------------------
    if (x > MATH_PI_DIV_4)
    {
        x = MATH_PI_DIV_2 - x;
        x2 = x*x;
        sum =  0.99999999999999995287+x2*(
              -0.49999999999999251135+x2*(
               0.04166666666647237368+x2*(
              -0.00138888888699833768+x2*(
               0.00002480157854004439+x2*(
              -0.00000027555234099211+x2*(
               0.00000000206304660116))))));
    }
    else
    {
        x2 = x*x;
        sum = x*(0.99999999999999999685+x2*(
                 -0.16666666666666616691+x2*(
                  0.00833333333332036780+x2*(
                 -0.00019841269828654287+x2*(
                  0.00000275573133777316+x2*(
                 -0.00000002505071717794+x2*(
                  0.00000000015894748309)))))));
    }

    if (sflag)
        return (-sum);
    else
        return (sum);
}

double SoftMath::tan(double x)
{
    // -----------------------------------------------------------------------
    //  Routine for the tangent. It puts x in the range 0-MATH_PI/4.
    //  Then it uses a 6,6 Maehly approximation.
    //  Derived and coded by J.A.M.Vermaseren 21-apr-1992.
    // -----------------------------------------------------------------------
    double sum,x2;
    int sflag, pflag;

    if (x < 0.0)
        { sflag = 1; x = -x; }
    else
        sflag = 0;
    if (x >= MATH_PI)
        x = fmod(x,MATH_PI);
    if (x >= MATH_PI_DIV_2)
        { x = MATH_PI - x; sflag = 1 - sflag; }
    if (x > MATH_PI_DIV_4)
        { pflag = 1; x = MATH_PI_DIV_2 - x; }
    else
        pflag = 0;
    x2 = x*x;
    sum = x*(-4797.60570220792587597407+x2*(
            615.45028715821815261900+x2*(
            -13.46133040142504243340+x2*
                0.03590005535592163977)))/(
            -4797.60570220792579574817+x2*(
            2214.65218789418096945538+x2*(
            -111.99796607144895992639+x2)));
    if (pflag) sum = 1.0/sum;
    if (sflag)
        return (-sum);
    else
        return (sum);
}

double SoftMath::exp(double x)
{
    // -----------------------------------------------------------------------
    //  routine for the evaluation of the exp (IEEE)
    //  Method:
    //          1: y = x/log(2)  -> exp(x) = 2^y;
    //          2: z = n+y, -0.5 <= y < 0.5
    //          3: exp(x) = exp(z*log(2))*2^n
    //  For the last series we need only terms till 13!
    //  Routine made by J.A.M.Vermaseren 20-apr-1992
    // -----------------------------------------------------------------------
    double y,z,scr;
    Int32 n, m;
    int sflag;
    y = x * MATH_1_DIV_LN_2;
    if (y < 0.0)
        { sflag = 1; y = -y; }
    else
        sflag = 0;
    scr = y + 0.5;
    n = *((Int32 *)(&scr));
    m = (n >> 20) - 0x3FF;
    if (m > 7)
    {
        fprintf(stderr,"Overflow in exponent of %e\n",x);
        exit(-1);
    }
    n &= 0xFFFFF;
    n |= 0x100000;
    n >>= (20-m);
    z = (y - n) * MATH_LN_2;
    if (sflag)
        { n = -n; z = -z; }
    scr = 0.99999999999999999693 + z * (
          0.99999999999999999850 + z * (
          0.50000000000000183922 + z * (
          0.16666666666666701804 + z * (
          0.04166666666648804594 + z * (
          0.00833333333330984930 + z * (
          0.00138888889523274996 + z * (
          0.00019841269908532560 + z * (
          0.00002480148546912669 + z * (
          0.00000275572255877329 + z * (
          0.00000027632644397330 + z * (
          0.00000002511466084085)))))))))));
    n <<= 20;
    *((Int32 *)(&scr)) += n;
    return (scr);
}


double SoftMath::log(double x)
{
    // -----------------------------------------------------------------------
    //  Routine for log(x) in IEEE double precision.
    //  Note that we have to hack around in the exponent!
    //  Steps: 1 x = y*2^n, with y between sqrt(2) and 1/sqrt(2)
    //         2 convert to z = (y-1)/(y+1)
    //         3 ln = 2*z*sum(i=0,9, z^(2*i)/(2*i+1)) + n*ln(2)
    //  Routine made by J.A.M. Vermaseren 20-apr-1992.
    // -----------------------------------------------------------------------
    double scr, y, z, z2, sum;
    Int16 n;
    if (x < 0.0)
    {
        fprintf(stderr,"log of a negative number: %d\n",x);
        exit(-1);
    }
    scr = x;
    n = ((*((Int16 *)(&scr))) & 0x7FF0) - 0x3FF0;
    *((Int16 *)(&scr)) -= n;
    n >>= 4;
    if (scr > MATH_SQRT_2)
    {
        *((Int16 *)(&scr)) -= 0x10;
        n++;
    }
    y = scr;
    z = (y - 1.0)/(y + 1.0);
    z2 = z*z;
    sum = 2.0 * z * (0.99999999999999999886 + z2 * (
                     0.33333333333333825805 + z2 * (
                     0.19999999999649522967 + z2 * (
                     0.14285714380662481221 + z2 * (
                     0.11111098494595009814 + z2 * (
                     0.09091817553294820737 + z2 * (
                     0.07656220982777783836 + z2 * (
                     0.07405280893003482131))))))))
          + n * 0.6931471805599453094;
    return (sum);
}

double SoftMath::floor(double x)
{
    // -----------------------------------------------------------------------
    //  Routine determines the largest integer that is smaller than x.
    //  The return value is a double again.
    //  We have to use the IEEE explicitly here.
    //  Routine made by J.A.M. Vermaseren 20-apr-1992
    // -----------------------------------------------------------------------

    double scr, y;
    int n;

    y = x;
    if (y < 0.0) y = -y;
    scr = y;
    n = *((Int16 *)(&scr));
    if (n < 0) n = -n;
    n -= 0x3FF0;
    n >>= 4;

    if (n < 0)                  // exponent < 0: trivial
    {
        if (x < 0.0)
            return (-1.0);
        else
            return (0.0);
    }

    // -----------------------------------------------------------------------
    //  There are 53 bits of which 52 after the period. This means that
    //  if n >= 52 we cannot do anything.
    // -----------------------------------------------------------------------

    if (n >= 52)
        return (x);

    if (n <= 20)
    {
        ((Int32 *)(&scr))[1] = 0;
        n = 20 - n;
        ((Int32 *)(&scr))[0] &= (-1) << n;
    }
    else
    {
        n = 52 - n;
        ((Int32 *)(&scr))[1] &= (-1) << n;
    }

    if (x < 0.0)
    {
        if (scr == y)                           // y is a negative integer
            x = -scr;
        else
            x = -scr - 1;
    }
    else
    {
        x = scr;
    }

    return (x);
}

double SoftMath::fmod(double x, double y)
{
    // -----------------------------------------------------------------------
    //  Routine determines the remainder of x/y with the same sign as x
    //  Routine made by J.A.M. Vermaseren 20-apr-1992
    // -----------------------------------------------------------------------
    double z;
    if (y == 0.0)
        return (x);
    if (y > 0.0)
    {
        if (x < y && x > -y)
            return (x);
    }
    else
    {
        if (x > y && x < -y)
            return (x);
    }
    z = x/y;
    z -= floor(z);
    z *= y;
    if (x < 0.0)
    {
        if (z > 0.0)
        {
            if (y > 0.0)
                z -= y;
            else
                z += y;
        }
    }
    else
    {
        if (z < 0.0)
        {
            if (y > 0.0)
                z += y;
            else
                z -= y;
        }
    }
    return (z);
}

// ===========================================================================


