// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       VectorN
//  TYPE:       c++ header
//  PROJECT:    VEGA Math Library
//  CONTENT:    vector and matrix of doubles
//  VERSION:    0.6.0
// ===========================================================================
//  AUTHORS:    rft     Robert F. Tobler
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 14:06:28  rft     last modification
//  28-jul-94 15:24:07  rft     created
// ===========================================================================
//  MANUAL:
//
//  DESCRIPTION
//      This module implements double Vector and Matrix classes. Although it
//      is recommended to work with vectors (matrices) with first index 0,
//      both classes are implemented in such a way that vectors and matrices
//      with different first indices can be created.
//
//	Both VectorN and MatrixNM implement subclasses calles SubVectorN and
//	SubMatrixNM, that do NOT own the actual memory they are using. Thus
//	a change to a submatrix will be reflected in the original matrix with
//	which it was initialized.
//
//	WARNING: If a VectorN (MatrixNM) is deleted before a SubVectorN
//	(SubMatrixNM) that was initialized with it, the SubVectorN
//	(SubMatrixNM) will contain an illegal pointer.
//
//  SEE ALSO
//      VEGA, MathLibrary
// ===========================================================================
#ifndef _MATH_VECTORN_H_
#define _MATH_VECTORN_H_

#include <VEGA/Misc/TTools.h>

#include "VEGA/Math/Globals.h"

#ifdef __GNUC__
#pragma interface
#endif

extern "C++"
{

#ifdef SGI
#define VECTOR2_NO_INLINE
#endif

#define ERROR_VECTOR_MISMATCH		1
	//
	//  A vector operation is attempted with two vectors that do not match
	//  in length (or base offset).
	//

#define ERROR_VECTOR_NOT_POWER_OF_2	2
	//
	//  Some vector operation require the vector length to be a power of
	//  two. If any other vector is supplied, this error is raised.
	//

// ============================================================= class VectorN

class VectorN
{
    friend class SubVectorN;

public:

    VectorN(
	);
    VectorN(
	int newLength			// implicit: newLower = 0
	);
    VectorN(
	int newLower,			// lowest valid index
	int newLength
	);
    VectorN(
	const double * values,		// implicit: newLower = 0
	int newLength
	);
    VectorN(
	const double * values,		// values[0] is placed in
	int newLower,			//  vector[newLower]
	int newLength
	);
    VectorN(
	const VectorN & vector
	);

    virtual ~VectorN(
	);

    double operator[](
	int index
	) const;
    double & operator[](
	int index
	);

    int getLength(
	) const;    
    int getLower(			// Get index of first valid element.
	) const;
    int getUpper(			// 1 + index of last valid element.
	) const;

    void setLower(			// Set index of first valid element
	int newLower			//  to newLower.
	);

    double get(
	int index
	) const;
    void set(
	int index,
	double value
	);

    VectorN & fill(			// Fill all elements with the same
	double value			//  specified value.
	);

    VectorN & operator=(
	const VectorN & vector
	);

    VectorN & operator*=(
	double scalar
	);
    VectorN & operator+=(
	const VectorN & vector
	);
    VectorN & operator-=(
	const VectorN & vector
	);


    friend VectorN operator-(
	const VectorN & vector
	);

    //
    //  Following operators only work on vectors of same length and lower
    //  index.
    //

    friend VectorN operator+(
	const VectorN & vector1,
	const VectorN & vector2
	);
    friend VectorN operator-(
	const VectorN & vector1,
	const VectorN & vector2
	);
    friend VectorN operator*(
	double scalar,
	const VectorN & vector
	);
    friend VectorN operator*(
	const VectorN & vector,
	double scalar
	);

    // sfmap: protect: double

    friend double dotProduct(
	const VectorN & vector1,
	const VectorN & vector2
	);

    VectorN & bitReverseIndex(
	);
    VectorN bitReverseIndexedCopy(
	) const;

    // sfmap: beginExcludeFrom: IVectorN

    VectorN & hartleyTransform(
	);
    VectorN hartleyTransformedCopy(
	) const;

    // sfmap: endExcludeFrom: IVectorN

protected:
    void make();
    void kill();

    // sfmap: beginExcludeFrom: IVectorN

    VectorN & fastHartleyTransform();

    // sfmap: endExcludeFrom: IVectorN

    double * getValuePtr();
    double * getLowerPtr();
    double * getUpperPtr();
    const double * getValuePtr() const;
    const double * getLowerPtr() const;
    const double * getUpperPtr() const;

    int lower;
    int length;
    double * valuePtr;
};

// ========================================================== class SubVectorN

class SubVectorN : public VectorN
{
public:
    SubVectorN(
	VectorN & vector
	);

    SubVectorN(				// lower is inherited
	VectorN & vector,
	int subLower,
	int subLength
	);

    SubVectorN(
	VectorN & vector,
	int subLower,
	int subLength,
	int newLower
	);

    SubVectorN(
	double * values,
	int newLength
	);

    SubVectorN(
	double * values,
	int newLower,
	int newLength
	);

    SubVectorN & operator=(
	const VectorN & vector
	);

    ~SubVectorN(
	);
};

// ============================================================ class MatrixNM

class MatrixNM
{
    friend class SubMatrixNM;

public:
    MatrixNM(
	);
    MatrixNM(
	const MatrixNM & matrix
	);
    MatrixNM(
	int newLengthN,
	int newLengthM
	);
    MatrixNM(
	int newLowerN,
	int newLowerM,
	int newLengthN,
	int newLengthM
	);
    MatrixNM(
	const double * values,
	int newLengthN,
	int newLengthM
	);
    MatrixNM(
	const double * values,
	int newLowerN,
	int newLowerM,
	int newLengthN,
	int newLengthM
	);
    virtual ~MatrixNM(
	);

    MatrixNM & operator=(
	const MatrixNM & matrix
	);

    const double * operator[](
	int index
	) const;
    double * operator[](
	int index
	);

    MatrixNM & fill(
	double value
	);

    MatrixNM & operator*=(
	double scalar
	);
    MatrixNM & operator+=(
	const MatrixNM & matrix
	);
    MatrixNM & operator-=(
	const MatrixNM & matrix
	);

    int getLengthN(
	) const;
    int getLengthM(
	) const;
    int getLowerN(
	) const;
    int getLowerM(
	) const;
    int getUpperN(
	) const;
    int getUpperM(
	) const;
    int getDeltaN(
	) const;
/*
    SubVectorN getSubRow(
	int row
	);
    SubVectorN getSubRow(
	int row,
	int colLower,
	int colLength
	);

    VectorN getColumn(
	int col
	);

    VectorN getColumn(
	int col,
	int rowLower,
	int rowLength
	);
*/

protected:
    void make();
    void kill();

    double * getValuePtr();
    double * getLowerNptr();
    double * getUpperNptr();
    double * getLowerMptr(double * value1ptr);
    double * getUpperMptr(double * value1ptr);

    const double * getValuePtr() const;
    const double * getLowerNptr() const;
    const double * getUpperNptr() const;
    const double * getLowerMptr(const double * value1ptr) const;
    const double * getUpperMptr(const double * value1ptr) const;

    int lowerN;				// lower index in major axis
    int lowerM;				// lower index in minor axis
    int lengthN;			// length in major axis
    int lengthM;			// length in minor axis
    int deltaN;				// difference of two
    					// matrix[i][j] and matrix[i+1][j]

    double * valuePtr;			// pointer to matrix[0][0]
};

// ========================================================= class SubMatrixNM

class SubMatrixNM : public MatrixNM
{
public:
    SubMatrixNM(
	MatrixNM & matrix
	);

    SubMatrixNM(			// lowerN and lowerM are inherited
	MatrixNM & matrix,
	int subLowerN,
	int subLowerM,
	int subLengthN,
	int subLengthM
	);

    SubMatrixNM(
	MatrixNM & matrix,
	int subLowerN,
	int subLowerM,
	int subLengthN,
	int subLengthM,
	int newLowerN,
	int newLowerM
	);

    SubMatrixNM(			// implicit: lowerN = lowerM = 0
	double * values,
	int newLengthN,
	int newLengthM
	);

    SubMatrixNM(
	double * values,
	int newLowerN,
	int newLowerM,
	int newLengthN,
	int newLengthM
	);

    SubMatrixNM & operator=(
	const MatrixNM & matrix
	);

    ~SubMatrixNM();
};

} // extern "C++"

#ifndef VECTORN1_NO_INLINE

#define INLINE_GLOBAL COMPILER_INLINE_GLOBAL
#define INLINE_MEMBER COMPILER_INLINE_MEMBER

#include "VEGA/Math/VectorN1.I"

#undef INLINE_GLOBAL
#undef INLINE_MEMBER

#endif

#ifndef VECTORN2_NO_INLINE

#define INLINE_GLOBAL COMPILER_INLINE_GLOBAL
#define INLINE_MEMBER COMPILER_INLINE_MEMBER

#include "VEGA/Math/VectorN2.I"

#undef INLINE_GLOBAL
#undef INLINE_MEMBER

#endif

#endif // _MATH_VECTORN_H_
// ===========================================================================
