// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       VectorN
//  TYPE:       c++ inline
//  PROJECT:    VEGA Math Library
//  CONTENT:    vector and matrix of doubles
//  VERSION:    0.6.0
// ===========================================================================
//  AUTHORS:    rft     Robert F. Tobler
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 14:06:28  rft     last modification
//  28-jul-94 15:24:07  rft     created
// ===========================================================================

extern "C++"
{

INLINE_MEMBER VectorN::VectorN()
:
    lower(0),
    length(0),
    valuePtr(NIL(double*))
{
}

INLINE_MEMBER void VectorN::make()
{
    if (length > 0)
	valuePtr = new double[length] - lower;
    else
	valuePtr = NIL(double *);
}

INLINE_MEMBER void VectorN::kill()
{
    if (isNotNIL(valuePtr))
	delete[] (valuePtr + lower);
}

INLINE_MEMBER int VectorN::getLength() const  { return length; }
INLINE_MEMBER int VectorN::getLower() const { return lower; }
INLINE_MEMBER int VectorN::getUpper() const { return lower + length; }

INLINE_MEMBER void VectorN::setLower(int newLower)
{
    valuePtr = valuePtr + lower - newLower;
    lower = newLower;
}

INLINE_MEMBER double * VectorN::getValuePtr()
{
    return valuePtr;
}

INLINE_MEMBER double * VectorN::getLowerPtr()
{
    return valuePtr + lower;
}

INLINE_MEMBER double * VectorN::getUpperPtr()
{
    return valuePtr + lower + length;
}

INLINE_MEMBER const double * VectorN::getValuePtr() const
{
    return valuePtr;
}

INLINE_MEMBER const double * VectorN::getLowerPtr() const
{
    return valuePtr + lower;
}

INLINE_MEMBER const double * VectorN::getUpperPtr() const
{
    return valuePtr + lower + length;
}

INLINE_MEMBER VectorN::VectorN(int newLength)
:
    lower(0),
    length(newLength)
{
    make();
}

INLINE_MEMBER VectorN::VectorN(int newLower, int newLength)
:
    lower(newLower),
    length(newLength)
{
    make();
}

INLINE_MEMBER VectorN::VectorN(
	const double * values,
	int newLower,
	int newLength
	)
:
    lower(newLower),
    length(newLength)
{
    make();
#ifdef __GNUC__
    copy((double *)values,getLength(),getLowerPtr());
    //   ^^^^^^^^^^
    //  this was put there due to a bug in gcc 2.5.8
    //
#else
    copy(values,getLength(),getLowerPtr());
#endif
}

INLINE_MEMBER VectorN::VectorN(
	const double * values,
	int newLength
	)
:
    lower(0),
    length(newLength)
{
    make();
#ifdef __GNUC__
    copy((double *)values,getLength(),getLowerPtr());
    //   ^^^^^^^^^^
    //  this was put there due to a bug in gcc 2.5.8
    //
#else
    copy(values,getLength(),getLowerPtr());
#endif
}

INLINE_MEMBER VectorN::VectorN(const VectorN & vector)
:
    lower(vector.getLower()),
    length(vector.getLength())
{
    make();
#ifdef __GNUC__
    copy((double*)vector.getLowerPtr(),getLength(),getLowerPtr());
    //   ^^^^^^^^^^
    //  this was put there due to a bug in gcc 2.5.8
    //
#else
    copy(vector.getLowerPtr(),getLength(),getLowerPtr());
#endif
}


INLINE_MEMBER double VectorN::operator[](int index) const
{
    return valuePtr[index];
}

INLINE_MEMBER double & VectorN::operator[](int index)
{
    return valuePtr[index];
}

INLINE_MEMBER double VectorN::get(int index) const
{
    return valuePtr[index];
}

INLINE_MEMBER void VectorN::set(int index, double value)
{
    valuePtr[index] = value;
}

INLINE_MEMBER VectorN & VectorN::operator=(const VectorN & vector)
{
    if (length != vector.getLength())
    {
	kill();
	lower = vector.getLower();
	length = vector.getLength();
	make();
    }
    else if (lower != vector.getLower())
    {
	valuePtr += lower - vector.getLower();
	lower = vector.getLower();
    }
#ifdef __GNUC__
    copy((double*)vector.getLowerPtr(),getLength(),getLowerPtr());
    //   ^^^^^^^^^^
    //  this was put there due to a bug in gcc 2.5.8
    //
#else
    copy(vector.getLowerPtr(),getLength(),getLowerPtr());
#endif
    return *this;
}

INLINE_MEMBER SubVectorN::SubVectorN(
	VectorN & vector
	)
:
    VectorN()
{
    lower = vector.getLower();
    length = vector.getLength();
    valuePtr = vector.getValuePtr();
}

INLINE_MEMBER SubVectorN::SubVectorN(
	VectorN & vector,
	int subLower,
	int subLength
	)
:
    VectorN()
{
    lower = vector.getLower();
    length = subLength;
    valuePtr = vector.getValuePtr() + subLower - lower;
}

INLINE_MEMBER SubVectorN::SubVectorN(
	VectorN & vector,
	int subLower,
	int subLength,
	int newLower
	)
:
    VectorN()
{
    lower = newLower;
    length = subLength;
    valuePtr = vector.getValuePtr() + subLower - newLower;
}

INLINE_MEMBER SubVectorN::SubVectorN(
	double * values,
	int newLength
	)
:
    VectorN()
{
    lower = 0;
    length = newLength;
    valuePtr = values;
}

INLINE_MEMBER SubVectorN::SubVectorN(
	double * values,
	int newLower,
	int newLength
	)
:
    VectorN()
{
    lower = newLower;
    length = newLength;
    valuePtr = values - newLower;
}

INLINE_MEMBER MatrixNM::MatrixNM()
:
    lowerN(0),
    lowerM(0),
    deltaN(0),
    lengthN(0),
    lengthM(0),
    valuePtr(NIL(double*))
{
}

INLINE_MEMBER void MatrixNM::make()
{
    if (lengthN > 0 && lengthM > 0)
    {
	deltaN = lengthM;
	valuePtr = (new double[lengthN * lengthM]) - lowerN * deltaN - lowerM;
    }
    else
    {
	valuePtr = NIL(double *);
    }
}

INLINE_MEMBER void MatrixNM::kill()
{
    if (isNotNIL(valuePtr))
    {
	delete [] (valuePtr + lowerN * deltaN + lowerM);
    }
}

INLINE_MEMBER MatrixNM::MatrixNM(
	int newLengthN,
	int newLengthM
	)
:
    lowerN(0), lowerM(0),
    lengthN(newLengthN), lengthM(newLengthM)
{
    make();
}

INLINE_MEMBER MatrixNM::MatrixNM(
    int newLowerN, int newLowerM,
    int newLengthN, int newLengthM
    )
:
    lowerN(newLowerN), lowerM(newLowerM),
    lengthN(newLengthN), lengthM(newLengthM)
{
    make();
}

INLINE_MEMBER int MatrixNM::getLengthN() const { return lengthN; }
INLINE_MEMBER int MatrixNM::getLengthM() const { return lengthM; }
INLINE_MEMBER int MatrixNM::getLowerN() const { return lowerN; }
INLINE_MEMBER int MatrixNM::getLowerM() const { return lowerM; }
INLINE_MEMBER int MatrixNM::getUpperN() const { return lowerN + lengthN; }
INLINE_MEMBER int MatrixNM::getUpperM() const { return lowerM + lengthM; }
INLINE_MEMBER int MatrixNM::getDeltaN() const { return deltaN; }

INLINE_MEMBER double * MatrixNM::getValuePtr()
{
    return valuePtr;
}

INLINE_MEMBER double * MatrixNM::getLowerNptr()
{
    return valuePtr + lowerN * deltaN;
}

INLINE_MEMBER double * MatrixNM::getUpperNptr()
{
    return valuePtr + (lowerN + lengthN) * deltaN;
}

INLINE_MEMBER double * MatrixNM::getLowerMptr(double * value1ptr)
{
    return value1ptr + lowerM;
}

INLINE_MEMBER double * MatrixNM::getUpperMptr(double * value1ptr)
{
    return value1ptr + lowerM + lengthM;
}

INLINE_MEMBER const double * MatrixNM::getValuePtr() const
{
    return valuePtr;
}

INLINE_MEMBER const double * MatrixNM::getLowerNptr() const
{
    return valuePtr + lowerN * deltaN;
}

INLINE_MEMBER const double * MatrixNM::getUpperNptr() const
{
    return valuePtr + (lowerN + lengthN) * deltaN;
}

INLINE_MEMBER const double * MatrixNM::getLowerMptr(
	const double * value1ptr
	) const
{
    return value1ptr + lowerM;
}

INLINE_MEMBER const double * MatrixNM::getUpperMptr(
	const double * value1ptr
	) const
{
    return value1ptr + lowerM + lengthM;
}

INLINE_MEMBER MatrixNM::MatrixNM(
    const double * values,
    int newLowerN, int newLowerM,
    int newLengthN, int newLengthM
    )
:
    lowerN(newLowerN), lowerM(newLowerM),
    lengthN(newLengthN), lengthM(newLengthM)
{
    make();
    
    if (isNotNIL(valuePtr))
#ifdef __GNUC__
	copy((double*)values,
	//   ^^^^^^^^^^
	//  this was put there due to a bug in gcc 2.5.8
	//
		getLengthN() * getLengthM(),
		getLowerMptr(getLowerNptr()));
#else
	copy(values,
	     getLengthN() * getLengthM(),
	     getLowerMptr(getLowerNptr()));
#endif
}

INLINE_MEMBER MatrixNM::MatrixNM(
    const double * values,
    int newLengthN, int newLengthM
    )
:
    lowerN(0), lowerM(0),
    lengthN(newLengthN), lengthM(newLengthM)
{
    make();
    
    if (isNotNIL(valuePtr))
#ifdef __GNUC__
	copy((double*)values,
	//   ^^^^^^^^^^
	//  this was put there due to a bug in gcc 2.5.8
	//
		getLengthN() * getLengthM(),
		getLowerMptr(getLowerNptr()));
#else
	copy(values,
	     getLengthN() * getLengthM(),
	     getLowerMptr(getLowerNptr()));
#endif
}

INLINE_MEMBER const double * MatrixNM::operator[](int index) const
{
    return valuePtr + index * deltaN;
}

INLINE_MEMBER double * MatrixNM::operator[](int index)
{
    return valuePtr + index * deltaN;
}

/*

INLINE_MEMBER SubVectorN SubMatrixNM::getRowSubVectorN(int row)
{
    return SubVectorN(getLowerMPtr(operator[](row)),getLengthM());
}

INLINE_MEMBER SubVectorN SubMatrixNM::getRowSubVectorN(
	int row,
	int colLower,
	int colLength,
	int newLower
	)
{
    return SubVectorN(getLowerMPtr(operator[](row)) + colLower,
		     newLower,colLength);
}

*/

INLINE_MEMBER SubMatrixNM::SubMatrixNM(
	MatrixNM & matrix
	)
:
    MatrixNM()
{
    lowerN = matrix.getLowerN();
    lowerM = matrix.getLowerM();
    deltaN = matrix.getDeltaN();
    lengthN = matrix.getLengthN();
    lengthM = matrix.getLengthM();
    valuePtr = matrix.getValuePtr();
}

INLINE_MEMBER SubMatrixNM::SubMatrixNM(
	MatrixNM & matrix,
	int subLowerN,
	int subLowerM,
	int subLengthN,
	int subLengthM
	)
:
    MatrixNM()
{
    lowerN = matrix.getLowerN();
    lowerM = matrix.getLowerM();
    deltaN = matrix.getDeltaN();
    lengthN = subLengthN;
    lengthM = subLengthM;

    if (lengthN > 0 && lengthM > 0)
	valuePtr = matrix.getValuePtr()
		    + (subLowerN - lowerN) * deltaN
		    + subLowerM - lowerM;
}

INLINE_MEMBER SubMatrixNM::SubMatrixNM(
	MatrixNM & matrix,
	int subLowerN,
	int subLowerM,
	int subLengthN,
	int subLengthM,
	int newLowerN,
	int newLowerM
	)
:
    MatrixNM()
{
    lowerN = newLowerN;
    lowerM = newLowerM;
    deltaN = matrix.getDeltaN();
    lengthN = subLengthN;
    lengthM = subLengthM;

    if (lengthN > 0 && lengthM > 0)
	valuePtr = matrix.getValuePtr()
		    + (subLowerN - lowerN) * deltaN
		    + subLowerM - lowerM;
}



INLINE_MEMBER SubMatrixNM::SubMatrixNM(
	double * values,
	int newLowerN,
	int newLowerM,
	int newLengthN,
	int newLengthM
	)
:
    MatrixNM()
{
    lowerN = newLowerN;
    lowerM = newLowerM;
    deltaN = newLengthM;
    lengthN = newLengthN;
    lengthM = newLengthM;

    if (lengthN > 0 && lengthM > 0)
	valuePtr = values - lowerN * deltaN - lowerM;
}

INLINE_MEMBER SubMatrixNM::SubMatrixNM(
	double * values,
	int newLengthN,
	int newLengthM
	)
:
    MatrixNM()
{
    lowerN = 0;
    lowerM = 0;
    deltaN = newLengthM;
    lengthN = newLengthN;
    lengthM = newLengthM;


    if (lengthN > 0 && lengthM > 0)
	valuePtr = values - lowerN * deltaN - lowerM;
}

// sfmap: beginExcludeFrom: IVectorN

INLINE_MEMBER VectorN & VectorN::hartleyTransform()
{
    bitReverseIndex();
    fastHartleyTransform();
    return *this;
}

INLINE_MEMBER VectorN VectorN::hartleyTransformedCopy() const
{
    VectorN result = bitReverseIndexedCopy();
    result.fastHartleyTransform();
    return result;
}

// sfmap: endExcludeFrom: IVectorN

} // extern "C++"

// ===========================================================================
