// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       VectorN
//  TYPE:       c++ inline
//  PROJECT:    VEGA Math Library
//  CONTENT:    vector and matrix of doubles
//  VERSION:    0.6.0
// ===========================================================================
//  AUTHORS:    rft     Robert F. Tobler
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 14:06:28  rft     last modification
//  28-jul-94 15:24:07  rft     created
// ===========================================================================

extern "C++"
{

INLINE_MEMBER VectorN & VectorN::fill(double value)
{
    double * pNow = getLowerPtr();
    double * pEnd = getUpperPtr();

    while (pNow < pEnd)
	*pNow++ = value;

    return *this;
}

INLINE_MEMBER VectorN & VectorN::operator*=(double scalar)
{
    double * pNow = getLowerPtr();
    double * pEnd = getUpperPtr();

    while (pNow < pEnd)
	*pNow++ *= scalar;

    return *this;
}

INLINE_MEMBER VectorN & VectorN::operator+=(const VectorN & vector)
{
    CH_VECTORN_CODE(
	EH_FATAL_CONDITION(VectorN_errors,getLength() != vector.getLength());
    )

    double * pNow = getLowerPtr();
    double * pEnd = getUpperPtr();
    const double * pOne = vector.getLowerPtr();

    while (pNow < pEnd)
	*pNow++ += *pOne++;

    return *this;
}

INLINE_MEMBER VectorN & VectorN::operator-=(const VectorN & vector)
{
    CH_VECTORN_CODE(
	EH_FATAL_CONDITION(VectorN_errors,getLength() != vector.getLength());
    )

    double * pNow = getLowerPtr();
    double * pEnd = getUpperPtr();
    const double * pOne = vector.getLowerPtr();

    while (pNow < pEnd)
	*pNow++ -= *pOne++;

    return *this;
}

// sfmap: protect: double

INLINE_GLOBAL double dotProduct(const VectorN & vector1, const VectorN & vector2)
{
    CH_VECTORN_CODE(
	EH_FATAL_CONDITION(VectorN_errors,
			   vector1.getLength() != vector2.getLength());
    )

    // sfmap: protect: double

    double result = 0;

    const double * pOne = vector1.getLowerPtr();
    const double * pEnd = vector1.getUpperPtr();
    const double * pTwo = vector2.getLowerPtr();


    // sfmap: protect: double double

    while (pOne < pEnd)
	result += (double)(*pOne++) * (double)(*pTwo++);

    return result;
}

INLINE_GLOBAL VectorN operator-(const VectorN & vector)
{
    VectorN result = VectorN(vector.getLower(), vector.getLength());

    double * pNow = result.getLowerPtr();
    double * pEnd = result.getUpperPtr();
    const double * pOne = vector.getLowerPtr();

    while (pNow < pEnd)
	*pNow++ = - *pOne++;

    return result;
}

INLINE_GLOBAL VectorN operator+(const VectorN & vector1, const VectorN & vector2)
{
    CH_VECTORN_CODE(
	EH_FATAL_CONDITION(VectorN_errors,
			   vector1.getLength() != vector2.getLength());
	EH_FATAL_CONDITION(VectorN_errors,
			   vector1.getLower() != vector2.getLower());
    )

    VectorN result = VectorN(vector1.getLower(),vector1.getLength());

    double * pNow = result.getLowerPtr();
    double * pEnd = result.getUpperPtr();
    const double * pOne = vector1.getLowerPtr();
    const double * pTwo = vector2.getLowerPtr();

    while (pNow < pEnd)
	*pNow++ = *pOne++ + *pTwo++;

    return result;
}

INLINE_GLOBAL VectorN operator-(const VectorN & vector1, const VectorN & vector2)
{
    CH_VECTORN_CODE(
	EH_FATAL_CONDITION(VectorN_errors,
			   vector1.getLength() != vector2.getLength());
	EH_FATAL_CONDITION(VectorN_errors,
			   vector1.getLower() != vector2.getLower());
    )

    VectorN result = VectorN(vector1.getLower(),vector1.getLength());

    double * pNow = result.getLowerPtr();
    double * pEnd = result.getUpperPtr();
    const double * pOne = vector1.getLowerPtr();
    const double * pTwo = vector2.getLowerPtr();

    while (pNow < pEnd)
	*pNow++ = *pOne++ - *pTwo++;

    return result;
}

INLINE_GLOBAL VectorN operator*(double scalar, const VectorN & vector)
{
    VectorN result = VectorN(vector.getLower(), vector.getLength());

    double * pNow = result.getLowerPtr();
    double * pEnd = result.getUpperPtr();
    const double * pOne = vector.getLowerPtr();

    while (pNow < pEnd)
	*pNow++ = scalar * *pOne++;

    return result;
}

INLINE_GLOBAL VectorN operator*(const VectorN & vector, double scalar)
{
    VectorN result = VectorN(vector.getLower(), vector.getLength());

    double * pNow = result.getLowerPtr();
    double * pEnd = result.getUpperPtr();
    const double * pOne = vector.getLowerPtr();

    while (pNow < pEnd)
	*pNow++ = *pOne++ * scalar;

    return result;
}

/*

INLINE_GLOBAL static void siftDownInHeap(
	double top, double * upperPtr, int t, int end
	)
{
    int i = t;
    int j = t+t;
    while (j >= end)
    {
	if (j > end && upperPtr[j] > upperPtr[j-1]) j--;
	if (top > upperPtr[j])
	{
	    upperPtr[i] = upperPtr[j];
	    i = j;
	    j <<= 1;
	}
	else
	{
	    j = end - 1;
	}
    }
    upperPtr[i] = top;
}

INLINE_GLOBAL VectorN sort()
{
    double * upperPtr = getUpperPtr();
    
    int end = -getLength();
    
    for (int t = end/2; t < 0; t++)
    {
	siftDownInHeap(upperPtr[t],upperPtr,t,end);
    }

    double top;
    
    while (end < -1)
    {
	top = upperPtr[end];
	upperPtr[end] = upperPtr[-1];
	end++;
	siftDownInHeap(top,upperPtr,-1,end);
    }
}

*/

INLINE_MEMBER SubVectorN & SubVectorN::operator=(const VectorN & vector)
{
    CH_VECTORN_CODE(
	EH_FATAL_CONDITION(VectorN_errors,getLength() != vector.getLength());
    )

#ifdef __GNUC__
    copy((double*)vector.getLowerPtr(),getLength(),getLowerPtr());
    //   ^^^^^^^^^^
    //  this was put there due to a bug in gcc 2.5.8
    //
#else
    copy(vector.getLowerPtr(),getLength(),getLowerPtr());
#endif
    return *this;
}

INLINE_MEMBER MatrixNM::MatrixNM(const MatrixNM & matrix)
:
    lowerN(matrix.getLowerN()),
    lowerM(matrix.getLowerM()),
    lengthN(matrix.getLengthN()),
    lengthM(matrix.getLengthM())
{
    make();

    double * pNowN = getLowerNptr();
    double * pEndN = getUpperNptr();
    const double * pOneN = matrix.getLowerNptr();

    while (pNowN < pEndN)
    {
#ifdef __GNUC__
	copy((double*)matrix.getLowerMptr(pOneN),
	//   ^^^^^^^^^^
	//  this was put there due to a bug in gcc 2.5.8
	//
	     getLengthM(),
	     getLowerMptr(pNowN));
#else
	copy(matrix.getLowerMptr(pOneN),
	     getLengthM(),
	     getLowerMptr(pNowN));
#endif
	pNowN += getDeltaN();
	pOneN += matrix.getDeltaN();
    }
}

INLINE_MEMBER MatrixNM & MatrixNM::fill(double value)
{
    double * pNowN = getLowerNptr();
    double * pEndN = getUpperNptr();
    
    while (pNowN < pEndN)
    {
	double * pNowM = getLowerMptr(pNowN);
	double * pEndM = getUpperMptr(pNowN);

	while (pNowM < pEndM)
	    *pNowM++ = value;

	pNowN += getDeltaN();
    }
    
    return *this;
}

INLINE_MEMBER MatrixNM & MatrixNM::operator=(const MatrixNM & matrix)
{
    if (this == &matrix)
	return *this;

    kill();
    lowerN = matrix.getLowerN();
    lowerM = matrix.getLowerM();
    lengthN = matrix.getLengthN();
    lengthM = matrix.getLengthM();
    make();

    double * pNowN = getLowerNptr();
    double * pEndN = getUpperNptr();
    const double * pOneN = matrix.getLowerNptr();

    while (pNowN < pEndN)
    {
#ifdef __GNUC__
	copy((double*)matrix.getLowerMptr(pOneN),
	//   ^^^^^^^^^^
	//  this was put there due to a bug in gcc 2.5.8
	//
	     getLengthM(),
	     getLowerMptr(pNowN));
#else
	copy(matrix.getLowerMptr(pOneN),
	     getLengthM(),
	     getLowerMptr(pNowN));
#endif
	pNowN += getDeltaN();
	pOneN += matrix.getDeltaN();
    }

    return *this;
}

INLINE_MEMBER MatrixNM & MatrixNM::operator*=(double scalar)
{
    double * pNowN = getLowerNptr();
    double * pEndN = getUpperNptr();
    
    while (pNowN < pEndN)
    {
	double * pNowM = getLowerMptr(pNowN);
	double * pEndM = getUpperMptr(pNowN);

	while (pNowM < pEndM)
	    *pNowM++ *= scalar;

	pNowN += getDeltaN();
    }
    
    return *this;
}

INLINE_MEMBER MatrixNM & MatrixNM::operator+=(const MatrixNM & matrix)
{
    CH_VECTORN_CODE(
	EH_FATAL_CONDITION(VectorN_errors,getLengthN() != matrix.getLengthN());
	EH_FATAL_CONDITION(VectorN_errors,getLengthM() != matrix.getLengthM());
    )

    double * pNowN = getLowerNptr();
    double * pEndN = getUpperNptr();
    const double * pOneN = matrix.getLowerNptr();
    
    while (pNowN < pEndN)
    {
	double * pNowM = getLowerMptr(pNowN);
	double * pEndM = getUpperMptr(pNowN);
	const double * pOneM = matrix.getLowerMptr(pOneN);

	while (pNowM < pEndM)
	    *pNowM++ += *pOneM++;

	pNowN += getDeltaN();
	pOneN += matrix.getDeltaN();
    }
    
    return *this;
}

INLINE_MEMBER MatrixNM & MatrixNM::operator-=(const MatrixNM & matrix)
{
    CH_VECTORN_CODE(
	EH_FATAL_CONDITION(VectorN_errors,getLengthN() != matrix.getLengthN());
	EH_FATAL_CONDITION(VectorN_errors,getLengthM() != matrix.getLengthM());
    )

    double * pNowN = getLowerNptr();
    double * pEndN = getUpperNptr();
    const double * pOneN = matrix.getLowerNptr();
    
    while (pNowN < pEndN)
    {
	double * pNowM = getLowerMptr(pNowN);
	double * pEndM = getUpperMptr(pNowN);
	const double * pOneM = matrix.getLowerMptr(pOneN);

	while (pNowM < pEndM)
	    *pNowM++ -= *pOneM++;

	pNowN += getDeltaN();
	pOneN += matrix.getDeltaN();
    }
    
    return *this;
}

INLINE_MEMBER SubMatrixNM & SubMatrixNM::operator=(const MatrixNM & matrix)
{
    CH_VECTORN_CODE(
	EH_FATAL_CONDITION(VectorN_errors,getLengthN() != matrix.getLengthN());
	EH_FATAL_CONDITION(VectorN_errors,getLengthM() != matrix.getLengthM());
    )

    double * pNowN = getLowerNptr();
    double * pEndN = getUpperNptr();
    const double * pOneN = matrix.getLowerNptr();

    while (pNowN < pEndN)
    {
#ifdef __GNUC__
	copy((double*)matrix.getLowerMptr(pOneN),
	//   ^^^^^^^^^^
	//  this was put there due to a bug in gcc 2.5.8
	//
	     getLengthM(),
	     getLowerMptr(pNowN));
#else
	copy(matrix.getLowerMptr(pOneN),
	     getLengthM(),
	     getLowerMptr(pNowN));
#endif
	pNowN += getDeltaN();
	pOneN += matrix.getDeltaN();
    }

    return *this;
}

} // extern "C++"

// ===========================================================================
