// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       Crc32
//  TYPE:       c++ inline
//  PROJECT:    VEGA Misc Library
//  CONTENT:    32-bit CRC
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    jlg     Jean-loup Gailly
//              rft     Robert F. Tobler
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 13:46:16  rft     last modification
//  07-Aug-93           rft     converted jlg's c-version to c++
//  07-Aug-93           rft     created
// ===========================================================================
//  DESCRIPTION:
//
//  First, the polynomial itself and its table of feedback terms.  The
//  polynomial is
//
//  x^32+x^26+x^23+x^22+x^16+x^12+x^11+x^10+x^8+x^7+x^5+x^4+x^2+x^1+x^0
//
//  Note that we take it "backwards" and put the highest-order term in
//  the lowest-order bit.  The X^32 term is "implied"; the LSB is the
//  X^31 term, etc.  The X^0 term (usually shown as "+1") results in
//  the MSB being 1.
//
//  Note that the usual hardware shift register implementation, which
//  is what we're using (we're merely optimizing it by doing eight-bit
//  chunks at a time) shifts bits into the lowest-order term.  In our
//  implementation, that means shifting towards the right.  Why do we
//  do it this way?  Because the calculated CRC must be transmitted in
//  order from highest-order term to lowest-order term.  UARTs transmit
//  characters in order from LSB to MSB.  By storing the CRC this way,
//  we hand it to the UART in the order low-byte to high-byte; the UART
//  sends each low-bit to hight-bit; and the result is transmission bit
//  by bit from highest- to lowest-order term without requiring any bit
//  shuffling on our part.  Reception works similarly.
//
//  The feedback terms table consists of 256, 32-bit entries.  Notes:
//
//	The table can be generated at runtime if desired; code to do so
//	is shown later.  It might not be obvious, but the feedback
//	terms simply represent the results of eight shift/xor opera-
//	tions for all combinations of data and CRC register values.
//
//	The values must be right-shifted by eight bits by the "updcrc"
//	logic; the shift must be unsigned (bring in zeroes).  On some
//	hardware you could probably optimize the shift in assembler by
//	using byte-swap instructions.
//	polynomial $edb88320
// ===========================================================================

extern "C++"
{

INLINE_MEMBER Crc32::Crc32(
        )
:
    crc(UINT32_MAX)
{
    TR_FUN0(Crc32_constructors,Crc32);
}

INLINE_MEMBER void Crc32::reset(
        )
{
    crc = UINT32_MAX;
}

INLINE_MEMBER UInt32 Crc32::value(
        )
{
    return crc ^ UINT32_MAX;            //  instead of ~c for 64-bit machines
}

INLINE_MEMBER void Crc32::update(
        const void * dataPtr,
        int length
        )
{
    CH_CODE(EH_FATAL_CONDITION(Crc32_errors,dataPtr == 0));
    CH_CODE(EH_FATAL_CONDITION(Crc32_errors,length < 0));

    register UInt32 c = crc;            // temporary variable
    register const Byte * bytePtr = (const Byte *)dataPtr;
    register const Byte * endPtr = bytePtr + length;

    while (bytePtr < endPtr)
    {
        c = crcTable[((Byte)c ^ (*bytePtr++)) & 0xff] ^ (c >> 8);
    }

    crc = c;
}

INLINE_MEMBER void Crc32::update(
        const char * charPtr
        )
{
    CH_CODE(EH_FATAL_CONDITION(Crc32_errors,charPtr == 0));

    register UInt32 c = crc;            // temporary variable
    register Byte ch;

    while ((ch = *charPtr) != (char)0)
    {
        c = crcTable[((Byte)c ^ ch) & 0xff] ^ (c >> 8);
        charPtr++;
    }

    crc = c;
}

INLINE_MEMBER void Crc32::write(
	void * dataPtr
	)
{
    CH_CODE(EH_FATAL_CONDITION(Crc32_errors,dataPtr == 0));

    register Byte *  bytePtr = (Byte *)dataPtr;
    UInt32 val = crc;
    
    *bytePtr++ = Byte(val); val >>= 8;
    *bytePtr++ = Byte(val); val >>= 8;
    *bytePtr++ = Byte(val); val >>= 8;
    *bytePtr   = Byte(val);
}


INLINE_MEMBER UInt32 Crc32::of(
        const void * dataPtr,
        int length
        )
{
    CH_CODE(EH_FATAL_CONDITION(Crc32_errors,dataPtr == 0));
    CH_CODE(EH_FATAL_CONDITION(Crc32_errors,length < 0));

    register UInt32 c = UINT32_MAX;
    register const Byte * bytePtr = (const Byte *)dataPtr;
    register const Byte * endPtr = bytePtr + length;

    while (bytePtr < endPtr)
    {
        c = crcTable[((Byte)c ^ (*bytePtr++)) & 0xff] ^ (c >> 8);
    }

    return c ^ UINT32_MAX;
}

INLINE_MEMBER UInt32 Crc32::of(
        const char * charPtr
        )
{
    CH_CODE(EH_FATAL_CONDITION(Crc32_errors,charPtr == 0));

    register UInt32 c = UINT32_MAX;
    register Byte ch;

    while ((ch = *charPtr) != 0)
    {
        c = crcTable[((Byte)c ^ ch) & 0xff] ^ (c >> 8);
        charPtr++;
    }

    return c ^ UINT32_MAX;
}

} // extern "C++"

// ===========================================================================
