// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       Crc32
//  TYPE:       c++ header
//  PROJECT:    VEGA Misc Library
//  CONTENT:    32-bit CRC
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    jlg     Jean-loup Gailly
//              rft     Robert F. Tobler
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 13:46:16  rft     last modification
//  07-aug-93           rft     converted jlg's c-version to c++
//  07-aug-93           rft     created
// ===========================================================================
//  MANUAL:
//
//  DESCRIPTION
//      Calculates the 32-bit CRC using a table of 256 precomputed CRC
//      values.
//
//      'int' is used to specify lengths of arbitrary data in byte. On
//      machines with 16-bit ints it is therefore necessary to calculate the
//      CRC of data that is larger than 64K in multiple chunks.
//      This functionality is supported by creating a Crc32 object and
//      updating the Crc with the update() function.
//
//  NOTE
//      This code is based on the c-version of crc32 found in gzip by
//      Jean-loup Gailly
//
//  SEE ALSO
//      VEGA, MiscLibrary
// ===========================================================================
#ifndef _MISC_CRC32_H_
#define _MISC_CRC32_H_

#include <VEGA/Base/System.h>

#include "VEGA/Misc/Globals.h"
#include "VEGA/Misc/Bits.h"

#ifdef __GNUC__
#pragma interface
#endif

extern "C++"
{

// =============================================================== class Crc32

class Crc32
{
public:

    Crc32(                              // Create Crc32 object.
        );

    void reset(                         // Reset crc value.
        );

    UInt32 value(                       // Get crc value.                     
        );

    void update(                        // Update crc to include
        const void * dataPtr,		//  data starting here
        int length			//  with this length (in bytes).
        );

    void update(                        // Update crc to include
        const char * charPtr		//  this (zero-terminated) string.
        );

    void write(				// Write CRC value into next 4 bytes
	void * dataPtr			//  starting at this address.
	);

    static UInt32 of(                   // Calculate crc for
        const void * dataPtr,		//  data starting here
        int length			//  with this length (in bytes).
        );

    static UInt32 of(                   // Calculate crc for
        const char * charPtr		//  this (zero-terminated) string.
        );

    static const UInt32 ok;

private:
    static const UInt32 crcTable[];
    UInt32 crc;
};

} // extern "C++"

#ifndef CRC32_NO_INLINE

#define INLINE_GLOBAL COMPILER_INLINE_GLOBAL
#define INLINE_MEMBER COMPILER_INLINE_MEMBER

#include "VEGA/Misc/Crc32.I"

#undef INLINE_GLOBAL
#undef INLINE_MEMBER

#endif

#endif // _MISC_CRC32_H_
// ===========================================================================
