// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       DynamicArray
//  TYPE:       c++ header
//  PROJECT:    VEGA Misc Library
//  CONTENT:    definition of template class DynamicArray
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    hl      Helwig L"offelmann
//              tg      Thomas Galla
//              rft     Robert F. Tobler
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 13:46:16  rft     last modification
//  29-aug-94 17:48:23  rft     revision: append -> enlareBy,
//                                          + appendElement(const T & element)
//                                          + DynamicArray<T>(int size);
//  27-jul-94 17:48:39  tg      added method isNotValidIndex
//  27-jul-94 15:28:02  hl      created
// ===========================================================================
//  MANUAL:
//
//  NAME
//      DynamicArray:
//
//  DESCRIPTION
//      The class DynamicArray ist a template for an array that can grow
//      at runtime according to the users needs. The access to the array
//      elements is similar to normal arrays in C or C++, but the array
//      boundaris are checked.
//
//      Two different subscription operators a provided:
//          One for constant objects, which returns a CONST REFERENCE to the
//          element, and a second one for non constant objects, which
//          returns a REFERENCE to the element (the element can be modified).
//
//      Several methods to query the size of the array and to set that size
//      are provided.
//
//  CHECKS
//      The subscription operator checks whether the index is within the
//      array bounds.
//
//  SEE ALSO
//      VEGA, MiscLibrary
// ===========================================================================

#ifndef _MISC_DYNAMIC_ARRAY_H_
# define _MISC_DYNAMIC_ARRAY_H_

#include "VEGA/Misc/Globals.h"

#include <VEGA/Misc/TTools.h>

extern "C++"
{

template <class T>
class DynamicArray
{

public:
    DynamicArray(			// Initialize without any elements.
        );
    DynamicArray(			// Initialize with
	int sizeNew			//  that many
        );				//  default elements.
    DynamicArray(
        CONST_TEMPLATE DynamicArray<T> & array
        );
    virtual ~DynamicArray(		// Do the garbage collection.
        );

    T & operator[](			// Return REFERENCE to element
        int index			//  at this index
        );
	//
	//  Note:  If there is no such element the program will terminate with
	//  a fatal error.
	//
	
    CONST_TEMPLATE T & operator[](	// Return CONST REFERENCE to element
        int index			//  at this index.
        ) const;
	//
	//  Note:  If there is no such element the program will terminate with
	//  a fatal error.
	//

    T getElement(			// Get a COPY of element
        int index			//  at this index.
        ) const;
    void setElement(			// Set element
        int       index,		//  at this index
        CONST_TEMPLATE T & newElement	//  to copy of this element.
        );

    void setSize(			// Resize array, so that
        int sizeNew			//  this many slots
        );				//  can be addressed.
	//
	// Note: Additional slots are filled with default Ts.
	//

    void enlargeBy(			// Enlarge array, so that
        int additionalSlots		//  this many additional slots
        );				//  can be addressed.
	//
	// Note: Additional slots are filled with default Ts.
	//

    void appendElement(			// Append a COPY of
        CONST_TEMPLATE T & newElement	//  this element
        );				//  at the end of the array.

    int getSize(			// Return the size of the array.
        ) const;

    DynamicArray<T> & operator=(
        CONST_TEMPLATE DynamicArray<T> & array
        );

    int isNotValidIndex(		// Checks, if
	int index			//  this index
	) const;			//  is invalid.

protected:
    T   * elements;
    int   size;				//  number of address-able elements
    int   allocatedSize;		//  currently allocated slots
};

} // extern "C++"

#ifdef COMPILER_TEMPLATE_CODE_IN_HEADER
#include "VEGA/Misc/DynamicArray.I"
#endif

#endif // #ifndef _MISC_DYNAMIC_ARRAY_H_ =============== eof DynamicArray.h ==

