// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       Crd2
//  TYPE:       c++ header
//  PROJECT:    VEGA Math Library
//  CONTENT:    base types for 2D coordinate manipulation using doubles
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rft     Robert F. Tobler
//              hl      Helwig L"offelman
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 14:06:28  rft     last modification
//  30-mar-94 10:30:36  rft     created
// ===========================================================================
//  MANUAL:
//
//  DESCRIPTION
//      Crd2 is the base type for two dimensional coordinate manipulation.
//      It is a set of three coordinates that can be transformed with the
//      associated matrix type called Mat2.
//
//      The Crd2 and Mat2 classes are based on doubles internally.
//
//  CHECKS
//      Range checking and attempts to normalize null-vectors will only be
//      caught if WITH_CRD2_CHECKS is defined.
//
//  ERRORHANDLERS
//      Crd2_errors
//
//  TRACERS
//      Crd2_constructors
//
//  SEE ALSO
//      VEGA, MathLibrary
// ===========================================================================
#ifndef _MATH_CRD2_H_
#define _MATH_CRD2_H_

#include <iostream.h>

#include "VEGA/Math/globals.h"
#include "VEGA/Math/math.h"

#ifdef __GNUC__
#pragma interface
#endif

extern "C++"
{

// ============================================ classes defined in this module

class Crd2;
class Mat2;

// =========================================== errors generated by this module

#define ERROR_CRD2_SINGULAR_MATRIX      1
        //
        //  An attempt was made to invert a singular matrix.
        //

// ================================================================ class Crd2

class Crd2
{
    friend class Mat2;

    friend class Viewer3D;		//  needs FAST access to members

public:

    // ---------------------------------------------------------- constructors

    Crd2(
        );
    Crd2(
        double x0new,
        double x1new
        );
    Crd2(
        const Crd2 & crd
        );
    ~Crd2(
        );

    // ------------------------------------------------------ member operators

    double operator[](
	int index
	) const;
    double & operator[](
	int index
	);
    Crd2 & operator+=(
	const Crd2 & crd
	);
    Crd2 & operator-=(
	const Crd2 & crd
	);
    Crd2 & operator*=(
	double scalar
	);
    Crd2 & operator/=(
	double scalar
	);
    Crd2 & operator*=(
	const Mat2 & mat
	);

    // ------------------------------------------------------ member functions

    double x0(
	) const;
    double x1(
	) const;
    void x0(
	double x0new
	);
    void x1(
	double x1new
	);

    void normalize(
	);

    // ------------------------------------------------------ friend operators

    friend ostream & operator<<(
	ostream & s,
	const Crd2 & crd
	);
    friend Crd2 operator-(
	const Crd2 & crd
	);
    friend Crd2 operator+(
	const Crd2 & crd1,
	const Crd2 & crd2
	);
    friend Crd2 operator-(
	const Crd2 & crd1,
	const Crd2 & crd2
	);
    friend Crd2 operator*(
	const Crd2 & crd,
	double scalar
	);
    friend Crd2 operator*(
	double scalar,
	const Crd2 & crd
	);
    friend Crd2 operator/(
	const Crd2 & crd,
	double scalar
	);
    friend Crd2 operator*(
	const Crd2 & crd,
	const Mat2 & mat
	);

    // ------------------------------------------------------ friend functions

    friend double squaredLength(
	const Crd2 & crd
	);
    friend double length(
	const Crd2 & crd
	);
    friend double dotProduct(
	const Crd2 & crd1,
	const Crd2 & crd2
	);
    friend Crd2 normalizedCopy(
	const Crd2 & crd
	);

    // ----------------------------------------------- componentwise functions
    //
    //  The following functions apply each operation componentwise.
    //

    void cwMultiplyBy(
	const Crd2 & crd
	);
    void cwDevideBy(
	const Crd2 & crd
	);

    friend Crd2 cwProduct(
	const Crd2 & crd1,
	const Crd2 & crd2
	);
    friend Crd2 cwQuotient(
	const Crd2 & crd1,
	const Crd2 & crd2
	);
    friend Crd2 cwMin(
	const Crd2 & crd1,
	const Crd2 & crd2
	);
    friend Crd2 cwMax(
	const Crd2 & crd1,
	const Crd2 & crd2
	);

    friend Crd2 cwReciprocalCopy(       // Componentwise multiplicative
        const Crd2 & crd		//  inverse. Zero components are not
        );				//  inverted.

    // ---------------------------------------------- transformation functions
    //
    //  The following transformation functions and methods are provided to
    //  facilitate the easy construction of complex matrices. They are mainly
    //  used in the Vec3D module to implement the special purpose
    //  transformation classes.
    //

    Crd2 & scale(
	const Crd2 & s
	);
    Crd2 & scale(
	double s0,
	double s1
	);
    Crd2 & rot(
	double cosAlpha,
	double sinAlpha
	);
    Crd2 & shear0(
	double x0s
	);
    Crd2 & shear1(
	double x1s
	);
    
    friend Crd2 prCrdScale(
	const Crd2 & crd,
	const Crd2 & s
	);
    friend Crd2 prCrdScale(
	const Crd2 & crd,
	double s0, double s1
	);
    friend Crd2 prCrdRot(
	const Crd2 & crd,
	double cosAlpha, double sinAlpha
	);
    friend Crd2 prCrdShear0(
	const Crd2 & crd,
	double x0s
	);
    friend Crd2 prCrdShear1(
	const Crd2 & crd,
	double x1s
	);

    // ----------------------------------------------- static member variables

    static const Crd2 base0;
    static const Crd2 base1;

protected:

    double x[2];
};

// =============================================================== class Mat2

class Mat2
{
    friend class Crd2;

    friend class Viewer3D;		//  needs FAST access to members
public:

    // --------------------------------------------- constructors / destructor

    Mat2(
        );
    Mat2(
        double x00, double x01,
        double x10, double x11
        );
    Mat2(
        const Mat2 & mat
        );
    ~Mat2(
        );

    // ------------------------------------------------------ member operators

    Mat2 & operator*=(
	double scalar
	);
    Mat2 & operator+=(
	const Mat2 & mat
	);
    Mat2 & operator*=(
	const Mat2 & mat
	);

    // ------------------------------------------------------ member functions

    double operator()(
	int i, int j
	) const;	// access element

    void invert(
	);
        //
        //  ERROR_CRD2_SINGULAR_MATRIX
        //

    void transpose(
	);

    // ------------------------------------------------------ friend operators

    friend Mat2 operator-(
	const Mat2 & mat
	);
    friend Mat2 operator+(
	const Mat2 & mat1,
	const Mat2 & mat2
	);
    friend Mat2 operator-(
	const Mat2 & mat1,
	const Mat2 & mat2
	);
    friend Mat2 operator*(
	const Mat2 & mat,
	double scalar
	);
    friend Mat2 operator*(
	double scalar,
	const Mat2 & mat
	);
    friend Crd2 operator*(
	const Crd2 & crd,
	const Mat2 & mat
	);
    friend Mat2 operator*(
	const Mat2 & mat1,
	const Mat2 & mat2
	);
    friend ostream & operator<<(
	ostream & s,
	const Mat2 & mat
	);

    // ------------------------------------------------------ friend functions

    friend double determinant(
	const Mat2 & mat
	);
    friend Mat2 inverseCopy(
	const Mat2 & mat,
	double det			//  det is NOT checked against zero!
	);

    friend Mat2 inverseCopy(
	const Mat2 & mat
	);
        //
        //  ERROR_CRD2_SINGULAR_MATRIX
        //

    friend Mat2 transposedCopy(
	const Mat2 & mat
	);

    // ---------------------------------------------- transformation functions
    //
    //  The following transformation functions and methods are provided to
    //  facilitate the easy construction of complex matrices. They are mainly
    //  used in the Vec3D module to implement the special purpose
    //  transformation classes.
    //

    Mat2 & scale(
	const Crd2 & s
	);
    Mat2 & scale(
	double s0,
	double s1
	);
    Mat2 & rot(
	double cosAlpha,
	double sinAlpha
	);
    Mat2 & shear0(
	double x0s
	);
    Mat2 & shear1(
	double x1s
	);
    
    friend Mat2 prMatScale(
	const Mat2 & mat,
	const Crd2 & s
	);
    friend Mat2 prMatScale(
	const Mat2 & mat,
	double s0, double s1
	);
    friend Mat2 prMatRot(
	const Mat2 & mat,
	double cosAlpha, double sinAlpha
	);
    friend Mat2 prMatShear0(
	const Mat2 & mat,
	double x0s
	);
    friend Mat2 prMatShear1(
	const Mat2 & mat,
	double x1s
	);

    friend Mat2 prScaleMat(
	const Crd2 & s,
	const Mat2 & mat
	);
    friend Mat2 prScaleMat(
	double s0, double s1,
	const Mat2 & mat
	);
    friend Mat2 prRotMat(
	double cosAlpha,
	double sinAlpha,
	const Mat2 & mat
	);
    friend Mat2 prShear0Mat(
	double x0s,
	const Mat2 & mat
	);
    friend Mat2 prShear1Mat(
	double x1s,
	const Mat2 & mat
	);

    // ----------------------------------------------- static member variables

    static const Mat2 identity;

protected:

    double x[3][3];
};

} // extern "C++"

#ifndef CRD2_NO_INLINE

#define INLINE_GLOBAL COMPILER_INLINE_GLOBAL
#define INLINE_MEMBER COMPILER_INLINE_MEMBER

#include "VEGA/Math/crd2.ipp"

#undef INLINE_GLOBAL
#undef INLINE_MEMBER

#endif

#endif // _MATH_CRD2_H_
// ===========================================================================
