// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       Crd3
//  TYPE:       c++ header
//  PROJECT:    VEGA Math Library
//  CONTENT:    base types for 3D coordinate manipulation using doubles
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rft     Robert F. Tobler
//              hl      Helwig L"offelman
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 14:06:28  rft     last modification
//  08-Aug-93           rft     created
// ===========================================================================
//  MANUAL:
//
//  DESCRIPTION
//      Crd3 is the base type for three dimensional coordinate manipulation.
//      It is a set of three coordinates that can be transformed with the
//      associated matrix type called Mat3.
//
//      The Crd3 and Mat3 classes are based on doubles internally.
//
//      In terms of matrix theory a Crd3 is a 1 by 3 line vector and a Mat3
//      is a 3 by 3 Matrix. Thus the transformation of a Crd3 by a Mat3 is
//      defined as follows.
//
//              Crd3    c;
//              Mat3    mat;
//
//              Crd3    crd = c * mat;
//
//                                  (mat(0,0), mat(0,1), mat(0,2))
//                                  (mat(1,0), mat(1,1), mat(1,2))
//                                  (mat(2,0), mat(2,1), mat(2,2))
//
//              (c[0],c[1],c[2])    (  crd[0],   crd[1],   crd[2])
//
//  CHECKS
//      Range checking and attempts to normalize null-vectors will only be
//      caught if WITH_CRD3_CHECKS is defined.
//
//  ERRORHANDLERS
//      Crd3_errors
//
//  TRACERS
//      Crd3_constructors
//
//  SEE ALSO
//      VEGA, MathLibrary
// ===========================================================================
#ifndef _MATH_CRD3_H_
#define _MATH_CRD3_H_

#include <iostream.h>

#include "VEGA/Math/globals.h"
#include "VEGA/Math/math.h"

#ifdef __GNUC__
#pragma interface
#endif

extern "C++"
{

// ============================================ classes defined in this module

class Crd3;
class Mat3;

// =========================================== errors generated by this module

#define ERROR_CRD3_SINGULAR_MATRIX      1
        //
        //  An attempt was made to invert a singular matrix.
        //

// ================================================================ class Crd3

class Crd3
{
    friend class Mat3;

    friend class Viewer3D;		//  needs FAST access to members
public:

    // ---------------------------------------------------------- constructors

    Crd3(
        );
    Crd3(
        double x0new,
        double x1new,
        double x2new
        );
    Crd3(
        const Crd3 & crd
        );
    ~Crd3(
        );

    // ------------------------------------------------------ member operators

    double operator[](
	int index
	) const;
    double & operator[](
	int index
	);
    Crd3 & operator+=(
	const Crd3 & crd
	);
    Crd3 & operator-=(
	const Crd3 & crd
	);
    Crd3 & operator*=(
	double scalar
	);
    Crd3 & operator/=(
	double scalar
	);
    Crd3 & operator*=(
	const Mat3 & mat
	);

    // ------------------------------------------------------ member functions

    double x0(
	) const;
    double x1(
	) const;
    double x2(
	) const;
    void x0(
	double x0new
	);
    void x1(
	double x1new
	);
    void x2(
	double x2new
	);

    void normalize(
	);

    // ------------------------------------------------------ friend operators

    friend ostream & operator<<(
	ostream & s,
	const Crd3 & crd
	);
    friend Crd3 operator-(
	const Crd3 & crd
	);
    friend Crd3 operator+(
	const Crd3 & crd1,
	const Crd3 & crd2
	);
    friend Crd3 operator-(
	const Crd3 & crd1,
	const Crd3 & crd2
	);
    friend Crd3 operator*(
	const Crd3 & crd,
	double scalar
	);
    friend Crd3 operator*(
	double scalar,
	const Crd3 & crd
	);
    friend Crd3 operator/(
	const Crd3 & crd,
	double scalar
	);
    friend Crd3 operator*(
	const Crd3 & crd,
	const Mat3 & mat
	);

    // ------------------------------------------------------ friend functions

    friend double squaredLength(
	const Crd3 & crd
	);
    friend double length(
	const Crd3 & crd
	);
    friend double dotProduct(
	const Crd3 & crd1,
	const Crd3 & crd2
	);
    friend Crd3 crossProduct(
	const Crd3 & crd1,		// (crd1, crd2, result) defines the
	const Crd3 & crd2		//  base of a right-handed coordinate
	);				//  system.

    friend Crd3 normalizedCopy(
	const Crd3 & crd
	);

    // ----------------------------------------------- componentwise functions
    //
    //  The following functions apply each operation componentwise.
    //

    void cwMultiplyBy(
	const Crd3 & crd
	);
    void cwDevideBy(
	const Crd3 & crd
	);

    friend Crd3 cwProduct(
	const Crd3 & crd1,
	const Crd3 & crd2
	);
    friend Crd3 cwQuotient(
	const Crd3 & crd1,
	const Crd3 & crd2
	);
    friend Crd3 cwMin(
	const Crd3 & crd1,
	const Crd3 & crd2
	);
    friend Crd3 cwMax(
	const Crd3 & crd1,
	const Crd3 & crd2
	);

    friend Crd3 cwReciprocalCopy(       // Componentwise multiplicative
        const Crd3 & crd		//  inverse. Zero components are not
        );				//  inverted.

    // ---------------------------------------------- transformation functions
    //
    //  The following transformation functions and methods are provided to
    //  facilitate the easy construction of complex matrices. They are mainly
    //  used in the Vec3D module to implement the special purpose
    //  transformation classes.
    //

    Crd3 & scale(
	const Crd3 & s
	);
    Crd3 & scale(
	double s0,
	double s1,
	double s2
	);
    Crd3 & rot0(
	double cosAlpha,
	double sinAlpha
	);
    Crd3 & rot1(
	double cosAlpha,
	double sinAlpha
	);
    Crd3 & rot2(
	double cosAlpha,
	double sinAlpha
	);
    Crd3 & shear01(
	double x0s,
	double x1s
	);
    Crd3 & shear20(
	double x2s,
	double x0s
	);
    Crd3 & shear12(
	double x1s,
	double x2s
	);
        
    friend Crd3 prCrdScale(
	const Crd3 & crd,
	const Crd3 & s
	);
    friend Crd3 prCrdScale(
	const Crd3 & crd,
	double s0, double s1, double s2
	);
    friend Crd3 prCrdRot0(
	const Crd3 & crd,
	double cosAlpha, double sinAlpha
	);
    friend Crd3 prCrdRot1(
	const Crd3 & crd,
	double cosAlpha, double sinAlpha
	);
    friend Crd3 prCrdRot2(
	const Crd3 & crd,
	double cosAlpha,
	double sinAlpha
	);
    friend Crd3 prCrdShear01(
	const Crd3 & crd,
	double x0s, double x1s
	);
    friend Crd3 prCrdShear20(
	const Crd3 & crd,
	double x0s, double x1s
	);
    friend Crd3 prCrdShear12(
	const Crd3 & crd,
	double x0s, double x1s
	);

    // ----------------------------------------------- static member variables

    static const Crd3 base0;
    static const Crd3 base1;
    static const Crd3 base2;

protected:

    double x[3];
};

// =============================================================== class Mat3

class Mat3
{
    friend class Crd3;

    friend class Viewer3D;		//  needs FAST access to members
public:

    // --------------------------------------------- constructors / destructor

    Mat3(
        );
    Mat3(
        double x00, double x01, double x02,
        double x10, double x11, double x12,
        double x20, double x21, double x22
        );
    Mat3(
        const Mat3 & mat
        );
    ~Mat3(
        );

    // ------------------------------------------------------ member operators

    Mat3 & operator+=(
	const Mat3 & mat
	);
    Mat3 & operator*=(
	double scalar
	);
    Mat3 & operator*=(
	const Mat3 & mat
	);

    // ------------------------------------------------------ member functions

    double operator()(			// Access element.
	int i, int j
	) const;

    void invert(
	);
        //
        //  ERROR_CRD3_SINGULAR_MATRIX
        //

    void transpose(
	);

    // ------------------------------------------------------ friend operators

    friend Mat3 operator-(
	const Mat3 & mat
	);
    friend Mat3 operator+(
	const Mat3 & mat1,
	const Mat3 & mat2
	);
    friend Mat3 operator-(
	const Mat3 & mat1,
	const Mat3 & mat2
	);
    friend Mat3 operator*(
	const Mat3 & mat,
	double scalar
	);
    friend Mat3 operator*(
	double scalar,
	const Mat3 & mat
	);
    friend Crd3 operator*(
	const Crd3 & crd,
	const Mat3 & mat
	);
    friend Mat3 operator*(
	const Mat3 & mat1,
	const Mat3 & mat2
	);
    friend ostream & operator<<(
	ostream & s,
	const Mat3 & mat
	);

    // ------------------------------------------------------ friend functions

    friend double determinant(
	const Mat3 & mat
	);
    friend Mat3 inverseCopy(
	const Mat3 & mat,
	double det			//  det is NOT checked against zero!
	);

    friend Mat3 inverseCopy(
	const Mat3 & mat
	);
        //
        //  ERROR_CRD3_SINGULAR_MATRIX
        //

    friend Mat3 transposedCopy(
	const Mat3 & mat
	);

    // ---------------------------------------------- transformation functions
    //
    //  The following transformation functions and methods are provided to
    //  facilitate the easy construction of complex matrices. They are mainly
    //  used in the Vec3D module to implement the special purpose
    //  transformation classes.
    //

    Mat3 & scale(
	const Crd3 & s
	);
    Mat3 & scale(
	double s0,
	double s1,
	double s2
	);
    Mat3 & rot0(
	double cosAlpha,
	double sinAlpha
	);
    Mat3 & rot1(
	double cosAlpha,
	double sinAlpha
	);
    Mat3 & rot2(
	double cosAlpha,
	double sinAlpha
	);
    Mat3 & shear01(
	double x0s,
	double x1s
	);
    Mat3 & shear20(
	double x2s,
	double x0s
	);
    Mat3 & shear12(
	double x1s,
	double x2s
	);
    
    friend Mat3 prMatScale(
	const Mat3 & mat,
	const Crd3 & s
	);
    friend Mat3 prMatScale(
	const Mat3 & mat,
	double s0, double s1, double s2
	);
    friend Mat3 prMatRot0(
	const Mat3 & mat,
	double cosAlpha, double sinAlpha
	);
    friend Mat3 prMatRot1(
	const Mat3 & mat,
	double cosAlpha, double sinAlpha
	);
    friend Mat3 prMatRot2(
	const Mat3 & mat,
	double cosAlpha, double sinAlpha
	);
    friend Mat3 prMatShear01(
	const Mat3 & mat,
	double x0s, double x1s
	);
    friend Mat3 prMatShear20(
	const Mat3 & mat,
	double x2s, double x0s
	);
    friend Mat3 prMatShear12(
	const Mat3 & mat,
	double x1s, double x2s
	);

    friend Mat3 prScaleMat(
	const Crd3 & s,
	const Mat3 & mat
	);
    friend Mat3 prScaleMat(
	double s0, double s1, double s2,
	const Mat3 & mat
	);
    friend Mat3 prRot0Mat(
	double cosAlpha, double sinAlpha,
	const Mat3 & mat
	);
    friend Mat3 prRot1Mat(
	double cosAlpha, double sinAlpha,
	const Mat3 & mat
	);
    friend Mat3 prRot2Mat(
	double cosAlpha, double sinAlpha,
	const Mat3 & mat
	);
    friend Mat3 prShear01Mat(
	double x0s, double x1s,
	const Mat3 & mat
	);
    friend Mat3 prShear20Mat(
	double x2s, double x0s,
	const Mat3 & mat
	);
    friend Mat3 prShear12Mat(
	double x1s, double x2s,
	const Mat3 & mat
	);

    // ----------------------------------------------- static member variables

    static const Mat3 identity;

protected:

    double x[3][3];
};

} // extern "C++"

#ifndef CRD3_NO_INLINE

#define INLINE_GLOBAL COMPILER_INLINE_GLOBAL
#define INLINE_MEMBER COMPILER_INLINE_MEMBER

#include "VEGA/Math/crd3.ipp"

#undef INLINE_GLOBAL
#undef INLINE_MEMBER

#endif

#endif // _MATH_CRD3_H_
// ===========================================================================

