// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       Crd3
//  TYPE:       c++ inline
//  PROJECT:    VEGA Math Library
//  CONTENT:    base types for 3D coordinate manipulation using doubles
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rft     Robert F. Tobler
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 14:06:28  rft     last modification
//  08-Aug-93           rft     created
// ===========================================================================

extern "C++"
{

// ================================================================ class Crd3

INLINE_MEMBER Crd3::Crd3()
{
    TR_FUN0(Crd3_constructors,Crd3);
}

INLINE_MEMBER Crd3::Crd3(double x0new, double x1new, double x2new)
{
    x[0] = x0new;
    x[1] = x1new;
    x[2] = x2new;

    TR_FUN3(Crd3_constructors,Crd3,x0new,x1new,x2new);
}

INLINE_MEMBER Crd3::Crd3(const Crd3 & crd)
{
    x[0] = crd.x[0];
    x[1] = crd.x[1];
    x[2] = crd.x[2];

    TR_FUN1(Crd3_constructors,Crd3,crd);
}

INLINE_MEMBER Crd3::~Crd3()
{
    TR_MESSAGE(Crd3_constructors,"~" << *this);
}

INLINE_MEMBER double Crd3::operator[](int index) const
{
    CH_CRD3_CODE(
	EH_FATAL_CONDITION(Crd3_errors, index < 0 || index > 2);
    
        if (index < 0 || index > 2)
            EH_FATAL(Crd3_errors,"index == " << index);
    )

    return x[index];
}

INLINE_MEMBER double & Crd3::operator[](int index)
{
    CH_CRD3_CODE(
	EH_FATAL_CONDITION(Crd3_errors, index < 0 || index > 2);
    
        if (index < 0 || index > 2)
            EH_FATAL(Crd3_errors,"index == " << index);
    )

    return x[index];
}

INLINE_MEMBER Crd3 & Crd3::operator+=(const Crd3 & crd)
{
    x[0] += crd.x[0];
    x[1] += crd.x[1];
    x[2] += crd.x[2];

    return *this;
}

INLINE_MEMBER Crd3 & Crd3::operator-=(const Crd3 & crd)
{
    x[0] -= crd.x[0];
    x[1] -= crd.x[1];
    x[2] -= crd.x[2];

    return *this;
}

INLINE_MEMBER Crd3 & Crd3::operator*=(double scalar)
{
    x[0] *= scalar;
    x[1] *= scalar;
    x[2] *= scalar;

    return *this;
}

INLINE_MEMBER Crd3 & Crd3::operator/=(double scalar)
{
    CH_CRD3_CODE(EH_FATAL_CONDITION(Crd3_errors,scalar == 0.0));

    x[0] /= scalar;
    x[1] /= scalar;
    x[2] /= scalar;

    return *this;
}

INLINE_MEMBER double Crd3::x0() const { return x[0]; }
INLINE_MEMBER double Crd3::x1() const { return x[1]; }
INLINE_MEMBER double Crd3::x2() const { return x[2]; }

INLINE_MEMBER void Crd3::x0(double x0new) { x[0] = x0new; }
INLINE_MEMBER void Crd3::x1(double x1new) { x[1] = x1new; }
INLINE_MEMBER void Crd3::x2(double x2new) { x[2] = x2new; }

INLINE_GLOBAL double squaredLength(const Crd3 & crd)
{
    return (square(crd.x[0]) + square(crd.x[1]) + square(crd.x[2]));
}

INLINE_GLOBAL double length(const Crd3 & crd)
{
    return sqrt(squaredLength(crd));
}

INLINE_MEMBER void Crd3::normalize()
{
    double crdLength = length(*this);

    *this /= crdLength;
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const Crd3 & crd)
{
    s << "Crd3(" << crd.x[0] << "," << crd.x[1] << "," << crd.x[2] << ")";
    return s;
}

INLINE_GLOBAL Crd3 operator-(const Crd3 & crd)
{
    return Crd3(-crd.x[0], -crd.x[1], -crd.x[2]);
}

INLINE_GLOBAL Crd3 operator+(const Crd3 & crd1, const Crd3 & crd2)
{
    return Crd3(crd1.x[0] + crd2.x[0],
                crd1.x[1] + crd2.x[1],
                crd1.x[2] + crd2.x[2]);
}

INLINE_GLOBAL Crd3 operator-(const Crd3 & crd1, const Crd3 & crd2)
{
    return Crd3(crd1.x[0] - crd2.x[0],
                crd1.x[1] - crd2.x[1],
                crd1.x[2] - crd2.x[2]);
}

INLINE_GLOBAL Crd3 operator*(const Crd3 & crd, double scalar)
{
    return Crd3(crd.x[0] * scalar,
                crd.x[1] * scalar,
                crd.x[2] * scalar);
}

INLINE_GLOBAL Crd3 operator*(double scalar, const Crd3 & crd)
{
    return Crd3(crd.x[0] * scalar,
                crd.x[1] * scalar,
                crd.x[2] * scalar);
}

INLINE_GLOBAL Crd3 operator/(const Crd3 & crd, double scalar)
{
    CH_CRD3_CODE(EH_FATAL_CONDITION(Crd3_errors,scalar == 0.0));

    return Crd3(crd.x[0] / scalar,
                crd.x[1] / scalar,
                crd.x[2] / scalar);
}

INLINE_GLOBAL double dotProduct(const Crd3 & crd1, const Crd3 & crd2)
{
    return (crd1.x[0] * crd2.x[0] +
            crd1.x[1] * crd2.x[1] +
            crd1.x[2] * crd2.x[2]);
}

INLINE_GLOBAL Crd3 crossProduct(const Crd3 & crd1, const Crd3 & crd2)
{
    return Crd3(crd1.x[1] * crd2.x[2] - crd1.x[2] * crd2.x[1],
                crd1.x[2] * crd2.x[0] - crd1.x[0] * crd2.x[2],
                crd1.x[0] * crd2.x[1] - crd1.x[1] * crd2.x[0]);
}

INLINE_GLOBAL Crd3 normalizedCopy(const Crd3 & crd)
{
    double crdLength = length(crd);

    return (crd / crdLength);
}

INLINE_MEMBER void Crd3::cwMultiplyBy(const Crd3 & crd)
{
    x[0] *= crd.x[0];
    x[1] *= crd.x[1];
    x[2] *= crd.x[2];
}

INLINE_MEMBER void Crd3::cwDevideBy(const Crd3 & crd)
{
    CH_CRD3_CODE(EH_FATAL_CONDITION(Crd3_errors,crd[0] == 0.0));
    CH_CRD3_CODE(EH_FATAL_CONDITION(Crd3_errors,crd[1] == 0.0));
    CH_CRD3_CODE(EH_FATAL_CONDITION(Crd3_errors,crd[2] == 0.0));

    x[0] /= crd.x[0];
    x[1] /= crd.x[1];
    x[2] /= crd.x[2];
}

INLINE_GLOBAL Crd3 cwReciprocalCopy(const Crd3 & crd)
{
    Crd3 newVec;
    
    if (crd.x[0] != 0.0)
        newVec.x[0] = 1.0 / crd.x[0];
    else
        newVec.x[0] = 0.0;

    if (crd.x[1] != 0.0)
        newVec.x[1] = 1.0 / crd.x[1];
    else
        newVec.x[1] = 0.0;

    if (crd.x[2] != 0.0)
        newVec.x[2] = 1.0 / crd.x[2];
    else
        newVec.x[2] = 0.0;

    return (newVec);
}

INLINE_GLOBAL Crd3 cwProduct(const Crd3 & crd1, const Crd3 & crd2)
{
    return (Crd3(crd1.x[0] * crd2.x[0],
                  crd1.x[1] * crd2.x[1],
                  crd1.x[2] * crd2.x[2]));
}

INLINE_GLOBAL Crd3 cwQuotient(const Crd3 & crd1, const Crd3 & crd2)
{
    CH_CRD3_CODE(EH_FATAL_CONDITION(Crd3_errors,crd2[0] == 0.0));
    CH_CRD3_CODE(EH_FATAL_CONDITION(Crd3_errors,crd2[1] == 0.0));
    CH_CRD3_CODE(EH_FATAL_CONDITION(Crd3_errors,crd2[2] == 0.0));

    return Crd3(crd1.x[0] / crd2.x[0],
                crd1.x[1] / crd2.x[1],
                crd1.x[2] / crd2.x[2]);
}

INLINE_GLOBAL Crd3 cwMin(const Crd3 & crd1, const Crd3 & crd2)
{
    return Crd3(min(crd1.x[0],crd2.x[0]),
                min(crd1.x[1],crd2.x[1]),
                min(crd1.x[2],crd2.x[2]));
}

INLINE_GLOBAL Crd3 cwMax(const Crd3 & crd1, const Crd3 & crd2)
{
    return Crd3(max(crd1.x[0],crd2.x[0]),
                max(crd1.x[1],crd2.x[1]),
                max(crd1.x[2],crd2.x[2]));
}

INLINE_MEMBER Crd3 & Crd3::scale(const Crd3 & s)
{
    x[0] *= s[0];
    x[1] *= s[1];
    x[2] *= s[2];

    return *this;
}

INLINE_MEMBER Crd3 & Crd3::scale(double s0, double s1, double s2)
{
    x[0] *= s0;
    x[1] *= s1;
    x[2] *= s2;

    return *this;
}

INLINE_MEMBER Crd3 & Crd3::rot0(double cosAlpha, double sinAlpha)
{
    double t = x[1];

    x[1] = x[1] * cosAlpha - x[2] * sinAlpha;
    x[2] = x[2] * cosAlpha +    t * sinAlpha;

    return *this;
}

INLINE_MEMBER Crd3 & Crd3::rot1(double cosAlpha, double sinAlpha)
{
    double t = x[2];
    
    x[2] = x[2] * cosAlpha - x[0] * sinAlpha;
    x[0] = x[0] * cosAlpha +    t * sinAlpha;

    return *this;
}

INLINE_MEMBER Crd3 & Crd3::rot2(double cosAlpha, double sinAlpha)
{
    double t = x[0];
    
    x[0] = x[0] * cosAlpha - x[1] * sinAlpha;
    x[1] = x[1] * cosAlpha +    t * sinAlpha;

    return *this;
}

INLINE_MEMBER Crd3 & Crd3::shear01(double x0s, double x1s)
{
    x[0] += x[2] * x0s;
    x[1] += x[2] * x1s;

    return *this;
}

INLINE_MEMBER Crd3 & Crd3::shear20(double x2s, double x0s)
{
    x[0] += x[1] * x0s;
    x[2] += x[1] * x2s;

    return *this;
}

INLINE_MEMBER Crd3 & Crd3::shear12(double x1s, double x2s)
{
    x[1] += x[0] * x1s;
    x[2] += x[0] * x2s;

    return *this;
}

INLINE_MEMBER Crd3 & Crd3::operator*=(const Mat3 & mat)
{
    double t0 = x[0];
    double t1 = x[1];

    x[0] = x[0] * mat.x[0][0] + x[1] * mat.x[1][0] + x[2] * mat.x[2][0];
    x[1] = t0   * mat.x[0][1] + x[1] * mat.x[1][1] + x[2] * mat.x[2][1];
    x[2] = t0   * mat.x[0][2] + t1   * mat.x[1][2] + x[2] * mat.x[2][2];
    
    return *this;
}

INLINE_GLOBAL Crd3 prCrdScale(const Crd3 & crd, const Crd3 & s)
{
    return Crd3(crd.x[0] * s[0], crd.x[1] * s[1], crd.x[2] * s[2]);
}

INLINE_GLOBAL Crd3 prCrdScale(const Crd3 & crd, double s0, double s1, double s2)
{
    return Crd3(crd.x[0] * s0, crd.x[1] * s1, crd.x[2] * s2);
}

INLINE_GLOBAL Crd3 prCrdRot0(const Crd3 & crd, double cosAlpha, double sinAlpha)
{
    return Crd3(crd.x[0],
                crd.x[1] * cosAlpha - crd.x[2] * sinAlpha,
                crd.x[2] * cosAlpha + crd.x[1] * sinAlpha);
}

INLINE_GLOBAL Crd3 prCrdRot1(const Crd3 & crd, double cosAlpha, double sinAlpha)
{
    return Crd3(crd.x[0] * cosAlpha + crd.x[2] * sinAlpha,
                crd.x[1],
                crd.x[2] * cosAlpha - crd.x[0] * sinAlpha);
}

INLINE_GLOBAL Crd3 prCrdRot2(const Crd3 & crd, double cosAlpha, double sinAlpha)
{
    return Crd3(crd.x[0] * cosAlpha - crd.x[1] * sinAlpha,
                crd.x[1] * cosAlpha + crd.x[0] * sinAlpha,
                crd.x[2]);
}

INLINE_GLOBAL Crd3 prCrdShear01(const Crd3 & crd, double x0s, double x1s)
{
    return Crd3(crd.x[0] + crd.x[2] * x0s,
                crd.x[1] + crd.x[2] * x1s,
                crd.x[2]);
}

INLINE_GLOBAL Crd3 prCrdShear20(const Crd3 & crd, double x2s, double x0s)
{
    return Crd3(crd.x[0] + crd.x[1] * x0s,
                crd.x[1],
                crd.x[2] + crd.x[1] * x2s);
}

INLINE_GLOBAL Crd3 prCrdShear12(const Crd3 & crd, double x1s, double x2s)
{
    return Crd3(crd.x[0],
                crd.x[1] + crd.x[0] * x1s,
                crd.x[2] + crd.x[0] * x2s);
}

INLINE_GLOBAL Crd3 operator*(const Crd3 & crd, const Mat3 & mat)
{
    return
        Crd3
        (
            crd.x[0] * mat.x[0][0] +
            crd.x[1] * mat.x[1][0] +
            crd.x[2] * mat.x[2][0],

            crd.x[0] * mat.x[0][1] +
            crd.x[1] * mat.x[1][1] +
            crd.x[2] * mat.x[2][1],

            crd.x[0] * mat.x[0][2] +
            crd.x[1] * mat.x[1][2] +
            crd.x[2] * mat.x[2][2]
        );
}

// ================================================================ class Mat3

INLINE_MEMBER Mat3::Mat3()
{
    TR_FUN0(Crd3_constructors,Mat3);
}

INLINE_MEMBER Mat3::Mat3(double x00, double x01, double x02,
                  double x10, double x11, double x12,
                  double x20, double x21, double x22)
{
    x[0][0] = x00; x[0][1] = x01; x[0][2] = x02;
    x[1][0] = x10; x[1][1] = x11; x[1][2] = x12;
    x[2][0] = x20; x[2][1] = x21; x[2][2] = x22;

    TR_FUN4(Crd3_constructors,Mat3,x00,x01,x02,"...");
}

INLINE_MEMBER Mat3::Mat3(const Mat3 & mat)
{
    x[0][0] = mat.x[0][0];
    x[0][1] = mat.x[0][1];
    x[0][2] = mat.x[0][2];
    x[1][0] = mat.x[1][0];
    x[1][1] = mat.x[1][1];
    x[1][2] = mat.x[1][2];
    x[2][0] = mat.x[2][0];
    x[2][1] = mat.x[2][1];
    x[2][2] = mat.x[2][2];
    TR_FUN1(Crd3_constructors,Mat3,mat);
}


INLINE_MEMBER Mat3::~Mat3()
{
    TR_MESSAGE(Crd3_constructors,"~" << *this);
}

INLINE_MEMBER double Mat3::operator()(int i, int j) const
{
    CH_CRD3_CODE(
        if (i < 0 || i > 2)
            EH_FATAL(Crd3_errors,"i == " << i);
        if (j < 0 || j > 2)
            EH_FATAL(Crd3_errors,"j == " << j);
    )
    return x[i][j];
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const Mat3 & mat)
{
    s << "Mat3("
	<< mat.x[0][0] << ',' << mat.x[0][1] << ',' << mat.x[0][2] << ','
	<< mat.x[1][0] << ',' << mat.x[1][1] << ',' << mat.x[1][2] << ','
	<< mat.x[2][0] << ',' << mat.x[2][1] << ',' << mat.x[2][2] << ')';
    return s;
}


INLINE_GLOBAL double determinant(const Mat3 & mat)
{
    return
        det3x3(
            mat.x[0][0], mat.x[0][1], mat.x[0][2],
            mat.x[1][0], mat.x[1][1], mat.x[1][2],
            mat.x[2][0], mat.x[2][1], mat.x[2][2]
        );
}

INLINE_GLOBAL Mat3 inverseCopy(const Mat3 & mat, double det)
{
    det = 1.0 / det;

    return
        Mat3
        (
            (mat.x[1][1] * mat.x[2][2] -
             mat.x[2][1] * mat.x[1][2]) * det,
            (mat.x[2][1] * mat.x[0][2] -
             mat.x[0][1] * mat.x[2][2]) * det,
            (mat.x[0][1] * mat.x[1][2] -
             mat.x[1][1] * mat.x[0][2]) * det,

            (mat.x[2][0] * mat.x[1][2] -
             mat.x[1][0] * mat.x[2][2]) * det,
            (mat.x[0][0] * mat.x[2][2] -
             mat.x[2][0] * mat.x[0][2]) * det,
            (mat.x[1][0] * mat.x[0][2] -
             mat.x[0][0] * mat.x[1][2]) * det,

            (mat.x[1][0] * mat.x[2][1] -
             mat.x[2][0] * mat.x[1][1]) * det,
            (mat.x[2][0] * mat.x[0][1] -
             mat.x[0][0] * mat.x[2][1]) * det,
            (mat.x[0][0] * mat.x[1][1] -
             mat.x[1][0] * mat.x[0][1]) * det
        );
}

INLINE_GLOBAL Mat3 inverseCopy(const Mat3 & mat)
{
    double det = determinant(mat);

    if (det == 0.0)
    {
        EH_ERROR(Crd3_errors, ERROR_CRD3_SINGULAR_MATRIX, "singular matrix");
        return mat;
    }

    return inverseCopy(mat,det);
}


INLINE_MEMBER void Mat3::invert()
{
    *this = inverseCopy(*this);
}

INLINE_GLOBAL Mat3 transposedCopy(const Mat3 & mat)
{
    return
        Mat3(
            mat.x[0][0], mat.x[1][0], mat.x[2][0],
            mat.x[0][1], mat.x[1][1], mat.x[2][1],
            mat.x[0][2], mat.x[1][2], mat.x[2][2]
        );
}

INLINE_MEMBER void Mat3::transpose()
{
    *this = transposedCopy(*this);
}

INLINE_MEMBER Mat3 & Mat3::operator*=(double s)
{
    x[0][0] *= s; x[0][1] *= s; x[0][2] *= s;
    x[1][0] *= s; x[1][1] *= s; x[1][2] *= s;
    x[2][0] *= s; x[2][1] *= s; x[2][2] *= s;
    return *this;
}

INLINE_MEMBER Mat3 & Mat3::scale(const Crd3 & s)
{
    x[0][0] *= s[0]; x[0][1] *= s[1]; x[0][2] *= s[2];
    x[1][0] *= s[0]; x[1][1] *= s[1]; x[1][2] *= s[2];
    x[2][0] *= s[0]; x[2][1] *= s[1]; x[2][2] *= s[2];
    return *this;
}

INLINE_MEMBER Mat3 & Mat3::scale(double s0, double s1, double s2)
{
    x[0][0] *= s0; x[0][1] *= s1; x[0][2] *= s2;
    x[1][0] *= s0; x[1][1] *= s1; x[1][2] *= s2;
    x[2][0] *= s0; x[2][1] *= s1; x[2][2] *= s2;
    return *this;
}

INLINE_MEMBER Mat3 & Mat3::rot0(double cosAlpha, double sinAlpha)
{
    double t;

    t = x[0][1];
    x[0][1] = t * cosAlpha - x[0][2] * sinAlpha;
    x[0][2] = t * sinAlpha + x[0][2] * cosAlpha;
    t = x[1][1];
    x[1][1] = t * cosAlpha - x[1][2] * sinAlpha;
    x[1][2] = t * sinAlpha + x[1][2] * cosAlpha;
    t = x[2][1];
    x[2][1] = t * cosAlpha - x[2][2] * sinAlpha;
    x[2][2] = t * sinAlpha + x[2][2] * cosAlpha;
    return *this;
}

INLINE_MEMBER Mat3 & Mat3::rot1(double cosAlpha, double sinAlpha)
{
    double t;

    t = x[0][0];
    x[0][0] =  t * cosAlpha + x[0][2] * sinAlpha;
    x[0][2] = -t * sinAlpha + x[0][2] * cosAlpha;
    t = x[1][0];
    x[1][0] =  t * cosAlpha + x[1][2] * sinAlpha;
    x[1][2] = -t * sinAlpha + x[1][2] * cosAlpha;
    t = x[2][0];
    x[2][0] =  t * cosAlpha + x[2][2] * sinAlpha;
    x[2][2] = -t * sinAlpha + x[2][2] * cosAlpha;
    return *this;
}

INLINE_MEMBER Mat3 & Mat3::rot2(double cosAlpha, double sinAlpha)
{
    double t;

    t = x[0][0];
    x[0][0] = t * cosAlpha - x[0][1] * sinAlpha;
    x[0][1] = t * sinAlpha + x[0][1] * cosAlpha;
    t = x[1][0];
    x[1][0] = t * cosAlpha - x[1][1] * sinAlpha;
    x[1][1] = t * sinAlpha + x[1][1] * cosAlpha;
    t = x[2][0];
    x[2][0] = t * cosAlpha - x[2][1] * sinAlpha;
    x[2][1] = t * sinAlpha + x[2][1] * cosAlpha;
    return *this;
}

INLINE_MEMBER Mat3 & Mat3::shear01(double x0s, double x1s)
{
    x[0][0] += x[0][2] * x0s;
    x[0][1] += x[0][2] * x1s;
    x[1][0] += x[1][2] * x0s;
    x[1][1] += x[1][2] * x1s;
    x[2][0] += x[2][2] * x0s;
    x[2][1] += x[2][2] * x1s;
    return *this;
}

INLINE_MEMBER Mat3 & Mat3::shear20(double x2s, double x0s)
{
    x[0][0] += x[0][1] * x0s;
    x[0][2] += x[0][1] * x2s;
    x[1][0] += x[1][1] * x0s;
    x[1][2] += x[1][1] * x2s;
    x[2][0] += x[2][1] * x0s;
    x[2][2] += x[2][1] * x2s;
    return *this;
}

INLINE_MEMBER Mat3 & Mat3::shear12(double x1s, double x2s)
{
    x[0][1] += x[0][0] * x1s;
    x[0][2] += x[0][0] * x2s;
    x[1][1] += x[1][0] * x1s;
    x[1][2] += x[1][0] * x2s;
    x[2][1] += x[2][0] * x1s;
    x[2][2] += x[2][0] * x2s;
    return *this;
}

INLINE_MEMBER Mat3 & Mat3::operator+=(const Mat3 & mat)
{
    x[0][0] += mat.x[0][0];
    x[0][1] += mat.x[0][1];
    x[0][2] += mat.x[0][2];
    x[1][0] += mat.x[1][0];
    x[1][1] += mat.x[1][1];
    x[1][2] += mat.x[1][2];
    x[2][0] += mat.x[2][0];
    x[2][1] += mat.x[2][1];
    x[2][2] += mat.x[2][2];

    return *this;
}

INLINE_GLOBAL Mat3 operator*(const Mat3 & mat1, const Mat3 & mat2)
{
    return
        Mat3
        (
            mat1.x[0][0] * mat2.x[0][0] +
            mat1.x[0][1] * mat2.x[1][0] +
            mat1.x[0][2] * mat2.x[2][0],
            mat1.x[0][0] * mat2.x[0][1] +
            mat1.x[0][1] * mat2.x[1][1] +
            mat1.x[0][2] * mat2.x[2][1],
            mat1.x[0][0] * mat2.x[0][2] +
            mat1.x[0][1] * mat2.x[1][2] +
            mat1.x[0][2] * mat2.x[2][2],

            mat1.x[1][0] * mat2.x[0][0] +
            mat1.x[1][1] * mat2.x[1][0] +
            mat1.x[1][2] * mat2.x[2][0],
            mat1.x[1][0] * mat2.x[0][1] +
            mat1.x[1][1] * mat2.x[1][1] +
            mat1.x[1][2] * mat2.x[2][1],
            mat1.x[1][0] * mat2.x[0][2] +
            mat1.x[1][1] * mat2.x[1][2] +
            mat1.x[1][2] * mat2.x[2][2],

            mat1.x[2][0] * mat2.x[0][0] +
            mat1.x[2][1] * mat2.x[1][0] +
            mat1.x[2][2] * mat2.x[2][0],
            mat1.x[2][0] * mat2.x[0][1] +
            mat1.x[2][1] * mat2.x[1][1] +
            mat1.x[2][2] * mat2.x[2][1],
            mat1.x[2][0] * mat2.x[0][2] +
            mat1.x[2][1] * mat2.x[1][2] +
            mat1.x[2][2] * mat2.x[2][2]
        );
}

INLINE_MEMBER Mat3 & Mat3::operator*=(const Mat3 & mat)
{
    *this = *this * mat;

    return *this;
}

INLINE_GLOBAL Mat3 operator*(const Mat3 & mat, double s)
{
    return
        Mat3
        (
            mat.x[0][0] * s, mat.x[0][1] * s, mat.x[0][2] * s,
            mat.x[1][0] * s, mat.x[1][1] * s, mat.x[1][2] * s,
            mat.x[2][0] * s, mat.x[2][1] * s, mat.x[2][2] * s
        );
}

INLINE_GLOBAL Mat3 prMatScale(const Mat3 & mat, const Crd3 & s)
{
    return
        Mat3
        (
            mat.x[0][0] * s[0], mat.x[0][1] * s[1], mat.x[0][2] * s[2],
            mat.x[1][0] * s[0], mat.x[1][1] * s[1], mat.x[1][2] * s[2],
            mat.x[2][0] * s[0], mat.x[2][1] * s[1], mat.x[2][2] * s[2]
        );
}

INLINE_GLOBAL Mat3 prMatScale(const Mat3 & mat, double s0, double s1, double s2)
{
    return
        Mat3
        (
            mat.x[0][0] * s0, mat.x[0][1] * s1, mat.x[0][2] * s2,
            mat.x[1][0] * s0, mat.x[1][1] * s1, mat.x[1][2] * s2,
            mat.x[2][0] * s0, mat.x[2][1] * s1, mat.x[2][2] * s2
        );
}

INLINE_GLOBAL Mat3 prMatRot0(const Mat3 & mat, double cosAlpha, double sinAlpha)
{
    return
        Mat3
        (
            mat.x[0][0],
            mat.x[0][1] * cosAlpha - mat.x[0][2] * sinAlpha,
            mat.x[0][2] * cosAlpha + mat.x[0][1] * sinAlpha,

            mat.x[1][0],
            mat.x[1][1] * cosAlpha - mat.x[1][2] * sinAlpha,
            mat.x[1][2] * cosAlpha + mat.x[1][1] * sinAlpha,

            mat.x[2][0],
            mat.x[2][1] * cosAlpha - mat.x[2][2] * sinAlpha,
            mat.x[2][2] * cosAlpha + mat.x[2][1] * sinAlpha
        );
}

INLINE_GLOBAL Mat3 prMatRot1(const Mat3 & mat, double cosAlpha, double sinAlpha)
{
    return
        Mat3
        (
            mat.x[0][0] * cosAlpha + mat.x[0][2] * sinAlpha,
            mat.x[0][1],
            mat.x[0][2] * cosAlpha - mat.x[0][0] * sinAlpha,

            mat.x[1][0] * cosAlpha + mat.x[1][2] * sinAlpha,
            mat.x[1][1],
            mat.x[1][2] * cosAlpha - mat.x[1][0] * sinAlpha,

            mat.x[2][0] * cosAlpha + mat.x[2][2] * sinAlpha,
            mat.x[2][1],
            mat.x[2][2] * cosAlpha - mat.x[2][0] * sinAlpha
        );
}

INLINE_GLOBAL Mat3 prMatRot2(const Mat3 & mat, double cosAlpha, double sinAlpha)
{
    return
        Mat3
        (
            mat.x[0][0] * cosAlpha - mat.x[0][1] * sinAlpha,
            mat.x[0][1] * cosAlpha + mat.x[0][0] * sinAlpha,
            mat.x[0][2],

            mat.x[1][0] * cosAlpha - mat.x[1][1] * sinAlpha,
            mat.x[1][1] * cosAlpha + mat.x[1][0] * sinAlpha,
            mat.x[1][2],

            mat.x[2][0] * cosAlpha - mat.x[2][1] * sinAlpha,
            mat.x[2][1] * cosAlpha + mat.x[2][0] * sinAlpha,
            mat.x[2][2]
        );
}

INLINE_GLOBAL Mat3 prMatShear01(const Mat3 & mat, double x0s, double x1s)
{
    return
        Mat3
        (
            mat.x[0][0] + mat.x[0][2] * x0s,
            mat.x[0][1] + mat.x[0][2] * x1s,
            mat.x[0][2],

            mat.x[1][0] + mat.x[1][2] * x0s,
            mat.x[1][1] + mat.x[1][2] * x1s,
            mat.x[1][2],

            mat.x[2][0] + mat.x[2][2] * x0s,
            mat.x[2][1] + mat.x[2][2] * x1s,
            mat.x[2][2]
        );
}

INLINE_GLOBAL Mat3 prMatShear20(const Mat3 & mat, double x2s, double x0s)
{
    return
        Mat3
        (
            mat.x[0][0] + mat.x[0][1] * x0s,
            mat.x[0][1],
            mat.x[0][2] + mat.x[0][1] * x2s,

            mat.x[1][0] + mat.x[1][1] * x0s,
            mat.x[1][1],
            mat.x[1][2] + mat.x[1][1] * x2s,

            mat.x[2][0] + mat.x[2][1] * x0s,
            mat.x[2][1],
            mat.x[2][2] + mat.x[2][1] * x2s
        );
}

INLINE_GLOBAL Mat3 prMatShear12(const Mat3 & mat, double x1s, double x2s)
{
    return
        Mat3
        (
            mat.x[0][0],
            mat.x[0][1] + mat.x[0][0] * x1s,
            mat.x[0][2] + mat.x[0][0] * x2s,

            mat.x[1][0],
            mat.x[1][1] + mat.x[1][0] * x1s,
            mat.x[1][2] + mat.x[1][0] * x2s,

            mat.x[2][0],
            mat.x[2][1] + mat.x[2][0] * x1s,
            mat.x[2][2] + mat.x[2][0] * x2s
        );
}

INLINE_GLOBAL Mat3 operator*(double s, const Mat3 & mat)
{
    return
        Mat3
        (
            s * mat.x[0][0], s * mat.x[0][1], s * mat.x[0][2],
            s * mat.x[1][0], s * mat.x[1][1], s * mat.x[1][2],
            s * mat.x[2][0], s * mat.x[2][1], s * mat.x[2][2]
        );
}

INLINE_GLOBAL Mat3 prScaleMat(const Crd3 & s, const Mat3 & mat)
{
    return
        Mat3
        (
            s[0] * mat.x[0][0], s[0] * mat.x[0][1], s[0] * mat.x[0][2],
            s[1] * mat.x[1][0], s[1] * mat.x[1][1], s[1] * mat.x[1][2],
            s[2] * mat.x[2][0], s[2] * mat.x[2][1], s[2] * mat.x[2][2]
        );
}

INLINE_GLOBAL Mat3 prScaleMat(double s0, double s1, double s2, const Mat3 & mat)
{
    return
        Mat3
        (
            s0 * mat.x[0][0], s0 * mat.x[0][1], s0 * mat.x[0][2],
            s1 * mat.x[1][0], s1 * mat.x[1][1], s1 * mat.x[1][2],
            s2 * mat.x[2][0], s2 * mat.x[2][1], s2 * mat.x[2][2]
        );
}

INLINE_GLOBAL Mat3 prRot0Mat(double cosAlpha, double sinAlpha, const Mat3 & mat)
{
    return
        Mat3
        (
            mat.x[0][0],
            mat.x[0][1],
            mat.x[0][2],

            cosAlpha * mat.x[1][0] + sinAlpha * mat.x[2][0],
            cosAlpha * mat.x[1][1] + sinAlpha * mat.x[2][1],
            cosAlpha * mat.x[1][2] + sinAlpha * mat.x[2][2],

            cosAlpha * mat.x[2][0] - sinAlpha * mat.x[1][0],
            cosAlpha * mat.x[2][1] - sinAlpha * mat.x[1][1],
            cosAlpha * mat.x[2][2] - sinAlpha * mat.x[1][2]
        );
}

INLINE_GLOBAL Mat3 prRot1Mat(double cosAlpha, double sinAlpha, const Mat3 & mat)
{
    return
        Mat3
        (
            cosAlpha * mat.x[0][0] - sinAlpha * mat.x[2][0],
            cosAlpha * mat.x[0][1] - sinAlpha * mat.x[2][1],
            cosAlpha * mat.x[0][2] - sinAlpha * mat.x[2][2],

            mat.x[1][0],
            mat.x[1][1],
            mat.x[1][2],

            cosAlpha * mat.x[2][0] + sinAlpha * mat.x[0][0],
            cosAlpha * mat.x[2][1] + sinAlpha * mat.x[0][1],
            cosAlpha * mat.x[2][2] + sinAlpha * mat.x[0][2]
        );
}

INLINE_GLOBAL Mat3 prRot2Mat(double cosAlpha, double sinAlpha, const Mat3 & mat)
{
    return
        Mat3
        (
            cosAlpha * mat.x[0][0] + sinAlpha * mat.x[1][0],
            cosAlpha * mat.x[0][1] + sinAlpha * mat.x[1][1],
            cosAlpha * mat.x[0][2] + sinAlpha * mat.x[1][2],

            cosAlpha * mat.x[1][0] - sinAlpha * mat.x[0][0],
            cosAlpha * mat.x[1][1] - sinAlpha * mat.x[0][1],
            cosAlpha * mat.x[1][2] - sinAlpha * mat.x[0][2],

            mat.x[2][0],
            mat.x[2][1],
            mat.x[2][2]
        );
}

INLINE_GLOBAL Mat3 prShear01Mat(double x0s, double x1s, const Mat3 & mat)
{
    return
        Mat3
        (
	    mat.x[0][0],
	    mat.x[0][1],
	    mat.x[0][2],
	    
	    mat.x[1][0],
	    mat.x[1][1],
	    mat.x[1][2],

            x0s * mat.x[0][0] + x1s * mat.x[1][0] + mat.x[2][0],
            x0s * mat.x[0][1] + x1s * mat.x[1][1] + mat.x[2][1],
            x0s * mat.x[0][2] + x1s * mat.x[1][2] + mat.x[2][2]
        );
}

INLINE_GLOBAL Mat3 prShear20Mat(double x2s, double x0s, const Mat3 & mat)
{
    return
        Mat3
        (
	    mat.x[0][0],
	    mat.x[0][1],
	    mat.x[0][2],
	    
            x0s * mat.x[0][0] + mat.x[1][0] + x2s * mat.x[2][0],
            x0s * mat.x[0][1] + mat.x[1][1] + x2s * mat.x[2][1],
            x0s * mat.x[0][2] + mat.x[1][2] + x2s * mat.x[2][2],

	    mat.x[2][0],
	    mat.x[2][1],
	    mat.x[2][2]
        );
}

INLINE_GLOBAL Mat3 prShear12Mat(double x1s, double x2s, const Mat3 & mat)
{
    return
        Mat3
        (
            mat.x[0][0] + x1s * mat.x[1][0] + x2s * mat.x[2][0],
            mat.x[0][1] + x1s * mat.x[1][1] + x2s * mat.x[2][1],
            mat.x[0][2] + x1s * mat.x[1][2] + x2s * mat.x[2][2],

	    mat.x[1][0],
	    mat.x[1][1],
	    mat.x[1][2],
	    
	    mat.x[2][0],
	    mat.x[2][1],
	    mat.x[2][2]
        );
}

INLINE_GLOBAL Mat3 operator-(const Mat3 & mat)
{
    return
        Mat3
        (
	    - mat.x[0][0], - mat.x[0][1], - mat.x[0][2],
	    - mat.x[1][0], - mat.x[1][1], - mat.x[1][2],
	    - mat.x[2][0], - mat.x[2][1], - mat.x[2][2]
	);
}

INLINE_GLOBAL Mat3 operator+(const Mat3 & mat1, const Mat3 & mat2)
{
    return
        Mat3
        (
	    mat1.x[0][0] + mat2.x[0][0],
	    mat1.x[0][1] + mat2.x[0][1],
	    mat1.x[0][2] + mat2.x[0][2],

	    mat1.x[1][0] + mat2.x[1][0],
	    mat1.x[1][1] + mat2.x[1][1],
	    mat1.x[1][2] + mat2.x[1][2],

	    mat1.x[2][0] + mat2.x[2][0],
	    mat1.x[2][1] + mat2.x[2][1],
	    mat1.x[2][2] + mat2.x[2][2]
	);
}

INLINE_GLOBAL Mat3 operator-(const Mat3 & mat1, const Mat3 & mat2)
{
    return
        Mat3
        (
	    mat1.x[0][0] - mat2.x[0][0],
	    mat1.x[0][1] - mat2.x[0][1],
	    mat1.x[0][2] - mat2.x[0][2],

	    mat1.x[1][0] - mat2.x[1][0],
	    mat1.x[1][1] - mat2.x[1][1],
	    mat1.x[1][2] - mat2.x[1][2],

	    mat1.x[2][0] - mat2.x[2][0],
	    mat1.x[2][1] - mat2.x[2][1],
	    mat1.x[2][2] - mat2.x[2][2]
	);
}

} // extern "C++"

// ===========================================================================
