// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       Crd4
//  TYPE:       c++ header
//  PROJECT:    VEGA Math Library
//  CONTENT:    base types for 4D coordinate manipulation using doubles
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rft     Robert F. Tobler
//              hl      Helwig L"offelman
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 14:06:28  rft     last modification
//  25-mar-94 23:09:46  rft     created
// ===========================================================================
//  MANUAL:
//
//  DESCRIPTION
//      Crd4 is the base type for four dimensional coordinate manipulation.
//      It is a set of four coordinates that can be transformed with the
//      associated matrix type called Mat4.
//
//      The Crd4 and Mat4 classes are based on doubles internally.
//
//  CHECKS
//      Range checking and attempts to normalize null-vectors will only be
//      caught if WITH_CRD4_CHECKS is defined.
//
//  ERRORHANDLERS
//      Crd4_errors
//
//  TRACERS
//      Crd4_constructors
//
//  SEE ALSO
//      VEGA, MathLibrary
// ===========================================================================
#ifndef _MATH_CRD4_H_
#define _MATH_CRD4_H_

#include <iostream.h>

#include "VEGA/Math/globals.h"
#include "VEGA/Math/math.h"

#ifdef __GNUC__
#pragma interface
#endif

extern "C++"
{

// ============================================ classes defined in this module

class Crd4;
class Mat4;

// =========================================== errors generated by this module

#define ERROR_CRD4_SINGULAR_MATRIX      1
        //
        //  An attempt was made to invert a singular matrix.
        //

// ================================================================ class Crd4

class Crd4
{
    friend class Mat4;

    friend class Viewer3D;		//  needs FAST access to members

public:

    // ---------------------------------------------------------- constructors

    Crd4(
        );
    Crd4(
        double x0new,
        double x1new,
        double x2new,
        double x3new
        );
    Crd4(
        const Crd4 & crd
        );
    ~Crd4(
        );

    // ------------------------------------------------------ member operators

    double operator[](
	int index
	) const;
    double & operator[](
	int index
	);
    Crd4 & operator+=(
	const Crd4 & crd
	);
    Crd4 & operator-=(
	const Crd4 & crd
	);
    Crd4 & operator*=(
	double scalar
	);
    Crd4 & operator/=(
	double scalar
	);
    Crd4 & operator*=(
	const Mat4 & mat
	);

    // ------------------------------------------------------ member functions

    double x0(
	) const;
    double x1(
	) const;
    double x2(
	) const;
    double x3(
	) const;
    void x0(
	double x0new
	);
    void x1(
	double x1new
	);
    void x2(
	double x2new
	);
    void x3(
	double x3new
	);

    void normalize(
	);

    // ------------------------------------------------------ friend operators

    friend ostream & operator<<(
	ostream & s,
	const Crd4 & crd
	);
    friend Crd4 operator-(
	const Crd4 & crd
	);
    friend Crd4 operator+(
	const Crd4 & crd1,
	const Crd4 & crd2
	);
    friend Crd4 operator-(
	const Crd4 & crd1,
	const Crd4 & crd2
	);
    friend Crd4 operator*(
	const Crd4 & crd,
	double scalar
	);
    friend Crd4 operator*(
	double scalar,
	const Crd4 & crd
	);
    friend Crd4 operator/(
	const Crd4 & crd,
	double scalar
	);
    friend Crd4 operator*(
	const Crd4 & crd,
	const Mat4 & mat
	);

    // ------------------------------------------------------ friend functions

    friend double squaredLength(
	const Crd4 & crd
	);
    friend double length(
	const Crd4 & crd
	);
    friend double dotProduct(
	const Crd4 & crd1,
	const Crd4 & crd2
	);
    friend Crd4 normalizedCopy(
	const Crd4 & crd
	);

    // ----------------------------------------------- componentwise functions
    //
    //  The following functions apply each operation componentwise.
    //

    void cwMultiplyBy(
	const Crd4 & crd
	);
    void cwDevideBy(
	const Crd4 & crd
	);

    friend Crd4 cwProduct(
	const Crd4 & crd1,
	const Crd4 & crd2
	);
    friend Crd4 cwQuotient(
	const Crd4 & crd1,
	const Crd4 & crd2
	);
    friend Crd4 cwMin(
	const Crd4 & crd1,
	const Crd4 & crd2
	);
    friend Crd4 cwMax(
	const Crd4 & crd1,
	const Crd4 & crd2
	);

    friend Crd4 cwReciprocalCopy(       // Componentwise multiplicative
        const Crd4 & crd		//  inverse. Zero components are not
        );				//  inverted.

    // ---------------------------------------------- transformation functions
    //
    //  The following transformation functions and methods are provided to
    //  facilitate the easy construction of complex matrices. They are mainly
    //  used in the Vec3D module to implement the special purpose
    //  transformation classes.
    //

    Crd4 & scale(
	const Crd4 & s
	);
    Crd4 & scale(
	double s0,
	double s1,
	double s2,
	double s3
	);

    friend Crd4 prCrdScale(
	const Crd4 & crd,
	const Crd4 & s
	);
    friend Crd4 prCrdScale(
        const Crd4 & crd,
        double s0, double s1, double s2, double s3
        );

    // ----------------------------------------------- static member variables

    static const Crd4 base0;
    static const Crd4 base1;
    static const Crd4 base2;
    static const Crd4 base3;

protected:

    double x[4];
};

// =============================================================== class Mat4

class Mat4
{
    friend class Crd4;

    friend class Viewer3D;		//  needs FAST access to members

public:

    // --------------------------------------------- constructors / destructor

    Mat4(
        );
    Mat4(
        double x00, double x01, double x02, double x03,
        double x10, double x11, double x12, double x13,
        double x20, double x21, double x22, double x23,
        double x30, double x31, double x32, double x33
        );
    Mat4(
        const Mat4 & mat
        );
    ~Mat4(
        );

    // ------------------------------------------------------ member operators

    Mat4 & operator*=(
	double scalar
	);
    Mat4 & operator+=(
	const Mat4 & mat
	);
    Mat4 & operator*=(
	const Mat4 & mat
	);

    // ------------------------------------------------------ member functions

    double operator()(			// Access element.
	int i, int j
	) const;

    void invert(
	);
        //
        //  ERROR_CRD4_SINGULAR_MATRIX
        //

    void transpose(
	);

    // ------------------------------------------------------ friend operators

    friend Mat4 operator-(
	const Mat4 & mat
	);
    friend Mat4 operator+(
	const Mat4 & mat1,
	const Mat4 & mat2
	);
    friend Mat4 operator-(
	const Mat4 & mat1,
	const Mat4 & mat2
	);
    friend Mat4 operator*(
	const Mat4 & mat,
	double scalar
	);
    friend Mat4 operator*(
	double scalar,
	const Mat4 & mat
	);
    friend Crd4 operator*(
	const Crd4 & crd,
	const Mat4 & mat
	);
    friend Mat4 operator*(
	const Mat4 & mat1,
	const Mat4 & mat2
	);
    friend ostream & operator<<(
	ostream & s,
	const Mat4 & mat
	);

    // ------------------------------------------------------ friend functions

    friend double determinant(
	const Mat4 & mat
	);

    friend Mat4 adjointCopy(
	const Mat4 & m
	);
        //
        //  Calculate the adjoint of a 4x4 matrix:
        //
        //  Let  a   denote the minor determinant of matrix A obtained
        //        ij
        //
        //  by deleting the ith row and jth column from A.
        //
        //                 i+j
        //  Let  b   = (-1)    a
        //        ij            ji
        //
        //  The matrix B = (b  ) is the adjoint of A
        //                   ij
        //

    friend Mat4 inverseCopy(
	const Mat4 & mat,
	double det		        //  det is NOT checked against zero!
	);

    friend Mat4 inverseCopy(
	const Mat4 & mat
	);
        //
        //  ERROR_CRD4_SINGULAR_MATRIX
        //

    friend Mat4 transposedCopy(
	const Mat4 & mat
	);

    // ---------------------------------------------- transformation functions
    //
    //  The following transformation functions and methods are provided to
    //  facilitate the easy construction of complex matrices. They are mainly
    //  used in the Vec3D module to implement the special purpose
    //  transformation classes.
    //

    Mat4 & scale(
	const Crd4 & s
	);
    Mat4 & scale(
	double s0,
	double s1,
	double s2,
	double s3
	);
    
    friend Mat4 prMatScale(
	const Mat4 & mat,
	const Crd4 & s
	);
    friend Mat4 prMatScale(
        const Mat4 & mat,
        double s0, double s1, double s2, double s3
        );

    friend Mat4 prScaleMat(
	const Crd4 & s,
	const Mat4 & mat
	);
    friend Mat4 prScaleMat(
        double s0, double s1, double s2, double s3,
        const Mat4 & mat
        );

    // ----------------------------------------------- static member variables

    static const Mat4 identity;

protected:

    double x[4][4];
};

} // extern "C++"

#ifndef CRD4_NO_INLINE

#define INLINE_GLOBAL COMPILER_INLINE_GLOBAL
#define INLINE_MEMBER COMPILER_INLINE_MEMBER

#include "VEGA/Math/crd4.ipp"

#undef INLINE_GLOBAL
#undef INLINE_MEMBER

#endif

#endif // _MATH_CRD4_H_
// ===========================================================================
