// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       FCrd2
//  TYPE:       c++ header
//  PROJECT:    VEGA Math Library
//  CONTENT:    base types for 2D coordinate manipulation using floats
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rft     Robert F. Tobler
//              hl      Helwig L"offelman
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 14:06:28  rft     last modification
//  30-mar-94 10:30:36  rft     created
// ===========================================================================
//  MANUAL:
//
//  DESCRIPTION
//      FCrd2 is the base type for two dimensional coordinate manipulation.
//      It is a set of three coordinates that can be transformed with the
//      associated matrix type called FMat2.
//
//      The FCrd2 and FMat2 classes are based on floats internally.
//
//  CHECKS
//      Range checking and attempts to normalize null-vectors will only be
//      caught if WITH_FCRD2_CHECKS is defined.
//
//  ERRORHANDLERS
//      FCrd2_errors
//
//  TRACERS
//      FCrd2_constructors
//
//  SEE ALSO
//      VEGA, MathLibrary
// ===========================================================================
#ifndef _MATH_FCRD2_H_
#define _MATH_FCRD2_H_

#include <iostream.h>

#include "VEGA/Math/globals.h"
#include "VEGA/Math/math.h"

#ifdef __GNUC__
#pragma interface
#endif

extern "C++"
{

// ============================================ classes defined in this module

class FCrd2;
class FMat2;

// =========================================== errors generated by this module

#define ERROR_FCRD2_SINGULAR_MATRIX      1
        //
        //  An attempt was made to invert a singular matrix.
        //

// ================================================================ class FCrd2

class FCrd2
{
    friend class FMat2;

    friend class Viewer3D;		//  needs FAST access to members

public:

    // ---------------------------------------------------------- constructors

    FCrd2(
        );
    FCrd2(
        float x0new,
        float x1new
        );
    FCrd2(
        const FCrd2 & crd
        );
    ~FCrd2(
        );

    // ------------------------------------------------------ member operators

    float operator[](
	int index
	) const;
    float & operator[](
	int index
	);
    FCrd2 & operator+=(
	const FCrd2 & crd
	);
    FCrd2 & operator-=(
	const FCrd2 & crd
	);
    FCrd2 & operator*=(
	float scalar
	);
    FCrd2 & operator/=(
	float scalar
	);
    FCrd2 & operator*=(
	const FMat2 & mat
	);

    // ------------------------------------------------------ member functions

    float x0(
	) const;
    float x1(
	) const;
    void x0(
	float x0new
	);
    void x1(
	float x1new
	);

    void normalize(
	);

    // ------------------------------------------------------ friend operators

    friend ostream & operator<<(
	ostream & s,
	const FCrd2 & crd
	);
    friend FCrd2 operator-(
	const FCrd2 & crd
	);
    friend FCrd2 operator+(
	const FCrd2 & crd1,
	const FCrd2 & crd2
	);
    friend FCrd2 operator-(
	const FCrd2 & crd1,
	const FCrd2 & crd2
	);
    friend FCrd2 operator*(
	const FCrd2 & crd,
	float scalar
	);
    friend FCrd2 operator*(
	float scalar,
	const FCrd2 & crd
	);
    friend FCrd2 operator/(
	const FCrd2 & crd,
	float scalar
	);
    friend FCrd2 operator*(
	const FCrd2 & crd,
	const FMat2 & mat
	);

    // ------------------------------------------------------ friend functions

    friend float squaredLength(
	const FCrd2 & crd
	);
    friend float length(
	const FCrd2 & crd
	);
    friend float dotProduct(
	const FCrd2 & crd1,
	const FCrd2 & crd2
	);
    friend FCrd2 normalizedCopy(
	const FCrd2 & crd
	);

    // ----------------------------------------------- componentwise functions
    //
    //  The following functions apply each operation componentwise.
    //

    void cwMultiplyBy(
	const FCrd2 & crd
	);
    void cwDevideBy(
	const FCrd2 & crd
	);

    friend FCrd2 cwProduct(
	const FCrd2 & crd1,
	const FCrd2 & crd2
	);
    friend FCrd2 cwQuotient(
	const FCrd2 & crd1,
	const FCrd2 & crd2
	);
    friend FCrd2 cwMin(
	const FCrd2 & crd1,
	const FCrd2 & crd2
	);
    friend FCrd2 cwMax(
	const FCrd2 & crd1,
	const FCrd2 & crd2
	);

    friend FCrd2 cwReciprocalCopy(       // Componentwise multiplicative
        const FCrd2 & crd		//  inverse. Zero components are not
        );				//  inverted.

    // ---------------------------------------------- transformation functions
    //
    //  The following transformation functions and methods are provided to
    //  facilitate the easy construction of complex matrices. They are mainly
    //  used in the Vec3D module to implement the special purpose
    //  transformation classes.
    //

    FCrd2 & scale(
	const FCrd2 & s
	);
    FCrd2 & scale(
	float s0,
	float s1
	);
    FCrd2 & rot(
	float cosAlpha,
	float sinAlpha
	);
    FCrd2 & shear0(
	float x0s
	);
    FCrd2 & shear1(
	float x1s
	);
    
    friend FCrd2 prCrdScale(
	const FCrd2 & crd,
	const FCrd2 & s
	);
    friend FCrd2 prCrdScale(
	const FCrd2 & crd,
	float s0, float s1
	);
    friend FCrd2 prCrdRot(
	const FCrd2 & crd,
	float cosAlpha, float sinAlpha
	);
    friend FCrd2 prCrdShear0(
	const FCrd2 & crd,
	float x0s
	);
    friend FCrd2 prCrdShear1(
	const FCrd2 & crd,
	float x1s
	);

    // ----------------------------------------------- static member variables

    static const FCrd2 base0;
    static const FCrd2 base1;

protected:

    float x[2];
};

// =============================================================== class FMat2

class FMat2
{
    friend class FCrd2;

    friend class Viewer3D;		//  needs FAST access to members
public:

    // --------------------------------------------- constructors / destructor

    FMat2(
        );
    FMat2(
        float x00, float x01,
        float x10, float x11
        );
    FMat2(
        const FMat2 & mat
        );
    ~FMat2(
        );

    // ------------------------------------------------------ member operators

    FMat2 & operator*=(
	float scalar
	);
    FMat2 & operator+=(
	const FMat2 & mat
	);
    FMat2 & operator*=(
	const FMat2 & mat
	);

    // ------------------------------------------------------ member functions

    float operator()(
	int i, int j
	) const;	// access element

    void invert(
	);
        //
        //  ERROR_FCRD2_SINGULAR_MATRIX
        //

    void transpose(
	);

    // ------------------------------------------------------ friend operators

    friend FMat2 operator-(
	const FMat2 & mat
	);
    friend FMat2 operator+(
	const FMat2 & mat1,
	const FMat2 & mat2
	);
    friend FMat2 operator-(
	const FMat2 & mat1,
	const FMat2 & mat2
	);
    friend FMat2 operator*(
	const FMat2 & mat,
	float scalar
	);
    friend FMat2 operator*(
	float scalar,
	const FMat2 & mat
	);
    friend FCrd2 operator*(
	const FCrd2 & crd,
	const FMat2 & mat
	);
    friend FMat2 operator*(
	const FMat2 & mat1,
	const FMat2 & mat2
	);
    friend ostream & operator<<(
	ostream & s,
	const FMat2 & mat
	);

    // ------------------------------------------------------ friend functions

    friend float determinant(
	const FMat2 & mat
	);
    friend FMat2 inverseCopy(
	const FMat2 & mat,
	float det			//  det is NOT checked against zero!
	);

    friend FMat2 inverseCopy(
	const FMat2 & mat
	);
        //
        //  ERROR_FCRD2_SINGULAR_MATRIX
        //

    friend FMat2 transposedCopy(
	const FMat2 & mat
	);

    // ---------------------------------------------- transformation functions
    //
    //  The following transformation functions and methods are provided to
    //  facilitate the easy construction of complex matrices. They are mainly
    //  used in the Vec3D module to implement the special purpose
    //  transformation classes.
    //

    FMat2 & scale(
	const FCrd2 & s
	);
    FMat2 & scale(
	float s0,
	float s1
	);
    FMat2 & rot(
	float cosAlpha,
	float sinAlpha
	);
    FMat2 & shear0(
	float x0s
	);
    FMat2 & shear1(
	float x1s
	);
    
    friend FMat2 prMatScale(
	const FMat2 & mat,
	const FCrd2 & s
	);
    friend FMat2 prMatScale(
	const FMat2 & mat,
	float s0, float s1
	);
    friend FMat2 prMatRot(
	const FMat2 & mat,
	float cosAlpha, float sinAlpha
	);
    friend FMat2 prMatShear0(
	const FMat2 & mat,
	float x0s
	);
    friend FMat2 prMatShear1(
	const FMat2 & mat,
	float x1s
	);

    friend FMat2 prScaleMat(
	const FCrd2 & s,
	const FMat2 & mat
	);
    friend FMat2 prScaleMat(
	float s0, float s1,
	const FMat2 & mat
	);
    friend FMat2 prRotMat(
	float cosAlpha,
	float sinAlpha,
	const FMat2 & mat
	);
    friend FMat2 prShear0Mat(
	float x0s,
	const FMat2 & mat
	);
    friend FMat2 prShear1Mat(
	float x1s,
	const FMat2 & mat
	);

    // ----------------------------------------------- static member variables

    static const FMat2 identity;

protected:

    float x[3][3];
};

} // extern "C++"

#ifndef CRD2_NO_INLINE

#define INLINE_GLOBAL COMPILER_INLINE_GLOBAL
#define INLINE_MEMBER COMPILER_INLINE_MEMBER

#include "VEGA/Math/fcrd2.ipp"

#undef INLINE_GLOBAL
#undef INLINE_MEMBER

#endif

#endif // _MATH_FCRD2_H_
// ===========================================================================
