// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       FCrd3
//  TYPE:       c++ inline
//  PROJECT:    VEGA Math Library
//  CONTENT:    base types for 3D coordinate manipulation using floats
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rft     Robert F. Tobler
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 14:06:28  rft     last modification
//  08-Aug-93           rft     created
// ===========================================================================

extern "C++"
{

// ================================================================ class FCrd3

INLINE_MEMBER FCrd3::FCrd3()
{
    TR_FUN0(FCrd3_constructors,FCrd3);
}

INLINE_MEMBER FCrd3::FCrd3(float x0new, float x1new, float x2new)
{
    x[0] = x0new;
    x[1] = x1new;
    x[2] = x2new;

    TR_FUN3(FCrd3_constructors,FCrd3,x0new,x1new,x2new);
}

INLINE_MEMBER FCrd3::FCrd3(const FCrd3 & crd)
{
    x[0] = crd.x[0];
    x[1] = crd.x[1];
    x[2] = crd.x[2];

    TR_FUN1(FCrd3_constructors,FCrd3,crd);
}

INLINE_MEMBER FCrd3::~FCrd3()
{
    TR_MESSAGE(FCrd3_constructors,"~" << *this);
}

INLINE_MEMBER float FCrd3::operator[](int index) const
{
    CH_FCRD3_CODE(
	EH_FATAL_CONDITION(FCrd3_errors, index < 0 || index > 2);
    
        if (index < 0 || index > 2)
            EH_FATAL(FCrd3_errors,"index == " << index);
    )

    return x[index];
}

INLINE_MEMBER float & FCrd3::operator[](int index)
{
    CH_FCRD3_CODE(
	EH_FATAL_CONDITION(FCrd3_errors, index < 0 || index > 2);
    
        if (index < 0 || index > 2)
            EH_FATAL(FCrd3_errors,"index == " << index);
    )

    return x[index];
}

INLINE_MEMBER FCrd3 & FCrd3::operator+=(const FCrd3 & crd)
{
    x[0] += crd.x[0];
    x[1] += crd.x[1];
    x[2] += crd.x[2];

    return *this;
}

INLINE_MEMBER FCrd3 & FCrd3::operator-=(const FCrd3 & crd)
{
    x[0] -= crd.x[0];
    x[1] -= crd.x[1];
    x[2] -= crd.x[2];

    return *this;
}

INLINE_MEMBER FCrd3 & FCrd3::operator*=(float scalar)
{
    x[0] *= scalar;
    x[1] *= scalar;
    x[2] *= scalar;

    return *this;
}

INLINE_MEMBER FCrd3 & FCrd3::operator/=(float scalar)
{
    CH_FCRD3_CODE(EH_FATAL_CONDITION(FCrd3_errors,scalar == 0.0));

    x[0] /= scalar;
    x[1] /= scalar;
    x[2] /= scalar;

    return *this;
}

INLINE_MEMBER float FCrd3::x0() const { return x[0]; }
INLINE_MEMBER float FCrd3::x1() const { return x[1]; }
INLINE_MEMBER float FCrd3::x2() const { return x[2]; }

INLINE_MEMBER void FCrd3::x0(float x0new) { x[0] = x0new; }
INLINE_MEMBER void FCrd3::x1(float x1new) { x[1] = x1new; }
INLINE_MEMBER void FCrd3::x2(float x2new) { x[2] = x2new; }

INLINE_GLOBAL float squaredLength(const FCrd3 & crd)
{
    return (square(crd.x[0]) + square(crd.x[1]) + square(crd.x[2]));
}

INLINE_GLOBAL float length(const FCrd3 & crd)
{
    return sqrt(squaredLength(crd));
}

INLINE_MEMBER void FCrd3::normalize()
{
    float crdLength = length(*this);

    *this /= crdLength;
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const FCrd3 & crd)
{
    s << "FCrd3(" << crd.x[0] << "," << crd.x[1] << "," << crd.x[2] << ")";
    return s;
}

INLINE_GLOBAL FCrd3 operator-(const FCrd3 & crd)
{
    return FCrd3(-crd.x[0], -crd.x[1], -crd.x[2]);
}

INLINE_GLOBAL FCrd3 operator+(const FCrd3 & crd1, const FCrd3 & crd2)
{
    return FCrd3(crd1.x[0] + crd2.x[0],
                crd1.x[1] + crd2.x[1],
                crd1.x[2] + crd2.x[2]);
}

INLINE_GLOBAL FCrd3 operator-(const FCrd3 & crd1, const FCrd3 & crd2)
{
    return FCrd3(crd1.x[0] - crd2.x[0],
                crd1.x[1] - crd2.x[1],
                crd1.x[2] - crd2.x[2]);
}

INLINE_GLOBAL FCrd3 operator*(const FCrd3 & crd, float scalar)
{
    return FCrd3(crd.x[0] * scalar,
                crd.x[1] * scalar,
                crd.x[2] * scalar);
}

INLINE_GLOBAL FCrd3 operator*(float scalar, const FCrd3 & crd)
{
    return FCrd3(crd.x[0] * scalar,
                crd.x[1] * scalar,
                crd.x[2] * scalar);
}

INLINE_GLOBAL FCrd3 operator/(const FCrd3 & crd, float scalar)
{
    CH_FCRD3_CODE(EH_FATAL_CONDITION(FCrd3_errors,scalar == 0.0));

    return FCrd3(crd.x[0] / scalar,
                crd.x[1] / scalar,
                crd.x[2] / scalar);
}

INLINE_GLOBAL float dotProduct(const FCrd3 & crd1, const FCrd3 & crd2)
{
    return (crd1.x[0] * crd2.x[0] +
            crd1.x[1] * crd2.x[1] +
            crd1.x[2] * crd2.x[2]);
}

INLINE_GLOBAL FCrd3 crossProduct(const FCrd3 & crd1, const FCrd3 & crd2)
{
    return FCrd3(crd1.x[1] * crd2.x[2] - crd1.x[2] * crd2.x[1],
                crd1.x[2] * crd2.x[0] - crd1.x[0] * crd2.x[2],
                crd1.x[0] * crd2.x[1] - crd1.x[1] * crd2.x[0]);
}

INLINE_GLOBAL FCrd3 normalizedCopy(const FCrd3 & crd)
{
    float crdLength = length(crd);

    return (crd / crdLength);
}

INLINE_MEMBER void FCrd3::cwMultiplyBy(const FCrd3 & crd)
{
    x[0] *= crd.x[0];
    x[1] *= crd.x[1];
    x[2] *= crd.x[2];
}

INLINE_MEMBER void FCrd3::cwDevideBy(const FCrd3 & crd)
{
    CH_FCRD3_CODE(EH_FATAL_CONDITION(FCrd3_errors,crd[0] == 0.0));
    CH_FCRD3_CODE(EH_FATAL_CONDITION(FCrd3_errors,crd[1] == 0.0));
    CH_FCRD3_CODE(EH_FATAL_CONDITION(FCrd3_errors,crd[2] == 0.0));

    x[0] /= crd.x[0];
    x[1] /= crd.x[1];
    x[2] /= crd.x[2];
}

INLINE_GLOBAL FCrd3 cwReciprocalCopy(const FCrd3 & crd)
{
    FCrd3 newVec;
    
    if (crd.x[0] != 0.0)
        newVec.x[0] = 1.0 / crd.x[0];
    else
        newVec.x[0] = 0.0;

    if (crd.x[1] != 0.0)
        newVec.x[1] = 1.0 / crd.x[1];
    else
        newVec.x[1] = 0.0;

    if (crd.x[2] != 0.0)
        newVec.x[2] = 1.0 / crd.x[2];
    else
        newVec.x[2] = 0.0;

    return (newVec);
}

INLINE_GLOBAL FCrd3 cwProduct(const FCrd3 & crd1, const FCrd3 & crd2)
{
    return (FCrd3(crd1.x[0] * crd2.x[0],
                  crd1.x[1] * crd2.x[1],
                  crd1.x[2] * crd2.x[2]));
}

INLINE_GLOBAL FCrd3 cwQuotient(const FCrd3 & crd1, const FCrd3 & crd2)
{
    CH_FCRD3_CODE(EH_FATAL_CONDITION(FCrd3_errors,crd2[0] == 0.0));
    CH_FCRD3_CODE(EH_FATAL_CONDITION(FCrd3_errors,crd2[1] == 0.0));
    CH_FCRD3_CODE(EH_FATAL_CONDITION(FCrd3_errors,crd2[2] == 0.0));

    return FCrd3(crd1.x[0] / crd2.x[0],
                crd1.x[1] / crd2.x[1],
                crd1.x[2] / crd2.x[2]);
}

INLINE_GLOBAL FCrd3 cwMin(const FCrd3 & crd1, const FCrd3 & crd2)
{
    return FCrd3(min(crd1.x[0],crd2.x[0]),
                min(crd1.x[1],crd2.x[1]),
                min(crd1.x[2],crd2.x[2]));
}

INLINE_GLOBAL FCrd3 cwMax(const FCrd3 & crd1, const FCrd3 & crd2)
{
    return FCrd3(max(crd1.x[0],crd2.x[0]),
                max(crd1.x[1],crd2.x[1]),
                max(crd1.x[2],crd2.x[2]));
}

INLINE_MEMBER FCrd3 & FCrd3::scale(const FCrd3 & s)
{
    x[0] *= s[0];
    x[1] *= s[1];
    x[2] *= s[2];

    return *this;
}

INLINE_MEMBER FCrd3 & FCrd3::scale(float s0, float s1, float s2)
{
    x[0] *= s0;
    x[1] *= s1;
    x[2] *= s2;

    return *this;
}

INLINE_MEMBER FCrd3 & FCrd3::rot0(float cosAlpha, float sinAlpha)
{
    float t = x[1];

    x[1] = x[1] * cosAlpha - x[2] * sinAlpha;
    x[2] = x[2] * cosAlpha +    t * sinAlpha;

    return *this;
}

INLINE_MEMBER FCrd3 & FCrd3::rot1(float cosAlpha, float sinAlpha)
{
    float t = x[2];
    
    x[2] = x[2] * cosAlpha - x[0] * sinAlpha;
    x[0] = x[0] * cosAlpha +    t * sinAlpha;

    return *this;
}

INLINE_MEMBER FCrd3 & FCrd3::rot2(float cosAlpha, float sinAlpha)
{
    float t = x[0];
    
    x[0] = x[0] * cosAlpha - x[1] * sinAlpha;
    x[1] = x[1] * cosAlpha +    t * sinAlpha;

    return *this;
}

INLINE_MEMBER FCrd3 & FCrd3::shear01(float x0s, float x1s)
{
    x[0] += x[2] * x0s;
    x[1] += x[2] * x1s;

    return *this;
}

INLINE_MEMBER FCrd3 & FCrd3::shear20(float x2s, float x0s)
{
    x[0] += x[1] * x0s;
    x[2] += x[1] * x2s;

    return *this;
}

INLINE_MEMBER FCrd3 & FCrd3::shear12(float x1s, float x2s)
{
    x[1] += x[0] * x1s;
    x[2] += x[0] * x2s;

    return *this;
}

INLINE_MEMBER FCrd3 & FCrd3::operator*=(const FMat3 & mat)
{
    float t0 = x[0];
    float t1 = x[1];

    x[0] = x[0] * mat.x[0][0] + x[1] * mat.x[1][0] + x[2] * mat.x[2][0];
    x[1] = t0   * mat.x[0][1] + x[1] * mat.x[1][1] + x[2] * mat.x[2][1];
    x[2] = t0   * mat.x[0][2] + t1   * mat.x[1][2] + x[2] * mat.x[2][2];
    
    return *this;
}

INLINE_GLOBAL FCrd3 prCrdScale(const FCrd3 & crd, const FCrd3 & s)
{
    return FCrd3(crd.x[0] * s[0], crd.x[1] * s[1], crd.x[2] * s[2]);
}

INLINE_GLOBAL FCrd3 prCrdScale(const FCrd3 & crd, float s0, float s1, float s2)
{
    return FCrd3(crd.x[0] * s0, crd.x[1] * s1, crd.x[2] * s2);
}

INLINE_GLOBAL FCrd3 prCrdRot0(const FCrd3 & crd, float cosAlpha, float sinAlpha)
{
    return FCrd3(crd.x[0],
                crd.x[1] * cosAlpha - crd.x[2] * sinAlpha,
                crd.x[2] * cosAlpha + crd.x[1] * sinAlpha);
}

INLINE_GLOBAL FCrd3 prCrdRot1(const FCrd3 & crd, float cosAlpha, float sinAlpha)
{
    return FCrd3(crd.x[0] * cosAlpha + crd.x[2] * sinAlpha,
                crd.x[1],
                crd.x[2] * cosAlpha - crd.x[0] * sinAlpha);
}

INLINE_GLOBAL FCrd3 prCrdRot2(const FCrd3 & crd, float cosAlpha, float sinAlpha)
{
    return FCrd3(crd.x[0] * cosAlpha - crd.x[1] * sinAlpha,
                crd.x[1] * cosAlpha + crd.x[0] * sinAlpha,
                crd.x[2]);
}

INLINE_GLOBAL FCrd3 prCrdShear01(const FCrd3 & crd, float x0s, float x1s)
{
    return FCrd3(crd.x[0] + crd.x[2] * x0s,
                crd.x[1] + crd.x[2] * x1s,
                crd.x[2]);
}

INLINE_GLOBAL FCrd3 prCrdShear20(const FCrd3 & crd, float x2s, float x0s)
{
    return FCrd3(crd.x[0] + crd.x[1] * x0s,
                crd.x[1],
                crd.x[2] + crd.x[1] * x2s);
}

INLINE_GLOBAL FCrd3 prCrdShear12(const FCrd3 & crd, float x1s, float x2s)
{
    return FCrd3(crd.x[0],
                crd.x[1] + crd.x[0] * x1s,
                crd.x[2] + crd.x[0] * x2s);
}

INLINE_GLOBAL FCrd3 operator*(const FCrd3 & crd, const FMat3 & mat)
{
    return
        FCrd3
        (
            crd.x[0] * mat.x[0][0] +
            crd.x[1] * mat.x[1][0] +
            crd.x[2] * mat.x[2][0],

            crd.x[0] * mat.x[0][1] +
            crd.x[1] * mat.x[1][1] +
            crd.x[2] * mat.x[2][1],

            crd.x[0] * mat.x[0][2] +
            crd.x[1] * mat.x[1][2] +
            crd.x[2] * mat.x[2][2]
        );
}

// ================================================================ class FMat3

INLINE_MEMBER FMat3::FMat3()
{
    TR_FUN0(FCrd3_constructors,FMat3);
}

INLINE_MEMBER FMat3::FMat3(float x00, float x01, float x02,
                  float x10, float x11, float x12,
                  float x20, float x21, float x22)
{
    x[0][0] = x00; x[0][1] = x01; x[0][2] = x02;
    x[1][0] = x10; x[1][1] = x11; x[1][2] = x12;
    x[2][0] = x20; x[2][1] = x21; x[2][2] = x22;

    TR_FUN4(FCrd3_constructors,FMat3,x00,x01,x02,"...");
}

INLINE_MEMBER FMat3::FMat3(const FMat3 & mat)
{
    x[0][0] = mat.x[0][0];
    x[0][1] = mat.x[0][1];
    x[0][2] = mat.x[0][2];
    x[1][0] = mat.x[1][0];
    x[1][1] = mat.x[1][1];
    x[1][2] = mat.x[1][2];
    x[2][0] = mat.x[2][0];
    x[2][1] = mat.x[2][1];
    x[2][2] = mat.x[2][2];
    TR_FUN1(FCrd3_constructors,FMat3,mat);
}


INLINE_MEMBER FMat3::~FMat3()
{
    TR_MESSAGE(FCrd3_constructors,"~" << *this);
}

INLINE_MEMBER float FMat3::operator()(int i, int j) const
{
    CH_FCRD3_CODE(
        if (i < 0 || i > 2)
            EH_FATAL(FCrd3_errors,"i == " << i);
        if (j < 0 || j > 2)
            EH_FATAL(FCrd3_errors,"j == " << j);
    )
    return x[i][j];
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const FMat3 & mat)
{
    s << "FMat3("
	<< mat.x[0][0] << ',' << mat.x[0][1] << ',' << mat.x[0][2] << ','
	<< mat.x[1][0] << ',' << mat.x[1][1] << ',' << mat.x[1][2] << ','
	<< mat.x[2][0] << ',' << mat.x[2][1] << ',' << mat.x[2][2] << ')';
    return s;
}


INLINE_GLOBAL float determinant(const FMat3 & mat)
{
    return
        det3x3(
            mat.x[0][0], mat.x[0][1], mat.x[0][2],
            mat.x[1][0], mat.x[1][1], mat.x[1][2],
            mat.x[2][0], mat.x[2][1], mat.x[2][2]
        );
}

INLINE_GLOBAL FMat3 inverseCopy(const FMat3 & mat, float det)
{
    det = 1.0 / det;

    return
        FMat3
        (
            (mat.x[1][1] * mat.x[2][2] -
             mat.x[2][1] * mat.x[1][2]) * det,
            (mat.x[2][1] * mat.x[0][2] -
             mat.x[0][1] * mat.x[2][2]) * det,
            (mat.x[0][1] * mat.x[1][2] -
             mat.x[1][1] * mat.x[0][2]) * det,

            (mat.x[2][0] * mat.x[1][2] -
             mat.x[1][0] * mat.x[2][2]) * det,
            (mat.x[0][0] * mat.x[2][2] -
             mat.x[2][0] * mat.x[0][2]) * det,
            (mat.x[1][0] * mat.x[0][2] -
             mat.x[0][0] * mat.x[1][2]) * det,

            (mat.x[1][0] * mat.x[2][1] -
             mat.x[2][0] * mat.x[1][1]) * det,
            (mat.x[2][0] * mat.x[0][1] -
             mat.x[0][0] * mat.x[2][1]) * det,
            (mat.x[0][0] * mat.x[1][1] -
             mat.x[1][0] * mat.x[0][1]) * det
        );
}

INLINE_GLOBAL FMat3 inverseCopy(const FMat3 & mat)
{
    float det = determinant(mat);

    if (det == 0.0)
    {
        EH_ERROR(FCrd3_errors, ERROR_FCRD3_SINGULAR_MATRIX, "singular matrix");
        return mat;
    }

    return inverseCopy(mat,det);
}


INLINE_MEMBER void FMat3::invert()
{
    *this = inverseCopy(*this);
}

INLINE_GLOBAL FMat3 transposedCopy(const FMat3 & mat)
{
    return
        FMat3(
            mat.x[0][0], mat.x[1][0], mat.x[2][0],
            mat.x[0][1], mat.x[1][1], mat.x[2][1],
            mat.x[0][2], mat.x[1][2], mat.x[2][2]
        );
}

INLINE_MEMBER void FMat3::transpose()
{
    *this = transposedCopy(*this);
}

INLINE_MEMBER FMat3 & FMat3::operator*=(float s)
{
    x[0][0] *= s; x[0][1] *= s; x[0][2] *= s;
    x[1][0] *= s; x[1][1] *= s; x[1][2] *= s;
    x[2][0] *= s; x[2][1] *= s; x[2][2] *= s;
    return *this;
}

INLINE_MEMBER FMat3 & FMat3::scale(const FCrd3 & s)
{
    x[0][0] *= s[0]; x[0][1] *= s[1]; x[0][2] *= s[2];
    x[1][0] *= s[0]; x[1][1] *= s[1]; x[1][2] *= s[2];
    x[2][0] *= s[0]; x[2][1] *= s[1]; x[2][2] *= s[2];
    return *this;
}

INLINE_MEMBER FMat3 & FMat3::scale(float s0, float s1, float s2)
{
    x[0][0] *= s0; x[0][1] *= s1; x[0][2] *= s2;
    x[1][0] *= s0; x[1][1] *= s1; x[1][2] *= s2;
    x[2][0] *= s0; x[2][1] *= s1; x[2][2] *= s2;
    return *this;
}

INLINE_MEMBER FMat3 & FMat3::rot0(float cosAlpha, float sinAlpha)
{
    float t;

    t = x[0][1];
    x[0][1] = t * cosAlpha - x[0][2] * sinAlpha;
    x[0][2] = t * sinAlpha + x[0][2] * cosAlpha;
    t = x[1][1];
    x[1][1] = t * cosAlpha - x[1][2] * sinAlpha;
    x[1][2] = t * sinAlpha + x[1][2] * cosAlpha;
    t = x[2][1];
    x[2][1] = t * cosAlpha - x[2][2] * sinAlpha;
    x[2][2] = t * sinAlpha + x[2][2] * cosAlpha;
    return *this;
}

INLINE_MEMBER FMat3 & FMat3::rot1(float cosAlpha, float sinAlpha)
{
    float t;

    t = x[0][0];
    x[0][0] =  t * cosAlpha + x[0][2] * sinAlpha;
    x[0][2] = -t * sinAlpha + x[0][2] * cosAlpha;
    t = x[1][0];
    x[1][0] =  t * cosAlpha + x[1][2] * sinAlpha;
    x[1][2] = -t * sinAlpha + x[1][2] * cosAlpha;
    t = x[2][0];
    x[2][0] =  t * cosAlpha + x[2][2] * sinAlpha;
    x[2][2] = -t * sinAlpha + x[2][2] * cosAlpha;
    return *this;
}

INLINE_MEMBER FMat3 & FMat3::rot2(float cosAlpha, float sinAlpha)
{
    float t;

    t = x[0][0];
    x[0][0] = t * cosAlpha - x[0][1] * sinAlpha;
    x[0][1] = t * sinAlpha + x[0][1] * cosAlpha;
    t = x[1][0];
    x[1][0] = t * cosAlpha - x[1][1] * sinAlpha;
    x[1][1] = t * sinAlpha + x[1][1] * cosAlpha;
    t = x[2][0];
    x[2][0] = t * cosAlpha - x[2][1] * sinAlpha;
    x[2][1] = t * sinAlpha + x[2][1] * cosAlpha;
    return *this;
}

INLINE_MEMBER FMat3 & FMat3::shear01(float x0s, float x1s)
{
    x[0][0] += x[0][2] * x0s;
    x[0][1] += x[0][2] * x1s;
    x[1][0] += x[1][2] * x0s;
    x[1][1] += x[1][2] * x1s;
    x[2][0] += x[2][2] * x0s;
    x[2][1] += x[2][2] * x1s;
    return *this;
}

INLINE_MEMBER FMat3 & FMat3::shear20(float x2s, float x0s)
{
    x[0][0] += x[0][1] * x0s;
    x[0][2] += x[0][1] * x2s;
    x[1][0] += x[1][1] * x0s;
    x[1][2] += x[1][1] * x2s;
    x[2][0] += x[2][1] * x0s;
    x[2][2] += x[2][1] * x2s;
    return *this;
}

INLINE_MEMBER FMat3 & FMat3::shear12(float x1s, float x2s)
{
    x[0][1] += x[0][0] * x1s;
    x[0][2] += x[0][0] * x2s;
    x[1][1] += x[1][0] * x1s;
    x[1][2] += x[1][0] * x2s;
    x[2][1] += x[2][0] * x1s;
    x[2][2] += x[2][0] * x2s;
    return *this;
}

INLINE_MEMBER FMat3 & FMat3::operator+=(const FMat3 & mat)
{
    x[0][0] += mat.x[0][0];
    x[0][1] += mat.x[0][1];
    x[0][2] += mat.x[0][2];
    x[1][0] += mat.x[1][0];
    x[1][1] += mat.x[1][1];
    x[1][2] += mat.x[1][2];
    x[2][0] += mat.x[2][0];
    x[2][1] += mat.x[2][1];
    x[2][2] += mat.x[2][2];

    return *this;
}

INLINE_GLOBAL FMat3 operator*(const FMat3 & mat1, const FMat3 & mat2)
{
    return
        FMat3
        (
            mat1.x[0][0] * mat2.x[0][0] +
            mat1.x[0][1] * mat2.x[1][0] +
            mat1.x[0][2] * mat2.x[2][0],
            mat1.x[0][0] * mat2.x[0][1] +
            mat1.x[0][1] * mat2.x[1][1] +
            mat1.x[0][2] * mat2.x[2][1],
            mat1.x[0][0] * mat2.x[0][2] +
            mat1.x[0][1] * mat2.x[1][2] +
            mat1.x[0][2] * mat2.x[2][2],

            mat1.x[1][0] * mat2.x[0][0] +
            mat1.x[1][1] * mat2.x[1][0] +
            mat1.x[1][2] * mat2.x[2][0],
            mat1.x[1][0] * mat2.x[0][1] +
            mat1.x[1][1] * mat2.x[1][1] +
            mat1.x[1][2] * mat2.x[2][1],
            mat1.x[1][0] * mat2.x[0][2] +
            mat1.x[1][1] * mat2.x[1][2] +
            mat1.x[1][2] * mat2.x[2][2],

            mat1.x[2][0] * mat2.x[0][0] +
            mat1.x[2][1] * mat2.x[1][0] +
            mat1.x[2][2] * mat2.x[2][0],
            mat1.x[2][0] * mat2.x[0][1] +
            mat1.x[2][1] * mat2.x[1][1] +
            mat1.x[2][2] * mat2.x[2][1],
            mat1.x[2][0] * mat2.x[0][2] +
            mat1.x[2][1] * mat2.x[1][2] +
            mat1.x[2][2] * mat2.x[2][2]
        );
}

INLINE_MEMBER FMat3 & FMat3::operator*=(const FMat3 & mat)
{
    *this = *this * mat;

    return *this;
}

INLINE_GLOBAL FMat3 operator*(const FMat3 & mat, float s)
{
    return
        FMat3
        (
            mat.x[0][0] * s, mat.x[0][1] * s, mat.x[0][2] * s,
            mat.x[1][0] * s, mat.x[1][1] * s, mat.x[1][2] * s,
            mat.x[2][0] * s, mat.x[2][1] * s, mat.x[2][2] * s
        );
}

INLINE_GLOBAL FMat3 prMatScale(const FMat3 & mat, const FCrd3 & s)
{
    return
        FMat3
        (
            mat.x[0][0] * s[0], mat.x[0][1] * s[1], mat.x[0][2] * s[2],
            mat.x[1][0] * s[0], mat.x[1][1] * s[1], mat.x[1][2] * s[2],
            mat.x[2][0] * s[0], mat.x[2][1] * s[1], mat.x[2][2] * s[2]
        );
}

INLINE_GLOBAL FMat3 prMatScale(const FMat3 & mat, float s0, float s1, float s2)
{
    return
        FMat3
        (
            mat.x[0][0] * s0, mat.x[0][1] * s1, mat.x[0][2] * s2,
            mat.x[1][0] * s0, mat.x[1][1] * s1, mat.x[1][2] * s2,
            mat.x[2][0] * s0, mat.x[2][1] * s1, mat.x[2][2] * s2
        );
}

INLINE_GLOBAL FMat3 prMatRot0(const FMat3 & mat, float cosAlpha, float sinAlpha)
{
    return
        FMat3
        (
            mat.x[0][0],
            mat.x[0][1] * cosAlpha - mat.x[0][2] * sinAlpha,
            mat.x[0][2] * cosAlpha + mat.x[0][1] * sinAlpha,

            mat.x[1][0],
            mat.x[1][1] * cosAlpha - mat.x[1][2] * sinAlpha,
            mat.x[1][2] * cosAlpha + mat.x[1][1] * sinAlpha,

            mat.x[2][0],
            mat.x[2][1] * cosAlpha - mat.x[2][2] * sinAlpha,
            mat.x[2][2] * cosAlpha + mat.x[2][1] * sinAlpha
        );
}

INLINE_GLOBAL FMat3 prMatRot1(const FMat3 & mat, float cosAlpha, float sinAlpha)
{
    return
        FMat3
        (
            mat.x[0][0] * cosAlpha + mat.x[0][2] * sinAlpha,
            mat.x[0][1],
            mat.x[0][2] * cosAlpha - mat.x[0][0] * sinAlpha,

            mat.x[1][0] * cosAlpha + mat.x[1][2] * sinAlpha,
            mat.x[1][1],
            mat.x[1][2] * cosAlpha - mat.x[1][0] * sinAlpha,

            mat.x[2][0] * cosAlpha + mat.x[2][2] * sinAlpha,
            mat.x[2][1],
            mat.x[2][2] * cosAlpha - mat.x[2][0] * sinAlpha
        );
}

INLINE_GLOBAL FMat3 prMatRot2(const FMat3 & mat, float cosAlpha, float sinAlpha)
{
    return
        FMat3
        (
            mat.x[0][0] * cosAlpha - mat.x[0][1] * sinAlpha,
            mat.x[0][1] * cosAlpha + mat.x[0][0] * sinAlpha,
            mat.x[0][2],

            mat.x[1][0] * cosAlpha - mat.x[1][1] * sinAlpha,
            mat.x[1][1] * cosAlpha + mat.x[1][0] * sinAlpha,
            mat.x[1][2],

            mat.x[2][0] * cosAlpha - mat.x[2][1] * sinAlpha,
            mat.x[2][1] * cosAlpha + mat.x[2][0] * sinAlpha,
            mat.x[2][2]
        );
}

INLINE_GLOBAL FMat3 prMatShear01(const FMat3 & mat, float x0s, float x1s)
{
    return
        FMat3
        (
            mat.x[0][0] + mat.x[0][2] * x0s,
            mat.x[0][1] + mat.x[0][2] * x1s,
            mat.x[0][2],

            mat.x[1][0] + mat.x[1][2] * x0s,
            mat.x[1][1] + mat.x[1][2] * x1s,
            mat.x[1][2],

            mat.x[2][0] + mat.x[2][2] * x0s,
            mat.x[2][1] + mat.x[2][2] * x1s,
            mat.x[2][2]
        );
}

INLINE_GLOBAL FMat3 prMatShear20(const FMat3 & mat, float x2s, float x0s)
{
    return
        FMat3
        (
            mat.x[0][0] + mat.x[0][1] * x0s,
            mat.x[0][1],
            mat.x[0][2] + mat.x[0][1] * x2s,

            mat.x[1][0] + mat.x[1][1] * x0s,
            mat.x[1][1],
            mat.x[1][2] + mat.x[1][1] * x2s,

            mat.x[2][0] + mat.x[2][1] * x0s,
            mat.x[2][1],
            mat.x[2][2] + mat.x[2][1] * x2s
        );
}

INLINE_GLOBAL FMat3 prMatShear12(const FMat3 & mat, float x1s, float x2s)
{
    return
        FMat3
        (
            mat.x[0][0],
            mat.x[0][1] + mat.x[0][0] * x1s,
            mat.x[0][2] + mat.x[0][0] * x2s,

            mat.x[1][0],
            mat.x[1][1] + mat.x[1][0] * x1s,
            mat.x[1][2] + mat.x[1][0] * x2s,

            mat.x[2][0],
            mat.x[2][1] + mat.x[2][0] * x1s,
            mat.x[2][2] + mat.x[2][0] * x2s
        );
}

INLINE_GLOBAL FMat3 operator*(float s, const FMat3 & mat)
{
    return
        FMat3
        (
            s * mat.x[0][0], s * mat.x[0][1], s * mat.x[0][2],
            s * mat.x[1][0], s * mat.x[1][1], s * mat.x[1][2],
            s * mat.x[2][0], s * mat.x[2][1], s * mat.x[2][2]
        );
}

INLINE_GLOBAL FMat3 prScaleMat(const FCrd3 & s, const FMat3 & mat)
{
    return
        FMat3
        (
            s[0] * mat.x[0][0], s[0] * mat.x[0][1], s[0] * mat.x[0][2],
            s[1] * mat.x[1][0], s[1] * mat.x[1][1], s[1] * mat.x[1][2],
            s[2] * mat.x[2][0], s[2] * mat.x[2][1], s[2] * mat.x[2][2]
        );
}

INLINE_GLOBAL FMat3 prScaleMat(float s0, float s1, float s2, const FMat3 & mat)
{
    return
        FMat3
        (
            s0 * mat.x[0][0], s0 * mat.x[0][1], s0 * mat.x[0][2],
            s1 * mat.x[1][0], s1 * mat.x[1][1], s1 * mat.x[1][2],
            s2 * mat.x[2][0], s2 * mat.x[2][1], s2 * mat.x[2][2]
        );
}

INLINE_GLOBAL FMat3 prRot0Mat(float cosAlpha, float sinAlpha, const FMat3 & mat)
{
    return
        FMat3
        (
            mat.x[0][0],
            mat.x[0][1],
            mat.x[0][2],

            cosAlpha * mat.x[1][0] + sinAlpha * mat.x[2][0],
            cosAlpha * mat.x[1][1] + sinAlpha * mat.x[2][1],
            cosAlpha * mat.x[1][2] + sinAlpha * mat.x[2][2],

            cosAlpha * mat.x[2][0] - sinAlpha * mat.x[1][0],
            cosAlpha * mat.x[2][1] - sinAlpha * mat.x[1][1],
            cosAlpha * mat.x[2][2] - sinAlpha * mat.x[1][2]
        );
}

INLINE_GLOBAL FMat3 prRot1Mat(float cosAlpha, float sinAlpha, const FMat3 & mat)
{
    return
        FMat3
        (
            cosAlpha * mat.x[0][0] - sinAlpha * mat.x[2][0],
            cosAlpha * mat.x[0][1] - sinAlpha * mat.x[2][1],
            cosAlpha * mat.x[0][2] - sinAlpha * mat.x[2][2],

            mat.x[1][0],
            mat.x[1][1],
            mat.x[1][2],

            cosAlpha * mat.x[2][0] + sinAlpha * mat.x[0][0],
            cosAlpha * mat.x[2][1] + sinAlpha * mat.x[0][1],
            cosAlpha * mat.x[2][2] + sinAlpha * mat.x[0][2]
        );
}

INLINE_GLOBAL FMat3 prRot2Mat(float cosAlpha, float sinAlpha, const FMat3 & mat)
{
    return
        FMat3
        (
            cosAlpha * mat.x[0][0] + sinAlpha * mat.x[1][0],
            cosAlpha * mat.x[0][1] + sinAlpha * mat.x[1][1],
            cosAlpha * mat.x[0][2] + sinAlpha * mat.x[1][2],

            cosAlpha * mat.x[1][0] - sinAlpha * mat.x[0][0],
            cosAlpha * mat.x[1][1] - sinAlpha * mat.x[0][1],
            cosAlpha * mat.x[1][2] - sinAlpha * mat.x[0][2],

            mat.x[2][0],
            mat.x[2][1],
            mat.x[2][2]
        );
}

INLINE_GLOBAL FMat3 prShear01Mat(float x0s, float x1s, const FMat3 & mat)
{
    return
        FMat3
        (
	    mat.x[0][0],
	    mat.x[0][1],
	    mat.x[0][2],
	    
	    mat.x[1][0],
	    mat.x[1][1],
	    mat.x[1][2],

            x0s * mat.x[0][0] + x1s * mat.x[1][0] + mat.x[2][0],
            x0s * mat.x[0][1] + x1s * mat.x[1][1] + mat.x[2][1],
            x0s * mat.x[0][2] + x1s * mat.x[1][2] + mat.x[2][2]
        );
}

INLINE_GLOBAL FMat3 prShear20Mat(float x2s, float x0s, const FMat3 & mat)
{
    return
        FMat3
        (
	    mat.x[0][0],
	    mat.x[0][1],
	    mat.x[0][2],
	    
            x0s * mat.x[0][0] + mat.x[1][0] + x2s * mat.x[2][0],
            x0s * mat.x[0][1] + mat.x[1][1] + x2s * mat.x[2][1],
            x0s * mat.x[0][2] + mat.x[1][2] + x2s * mat.x[2][2],

	    mat.x[2][0],
	    mat.x[2][1],
	    mat.x[2][2]
        );
}

INLINE_GLOBAL FMat3 prShear12Mat(float x1s, float x2s, const FMat3 & mat)
{
    return
        FMat3
        (
            mat.x[0][0] + x1s * mat.x[1][0] + x2s * mat.x[2][0],
            mat.x[0][1] + x1s * mat.x[1][1] + x2s * mat.x[2][1],
            mat.x[0][2] + x1s * mat.x[1][2] + x2s * mat.x[2][2],

	    mat.x[1][0],
	    mat.x[1][1],
	    mat.x[1][2],
	    
	    mat.x[2][0],
	    mat.x[2][1],
	    mat.x[2][2]
        );
}

INLINE_GLOBAL FMat3 operator-(const FMat3 & mat)
{
    return
        FMat3
        (
	    - mat.x[0][0], - mat.x[0][1], - mat.x[0][2],
	    - mat.x[1][0], - mat.x[1][1], - mat.x[1][2],
	    - mat.x[2][0], - mat.x[2][1], - mat.x[2][2]
	);
}

INLINE_GLOBAL FMat3 operator+(const FMat3 & mat1, const FMat3 & mat2)
{
    return
        FMat3
        (
	    mat1.x[0][0] + mat2.x[0][0],
	    mat1.x[0][1] + mat2.x[0][1],
	    mat1.x[0][2] + mat2.x[0][2],

	    mat1.x[1][0] + mat2.x[1][0],
	    mat1.x[1][1] + mat2.x[1][1],
	    mat1.x[1][2] + mat2.x[1][2],

	    mat1.x[2][0] + mat2.x[2][0],
	    mat1.x[2][1] + mat2.x[2][1],
	    mat1.x[2][2] + mat2.x[2][2]
	);
}

INLINE_GLOBAL FMat3 operator-(const FMat3 & mat1, const FMat3 & mat2)
{
    return
        FMat3
        (
	    mat1.x[0][0] - mat2.x[0][0],
	    mat1.x[0][1] - mat2.x[0][1],
	    mat1.x[0][2] - mat2.x[0][2],

	    mat1.x[1][0] - mat2.x[1][0],
	    mat1.x[1][1] - mat2.x[1][1],
	    mat1.x[1][2] - mat2.x[1][2],

	    mat1.x[2][0] - mat2.x[2][0],
	    mat1.x[2][1] - mat2.x[2][1],
	    mat1.x[2][2] - mat2.x[2][2]
	);
}

} // extern "C++"

// ===========================================================================
