// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       FCrd4
//  TYPE:       c++ header
//  PROJECT:    VEGA Math Library
//  CONTENT:    base types for 4D coordinate manipulation using floats
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rft     Robert F. Tobler
//              hl      Helwig L"offelman
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 14:06:28  rft     last modification
//  25-mar-94 23:09:46  rft     created
// ===========================================================================
//  MANUAL:
//
//  DESCRIPTION
//      FCrd4 is the base type for four dimensional coordinate manipulation.
//      It is a set of four coordinates that can be transformed with the
//      associated matrix type called FMat4.
//
//      The FCrd4 and FMat4 classes are based on floats internally.
//
//  CHECKS
//      Range checking and attempts to normalize null-vectors will only be
//      caught if WITH_FCRD4_CHECKS is defined.
//
//  ERRORHANDLERS
//      FCrd4_errors
//
//  TRACERS
//      FCrd4_constructors
//
//  SEE ALSO
//      VEGA, MathLibrary
// ===========================================================================
#ifndef _MATH_FCRD4_H_
#define _MATH_FCRD4_H_

#include <iostream.h>

#include "VEGA/Math/globals.h"
#include "VEGA/Math/math.h"

#ifdef __GNUC__
#pragma interface
#endif

extern "C++"
{

// ============================================ classes defined in this module

class FCrd4;
class FMat4;

// =========================================== errors generated by this module

#define ERROR_FCRD4_SINGULAR_MATRIX      1
        //
        //  An attempt was made to invert a singular matrix.
        //

// ================================================================ class FCrd4

class FCrd4
{
    friend class FMat4;

    friend class Viewer3D;		//  needs FAST access to members

public:

    // ---------------------------------------------------------- constructors

    FCrd4(
        );
    FCrd4(
        float x0new,
        float x1new,
        float x2new,
        float x3new
        );
    FCrd4(
        const FCrd4 & crd
        );
    ~FCrd4(
        );

    // ------------------------------------------------------ member operators

    float operator[](
	int index
	) const;
    float & operator[](
	int index
	);
    FCrd4 & operator+=(
	const FCrd4 & crd
	);
    FCrd4 & operator-=(
	const FCrd4 & crd
	);
    FCrd4 & operator*=(
	float scalar
	);
    FCrd4 & operator/=(
	float scalar
	);
    FCrd4 & operator*=(
	const FMat4 & mat
	);

    // ------------------------------------------------------ member functions

    float x0(
	) const;
    float x1(
	) const;
    float x2(
	) const;
    float x3(
	) const;
    void x0(
	float x0new
	);
    void x1(
	float x1new
	);
    void x2(
	float x2new
	);
    void x3(
	float x3new
	);

    void normalize(
	);

    // ------------------------------------------------------ friend operators

    friend ostream & operator<<(
	ostream & s,
	const FCrd4 & crd
	);
    friend FCrd4 operator-(
	const FCrd4 & crd
	);
    friend FCrd4 operator+(
	const FCrd4 & crd1,
	const FCrd4 & crd2
	);
    friend FCrd4 operator-(
	const FCrd4 & crd1,
	const FCrd4 & crd2
	);
    friend FCrd4 operator*(
	const FCrd4 & crd,
	float scalar
	);
    friend FCrd4 operator*(
	float scalar,
	const FCrd4 & crd
	);
    friend FCrd4 operator/(
	const FCrd4 & crd,
	float scalar
	);
    friend FCrd4 operator*(
	const FCrd4 & crd,
	const FMat4 & mat
	);

    // ------------------------------------------------------ friend functions

    friend float squaredLength(
	const FCrd4 & crd
	);
    friend float length(
	const FCrd4 & crd
	);
    friend float dotProduct(
	const FCrd4 & crd1,
	const FCrd4 & crd2
	);
    friend FCrd4 normalizedCopy(
	const FCrd4 & crd
	);

    // ----------------------------------------------- componentwise functions
    //
    //  The following functions apply each operation componentwise.
    //

    void cwMultiplyBy(
	const FCrd4 & crd
	);
    void cwDevideBy(
	const FCrd4 & crd
	);

    friend FCrd4 cwProduct(
	const FCrd4 & crd1,
	const FCrd4 & crd2
	);
    friend FCrd4 cwQuotient(
	const FCrd4 & crd1,
	const FCrd4 & crd2
	);
    friend FCrd4 cwMin(
	const FCrd4 & crd1,
	const FCrd4 & crd2
	);
    friend FCrd4 cwMax(
	const FCrd4 & crd1,
	const FCrd4 & crd2
	);

    friend FCrd4 cwReciprocalCopy(       // Componentwise multiplicative
        const FCrd4 & crd		//  inverse. Zero components are not
        );				//  inverted.

    // ---------------------------------------------- transformation functions
    //
    //  The following transformation functions and methods are provided to
    //  facilitate the easy construction of complex matrices. They are mainly
    //  used in the Vec3D module to implement the special purpose
    //  transformation classes.
    //

    FCrd4 & scale(
	const FCrd4 & s
	);
    FCrd4 & scale(
	float s0,
	float s1,
	float s2,
	float s3
	);

    friend FCrd4 prCrdScale(
	const FCrd4 & crd,
	const FCrd4 & s
	);
    friend FCrd4 prCrdScale(
        const FCrd4 & crd,
        float s0, float s1, float s2, float s3
        );

    // ----------------------------------------------- static member variables

    static const FCrd4 base0;
    static const FCrd4 base1;
    static const FCrd4 base2;
    static const FCrd4 base3;

protected:

    float x[4];
};

// =============================================================== class FMat4

class FMat4
{
    friend class FCrd4;

    friend class Viewer3D;		//  needs FAST access to members

public:

    // --------------------------------------------- constructors / destructor

    FMat4(
        );
    FMat4(
        float x00, float x01, float x02, float x03,
        float x10, float x11, float x12, float x13,
        float x20, float x21, float x22, float x23,
        float x30, float x31, float x32, float x33
        );
    FMat4(
        const FMat4 & mat
        );
    ~FMat4(
        );

    // ------------------------------------------------------ member operators

    FMat4 & operator*=(
	float scalar
	);
    FMat4 & operator+=(
	const FMat4 & mat
	);
    FMat4 & operator*=(
	const FMat4 & mat
	);

    // ------------------------------------------------------ member functions

    float operator()(			// Access element.
	int i, int j
	) const;

    void invert(
	);
        //
        //  ERROR_FCRD4_SINGULAR_MATRIX
        //

    void transpose(
	);

    // ------------------------------------------------------ friend operators

    friend FMat4 operator-(
	const FMat4 & mat
	);
    friend FMat4 operator+(
	const FMat4 & mat1,
	const FMat4 & mat2
	);
    friend FMat4 operator-(
	const FMat4 & mat1,
	const FMat4 & mat2
	);
    friend FMat4 operator*(
	const FMat4 & mat,
	float scalar
	);
    friend FMat4 operator*(
	float scalar,
	const FMat4 & mat
	);
    friend FCrd4 operator*(
	const FCrd4 & crd,
	const FMat4 & mat
	);
    friend FMat4 operator*(
	const FMat4 & mat1,
	const FMat4 & mat2
	);
    friend ostream & operator<<(
	ostream & s,
	const FMat4 & mat
	);

    // ------------------------------------------------------ friend functions

    friend float determinant(
	const FMat4 & mat
	);

    friend FMat4 adjointCopy(
	const FMat4 & m
	);
        //
        //  Calculate the adjoint of a 4x4 matrix:
        //
        //  Let  a   denote the minor determinant of matrix A obtained
        //        ij
        //
        //  by deleting the ith row and jth column from A.
        //
        //                 i+j
        //  Let  b   = (-1)    a
        //        ij            ji
        //
        //  The matrix B = (b  ) is the adjoint of A
        //                   ij
        //

    friend FMat4 inverseCopy(
	const FMat4 & mat,
	float det		        //  det is NOT checked against zero!
	);

    friend FMat4 inverseCopy(
	const FMat4 & mat
	);
        //
        //  ERROR_FCRD4_SINGULAR_MATRIX
        //

    friend FMat4 transposedCopy(
	const FMat4 & mat
	);

    // ---------------------------------------------- transformation functions
    //
    //  The following transformation functions and methods are provided to
    //  facilitate the easy construction of complex matrices. They are mainly
    //  used in the Vec3D module to implement the special purpose
    //  transformation classes.
    //

    FMat4 & scale(
	const FCrd4 & s
	);
    FMat4 & scale(
	float s0,
	float s1,
	float s2,
	float s3
	);
    
    friend FMat4 prMatScale(
	const FMat4 & mat,
	const FCrd4 & s
	);
    friend FMat4 prMatScale(
        const FMat4 & mat,
        float s0, float s1, float s2, float s3
        );

    friend FMat4 prScaleMat(
	const FCrd4 & s,
	const FMat4 & mat
	);
    friend FMat4 prScaleMat(
        float s0, float s1, float s2, float s3,
        const FMat4 & mat
        );

    // ----------------------------------------------- static member variables

    static const FMat4 identity;

protected:

    float x[4][4];
};

} // extern "C++"

#ifndef CRD4_NO_INLINE

#define INLINE_GLOBAL COMPILER_INLINE_GLOBAL
#define INLINE_MEMBER COMPILER_INLINE_MEMBER

#include "VEGA/Math/fcrd4.ipp"

#undef INLINE_GLOBAL
#undef INLINE_MEMBER

#endif

#endif // _MATH_FCRD4_H_
// ===========================================================================
