// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       FCrd4
//  TYPE:       c++ inline
//  PROJECT:    VEGA Math Library
//  CONTENT:    base types for 4D coordinate manipulation using floats
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rft     Robert F. Tobler
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 14:06:28  rft     last modification
//  25-mar-94 23:09:46  rft     created
// ===========================================================================

extern "C++"
{

// ================================================================ class FCrd4

INLINE_MEMBER FCrd4::FCrd4()
{
    TR_FUN0(FCrd4_constructors,FCrd4);
}

INLINE_MEMBER FCrd4::FCrd4(
	float x0new, float x1new, float x2new, float x3new
	)
{
    x[0] = x0new;
    x[1] = x1new;
    x[2] = x2new;
    x[3] = x3new;

    TR_FUN4(FCrd4_constructors,FCrd4,x0new,x1new,x2new,x3new);
}

INLINE_MEMBER FCrd4::FCrd4(const FCrd4 & crd)
{
    x[0] = crd.x[0];
    x[1] = crd.x[1];
    x[2] = crd.x[2];
    x[3] = crd.x[3];

    TR_FUN1(FCrd4_constructors,FCrd4,crd);
}

INLINE_MEMBER FCrd4::~FCrd4()
{
    TR_MESSAGE(FCrd4_constructors,"~" << *this);
}

INLINE_MEMBER float FCrd4::operator[](int index) const
{
    CH_FCRD4_CODE(
        if (index < 0 || index > 3)
            EH_FATAL(FCrd4_errors,"index == " << index);
    )

    return x[index];
}

INLINE_MEMBER float & FCrd4::operator[](int index)
{
    CH_FCRD4_CODE(
        if (index < 0 || index > 3)
            EH_FATAL(FCrd4_errors,"index == " << index);
    )

    return x[index];
}

INLINE_MEMBER FCrd4 & FCrd4::operator+=(const FCrd4 & crd)
{
    x[0] += crd.x[0];
    x[1] += crd.x[1];
    x[2] += crd.x[2];
    x[3] += crd.x[3];

    return *this;
}

INLINE_MEMBER FCrd4 & FCrd4::operator-=(const FCrd4 & crd)
{
    x[0] -= crd.x[0];
    x[1] -= crd.x[1];
    x[2] -= crd.x[2];
    x[3] -= crd.x[3];

    return *this;
}

INLINE_MEMBER FCrd4 & FCrd4::operator*=(float scalar)
{
    x[0] *= scalar;
    x[1] *= scalar;
    x[2] *= scalar;
    x[3] *= scalar;

    return *this;
}

INLINE_MEMBER FCrd4 & FCrd4::operator/=(float scalar)
{
    CH_FCRD4_CODE(EH_FATAL_CONDITION(FCrd4_errors,scalar == 0.0));

    x[0] /= scalar;
    x[1] /= scalar;
    x[2] /= scalar;
    x[3] /= scalar;

    return *this;
}

INLINE_MEMBER float FCrd4::x0() const { return x[0]; }
INLINE_MEMBER float FCrd4::x1() const { return x[1]; }
INLINE_MEMBER float FCrd4::x2() const { return x[2]; }
INLINE_MEMBER float FCrd4::x3() const { return x[3]; }

INLINE_MEMBER void FCrd4::x0(float x0new) { x[0] = x0new; }
INLINE_MEMBER void FCrd4::x1(float x1new) { x[1] = x1new; }
INLINE_MEMBER void FCrd4::x2(float x2new) { x[2] = x2new; }
INLINE_MEMBER void FCrd4::x3(float x3new) { x[3] = x3new; }

INLINE_GLOBAL float squaredLength(const FCrd4 & crd)
{
    return (square(crd.x[0]) + square(crd.x[1]) +
            square(crd.x[2]) + square(crd.x[3]));
}

INLINE_GLOBAL float length(const FCrd4 & crd)
{
    return sqrt(squaredLength(crd));
}

INLINE_MEMBER void FCrd4::normalize()
{
    float crdLength = length(*this);

    *this /= crdLength;
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const FCrd4 & crd)
{
    s << "FCrd4(" << crd.x[0] << "," << crd.x[1] << "," << crd.x[2]
      << "," << crd.x[3] << ")";
    return s;
}

INLINE_GLOBAL FCrd4 operator-(const FCrd4 & crd)
{
    return FCrd4(-crd.x[0], -crd.x[1], -crd.x[2], -crd.x[3]);
}

INLINE_GLOBAL FCrd4 operator+(const FCrd4 & crd1, const FCrd4 & crd2)
{
    return FCrd4(crd1.x[0] + crd2.x[0],
                crd1.x[1] + crd2.x[1],
                crd1.x[2] + crd2.x[2],
                crd1.x[3] + crd2.x[3]);
}

INLINE_GLOBAL FCrd4 operator-(const FCrd4 & crd1, const FCrd4 & crd2)
{
    return FCrd4(crd1.x[0] - crd2.x[0],
                crd1.x[1] - crd2.x[1],
                crd1.x[2] - crd2.x[2],
                crd1.x[3] - crd2.x[3]);
}

INLINE_GLOBAL FCrd4 operator*(const FCrd4 & crd, float scalar)
{
    return FCrd4(crd.x[0] * scalar,
                crd.x[1] * scalar,
                crd.x[2] * scalar,
                crd.x[3] * scalar);
}

INLINE_GLOBAL FCrd4 operator*(float scalar, const FCrd4 & crd)
{
    return FCrd4(crd.x[0] * scalar,
                crd.x[1] * scalar,
                crd.x[2] * scalar,
                crd.x[3] * scalar);
}

INLINE_GLOBAL FCrd4 operator/(const FCrd4 & crd, float scalar)
{
    CH_FCRD4_CODE(EH_FATAL_CONDITION(FCrd4_errors,scalar == 0.0));

    return FCrd4(crd.x[0] / scalar,
                crd.x[1] / scalar,
                crd.x[2] / scalar,
                crd.x[3] / scalar);
}

INLINE_GLOBAL float dotProduct(const FCrd4 & crd1, const FCrd4 & crd2)
{
    return (crd1.x[0] * crd2.x[0] +
            crd1.x[1] * crd2.x[1] +
            crd1.x[2] * crd2.x[2] +
            crd1.x[3] * crd2.x[3]);
}

INLINE_GLOBAL FCrd4 normalizedCopy(const FCrd4 & crd)
{
    float crdLength = length(crd);

    return (crd / crdLength);
}

INLINE_MEMBER void FCrd4::cwMultiplyBy(const FCrd4 & crd)
{
    x[0] *= crd.x[0];
    x[1] *= crd.x[1];
    x[2] *= crd.x[2];
    x[3] *= crd.x[3];
}

INLINE_MEMBER void FCrd4::cwDevideBy(const FCrd4 & crd)
{
    CH_FCRD4_CODE(EH_FATAL_CONDITION(FCrd4_errors,crd[0] == 0.0));
    CH_FCRD4_CODE(EH_FATAL_CONDITION(FCrd4_errors,crd[1] == 0.0));
    CH_FCRD4_CODE(EH_FATAL_CONDITION(FCrd4_errors,crd[2] == 0.0));
    CH_FCRD4_CODE(EH_FATAL_CONDITION(FCrd4_errors,crd[3] == 0.0));

    x[0] /= crd.x[0];
    x[1] /= crd.x[1];
    x[2] /= crd.x[2];
    x[3] /= crd.x[3];
}

INLINE_GLOBAL FCrd4 cwReciprocalCopy(const FCrd4 & crd)
{
    FCrd4 newVec;
    
    if (crd.x[0] != 0.0)
        newVec.x[0] = 1.0 / crd.x[0];
    else
        newVec.x[0] = 0.0;

    if (crd.x[1] != 0.0)
        newVec.x[1] = 1.0 / crd.x[1];
    else
        newVec.x[1] = 0.0;

    if (crd.x[2] != 0.0)
        newVec.x[2] = 1.0 / crd.x[2];
    else
        newVec.x[2] = 0.0;

    if (crd.x[3] != 0.0)
        newVec.x[3] = 1.0 / crd.x[3];
    else
        newVec.x[3] = 0.0;

    return (newVec);
}

INLINE_GLOBAL FCrd4 cwProduct(const FCrd4 & crd1, const FCrd4 & crd2)
{
    return (FCrd4(crd1.x[0] * crd2.x[0],
                 crd1.x[1] * crd2.x[1],
                 crd1.x[2] * crd2.x[2],
                 crd1.x[3] * crd2.x[3]));
}

INLINE_GLOBAL FCrd4 cwQuotient(const FCrd4 & crd1, const FCrd4 & crd2)
{
    CH_FCRD4_CODE(EH_FATAL_CONDITION(FCrd4_errors,crd2[0] == 0.0));
    CH_FCRD4_CODE(EH_FATAL_CONDITION(FCrd4_errors,crd2[1] == 0.0));
    CH_FCRD4_CODE(EH_FATAL_CONDITION(FCrd4_errors,crd2[2] == 0.0));

    return FCrd4(crd1.x[0] / crd2.x[0],
                crd1.x[1] / crd2.x[1],
                crd1.x[2] / crd2.x[2],
                crd1.x[3] / crd2.x[3]);
}

INLINE_GLOBAL FCrd4 cwMin(const FCrd4 & crd1, const FCrd4 & crd2)
{
    return FCrd4(min(crd1.x[0],crd2.x[0]),
                min(crd1.x[1],crd2.x[1]),
                min(crd1.x[2],crd2.x[2]),
                min(crd1.x[3],crd2.x[3]));
}

INLINE_GLOBAL FCrd4 cwMax(const FCrd4 & crd1, const FCrd4 & crd2)
{
    return FCrd4(max(crd1.x[0],crd2.x[0]),
                max(crd1.x[1],crd2.x[1]),
                max(crd1.x[2],crd2.x[2]),
                max(crd1.x[3],crd2.x[3]));
}

INLINE_MEMBER FCrd4 & FCrd4::scale(const FCrd4 & s)
{
    x[0] *= s[0];
    x[1] *= s[1];
    x[2] *= s[2];
    x[3] *= s[3];

    return *this;
}

INLINE_MEMBER FCrd4 & FCrd4::scale(float s0, float s1, float s2, float s3)
{
    x[0] *= s0;
    x[1] *= s1;
    x[2] *= s2;
    x[3] *= s3;

    return *this;
}

INLINE_MEMBER FCrd4 & FCrd4::operator*=(const FMat4 & mat)
{
    float t0 = x[0];
    float t1 = x[1];
    float t2 = x[2];

    x[0] = x[0] * mat.x[0][0] + x[1] * mat.x[1][0] +
           x[2] * mat.x[2][0] + x[3] * mat.x[3][0];
    x[1] = t0   * mat.x[0][1] + x[1] * mat.x[1][1] +
           x[2] * mat.x[2][1] + x[3] * mat.x[3][1];
    x[2] = t0   * mat.x[0][2] + t1   * mat.x[1][2] +
           x[2] * mat.x[2][2] + x[3] * mat.x[3][2];
    x[3] = t0   * mat.x[0][3] + t1   * mat.x[1][3] +
           t2   * mat.x[2][3] + x[3] * mat.x[3][3];
    
    return *this;
}

INLINE_GLOBAL FCrd4 prCrdScale(const FCrd4 & crd, const FCrd4 & s)
{
    return FCrd4(crd.x[0] * s[0], crd.x[1] * s[1],
                crd.x[2] * s[2], crd.x[3] * s[3]);
}

INLINE_GLOBAL FCrd4 prCrdScale(const FCrd4 & crd,
        float s0, float s1, float s2, float s3)
{
    return FCrd4(crd.x[0] * s0, crd.x[1] * s1,
                crd.x[2] * s2, crd.x[3] * s3);
}

INLINE_GLOBAL FCrd4 operator*(const FCrd4 & crd, const FMat4 & mat)
{
    return
        FCrd4
        (
            crd.x[0] * mat.x[0][0] +
            crd.x[1] * mat.x[1][0] +
            crd.x[2] * mat.x[2][0] +
            crd.x[3] * mat.x[3][0],

            crd.x[0] * mat.x[0][1] +
            crd.x[1] * mat.x[1][1] +
            crd.x[2] * mat.x[2][1] +
            crd.x[3] * mat.x[3][1],

            crd.x[0] * mat.x[0][2] +
            crd.x[1] * mat.x[1][2] +
            crd.x[2] * mat.x[2][2] +
            crd.x[3] * mat.x[3][2],

            crd.x[0] * mat.x[0][3] +
            crd.x[1] * mat.x[1][3] +
            crd.x[2] * mat.x[2][3] +
            crd.x[3] * mat.x[3][3]
        );
}

// ================================================================ class FMat4

INLINE_MEMBER FMat4::FMat4()
{
    TR_FUN0(FCrd4_constructors,FMat4);
}

INLINE_MEMBER FMat4::FMat4(float x00, float x01, float x02, float x03,
                  float x10, float x11, float x12, float x13,
                  float x20, float x21, float x22, float x23,
                  float x30, float x31, float x32, float x33)
{
    x[0][0] = x00; x[0][1] = x01; x[0][2] = x02; x[0][3] = x03;
    x[1][0] = x10; x[1][1] = x11; x[1][2] = x12; x[1][3] = x13;
    x[2][0] = x20; x[2][1] = x21; x[2][2] = x22; x[2][3] = x23;
    x[3][0] = x30; x[3][1] = x31; x[3][2] = x32; x[3][3] = x33;

    TR_FUN4(FCrd4_constructors,FMat4,x00,x01,x02,"...");
}

INLINE_MEMBER FMat4::FMat4(const FMat4 & mat)
{
    x[0][0] = mat.x[0][0];
    x[0][1] = mat.x[0][1];
    x[0][2] = mat.x[0][2];
    x[0][3] = mat.x[0][3];

    x[1][0] = mat.x[1][0];
    x[1][1] = mat.x[1][1];
    x[1][2] = mat.x[1][2];
    x[1][3] = mat.x[1][3];

    x[2][0] = mat.x[2][0];
    x[2][1] = mat.x[2][1];
    x[2][2] = mat.x[2][2];
    x[2][3] = mat.x[2][3];

    x[3][0] = mat.x[3][0];
    x[3][1] = mat.x[3][1];
    x[3][2] = mat.x[3][2];
    x[3][3] = mat.x[3][3];

    TR_FUN1(FCrd4_constructors,FMat4,mat);
}


INLINE_MEMBER FMat4::~FMat4()
{
    TR_MESSAGE(FCrd4_constructors,"~" << *this);
}

INLINE_MEMBER float FMat4::operator()(int i, int j) const
{
    CH_FCRD4_CODE(
        if (i < 0 || i > 3)
            EH_FATAL(Crd3_errors,"i == " << i);
        if (j < 0 || j > 3)
            EH_FATAL(Crd3_errors,"j == " << j);
    )
    return x[i][j];
}

INLINE_GLOBAL ostream & operator<<(ostream & s, const FMat4 & mat)
{
    s << "FMat4("
	<< mat.x[0][0] << ',' << mat.x[0][1] << ','
	<< mat.x[0][2] << ',' << mat.x[0][3] << ','
	<< mat.x[1][0] << ',' << mat.x[1][1] << ','
	<< mat.x[1][2] << ',' << mat.x[1][3] << ','
	<< mat.x[2][0] << ',' << mat.x[2][1] << ','
	<< mat.x[2][2] << ',' << mat.x[2][3] << ','
	<< mat.x[3][0] << ',' << mat.x[3][1] << ','
	<< mat.x[3][2] << ',' << mat.x[3][3] << ')';
    return s;
}

INLINE_GLOBAL float determinant(const FMat4 & m)
{
    return (  m.x[0][0] * det3x3(m.x[1][1], m.x[1][2], m.x[1][3],
                                 m.x[2][1], m.x[2][2], m.x[2][3],
                                 m.x[3][1], m.x[3][2], m.x[3][3])

            - m.x[0][1] * det3x3(m.x[1][0], m.x[1][2], m.x[1][3],
                                 m.x[2][0], m.x[2][2], m.x[2][3],
                                 m.x[3][0], m.x[3][2], m.x[3][3])

            + m.x[0][2] * det3x3(m.x[1][0], m.x[1][1], m.x[1][3],
                                 m.x[2][0], m.x[2][1], m.x[2][3],
                                 m.x[3][0], m.x[3][1], m.x[3][3])

            - m.x[0][3] * det3x3(m.x[1][0], m.x[1][1], m.x[1][2],
                                 m.x[2][0], m.x[2][1], m.x[2][2],
                                 m.x[3][0], m.x[3][1], m.x[3][2]));
}

INLINE_GLOBAL FMat4 adjointCopy(const FMat4 & m)
{
    FMat4 a;
    //
    //  Row column labeling reversed since we transpose rows & columns.
    //

    a.x[0][0] =   det3x3(m.x[1][1], m.x[1][2], m.x[1][3],
                         m.x[2][1], m.x[2][2], m.x[2][3],
                         m.x[3][1], m.x[3][2], m.x[3][3]);
    a.x[1][0] = - det3x3(m.x[1][0], m.x[1][2], m.x[1][3],
                         m.x[2][0], m.x[2][2], m.x[2][3],
                         m.x[3][0], m.x[3][2], m.x[3][3]);
    a.x[2][0] =   det3x3(m.x[1][0], m.x[1][1], m.x[1][3],
                         m.x[2][0], m.x[2][1], m.x[2][3],
                         m.x[3][0], m.x[3][1], m.x[3][3]);
    a.x[3][0] = - det3x3(m.x[1][0], m.x[1][1], m.x[1][2],
                         m.x[2][0], m.x[2][1], m.x[2][2],
                         m.x[3][0], m.x[3][1], m.x[3][2]);
        
    a.x[0][1] = - det3x3(m.x[0][1], m.x[0][2], m.x[0][3],
                         m.x[2][1], m.x[2][2], m.x[2][3],
                         m.x[3][1], m.x[3][2], m.x[3][3]);
    a.x[1][1] =   det3x3(m.x[0][0], m.x[0][2], m.x[0][3],
                         m.x[2][0], m.x[2][2], m.x[2][3],
                         m.x[3][0], m.x[3][2], m.x[3][3]);
    a.x[2][1] = - det3x3(m.x[0][0], m.x[0][1], m.x[0][3],
                         m.x[2][0], m.x[2][1], m.x[2][3],
                         m.x[3][0], m.x[3][1], m.x[3][3]);
    a.x[3][1] =   det3x3(m.x[0][0], m.x[0][1], m.x[0][2],
                         m.x[2][0], m.x[2][1], m.x[2][2],
                         m.x[3][0], m.x[3][1], m.x[3][2]);

    a.x[0][2] =   det3x3(m.x[0][1], m.x[0][2], m.x[0][3],
                         m.x[1][1], m.x[1][2], m.x[1][3],
                         m.x[3][1], m.x[3][2], m.x[3][3]);
    a.x[1][2] = - det3x3(m.x[0][0], m.x[0][2], m.x[0][3],
                         m.x[1][0], m.x[1][2], m.x[1][3],
                         m.x[3][0], m.x[3][2], m.x[3][3]);
    a.x[2][2] =   det3x3(m.x[0][0], m.x[0][1], m.x[0][3],
                         m.x[1][0], m.x[1][1], m.x[1][3],
                         m.x[3][0], m.x[3][1], m.x[3][3]);
    a.x[3][2] = - det3x3(m.x[0][0], m.x[0][1], m.x[0][2],
                         m.x[1][0], m.x[1][1], m.x[1][2],
                         m.x[3][0], m.x[3][1], m.x[3][2]);
        
    a.x[0][3] = - det3x3(m.x[0][1], m.x[0][2], m.x[0][3],
                         m.x[1][1], m.x[1][2], m.x[1][3],
                         m.x[2][1], m.x[2][2], m.x[2][3]);
    a.x[1][3] =   det3x3(m.x[0][0], m.x[0][2], m.x[0][3],
                         m.x[1][0], m.x[1][2], m.x[1][3],
                         m.x[2][0], m.x[2][2], m.x[2][3]);
    a.x[2][3] = - det3x3(m.x[0][0], m.x[0][1], m.x[0][3],
                         m.x[1][0], m.x[1][1], m.x[1][3],
                         m.x[2][0], m.x[2][1], m.x[2][3]);
    a.x[3][3] =   det3x3(m.x[0][0], m.x[0][1], m.x[0][2],
                         m.x[1][0], m.x[1][1], m.x[1][2],
                         m.x[2][0], m.x[2][1], m.x[2][2]);
        
    return a;
}

INLINE_GLOBAL FMat4 operator*(const FMat4 & mat, float s)
{
    return
        FMat4
        (
            mat.x[0][0] * s, mat.x[0][1] * s,
            mat.x[0][2] * s, mat.x[0][3] * s,
            mat.x[1][0] * s, mat.x[1][1] * s,
            mat.x[1][2] * s, mat.x[1][3] * s,
            mat.x[2][0] * s, mat.x[2][1] * s,
            mat.x[2][2] * s, mat.x[2][3] * s,
            mat.x[3][0] * s, mat.x[3][1] * s,
            mat.x[3][2] * s, mat.x[3][3] * s
        );
}

INLINE_GLOBAL FMat4 inverseCopy(const FMat4 & mat, float det)
{
    det = 1.0 / det;

    return
        adjointCopy(mat) * det;
}

INLINE_GLOBAL FMat4 inverseCopy(const FMat4 & mat)
{
    float det = determinant(mat);

    if (det == 0.0)
    {
        EH_ERROR(FCrd4_errors, ERROR_FCRD4_SINGULAR_MATRIX, "singular matrix");
        return mat;
    }

    return inverseCopy(mat,det);
}

INLINE_MEMBER void FMat4::invert()
{
    *this = inverseCopy(*this);
}

INLINE_GLOBAL FMat4 transposedCopy(const FMat4 & mat)
{
    return
        FMat4(
            mat.x[0][0], mat.x[1][0], mat.x[2][0], mat.x[3][0],
            mat.x[0][1], mat.x[1][1], mat.x[2][1], mat.x[3][1],
            mat.x[0][2], mat.x[1][2], mat.x[2][2], mat.x[3][2],
            mat.x[0][3], mat.x[1][3], mat.x[2][3], mat.x[3][3]
        );
}

INLINE_MEMBER void FMat4::transpose()
{
    *this = transposedCopy(*this);
}

INLINE_MEMBER FMat4 & FMat4::operator*=(float s)
{
    x[0][0] *= s; x[0][1] *= s; x[0][2] *= s; x[0][3] *= s;
    x[1][0] *= s; x[1][1] *= s; x[1][2] *= s; x[1][3] *= s;
    x[2][0] *= s; x[2][1] *= s; x[2][2] *= s; x[2][3] *= s;
    x[3][0] *= s; x[3][1] *= s; x[3][2] *= s; x[3][3] *= s;
    return *this;
}

INLINE_MEMBER FMat4 & FMat4::scale(const FCrd4 & s)
{
    x[0][0] *= s[0]; x[0][1] *= s[1]; x[0][2] *= s[2]; x[0][3] *= s[3];
    x[1][0] *= s[0]; x[1][1] *= s[1]; x[1][2] *= s[2]; x[1][3] *= s[3];
    x[2][0] *= s[0]; x[2][1] *= s[1]; x[2][2] *= s[2]; x[2][3] *= s[3];
    x[3][0] *= s[0]; x[3][1] *= s[1]; x[3][2] *= s[2]; x[3][3] *= s[3];
    return *this;
}

INLINE_MEMBER FMat4 & FMat4::scale(float s0, float s1, float s2, float s3)
{
    x[0][0] *= s0; x[0][1] *= s1; x[0][2] *= s2; x[0][3] *= s3;
    x[1][0] *= s0; x[1][1] *= s1; x[1][2] *= s2; x[1][3] *= s3;
    x[2][0] *= s0; x[2][1] *= s1; x[2][2] *= s2; x[2][3] *= s3;
    x[3][0] *= s0; x[3][1] *= s1; x[3][2] *= s2; x[3][3] *= s3;
    return *this;
}

INLINE_MEMBER FMat4 & FMat4::operator+=(const FMat4 & mat)
{
    x[0][0] += mat.x[0][0];
    x[0][1] += mat.x[0][1];
    x[0][2] += mat.x[0][2];
    x[0][3] += mat.x[0][3];

    x[1][0] += mat.x[1][0];
    x[1][1] += mat.x[1][1];
    x[1][2] += mat.x[1][2];
    x[1][3] += mat.x[1][3];

    x[2][0] += mat.x[2][0];
    x[2][1] += mat.x[2][1];
    x[2][2] += mat.x[2][2];
    x[2][3] += mat.x[2][3];

    x[3][0] += mat.x[3][0];
    x[3][1] += mat.x[3][1];
    x[3][2] += mat.x[3][2];
    x[3][3] += mat.x[3][3];

    return *this;
}

INLINE_GLOBAL FMat4 operator*(const FMat4 & mat1, const FMat4 & mat2)
{
    return
        FMat4
        (
            mat1.x[0][0] * mat2.x[0][0] + mat1.x[0][1] * mat2.x[1][0] +
            mat1.x[0][2] * mat2.x[2][0] + mat1.x[0][3] * mat2.x[3][0],
            mat1.x[0][0] * mat2.x[0][1] + mat1.x[0][1] * mat2.x[1][1] +
            mat1.x[0][2] * mat2.x[2][1] + mat1.x[0][3] * mat2.x[3][1],
            mat1.x[0][0] * mat2.x[0][2] + mat1.x[0][1] * mat2.x[1][2] +
            mat1.x[0][2] * mat2.x[2][2] + mat1.x[0][3] * mat2.x[3][2],
            mat1.x[0][0] * mat2.x[0][3] + mat1.x[0][1] * mat2.x[1][3] +
            mat1.x[0][2] * mat2.x[2][3] + mat1.x[0][3] * mat2.x[3][3],

            mat1.x[1][0] * mat2.x[0][0] + mat1.x[1][1] * mat2.x[1][0] +
            mat1.x[1][2] * mat2.x[2][0] + mat1.x[1][3] * mat2.x[3][0],
            mat1.x[1][0] * mat2.x[0][1] + mat1.x[1][1] * mat2.x[1][1] +
            mat1.x[1][2] * mat2.x[2][1] + mat1.x[1][3] * mat2.x[3][1],
            mat1.x[1][0] * mat2.x[0][2] + mat1.x[1][1] * mat2.x[1][2] +
            mat1.x[1][2] * mat2.x[2][2] + mat1.x[1][3] * mat2.x[3][2],
            mat1.x[1][0] * mat2.x[0][3] + mat1.x[1][1] * mat2.x[1][3] +
            mat1.x[1][2] * mat2.x[2][3] + mat1.x[1][3] * mat2.x[3][3],

            mat1.x[2][0] * mat2.x[0][0] + mat1.x[2][1] * mat2.x[1][0] +
            mat1.x[2][2] * mat2.x[2][0] + mat1.x[2][3] * mat2.x[3][0],
            mat1.x[2][0] * mat2.x[0][1] + mat1.x[2][1] * mat2.x[1][1] +
            mat1.x[2][2] * mat2.x[2][1] + mat1.x[2][3] * mat2.x[3][1],
            mat1.x[2][0] * mat2.x[0][2] + mat1.x[2][1] * mat2.x[1][2] +
            mat1.x[2][2] * mat2.x[2][2] + mat1.x[2][3] * mat2.x[3][2],
            mat1.x[2][0] * mat2.x[0][3] + mat1.x[2][1] * mat2.x[1][3] +
            mat1.x[2][2] * mat2.x[2][3] + mat1.x[2][3] * mat2.x[3][3],

            mat1.x[3][0] * mat2.x[0][0] + mat1.x[3][1] * mat2.x[1][0] +
            mat1.x[3][2] * mat2.x[2][0] + mat1.x[3][3] * mat2.x[3][0],
            mat1.x[3][0] * mat2.x[0][1] + mat1.x[3][1] * mat2.x[1][1] +
            mat1.x[3][2] * mat2.x[2][1] + mat1.x[3][3] * mat2.x[3][1],
            mat1.x[3][0] * mat2.x[0][2] + mat1.x[3][1] * mat2.x[1][2] +
            mat1.x[3][2] * mat2.x[2][2] + mat1.x[3][3] * mat2.x[3][2],
            mat1.x[3][0] * mat2.x[0][3] + mat1.x[3][1] * mat2.x[1][3] +
            mat1.x[3][2] * mat2.x[2][3] + mat1.x[3][3] * mat2.x[3][3]
        );
}

INLINE_MEMBER FMat4 & FMat4::operator*=(const FMat4 & mat)
{
    *this = *this * mat;

    return *this;
}

INLINE_GLOBAL FMat4 prMatScale(const FMat4 & mat, const FCrd4 & s)
{
    return
        FMat4
        (
            mat.x[0][0] * s[0], mat.x[0][1] * s[1],
            mat.x[0][2] * s[2], mat.x[0][3] * s[3],
            mat.x[1][0] * s[0], mat.x[1][1] * s[1],
            mat.x[1][2] * s[2], mat.x[1][3] * s[3],
            mat.x[2][0] * s[0], mat.x[2][1] * s[1],
            mat.x[2][2] * s[2], mat.x[2][3] * s[3],
            mat.x[3][0] * s[0], mat.x[3][1] * s[1],
            mat.x[3][2] * s[2], mat.x[3][3] * s[3]
        );
}

INLINE_GLOBAL FMat4 prMatScale(
        const FMat4 & mat,
        float s0, float s1, float s2, float s3)
{
    return
        FMat4
        (
            mat.x[0][0] * s0, mat.x[0][1] * s1,
            mat.x[0][2] * s2, mat.x[0][3] * s3,
            mat.x[1][0] * s0, mat.x[1][1] * s1,
            mat.x[1][2] * s2, mat.x[1][3] * s3,
            mat.x[2][0] * s0, mat.x[2][1] * s1,
            mat.x[2][2] * s2, mat.x[2][3] * s3,
            mat.x[3][0] * s0, mat.x[3][1] * s1,
            mat.x[3][2] * s2, mat.x[3][3] * s3
        );
}

INLINE_GLOBAL FMat4 operator*(float s, const FMat4 & mat)
{
    return
        FMat4
        (
            s * mat.x[0][0], s * mat.x[0][1],
            s * mat.x[0][2], s * mat.x[0][3],
            s * mat.x[1][0], s * mat.x[1][1],
            s * mat.x[1][2], s * mat.x[1][3],
            s * mat.x[2][0], s * mat.x[2][1],
            s * mat.x[2][2], s * mat.x[2][3],
            s * mat.x[3][0], s * mat.x[3][1],
            s * mat.x[3][2], s * mat.x[3][3]
        );
}

INLINE_GLOBAL FMat4 prScaleMat(const FCrd4 & s, const FMat4 & mat)
{
    return
        FMat4
        (
            s[0] * mat.x[0][0], s[0] * mat.x[0][1],
            s[0] * mat.x[0][2], s[0] * mat.x[0][3],
            s[1] * mat.x[1][0], s[1] * mat.x[1][1],
            s[1] * mat.x[1][2], s[1] * mat.x[1][3],
            s[2] * mat.x[2][0], s[2] * mat.x[2][1],
            s[2] * mat.x[2][2], s[2] * mat.x[2][3],
            s[3] * mat.x[3][0], s[3] * mat.x[3][1],
            s[3] * mat.x[3][2], s[3] * mat.x[3][3]
        );
}

INLINE_GLOBAL FMat4 prScaleMat(
        float s0, float s1, float s2, float s3,
        const FMat4 & mat
        )
{
    return
        FMat4
        (
            s0 * mat.x[0][0], s0 * mat.x[0][1],
            s0 * mat.x[0][2], s0 * mat.x[0][3],
            s1 * mat.x[1][0], s1 * mat.x[1][1],
            s1 * mat.x[1][2], s1 * mat.x[1][3],
            s2 * mat.x[2][0], s2 * mat.x[2][1],
            s2 * mat.x[2][2], s2 * mat.x[2][3],
            s3 * mat.x[3][0], s3 * mat.x[3][1],
            s3 * mat.x[3][2], s3 * mat.x[3][3]
        );
}

INLINE_GLOBAL FMat4 operator-(const FMat4 & mat)
{
    return
        FMat4
        (
	    - mat.x[0][0], - mat.x[0][1], - mat.x[0][2], - mat.x[0][3],
	    - mat.x[1][0], - mat.x[1][1], - mat.x[1][2], - mat.x[1][3],
	    - mat.x[2][0], - mat.x[2][1], - mat.x[2][2], - mat.x[2][3],
	    - mat.x[3][0], - mat.x[3][1], - mat.x[3][2], - mat.x[3][3]
        );
}

INLINE_GLOBAL FMat4 operator+(const FMat4 & mat1, const FMat4 & mat2)
{
    return
        FMat4
        (
	    mat1.x[0][0] + mat2.x[0][0],
	    mat1.x[0][1] + mat2.x[0][1],
	    mat1.x[0][2] + mat2.x[0][2],
	    mat1.x[0][3] + mat2.x[0][3],
	
	    mat1.x[1][0] + mat2.x[1][0],
	    mat1.x[1][1] + mat2.x[1][1],
	    mat1.x[1][2] + mat2.x[1][2],
	    mat1.x[1][3] + mat2.x[1][3],
	
	    mat1.x[2][0] + mat2.x[2][0],
	    mat1.x[2][1] + mat2.x[2][1],
	    mat1.x[2][2] + mat2.x[2][2],
	    mat1.x[2][3] + mat2.x[2][3],
	
	    mat1.x[3][0] + mat2.x[3][0],
	    mat1.x[3][1] + mat2.x[3][1],
	    mat1.x[3][2] + mat2.x[3][2],
	    mat1.x[3][3] + mat2.x[3][3]
        );
}

INLINE_GLOBAL FMat4 operator-(const FMat4 & mat1, const FMat4 & mat2)
{
    return
        FMat4
        (
	    mat1.x[0][0] - mat2.x[0][0],
	    mat1.x[0][1] - mat2.x[0][1],
	    mat1.x[0][2] - mat2.x[0][2],
	    mat1.x[0][3] - mat2.x[0][3],
	
	    mat1.x[1][0] - mat2.x[1][0],
	    mat1.x[1][1] - mat2.x[1][1],
	    mat1.x[1][2] - mat2.x[1][2],
	    mat1.x[1][3] - mat2.x[1][3],
	
	    mat1.x[2][0] - mat2.x[2][0],
	    mat1.x[2][1] - mat2.x[2][1],
	    mat1.x[2][2] - mat2.x[2][2],
	    mat1.x[2][3] - mat2.x[2][3],
	
	    mat1.x[3][0] - mat2.x[3][0],
	    mat1.x[3][1] - mat2.x[3][1],
	    mat1.x[3][2] - mat2.x[3][2],
	    mat1.x[3][3] - mat2.x[3][3]
        );
}

} // extern "C++"

// ===========================================================================
