// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       FVectorN
//  TYPE:       c++ inline
//  PROJECT:    VEGA Math Library
//  CONTENT:    vector and matrix of floats
//  VERSION:    0.6.0
// ===========================================================================
//  AUTHORS:    rft     Robert F. Tobler
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 14:06:28  rft     last modification
//  28-jul-94 15:24:07  rft     created
// ===========================================================================

extern "C++"
{

INLINE_MEMBER FVectorN::FVectorN()
:
    lower(0),
    length(0),
    valuePtr(NIL(float*))
{
}

INLINE_MEMBER void FVectorN::make()
{
    if (length > 0)
	valuePtr = new float[length] - lower;
    else
	valuePtr = NIL(float *);
}

INLINE_MEMBER void FVectorN::kill()
{
    if (isNotNIL(valuePtr))
	delete[] (valuePtr + lower);
}

INLINE_MEMBER int FVectorN::getLength() const  { return length; }
INLINE_MEMBER int FVectorN::getLower() const { return lower; }
INLINE_MEMBER int FVectorN::getUpper() const { return lower + length; }

INLINE_MEMBER void FVectorN::setLower(int newLower)
{
    valuePtr = valuePtr + lower - newLower;
    lower = newLower;
}

INLINE_MEMBER float * FVectorN::getValuePtr()
{
    return valuePtr;
}

INLINE_MEMBER float * FVectorN::getLowerPtr()
{
    return valuePtr + lower;
}

INLINE_MEMBER float * FVectorN::getUpperPtr()
{
    return valuePtr + lower + length;
}

INLINE_MEMBER const float * FVectorN::getValuePtr() const
{
    return valuePtr;
}

INLINE_MEMBER const float * FVectorN::getLowerPtr() const
{
    return valuePtr + lower;
}

INLINE_MEMBER const float * FVectorN::getUpperPtr() const
{
    return valuePtr + lower + length;
}

INLINE_MEMBER FVectorN::FVectorN(int newLength)
:
    lower(0),
    length(newLength)
{
    make();
}

INLINE_MEMBER FVectorN::FVectorN(int newLower, int newLength)
:
    lower(newLower),
    length(newLength)
{
    make();
}

INLINE_MEMBER FVectorN::FVectorN(
	const float * values,
	int newLower,
	int newLength
	)
:
    lower(newLower),
    length(newLength)
{
    make();
#ifdef __GNUC__
    copy((float *)values,getLength(),getLowerPtr());
    //   ^^^^^^^^^^
    //  this was put there due to a bug in gcc 2.5.8
    //
#else
    copy(values,getLength(),getLowerPtr());
#endif
}

INLINE_MEMBER FVectorN::FVectorN(
	const float * values,
	int newLength
	)
:
    lower(0),
    length(newLength)
{
    make();
#ifdef __GNUC__
    copy((float *)values,getLength(),getLowerPtr());
    //   ^^^^^^^^^^
    //  this was put there due to a bug in gcc 2.5.8
    //
#else
    copy(values,getLength(),getLowerPtr());
#endif
}

INLINE_MEMBER FVectorN::FVectorN(const FVectorN & vector)
:
    lower(vector.getLower()),
    length(vector.getLength())
{
    make();
#ifdef __GNUC__
    copy((float*)vector.getLowerPtr(),getLength(),getLowerPtr());
    //   ^^^^^^^^^^
    //  this was put there due to a bug in gcc 2.5.8
    //
#else
    copy(vector.getLowerPtr(),getLength(),getLowerPtr());
#endif
}


INLINE_MEMBER float FVectorN::operator[](int index) const
{
    return valuePtr[index];
}

INLINE_MEMBER float & FVectorN::operator[](int index)
{
    return valuePtr[index];
}

INLINE_MEMBER float FVectorN::get(int index) const
{
    return valuePtr[index];
}

INLINE_MEMBER void FVectorN::set(int index, float value)
{
    valuePtr[index] = value;
}

INLINE_MEMBER FVectorN & FVectorN::operator=(const FVectorN & vector)
{
    if (length != vector.getLength())
    {
	kill();
	lower = vector.getLower();
	length = vector.getLength();
	make();
    }
    else if (lower != vector.getLower())
    {
	valuePtr += lower - vector.getLower();
	lower = vector.getLower();
    }
#ifdef __GNUC__
    copy((float*)vector.getLowerPtr(),getLength(),getLowerPtr());
    //   ^^^^^^^^^^
    //  this was put there due to a bug in gcc 2.5.8
    //
#else
    copy(vector.getLowerPtr(),getLength(),getLowerPtr());
#endif
    return *this;
}

INLINE_MEMBER SubFVectorN::SubFVectorN(
	FVectorN & vector
	)
:
    FVectorN()
{
    lower = vector.getLower();
    length = vector.getLength();
    valuePtr = vector.getValuePtr();
}

INLINE_MEMBER SubFVectorN::SubFVectorN(
	FVectorN & vector,
	int subLower,
	int subLength
	)
:
    FVectorN()
{
    lower = vector.getLower();
    length = subLength;
    valuePtr = vector.getValuePtr() + subLower - lower;
}

INLINE_MEMBER SubFVectorN::SubFVectorN(
	FVectorN & vector,
	int subLower,
	int subLength,
	int newLower
	)
:
    FVectorN()
{
    lower = newLower;
    length = subLength;
    valuePtr = vector.getValuePtr() + subLower - newLower;
}

INLINE_MEMBER SubFVectorN::SubFVectorN(
	float * values,
	int newLength
	)
:
    FVectorN()
{
    lower = 0;
    length = newLength;
    valuePtr = values;
}

INLINE_MEMBER SubFVectorN::SubFVectorN(
	float * values,
	int newLower,
	int newLength
	)
:
    FVectorN()
{
    lower = newLower;
    length = newLength;
    valuePtr = values - newLower;
}

INLINE_MEMBER FMatrixNM::FMatrixNM()
:
    lowerN(0),
    lowerM(0),
    deltaN(0),
    lengthN(0),
    lengthM(0),
    valuePtr(NIL(float*))
{
}

INLINE_MEMBER void FMatrixNM::make()
{
    if (lengthN > 0 && lengthM > 0)
    {
	deltaN = lengthM;
	valuePtr = (new float[lengthN * lengthM]) - lowerN * deltaN - lowerM;
    }
    else
    {
	valuePtr = NIL(float *);
    }
}

INLINE_MEMBER void FMatrixNM::kill()
{
    if (isNotNIL(valuePtr))
    {
	delete [] (valuePtr + lowerN * deltaN + lowerM);
    }
}

INLINE_MEMBER FMatrixNM::FMatrixNM(
	int newLengthN,
	int newLengthM
	)
:
    lowerN(0), lowerM(0),
    lengthN(newLengthN), lengthM(newLengthM)
{
    make();
}

INLINE_MEMBER FMatrixNM::FMatrixNM(
    int newLowerN, int newLowerM,
    int newLengthN, int newLengthM
    )
:
    lowerN(newLowerN), lowerM(newLowerM),
    lengthN(newLengthN), lengthM(newLengthM)
{
    make();
}

INLINE_MEMBER int FMatrixNM::getLengthN() const { return lengthN; }
INLINE_MEMBER int FMatrixNM::getLengthM() const { return lengthM; }
INLINE_MEMBER int FMatrixNM::getLowerN() const { return lowerN; }
INLINE_MEMBER int FMatrixNM::getLowerM() const { return lowerM; }
INLINE_MEMBER int FMatrixNM::getUpperN() const { return lowerN + lengthN; }
INLINE_MEMBER int FMatrixNM::getUpperM() const { return lowerM + lengthM; }
INLINE_MEMBER int FMatrixNM::getDeltaN() const { return deltaN; }

INLINE_MEMBER float * FMatrixNM::getValuePtr()
{
    return valuePtr;
}

INLINE_MEMBER float * FMatrixNM::getLowerNptr()
{
    return valuePtr + lowerN * deltaN;
}

INLINE_MEMBER float * FMatrixNM::getUpperNptr()
{
    return valuePtr + (lowerN + lengthN) * deltaN;
}

INLINE_MEMBER float * FMatrixNM::getLowerMptr(float * value1ptr)
{
    return value1ptr + lowerM;
}

INLINE_MEMBER float * FMatrixNM::getUpperMptr(float * value1ptr)
{
    return value1ptr + lowerM + lengthM;
}

INLINE_MEMBER const float * FMatrixNM::getValuePtr() const
{
    return valuePtr;
}

INLINE_MEMBER const float * FMatrixNM::getLowerNptr() const
{
    return valuePtr + lowerN * deltaN;
}

INLINE_MEMBER const float * FMatrixNM::getUpperNptr() const
{
    return valuePtr + (lowerN + lengthN) * deltaN;
}

INLINE_MEMBER const float * FMatrixNM::getLowerMptr(
	const float * value1ptr
	) const
{
    return value1ptr + lowerM;
}

INLINE_MEMBER const float * FMatrixNM::getUpperMptr(
	const float * value1ptr
	) const
{
    return value1ptr + lowerM + lengthM;
}

INLINE_MEMBER FMatrixNM::FMatrixNM(
    const float * values,
    int newLowerN, int newLowerM,
    int newLengthN, int newLengthM
    )
:
    lowerN(newLowerN), lowerM(newLowerM),
    lengthN(newLengthN), lengthM(newLengthM)
{
    make();
    
    if (isNotNIL(valuePtr))
#ifdef __GNUC__
	copy((float*)values,
	//   ^^^^^^^^^^
	//  this was put there due to a bug in gcc 2.5.8
	//
		getLengthN() * getLengthM(),
		getLowerMptr(getLowerNptr()));
#else
	copy(values,
	     getLengthN() * getLengthM(),
	     getLowerMptr(getLowerNptr()));
#endif
}

INLINE_MEMBER FMatrixNM::FMatrixNM(
    const float * values,
    int newLengthN, int newLengthM
    )
:
    lowerN(0), lowerM(0),
    lengthN(newLengthN), lengthM(newLengthM)
{
    make();
    
    if (isNotNIL(valuePtr))
#ifdef __GNUC__
	copy((float*)values,
	//   ^^^^^^^^^^
	//  this was put there due to a bug in gcc 2.5.8
	//
		getLengthN() * getLengthM(),
		getLowerMptr(getLowerNptr()));
#else
	copy(values,
	     getLengthN() * getLengthM(),
	     getLowerMptr(getLowerNptr()));
#endif
}

INLINE_MEMBER const float * FMatrixNM::operator[](int index) const
{
    return valuePtr + index * deltaN;
}

INLINE_MEMBER float * FMatrixNM::operator[](int index)
{
    return valuePtr + index * deltaN;
}

/*

INLINE_MEMBER SubFVectorN SubFMatrixNM::getRowSubVectorN(int row)
{
    return SubFVectorN(getLowerMPtr(operator[](row)),getLengthM());
}

INLINE_MEMBER SubFVectorN SubFMatrixNM::getRowSubVectorN(
	int row,
	int colLower,
	int colLength,
	int newLower
	)
{
    return SubFVectorN(getLowerMPtr(operator[](row)) + colLower,
		     newLower,colLength);
}

*/

INLINE_MEMBER SubFMatrixNM::SubFMatrixNM(
	FMatrixNM & matrix
	)
:
    FMatrixNM()
{
    lowerN = matrix.getLowerN();
    lowerM = matrix.getLowerM();
    deltaN = matrix.getDeltaN();
    lengthN = matrix.getLengthN();
    lengthM = matrix.getLengthM();
    valuePtr = matrix.getValuePtr();
}

INLINE_MEMBER SubFMatrixNM::SubFMatrixNM(
	FMatrixNM & matrix,
	int subLowerN,
	int subLowerM,
	int subLengthN,
	int subLengthM
	)
:
    FMatrixNM()
{
    lowerN = matrix.getLowerN();
    lowerM = matrix.getLowerM();
    deltaN = matrix.getDeltaN();
    lengthN = subLengthN;
    lengthM = subLengthM;

    if (lengthN > 0 && lengthM > 0)
	valuePtr = matrix.getValuePtr()
		    + (subLowerN - lowerN) * deltaN
		    + subLowerM - lowerM;
}

INLINE_MEMBER SubFMatrixNM::SubFMatrixNM(
	FMatrixNM & matrix,
	int subLowerN,
	int subLowerM,
	int subLengthN,
	int subLengthM,
	int newLowerN,
	int newLowerM
	)
:
    FMatrixNM()
{
    lowerN = newLowerN;
    lowerM = newLowerM;
    deltaN = matrix.getDeltaN();
    lengthN = subLengthN;
    lengthM = subLengthM;

    if (lengthN > 0 && lengthM > 0)
	valuePtr = matrix.getValuePtr()
		    + (subLowerN - lowerN) * deltaN
		    + subLowerM - lowerM;
}



INLINE_MEMBER SubFMatrixNM::SubFMatrixNM(
	float * values,
	int newLowerN,
	int newLowerM,
	int newLengthN,
	int newLengthM
	)
:
    FMatrixNM()
{
    lowerN = newLowerN;
    lowerM = newLowerM;
    deltaN = newLengthM;
    lengthN = newLengthN;
    lengthM = newLengthM;

    if (lengthN > 0 && lengthM > 0)
	valuePtr = values - lowerN * deltaN - lowerM;
}

INLINE_MEMBER SubFMatrixNM::SubFMatrixNM(
	float * values,
	int newLengthN,
	int newLengthM
	)
:
    FMatrixNM()
{
    lowerN = 0;
    lowerM = 0;
    deltaN = newLengthM;
    lengthN = newLengthN;
    lengthM = newLengthM;


    if (lengthN > 0 && lengthM > 0)
	valuePtr = values - lowerN * deltaN - lowerM;
}


INLINE_MEMBER FVectorN & FVectorN::hartleyTransform()
{
    bitReverseIndex();
    fastHartleyTransform();
    return *this;
}

INLINE_MEMBER FVectorN FVectorN::hartleyTransformedCopy() const
{
    FVectorN result = bitReverseIndexedCopy();
    result.fastHartleyTransform();
    return result;
}


} // extern "C++"

// ===========================================================================
