// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       IVectorN
//  TYPE:       c++ header
//  PROJECT:    VEGA Math Library
//  CONTENT:    vector and matrix of ints
//  VERSION:    0.6.0
// ===========================================================================
//  AUTHORS:    rft     Robert F. Tobler
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 14:06:28  rft     last modification
//  28-jul-94 15:24:07  rft     created
// ===========================================================================
//  MANUAL:
//
//  DESCRIPTION
//      This module implements int Vector and Matrix classes. Although it
//      is recommended to work with vectors (matrices) with first index 0,
//      both classes are implemented in such a way that vectors and matrices
//      with different first indices can be created.
//
//	Both IVectorN and IMatrixNM implement subclasses calles SubIVectorN and
//	SubIMatrixNM, that do NOT own the actual memory they are using. Thus
//	a change to a submatrix will be reflected in the original matrix with
//	which it was initialized.
//
//	WARNING: If a IVectorN (IMatrixNM) is deleted before a SubIVectorN
//	(SubIMatrixNM) that was initialized with it, the SubIVectorN
//	(SubIMatrixNM) will contain an illegal pointer.
//
//  SEE ALSO
//      VEGA, MathLibrary
// ===========================================================================
#ifndef _MATH_IVECTORN_H_
#define _MATH_IVECTORN_H_

#include <VEGA/Misc/ttools.h>

#include "VEGA/Math/globals.h"

#ifdef __GNUC__
#pragma interface
#endif

extern "C++"
{

#ifdef SGI
#define VECTOR2_NO_INLINE
#endif

#define ERROR_VECTOR_MISMATCH		1
	//
	//  A vector operation is attempted with two vectors that do not match
	//  in length (or base offset).
	//

#define ERROR_VECTOR_NOT_POWER_OF_2	2
	//
	//  Some vector operation require the vector length to be a power of
	//  two. If any other vector is supplied, this error is raised.
	//

// ============================================================= class IVectorN

class IVectorN
{
    friend class SubIVectorN;

public:

    IVectorN(
	);
    IVectorN(
	int newLength			// implicit: newLower = 0
	);
    IVectorN(
	int newLower,			// lowest valid index
	int newLength
	);
    IVectorN(
	const int * values,		// implicit: newLower = 0
	int newLength
	);
    IVectorN(
	const int * values,		// values[0] is placed in
	int newLower,			//  vector[newLower]
	int newLength
	);
    IVectorN(
	const IVectorN & vector
	);

    virtual ~IVectorN(
	);

    int operator[](
	int index
	) const;
    int & operator[](
	int index
	);

    int getLength(
	) const;    
    int getLower(			// Get index of first valid element.
	) const;
    int getUpper(			// 1 + index of last valid element.
	) const;

    void setLower(			// Set index of first valid element
	int newLower			//  to newLower.
	);

    int get(
	int index
	) const;
    void set(
	int index,
	int value
	);

    IVectorN & fill(			// Fill all elements with the same
	int value			//  specified value.
	);

    IVectorN & operator=(
	const IVectorN & vector
	);

    IVectorN & operator*=(
	int scalar
	);
    IVectorN & operator+=(
	const IVectorN & vector
	);
    IVectorN & operator-=(
	const IVectorN & vector
	);


    friend IVectorN operator-(
	const IVectorN & vector
	);

    //
    //  Following operators only work on vectors of same length and lower
    //  index.
    //

    friend IVectorN operator+(
	const IVectorN & vector1,
	const IVectorN & vector2
	);
    friend IVectorN operator-(
	const IVectorN & vector1,
	const IVectorN & vector2
	);
    friend IVectorN operator*(
	int scalar,
	const IVectorN & vector
	);
    friend IVectorN operator*(
	const IVectorN & vector,
	int scalar
	);

    // sfmap: protect: double

    friend double dotProduct(
	const IVectorN & vector1,
	const IVectorN & vector2
	);

    IVectorN & bitReverseIndex(
	);
    IVectorN bitReverseIndexedCopy(
	) const;


protected:
    void make();
    void kill();


    int * getValuePtr();
    int * getLowerPtr();
    int * getUpperPtr();
    const int * getValuePtr() const;
    const int * getLowerPtr() const;
    const int * getUpperPtr() const;

    int lower;
    int length;
    int * valuePtr;
};

// ========================================================== class SubIVectorN

class SubIVectorN : public IVectorN
{
public:
    SubIVectorN(
	IVectorN & vector
	);

    SubIVectorN(				// lower is inherited
	IVectorN & vector,
	int subLower,
	int subLength
	);

    SubIVectorN(
	IVectorN & vector,
	int subLower,
	int subLength,
	int newLower
	);

    SubIVectorN(
	int * values,
	int newLength
	);

    SubIVectorN(
	int * values,
	int newLower,
	int newLength
	);

    SubIVectorN & operator=(
	const IVectorN & vector
	);

    ~SubIVectorN(
	);
};

// ============================================================ class IMatrixNM

class IMatrixNM
{
    friend class SubIMatrixNM;

public:
    IMatrixNM(
	);
    IMatrixNM(
	const IMatrixNM & matrix
	);
    IMatrixNM(
	int newLengthN,
	int newLengthM
	);
    IMatrixNM(
	int newLowerN,
	int newLowerM,
	int newLengthN,
	int newLengthM
	);
    IMatrixNM(
	const int * values,
	int newLengthN,
	int newLengthM
	);
    IMatrixNM(
	const int * values,
	int newLowerN,
	int newLowerM,
	int newLengthN,
	int newLengthM
	);
    virtual ~IMatrixNM(
	);

    IMatrixNM & operator=(
	const IMatrixNM & matrix
	);

    const int * operator[](
	int index
	) const;
    int * operator[](
	int index
	);

    IMatrixNM & fill(
	int value
	);

    IMatrixNM & operator*=(
	int scalar
	);
    IMatrixNM & operator+=(
	const IMatrixNM & matrix
	);
    IMatrixNM & operator-=(
	const IMatrixNM & matrix
	);

    int getLengthN(
	) const;
    int getLengthM(
	) const;
    int getLowerN(
	) const;
    int getLowerM(
	) const;
    int getUpperN(
	) const;
    int getUpperM(
	) const;
    int getDeltaN(
	) const;
/*
    SubIVectorN getSubRow(
	int row
	);
    SubIVectorN getSubRow(
	int row,
	int colLower,
	int colLength
	);

    IVectorN getColumn(
	int col
	);

    IVectorN getColumn(
	int col,
	int rowLower,
	int rowLength
	);
*/

protected:
    void make();
    void kill();

    int * getValuePtr();
    int * getLowerNptr();
    int * getUpperNptr();
    int * getLowerMptr(int * value1ptr);
    int * getUpperMptr(int * value1ptr);

    const int * getValuePtr() const;
    const int * getLowerNptr() const;
    const int * getUpperNptr() const;
    const int * getLowerMptr(const int * value1ptr) const;
    const int * getUpperMptr(const int * value1ptr) const;

    int lowerN;				// lower index in major axis
    int lowerM;				// lower index in minor axis
    int lengthN;			// length in major axis
    int lengthM;			// length in minor axis
    int deltaN;				// difference of two
    					// matrix[i][j] and matrix[i+1][j]

    int * valuePtr;			// pointer to matrix[0][0]
};

// ========================================================= class SubIMatrixNM

class SubIMatrixNM : public IMatrixNM
{
public:
    SubIMatrixNM(
	IMatrixNM & matrix
	);

    SubIMatrixNM(			// lowerN and lowerM are inherited
	IMatrixNM & matrix,
	int subLowerN,
	int subLowerM,
	int subLengthN,
	int subLengthM
	);

    SubIMatrixNM(
	IMatrixNM & matrix,
	int subLowerN,
	int subLowerM,
	int subLengthN,
	int subLengthM,
	int newLowerN,
	int newLowerM
	);

    SubIMatrixNM(			// implicit: lowerN = lowerM = 0
	int * values,
	int newLengthN,
	int newLengthM
	);

    SubIMatrixNM(
	int * values,
	int newLowerN,
	int newLowerM,
	int newLengthN,
	int newLengthM
	);

    SubIMatrixNM & operator=(
	const IMatrixNM & matrix
	);

    ~SubIMatrixNM();
};

} // extern "C++"

#ifndef IVECTORN1_NO_INLINE

#define INLINE_GLOBAL COMPILER_INLINE_GLOBAL
#define INLINE_MEMBER COMPILER_INLINE_MEMBER

#include "VEGA/Math/ivecton1.ipp"

#undef INLINE_GLOBAL
#undef INLINE_MEMBER

#endif

#ifndef IVECTORN2_NO_INLINE

#define INLINE_GLOBAL COMPILER_INLINE_GLOBAL
#define INLINE_MEMBER COMPILER_INLINE_MEMBER

#include "VEGA/Math/ivecton2.ipp"

#undef INLINE_GLOBAL
#undef INLINE_MEMBER

#endif

#endif // _MATH_IVECTORN_H_
// ===========================================================================
