// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       IVectorN
//  TYPE:       c++ inline
//  PROJECT:    VEGA Math Library
//  CONTENT:    vector and matrix of ints
//  VERSION:    0.6.0
// ===========================================================================
//  AUTHORS:    rft     Robert F. Tobler
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 14:06:28  rft     last modification
//  28-jul-94 15:24:07  rft     created
// ===========================================================================

extern "C++"
{

INLINE_MEMBER IVectorN::IVectorN()
:
    lower(0),
    length(0),
    valuePtr(NIL(int*))
{
}

INLINE_MEMBER void IVectorN::make()
{
    if (length > 0)
	valuePtr = new int[length] - lower;
    else
	valuePtr = NIL(int *);
}

INLINE_MEMBER void IVectorN::kill()
{
    if (isNotNIL(valuePtr))
	delete[] (valuePtr + lower);
}

INLINE_MEMBER int IVectorN::getLength() const  { return length; }
INLINE_MEMBER int IVectorN::getLower() const { return lower; }
INLINE_MEMBER int IVectorN::getUpper() const { return lower + length; }

INLINE_MEMBER void IVectorN::setLower(int newLower)
{
    valuePtr = valuePtr + lower - newLower;
    lower = newLower;
}

INLINE_MEMBER int * IVectorN::getValuePtr()
{
    return valuePtr;
}

INLINE_MEMBER int * IVectorN::getLowerPtr()
{
    return valuePtr + lower;
}

INLINE_MEMBER int * IVectorN::getUpperPtr()
{
    return valuePtr + lower + length;
}

INLINE_MEMBER const int * IVectorN::getValuePtr() const
{
    return valuePtr;
}

INLINE_MEMBER const int * IVectorN::getLowerPtr() const
{
    return valuePtr + lower;
}

INLINE_MEMBER const int * IVectorN::getUpperPtr() const
{
    return valuePtr + lower + length;
}

INLINE_MEMBER IVectorN::IVectorN(int newLength)
:
    lower(0),
    length(newLength)
{
    make();
}

INLINE_MEMBER IVectorN::IVectorN(int newLower, int newLength)
:
    lower(newLower),
    length(newLength)
{
    make();
}

INLINE_MEMBER IVectorN::IVectorN(
	const int * values,
	int newLower,
	int newLength
	)
:
    lower(newLower),
    length(newLength)
{
    make();
#ifdef __GNUC__
    copy((int *)values,getLength(),getLowerPtr());
    //   ^^^^^^^^^^
    //  this was put there due to a bug in gcc 2.5.8
    //
#else
    copy(values,getLength(),getLowerPtr());
#endif
}

INLINE_MEMBER IVectorN::IVectorN(
	const int * values,
	int newLength
	)
:
    lower(0),
    length(newLength)
{
    make();
#ifdef __GNUC__
    copy((int *)values,getLength(),getLowerPtr());
    //   ^^^^^^^^^^
    //  this was put there due to a bug in gcc 2.5.8
    //
#else
    copy(values,getLength(),getLowerPtr());
#endif
}

INLINE_MEMBER IVectorN::IVectorN(const IVectorN & vector)
:
    lower(vector.getLower()),
    length(vector.getLength())
{
    make();
#ifdef __GNUC__
    copy((int*)vector.getLowerPtr(),getLength(),getLowerPtr());
    //   ^^^^^^^^^^
    //  this was put there due to a bug in gcc 2.5.8
    //
#else
    copy(vector.getLowerPtr(),getLength(),getLowerPtr());
#endif
}


INLINE_MEMBER int IVectorN::operator[](int index) const
{
    return valuePtr[index];
}

INLINE_MEMBER int & IVectorN::operator[](int index)
{
    return valuePtr[index];
}

INLINE_MEMBER int IVectorN::get(int index) const
{
    return valuePtr[index];
}

INLINE_MEMBER void IVectorN::set(int index, int value)
{
    valuePtr[index] = value;
}

INLINE_MEMBER IVectorN & IVectorN::operator=(const IVectorN & vector)
{
    if (length != vector.getLength())
    {
	kill();
	lower = vector.getLower();
	length = vector.getLength();
	make();
    }
    else if (lower != vector.getLower())
    {
	valuePtr += lower - vector.getLower();
	lower = vector.getLower();
    }
#ifdef __GNUC__
    copy((int*)vector.getLowerPtr(),getLength(),getLowerPtr());
    //   ^^^^^^^^^^
    //  this was put there due to a bug in gcc 2.5.8
    //
#else
    copy(vector.getLowerPtr(),getLength(),getLowerPtr());
#endif
    return *this;
}

INLINE_MEMBER SubIVectorN::SubIVectorN(
	IVectorN & vector
	)
:
    IVectorN()
{
    lower = vector.getLower();
    length = vector.getLength();
    valuePtr = vector.getValuePtr();
}

INLINE_MEMBER SubIVectorN::SubIVectorN(
	IVectorN & vector,
	int subLower,
	int subLength
	)
:
    IVectorN()
{
    lower = vector.getLower();
    length = subLength;
    valuePtr = vector.getValuePtr() + subLower - lower;
}

INLINE_MEMBER SubIVectorN::SubIVectorN(
	IVectorN & vector,
	int subLower,
	int subLength,
	int newLower
	)
:
    IVectorN()
{
    lower = newLower;
    length = subLength;
    valuePtr = vector.getValuePtr() + subLower - newLower;
}

INLINE_MEMBER SubIVectorN::SubIVectorN(
	int * values,
	int newLength
	)
:
    IVectorN()
{
    lower = 0;
    length = newLength;
    valuePtr = values;
}

INLINE_MEMBER SubIVectorN::SubIVectorN(
	int * values,
	int newLower,
	int newLength
	)
:
    IVectorN()
{
    lower = newLower;
    length = newLength;
    valuePtr = values - newLower;
}

INLINE_MEMBER IMatrixNM::IMatrixNM()
:
    lowerN(0),
    lowerM(0),
    deltaN(0),
    lengthN(0),
    lengthM(0),
    valuePtr(NIL(int*))
{
}

INLINE_MEMBER void IMatrixNM::make()
{
    if (lengthN > 0 && lengthM > 0)
    {
	deltaN = lengthM;
	valuePtr = (new int[lengthN * lengthM]) - lowerN * deltaN - lowerM;
    }
    else
    {
	valuePtr = NIL(int *);
    }
}

INLINE_MEMBER void IMatrixNM::kill()
{
    if (isNotNIL(valuePtr))
    {
	delete [] (valuePtr + lowerN * deltaN + lowerM);
    }
}

INLINE_MEMBER IMatrixNM::IMatrixNM(
	int newLengthN,
	int newLengthM
	)
:
    lowerN(0), lowerM(0),
    lengthN(newLengthN), lengthM(newLengthM)
{
    make();
}

INLINE_MEMBER IMatrixNM::IMatrixNM(
    int newLowerN, int newLowerM,
    int newLengthN, int newLengthM
    )
:
    lowerN(newLowerN), lowerM(newLowerM),
    lengthN(newLengthN), lengthM(newLengthM)
{
    make();
}

INLINE_MEMBER int IMatrixNM::getLengthN() const { return lengthN; }
INLINE_MEMBER int IMatrixNM::getLengthM() const { return lengthM; }
INLINE_MEMBER int IMatrixNM::getLowerN() const { return lowerN; }
INLINE_MEMBER int IMatrixNM::getLowerM() const { return lowerM; }
INLINE_MEMBER int IMatrixNM::getUpperN() const { return lowerN + lengthN; }
INLINE_MEMBER int IMatrixNM::getUpperM() const { return lowerM + lengthM; }
INLINE_MEMBER int IMatrixNM::getDeltaN() const { return deltaN; }

INLINE_MEMBER int * IMatrixNM::getValuePtr()
{
    return valuePtr;
}

INLINE_MEMBER int * IMatrixNM::getLowerNptr()
{
    return valuePtr + lowerN * deltaN;
}

INLINE_MEMBER int * IMatrixNM::getUpperNptr()
{
    return valuePtr + (lowerN + lengthN) * deltaN;
}

INLINE_MEMBER int * IMatrixNM::getLowerMptr(int * value1ptr)
{
    return value1ptr + lowerM;
}

INLINE_MEMBER int * IMatrixNM::getUpperMptr(int * value1ptr)
{
    return value1ptr + lowerM + lengthM;
}

INLINE_MEMBER const int * IMatrixNM::getValuePtr() const
{
    return valuePtr;
}

INLINE_MEMBER const int * IMatrixNM::getLowerNptr() const
{
    return valuePtr + lowerN * deltaN;
}

INLINE_MEMBER const int * IMatrixNM::getUpperNptr() const
{
    return valuePtr + (lowerN + lengthN) * deltaN;
}

INLINE_MEMBER const int * IMatrixNM::getLowerMptr(
	const int * value1ptr
	) const
{
    return value1ptr + lowerM;
}

INLINE_MEMBER const int * IMatrixNM::getUpperMptr(
	const int * value1ptr
	) const
{
    return value1ptr + lowerM + lengthM;
}

INLINE_MEMBER IMatrixNM::IMatrixNM(
    const int * values,
    int newLowerN, int newLowerM,
    int newLengthN, int newLengthM
    )
:
    lowerN(newLowerN), lowerM(newLowerM),
    lengthN(newLengthN), lengthM(newLengthM)
{
    make();
    
    if (isNotNIL(valuePtr))
#ifdef __GNUC__
	copy((int*)values,
	//   ^^^^^^^^^^
	//  this was put there due to a bug in gcc 2.5.8
	//
		getLengthN() * getLengthM(),
		getLowerMptr(getLowerNptr()));
#else
	copy(values,
	     getLengthN() * getLengthM(),
	     getLowerMptr(getLowerNptr()));
#endif
}

INLINE_MEMBER IMatrixNM::IMatrixNM(
    const int * values,
    int newLengthN, int newLengthM
    )
:
    lowerN(0), lowerM(0),
    lengthN(newLengthN), lengthM(newLengthM)
{
    make();
    
    if (isNotNIL(valuePtr))
#ifdef __GNUC__
	copy((int*)values,
	//   ^^^^^^^^^^
	//  this was put there due to a bug in gcc 2.5.8
	//
		getLengthN() * getLengthM(),
		getLowerMptr(getLowerNptr()));
#else
	copy(values,
	     getLengthN() * getLengthM(),
	     getLowerMptr(getLowerNptr()));
#endif
}

INLINE_MEMBER const int * IMatrixNM::operator[](int index) const
{
    return valuePtr + index * deltaN;
}

INLINE_MEMBER int * IMatrixNM::operator[](int index)
{
    return valuePtr + index * deltaN;
}

/*

INLINE_MEMBER SubIVectorN SubIMatrixNM::getRowSubVectorN(int row)
{
    return SubIVectorN(getLowerMPtr(operator[](row)),getLengthM());
}

INLINE_MEMBER SubIVectorN SubIMatrixNM::getRowSubVectorN(
	int row,
	int colLower,
	int colLength,
	int newLower
	)
{
    return SubIVectorN(getLowerMPtr(operator[](row)) + colLower,
		     newLower,colLength);
}

*/

INLINE_MEMBER SubIMatrixNM::SubIMatrixNM(
	IMatrixNM & matrix
	)
:
    IMatrixNM()
{
    lowerN = matrix.getLowerN();
    lowerM = matrix.getLowerM();
    deltaN = matrix.getDeltaN();
    lengthN = matrix.getLengthN();
    lengthM = matrix.getLengthM();
    valuePtr = matrix.getValuePtr();
}

INLINE_MEMBER SubIMatrixNM::SubIMatrixNM(
	IMatrixNM & matrix,
	int subLowerN,
	int subLowerM,
	int subLengthN,
	int subLengthM
	)
:
    IMatrixNM()
{
    lowerN = matrix.getLowerN();
    lowerM = matrix.getLowerM();
    deltaN = matrix.getDeltaN();
    lengthN = subLengthN;
    lengthM = subLengthM;

    if (lengthN > 0 && lengthM > 0)
	valuePtr = matrix.getValuePtr()
		    + (subLowerN - lowerN) * deltaN
		    + subLowerM - lowerM;
}

INLINE_MEMBER SubIMatrixNM::SubIMatrixNM(
	IMatrixNM & matrix,
	int subLowerN,
	int subLowerM,
	int subLengthN,
	int subLengthM,
	int newLowerN,
	int newLowerM
	)
:
    IMatrixNM()
{
    lowerN = newLowerN;
    lowerM = newLowerM;
    deltaN = matrix.getDeltaN();
    lengthN = subLengthN;
    lengthM = subLengthM;

    if (lengthN > 0 && lengthM > 0)
	valuePtr = matrix.getValuePtr()
		    + (subLowerN - lowerN) * deltaN
		    + subLowerM - lowerM;
}



INLINE_MEMBER SubIMatrixNM::SubIMatrixNM(
	int * values,
	int newLowerN,
	int newLowerM,
	int newLengthN,
	int newLengthM
	)
:
    IMatrixNM()
{
    lowerN = newLowerN;
    lowerM = newLowerM;
    deltaN = newLengthM;
    lengthN = newLengthN;
    lengthM = newLengthM;

    if (lengthN > 0 && lengthM > 0)
	valuePtr = values - lowerN * deltaN - lowerM;
}

INLINE_MEMBER SubIMatrixNM::SubIMatrixNM(
	int * values,
	int newLengthN,
	int newLengthM
	)
:
    IMatrixNM()
{
    lowerN = 0;
    lowerM = 0;
    deltaN = newLengthM;
    lengthN = newLengthN;
    lengthM = newLengthM;


    if (lengthN > 0 && lengthM > 0)
	valuePtr = values - lowerN * deltaN - lowerM;
}


} // extern "C++"

// ===========================================================================
