// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       IVectorN
//  TYPE:       c++ inline
//  PROJECT:    VEGA Math Library
//  CONTENT:    vector and matrix of ints
//  VERSION:    0.6.0
// ===========================================================================
//  AUTHORS:    rft     Robert F. Tobler
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 14:06:28  rft     last modification
//  28-jul-94 15:24:07  rft     created
// ===========================================================================

extern "C++"
{

INLINE_MEMBER IVectorN & IVectorN::fill(int value)
{
    int * pNow = getLowerPtr();
    int * pEnd = getUpperPtr();

    while (pNow < pEnd)
	*pNow++ = value;

    return *this;
}

INLINE_MEMBER IVectorN & IVectorN::operator*=(int scalar)
{
    int * pNow = getLowerPtr();
    int * pEnd = getUpperPtr();

    while (pNow < pEnd)
	*pNow++ *= scalar;

    return *this;
}

INLINE_MEMBER IVectorN & IVectorN::operator+=(const IVectorN & vector)
{
    CH_IVECTORN_CODE(
	EH_FATAL_CONDITION(IVectorN_errors,getLength() != vector.getLength());
    )

    int * pNow = getLowerPtr();
    int * pEnd = getUpperPtr();
    const int * pOne = vector.getLowerPtr();

    while (pNow < pEnd)
	*pNow++ += *pOne++;

    return *this;
}

INLINE_MEMBER IVectorN & IVectorN::operator-=(const IVectorN & vector)
{
    CH_IVECTORN_CODE(
	EH_FATAL_CONDITION(IVectorN_errors,getLength() != vector.getLength());
    )

    int * pNow = getLowerPtr();
    int * pEnd = getUpperPtr();
    const int * pOne = vector.getLowerPtr();

    while (pNow < pEnd)
	*pNow++ -= *pOne++;

    return *this;
}

// sfmap: protect: double

INLINE_GLOBAL double dotProduct(const IVectorN & vector1, const IVectorN & vector2)
{
    CH_IVECTORN_CODE(
	EH_FATAL_CONDITION(IVectorN_errors,
			   vector1.getLength() != vector2.getLength());
    )

    // sfmap: protect: double

    double result = 0;

    const int * pOne = vector1.getLowerPtr();
    const int * pEnd = vector1.getUpperPtr();
    const int * pTwo = vector2.getLowerPtr();


    // sfmap: protect: double double

    while (pOne < pEnd)
	result += (double)(*pOne++) * (double)(*pTwo++);

    return result;
}

INLINE_GLOBAL IVectorN operator-(const IVectorN & vector)
{
    IVectorN result = IVectorN(vector.getLower(), vector.getLength());

    int * pNow = result.getLowerPtr();
    int * pEnd = result.getUpperPtr();
    const int * pOne = vector.getLowerPtr();

    while (pNow < pEnd)
	*pNow++ = - *pOne++;

    return result;
}

INLINE_GLOBAL IVectorN operator+(const IVectorN & vector1, const IVectorN & vector2)
{
    CH_IVECTORN_CODE(
	EH_FATAL_CONDITION(IVectorN_errors,
			   vector1.getLength() != vector2.getLength());
	EH_FATAL_CONDITION(IVectorN_errors,
			   vector1.getLower() != vector2.getLower());
    )

    IVectorN result = IVectorN(vector1.getLower(),vector1.getLength());

    int * pNow = result.getLowerPtr();
    int * pEnd = result.getUpperPtr();
    const int * pOne = vector1.getLowerPtr();
    const int * pTwo = vector2.getLowerPtr();

    while (pNow < pEnd)
	*pNow++ = *pOne++ + *pTwo++;

    return result;
}

INLINE_GLOBAL IVectorN operator-(const IVectorN & vector1, const IVectorN & vector2)
{
    CH_IVECTORN_CODE(
	EH_FATAL_CONDITION(IVectorN_errors,
			   vector1.getLength() != vector2.getLength());
	EH_FATAL_CONDITION(IVectorN_errors,
			   vector1.getLower() != vector2.getLower());
    )

    IVectorN result = IVectorN(vector1.getLower(),vector1.getLength());

    int * pNow = result.getLowerPtr();
    int * pEnd = result.getUpperPtr();
    const int * pOne = vector1.getLowerPtr();
    const int * pTwo = vector2.getLowerPtr();

    while (pNow < pEnd)
	*pNow++ = *pOne++ - *pTwo++;

    return result;
}

INLINE_GLOBAL IVectorN operator*(int scalar, const IVectorN & vector)
{
    IVectorN result = IVectorN(vector.getLower(), vector.getLength());

    int * pNow = result.getLowerPtr();
    int * pEnd = result.getUpperPtr();
    const int * pOne = vector.getLowerPtr();

    while (pNow < pEnd)
	*pNow++ = scalar * *pOne++;

    return result;
}

INLINE_GLOBAL IVectorN operator*(const IVectorN & vector, int scalar)
{
    IVectorN result = IVectorN(vector.getLower(), vector.getLength());

    int * pNow = result.getLowerPtr();
    int * pEnd = result.getUpperPtr();
    const int * pOne = vector.getLowerPtr();

    while (pNow < pEnd)
	*pNow++ = *pOne++ * scalar;

    return result;
}

/*

INLINE_GLOBAL static void siftDownInHeap(
	int top, int * upperPtr, int t, int end
	)
{
    int i = t;
    int j = t+t;
    while (j >= end)
    {
	if (j > end && upperPtr[j] > upperPtr[j-1]) j--;
	if (top > upperPtr[j])
	{
	    upperPtr[i] = upperPtr[j];
	    i = j;
	    j <<= 1;
	}
	else
	{
	    j = end - 1;
	}
    }
    upperPtr[i] = top;
}

INLINE_GLOBAL IVectorN sort()
{
    int * upperPtr = getUpperPtr();
    
    int end = -getLength();
    
    for (int t = end/2; t < 0; t++)
    {
	siftDownInHeap(upperPtr[t],upperPtr,t,end);
    }

    int top;
    
    while (end < -1)
    {
	top = upperPtr[end];
	upperPtr[end] = upperPtr[-1];
	end++;
	siftDownInHeap(top,upperPtr,-1,end);
    }
}

*/

INLINE_MEMBER SubIVectorN & SubIVectorN::operator=(const IVectorN & vector)
{
    CH_IVECTORN_CODE(
	EH_FATAL_CONDITION(IVectorN_errors,getLength() != vector.getLength());
    )

#ifdef __GNUC__
    copy((int*)vector.getLowerPtr(),getLength(),getLowerPtr());
    //   ^^^^^^^^^^
    //  this was put there due to a bug in gcc 2.5.8
    //
#else
    copy(vector.getLowerPtr(),getLength(),getLowerPtr());
#endif
    return *this;
}

INLINE_MEMBER IMatrixNM::IMatrixNM(const IMatrixNM & matrix)
:
    lowerN(matrix.getLowerN()),
    lowerM(matrix.getLowerM()),
    lengthN(matrix.getLengthN()),
    lengthM(matrix.getLengthM())
{
    make();

    int * pNowN = getLowerNptr();
    int * pEndN = getUpperNptr();
    const int * pOneN = matrix.getLowerNptr();

    while (pNowN < pEndN)
    {
#ifdef __GNUC__
	copy((int*)matrix.getLowerMptr(pOneN),
	//   ^^^^^^^^^^
	//  this was put there due to a bug in gcc 2.5.8
	//
	     getLengthM(),
	     getLowerMptr(pNowN));
#else
	copy(matrix.getLowerMptr(pOneN),
	     getLengthM(),
	     getLowerMptr(pNowN));
#endif
	pNowN += getDeltaN();
	pOneN += matrix.getDeltaN();
    }
}

INLINE_MEMBER IMatrixNM & IMatrixNM::fill(int value)
{
    int * pNowN = getLowerNptr();
    int * pEndN = getUpperNptr();
    
    while (pNowN < pEndN)
    {
	int * pNowM = getLowerMptr(pNowN);
	int * pEndM = getUpperMptr(pNowN);

	while (pNowM < pEndM)
	    *pNowM++ = value;

	pNowN += getDeltaN();
    }
    
    return *this;
}

INLINE_MEMBER IMatrixNM & IMatrixNM::operator=(const IMatrixNM & matrix)
{
    if (this == &matrix)
	return *this;

    kill();
    lowerN = matrix.getLowerN();
    lowerM = matrix.getLowerM();
    lengthN = matrix.getLengthN();
    lengthM = matrix.getLengthM();
    make();

    int * pNowN = getLowerNptr();
    int * pEndN = getUpperNptr();
    const int * pOneN = matrix.getLowerNptr();

    while (pNowN < pEndN)
    {
#ifdef __GNUC__
	copy((int*)matrix.getLowerMptr(pOneN),
	//   ^^^^^^^^^^
	//  this was put there due to a bug in gcc 2.5.8
	//
	     getLengthM(),
	     getLowerMptr(pNowN));
#else
	copy(matrix.getLowerMptr(pOneN),
	     getLengthM(),
	     getLowerMptr(pNowN));
#endif
	pNowN += getDeltaN();
	pOneN += matrix.getDeltaN();
    }

    return *this;
}

INLINE_MEMBER IMatrixNM & IMatrixNM::operator*=(int scalar)
{
    int * pNowN = getLowerNptr();
    int * pEndN = getUpperNptr();
    
    while (pNowN < pEndN)
    {
	int * pNowM = getLowerMptr(pNowN);
	int * pEndM = getUpperMptr(pNowN);

	while (pNowM < pEndM)
	    *pNowM++ *= scalar;

	pNowN += getDeltaN();
    }
    
    return *this;
}

INLINE_MEMBER IMatrixNM & IMatrixNM::operator+=(const IMatrixNM & matrix)
{
    CH_IVECTORN_CODE(
	EH_FATAL_CONDITION(IVectorN_errors,getLengthN() != matrix.getLengthN());
	EH_FATAL_CONDITION(IVectorN_errors,getLengthM() != matrix.getLengthM());
    )

    int * pNowN = getLowerNptr();
    int * pEndN = getUpperNptr();
    const int * pOneN = matrix.getLowerNptr();
    
    while (pNowN < pEndN)
    {
	int * pNowM = getLowerMptr(pNowN);
	int * pEndM = getUpperMptr(pNowN);
	const int * pOneM = matrix.getLowerMptr(pOneN);

	while (pNowM < pEndM)
	    *pNowM++ += *pOneM++;

	pNowN += getDeltaN();
	pOneN += matrix.getDeltaN();
    }
    
    return *this;
}

INLINE_MEMBER IMatrixNM & IMatrixNM::operator-=(const IMatrixNM & matrix)
{
    CH_IVECTORN_CODE(
	EH_FATAL_CONDITION(IVectorN_errors,getLengthN() != matrix.getLengthN());
	EH_FATAL_CONDITION(IVectorN_errors,getLengthM() != matrix.getLengthM());
    )

    int * pNowN = getLowerNptr();
    int * pEndN = getUpperNptr();
    const int * pOneN = matrix.getLowerNptr();
    
    while (pNowN < pEndN)
    {
	int * pNowM = getLowerMptr(pNowN);
	int * pEndM = getUpperMptr(pNowN);
	const int * pOneM = matrix.getLowerMptr(pOneN);

	while (pNowM < pEndM)
	    *pNowM++ -= *pOneM++;

	pNowN += getDeltaN();
	pOneN += matrix.getDeltaN();
    }
    
    return *this;
}

INLINE_MEMBER SubIMatrixNM & SubIMatrixNM::operator=(const IMatrixNM & matrix)
{
    CH_IVECTORN_CODE(
	EH_FATAL_CONDITION(IVectorN_errors,getLengthN() != matrix.getLengthN());
	EH_FATAL_CONDITION(IVectorN_errors,getLengthM() != matrix.getLengthM());
    )

    int * pNowN = getLowerNptr();
    int * pEndN = getUpperNptr();
    const int * pOneN = matrix.getLowerNptr();

    while (pNowN < pEndN)
    {
#ifdef __GNUC__
	copy((int*)matrix.getLowerMptr(pOneN),
	//   ^^^^^^^^^^
	//  this was put there due to a bug in gcc 2.5.8
	//
	     getLengthM(),
	     getLowerMptr(pNowN));
#else
	copy(matrix.getLowerMptr(pOneN),
	     getLengthM(),
	     getLowerMptr(pNowN));
#endif
	pNowN += getDeltaN();
	pOneN += matrix.getDeltaN();
    }

    return *this;
}

} // extern "C++"

// ===========================================================================
