// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       Math
//  TYPE:       c++ header
//  PROJECT:    VEGA Math Library
//  CONTENT:    mathematical functions
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    rft     Robert F. Tobler
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 14:06:28  rft     last modification
//  28-Aug-93           rft     created
// ===========================================================================
//  MANUAL:
//
//  DESCRIPTION
//      This module defines mathematical constants and functions. It uses the
//      ANSI standard header <math.h>, but adds additional functionality.
//
//      The mathematical constants defined in this module are macros named
//      according to the following schemes:
//
//              MATH_<val>
//              MATH_<unop>_<val>
//              MATH_<val>_<binop>_<val>
//              MATH_<val>_<binop>_<unop>_<val>
//
//      Valid substitutions for <val> are:
//
//              1  2  3  4  PI  E  GOLDEN
//
//      The symbol GOLDEN denotes the value of the golden ratio:
//      (1 + sqrt(5))/2.
//
//      Valid substitutions for <unop> are:
//
//              LOG2  LOG10  LN  SQRT
//
//      Valid substitutions for <binop> are:
//
//              DIV  MUL
//
//      Not all of the valid combinations are defined yet. If you need an
//      additional mathematical constant contact the library administrator.
//
//      Two constants that do not follow this scheme are:
//
//              MATH_DEG_TO_RAD
//              MATH_RAD_TO_DEG
//
//      These allow to convert between radians and degrees.
//
//  SEE ALSO
//      VEGA, MathLibrary
// ===========================================================================
#ifndef _MATH_MATH_H_
#define _MATH_MATH_H_

extern "C"
{
#include <math.h>
}

#include "VEGA/Math/globals.h"

#ifdef __GNUC__
#pragma interface
#endif

extern "C++"
{

// ===========================================================================

#define MATH_1_DIV_3            0.3333333333333333333333333333333333333333
#define MATH_E                  2.718281828459045235360287471352662497757
#define MATH_LOG2_E             1.442695040888963407359924681001892137427
#define MATH_LOG10_E            0.4342944819032518276511289189166050822944
#define MATH_PI                 3.141592653589793238462643383279502884197
#define MATH_2_MUL_PI           6.28318530717958647692528676655900576839
#define MATH_PI_DIV_2           1.570796326794896619231321691639751442099
#define MATH_PI_DIV_3           1.047197551196597746154214461093167628066
#define MATH_PI_DIV_4           0.785398163397448309615660845819875721049
#define MATH_1_DIV_PI           0.3183098861837906715377675267450287240689
#define MATH_2_DIV_PI           0.636619772367581343075535053490057448138
#define MATH_2_DIV_SQRT_PI      1.128379167095512573896158903121545171688
#define MATH_SQRT_2             1.41421356237309504880168872420969807857
#define MATH_1_DIV_SQRT_2       0.7071067811865475244008443621048490392848
#define MATH_SQRT_3             1.732050807568877293527446341505872366943
#define MATH_LN_2               0.6931471805599453094172321214581765680755
#define MATH_LN_10              2.302585092994045684017991454684364207601
#define MATH_1_DIV_LN_2         1.442695040888963407359924681001892137427
#define MATH_1_DIV_LN_10        0.4342944819032518276511289189166050822944
#define MATH_DEG_TO_RAD         0.01745329251994329576923690768488612713443
#define MATH_RAD_TO_DEG         57.2957795130823208767981548141051703324
#define MATH_GOLDEN             1.61803398874989484820458683436563811772
#define MATH_1_DIV_GOLDEN       0.61803398874989484820458683436563811772

#define DEGREES                 * MATH_DEG_TO_RAD
        //
        //  This defines a postfix operator with the precedence of the
        //  multiplication operator, that converts degrees into radians.
	//
	//  Use it like this:
	//
	//      cos(45 DEGREES)
        //
	//  But beware:
	//
	//      cos(40 + 5 DEGREES)
	//
	//  means 40 RADIANS + 5 DEGREES

#define DEGREE			DEGREES
	//
	//  Companion to the DEGREES macro to allow 'cos(1 DEGREE)'.
	//

#define RADIANS
	//
	//  This is defined to allow expressions like:
	//
	//      cos(40 RADIANS + 5 DEGREES)
	//

#define RADIAN			RADIANS
	//
	//  Companion to the RADIANS macro to allow 'cos(1 RADIAN)'.
	//

#define FLOAT_TINY              4.76837158203125E-7
        //
        //  This is a small float that is used in int isTiny(float x) which
        //  should be used instead of comparison against zero.
        //
	//  For  floats this constant was set to 2^-21, resulting in
	//  FLOAT_TINY = 4 * float_eps.
	//

#define FLOAT_HUGE              2.147483648E+9
        //
        //  This is a huge float that can be regarded as infinity for this
	//  type of floating point arithmetic.
        //
	//  For  floats this constant was set to 2^31, so that multiplying
	//  its exponent by 4 yields a float that is still valid.
	//

#define DOUBLE_TINY             8.8817841970012523233891E-16
        //
        //  This is a small double that is used in int isTiny(double x) which
        //  should be used instead of comparison against zero.
        //
	//  For  IEEE floats this constant was set to 2^-50, resulting in
	//  DOUBLE_TINY = 4 * double_eps.
	//

#define DOUBLE_HUGE             5.78960446186580977117855E+76
        //
        //  This is a huge double that can be regarded as infinity for this
	//  type of floating point arithmetic.
        //
	//  For  doubles this constant was set to 2^255, so that
	//  multiplying its exponent by 4 yields a double that is still valid.
	//

extern float float_eps;
	//
	//  This is the machine epsilon for floats on this machine.
	//

extern float float_onePlusEps;

extern double double_eps;
	//
	//  This is the machine epsilon for doubles on this machine.
	//

extern double double_onePlusEps;

float  abs(
	float  a
	);
double abs(
	double a
	);
float  max(
	float  a,
	float  b
	);
double max(
	double a,
	double b
	);
float  min(
	float  a,
	float  b
	);
double min(
	double a,
	double b
	);

int sign(
	float x
	);
int sign(
	double x
	);
float square(
	float x
	);
float cube(
	float x
	);
double square(
	double x
	);
double cube(
	double x
	);
double round(
	double x
	);
double fract(
	double x
	);

#ifdef __BCPLUSPLUS__
double cbrt(
	double x
	);
#endif


float clamp(
	float x,
	float lo,
	float hi
	);
double clamp(
	double x,
	double lo,
	double hi
	);

// ---------------------------------------------------------------------------
//  The following functions compare the absolute value of a float/double
//  against a small float/double constant (FLOAT_TINY/DOUBLE_TINY). They
//  should be used instad of comparisons against 0.0.
//

int isTiny(
	float x
	);
int isTiny(
	double x
	);

// ---------------------------------------------------------------------------
//  The following functions compare the absolute value of a float/double
//  against a huge float/double constant (FLOAT_HUGE/DOUBLE_HUGE).
//

int isHuge(
	float x
	);
int isHuge(
	double x
	);

// ---------------------------------------------------------------------------
//  Calculate the determinant of a 3x3 matrix.
//

double det2x2(
        double x00, double x01,
        double x10, double x11
        );

// ---------------------------------------------------------------------------
//  Calculate the determinant of a 3x3 matrix in the form
//
//      | x00,  x01,  x02 |
//      | x10,  x11,  x12 |
//      | x20,  x21,  x22 |
//

double det3x3(
        double x00, double x10, double x20,
        double x01, double x11, double x21,
        double x02, double x12, double x22
        );

class SoftMath
{
public:

    static double acos(
	double x
	);
    static double asin(
	double x
	);
    static double atan(
	double x
	);
    static double atan2(
	double y,
	double x
	);
    static double cos(
	double x
	);
    static double sin(
	double x
	);
    static double tan(
	double x
	);
    static double cosh(
	double x
	);
    static double sinh(
	double x
	);
    static double tanh(
	double x
	);
    static double exp(
	double x
	);
    static double log(
	double x
	);
    static double log10(
	double x
	);
    static double pow(
	double x,
	double y
	);
    static double floor(
	double x
	);
    static double fmod(
	double x,
	double y
	);
    static double fabs(
	double x
	);

};

} // extern "C++"

#ifndef MATH_NO_INLINE

#define INLINE_GLOBAL COMPILER_INLINE_GLOBAL
#define INLINE_MEMBER COMPILER_INLINE_MEMBER

#include "VEGA/Math/math.ipp"

#undef INLINE_GLOBAL
#undef INLINE_MEMBER

#endif

#endif // _MATH_MATH_H_
// ===========================================================================
