// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       SimplePolynomial
//  TYPE:       c++ code
//  PROJECT:    VEGA Math Library
//  CONTENT:    manipulation routines for single variable polynomials
//  VERSION:    0.4.0
// ===========================================================================
//  AUTHORS:    rft     Robert F. Tobler
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 14:06:28  rft     last modification
//  28-Aug-93           rft     created
// ===========================================================================

#include "VEGA/Math/math.h"

#include "VEGA/Math/smplpol.h"

SimplePolynomial::SimplePolynomial(
    int newOrder,
    const double * newCoeffPtr
    )
:
    order(newOrder),
    coefficientTable(new double[newOrder])
{
#ifdef __GNUC__
    copy((double *)newCoeffPtr,newOrder, coefficientTable);
    //   ^^^^^^^^^^
    //  this was put there due to a bug in gcc 2.5.8
    //
#else
    copy(newCoeffPtr,newOrder, coefficientTable);
#endif
}

SimplePolynomial::~SimplePolynomial()
{
    if (isNotNIL(coefficientTable))
	delete[] coefficientTable;
}

double SimplePolynomial::evaluateAt(double x)
{
    double f = coefficientTable[order];

    for (int i = order - 1; i >= 0; i--)
    {
	f = f * x + coefficientTable[i];
    }

    return f;
}

int SimplePolynomial::getOrder() const { return order; }


VectorN SimplePolynomial::uniqueRoots()
{
    int n;
    double roots[4];

    switch (getOrder())
    {
        case 0: n = 0; break;
	case 1:	n = uniqueRootsO1(coefficientTable,roots); break;
	case 2: n = uniqueRootsO2(coefficientTable,roots); break;
	case 3: n = uniqueRootsO3(coefficientTable,roots); break;
	case 4: n = uniqueRootsO4(coefficientTable,roots); break;
	default:
	    n = 0;
	    EH_ERROR(SimplePolynomial_errors,ERROR_POLYNOMIAL_ORDER_TOO_HIGH,
		     "SimplePolynomial::uniqueRoots() too complex");
    }
    return VectorN(roots,n);
}

/* TODO:

VectorN SimplePolynomial::uniqueRoots(Range range)
{
    if (getOrder() < 5)
	return uniqueRoots().insideRange(range);
    else
	return sturmSequenceRoots(range);
}
*/

// ========================================= Unique Roots of SimplePolynomials

int SimplePolynomial::uniqueRootsO1(const double * pol, double * root)
{
    CH_CODE(EH_FATAL_CONDITION(SimplePolynomial_errors,pol == 0));
    CH_CODE(EH_FATAL_CONDITION(SimplePolynomial_errors,root == 0));

    if (isTiny(pol[1]))				// constant ?
	return (0);

    root[0] = -pol[0] / pol[1];			// solve linear equation
    return (1);
}

int SimplePolynomial::uniqueRootsN2(const double *pol, double *root)
{
    CH_CODE(EH_FATAL_CONDITION(SimplePolynomial_errors,pol == 0));
    CH_CODE(EH_FATAL_CONDITION(SimplePolynomial_errors,root == 0));

    double p2 = pol[1] / 2.0;
    double d = p2 * p2 - pol[0];

    if (isTiny(d) || d < 0.0)
        return (0);

    if (p2 > 0.0)                               // prevent cancellation
    {
        root[0] = -(p2 + sqrt(d));
        root[1] = pol[0] / root[0];             // VIETA
    }
    else
    {
        root[1] = sqrt(d) - p2;
        root[0] = pol[0] / root[1];             // VIETA
    }
    return (2);
}

int SimplePolynomial::uniqueRootsO2(const double *pol, double *root)
{
    CH_CODE(EH_FATAL_CONDITION(SimplePolynomial_errors,pol == 0));
    CH_CODE(EH_FATAL_CONDITION(SimplePolynomial_errors,root == 0));

    if (isTiny(pol[2]))                         // linear ?
    {
        if (isTiny(pol[1]))                     // constant ?
            return (0);

        root[0] = -pol[0] / pol[1];             // solve linear equation
        return (1);
    }
    else                                        // solve quadric equation
    {
        double d = pol[1] * pol[1] - 4 * pol[2] * pol[0];

        if (isTiny(d) || d < 0.0)               // double or imaginary roots ?
            return (0);

        if (pol[1] > 0.0)                       // prevent cancellation
        {
            d = -(sqrt(d) + pol[1]);
            if (pol[2] > 0.0)
            {
                root[0] = d / (pol[2] + pol[2]); // VIETA
                root[1] = (pol[0] + pol[0]) / d;
            }
            else
            {
                root[1] = d / (pol[2] + pol[2]); // VIETA
                root[0] = (pol[0] + pol[0]) / d;
            }
        }
        else
        {
            d = sqrt(d) - pol[1];
            if (pol[2] < 0.0)
            {
                root[0] = d / (pol[2] + pol[2]); // VIETA
                root[1] = (pol[0] + pol[0]) / d;
            }
            else
            {
                root[1] = d / (pol[2] + pol[2]); // VIETA
                root[0] = (pol[0] + pol[0]) / d;
            }
        }
        return (2);
    }
}

double SimplePolynomial::oneRootN3(const double *pol)
{
    CH_CODE(EH_FATAL_CONDITION(SimplePolynomial_errors,pol == 0));

    double  d, p, q, p3;

    // ------- polynom is normalized: x^3 + pol[2] x^2 + pol[1] x + pol[0] = 0

    // ---------- eliminate quadric term (x = y - pol[2]/3): x^3 + p x + q = 0
    d = pol[2] * pol[2];

    p = 1.0/3.0 * (-1.0/3.0 * d + pol[1]);
    q = 1.0/2.0 * ((2.0/27.0 * d - 1.0/3.0 * pol[1]) * pol[2] + pol[0]);

    // ------------------------------------------------- use Cardano's formula
    p3 = p * p * p;
    d = q * q + p3;

    if (isTiny(d))
    {
        if (isTiny(q))                          // one triple solution
            return (-1.0/3.0 * pol[2]);
        else                                    // one single and one double 
        {                                       // solution
            q = cbrt(-q);

            return (q + q - 1.0/3.0 * pol[2]);
        }
    }
    else if (d < 0.0)                           // Casus irreducibilis:
    {                                           // three real solutions

        return (2.0 * sqrt(-p) * cos(1.0/3.0 * acos(-q / sqrt(-p3)))
                                        - 1.0/3.0 * pol[2]);
    }
    else
    {
        d = sqrt(d);                            // use d as temporary

        return (cbrt(d - q) - cbrt(d + q) - 1.0/3.0 * pol[2]);
    }
}



int SimplePolynomial::uniqueRootsN3(const double *pol, double *root)
{
    CH_CODE(EH_FATAL_CONDITION(SimplePolynomial_errors,pol == 0));
    CH_CODE(EH_FATAL_CONDITION(SimplePolynomial_errors,root == 0));

    double  d, p, q, p3;

    // ------- polynom is normalized: x^3 + pol[2] x^2 + pol[1] x + pol[0] = 0

    // ---------- eliminate quadric term (x = y - pol[2]/3): x^3 + p x + q = 0
    d = pol[2] * pol[2];

    p = 1.0/3.0 * (-1.0/3.0 * d + pol[1]);
    q = 1.0/2.0 * ((2.0/27.0 * d  - 1.0/3.0 * pol[1]) * pol[2] + pol[0]);

    // ------------------------------------------------- use Cardano's formula
    p3 = p * p * p;
    d = q * q + p3;

    if (isTiny(d))
    {
        if (isTiny(q))                          // one triple solution
        {
            root[0] = - 1.0/3.0 * pol[2];
            return (1);
        }
        else                                    // one single and one double
        {                                       // solution
            double u;

            u = cbrt(-q);

            root[0] = u + u - 1.0/3.0 * pol[2];
            return (1);
        }
    }
    else if (d < 0.0)                           // Casus irreducibilis:
    {                                           // three real solutions
        double phi;
        double t;

        phi = 1.0/3.0 * acos(-q / sqrt(-p3));
        t = 2.0 * sqrt(-p);

        root[0] =   t * cos(phi)                 - 1.0/3.0 * pol[2];
        root[1] = - t * cos(phi + MATH_PI_DIV_3) - 1.0/3.0 * pol[2];
        root[2] = - t * cos(phi - MATH_PI_DIV_3) - 1.0/3.0 * pol[2];
        sort3ascending(root);
        return (3);
    }
    else
    {
        d = sqrt(d);                            // use d as temporary

        root[0] = cbrt(d - q) - cbrt(d + q) - 1.0/3.0 * pol[2];
        return (1);
    }
}

int SimplePolynomial::uniqueRootsO3(const double *pol, double *root)
{
    CH_CODE(EH_FATAL_CONDITION(SimplePolynomial_errors,pol == 0));
    CH_CODE(EH_FATAL_CONDITION(SimplePolynomial_errors,root == 0));
    CH_CODE(EH_FATAL_CONDITION(SimplePolynomial_errors,isTiny(pol[3])));

    double  a, b, c, d, p, q, p3;

    // -------------------------------- normal form: x^3 + a x^2 + b x + c = 0
    a = pol[2] / pol[3];
    b = pol[1] / pol[3];
    c = pol[0] / pol[3];

    // --------------- eliminate quadric term (x = y - a/3): x^3 + p x + q = 0
    d = a * a;

    p = 1.0/3.0 * (-1.0/3.0 * d + b);
    q = 1.0/2.0 * ((2.0/27.0 * d - 1.0/3.0 * b) * a + c);

    // ------------------------------------------------- use Cardano's formula
    p3 = p * p * p;
    d = q * q + p3;

    if (isTiny(d))
    {
        if (isTiny(q))                          // one triple solution
        {
            root[0] = - 1.0/3.0 * a;
            return (1);
        }
        else                                    // one single and one double
        {                                       // solution
            double u;

            u = cbrt(-q);

            root[0] = u + u - 1.0/3.0 * a;
            return (1);
        }
    }
    else if (d < 0.0)                           // Casus irreducibilis:
    {                                           // three real solutions
        double phi;
        double t;

        phi = 1.0/3.0 * acos(-q / sqrt(-p3));
        t = 2.0 * sqrt(-p);

        root[0] =   t * cos(phi)                 - 1.0/3.0 * a;
        root[1] = - t * cos(phi + MATH_PI_DIV_3) - 1.0/3.0 * a;
        root[2] = - t * cos(phi - MATH_PI_DIV_3) - 1.0/3.0 * a;
        sort3ascending(root);
        return (3);
    }
    else
    {
        d = sqrt(d);                            // use d as temporary

        root[0] = cbrt(d - q) - cbrt(d + q) - 1.0/3.0 * a;
        return (1);
    }
}

int SimplePolynomial::uniqueRootsO4(const double *pol, double *root)
{
    CH_CODE(EH_FATAL_CONDITION(SimplePolynomial_errors,pol == 0));
    CH_CODE(EH_FATAL_CONDITION(SimplePolynomial_errors,root == 0));
    CH_CODE(EH_FATAL_CONDITION(SimplePolynomial_errors,isTiny(pol[4])));

    double  poly[4];
    double  a,b,c,d,e,p,q,r,u,v,z;
    int    i,num,n1,n2;

    // ------------------------ normal form: x^4 + a x^3 + b x^2 + c x + d = 0
    a = pol[3] / pol[4];
    b = pol[2] / pol[4];
    c = pol[1] / pol[4];
    d = pol[0] / pol[4];

    // -------- eliminate cubic term (x = y - a /4): x^4 + p x^2 + q x + r = 0
    e = a * a;
    p = - 3.0/8.0 * e + b;
    q = (1.0/8.0 * e - 1.0/2.0 * b) * a + c;
    r = (1.0/16.0 * b - 3.0/256.0 * e) * e - 1.0/4.0 * a * c + d;

    if (isTiny(r))
    {
        // --------------------------- no absolute term: y (y^3 + p y + q) = 0
        poly[0] = q;
        poly[1] = p;
        poly[2] = 0.0;
        poly[3] = 1.0;

        num = uniqueRootsN3(poly,root);
        for (i = num; i > 0 && root[i - 1] > 0.0; i--)
            root[i] = root[i - 1];              // insert the 0.0 root
        root[i] = 0.0;
        num++;
    }
    else
    {
        // ------------------------------------- solve the resolvent cubic ...
        poly[0] = 1.0/2.0 * r * p - 1.0/8.0 * q * q;
        poly[1] = - r;
        poly[2] = -1.0/2.0 * p;
        poly[3] = 1.0;

        // ---------------------------- ... and take the one real solution ...
        z = oneRootN3(poly);

        // ---------------------------- ... to build two quadric equations ...
        u = z * z - r;
        v = 2.0 * z - p;

        if (isTiny(u))
            u = 0.0;
        else if (u > 0.0)
            u = sqrt(u);
        else
            return (0);

        if (isTiny(v))
            v = 0.0;
        else if (v > 0.0)
            v = sqrt(v);
        else
            return (0);

        root[0] = z - u;                // use the root field for polynomial
        root[1] = (q < 0.0) ? -v : v;
        root[2] = 1.0;

        n1 = uniqueRootsN2(root,poly);  // and the poly field for roots

        root[0] = z + u;                // the same misuse again
        root[1] = - root[1];

        n2 = uniqueRootsN2(root,poly + 2);
#ifdef __BCPLUSPLUS__
        num = mergeAscending((const double *)poly,n1,
			     (const double *)(poly+2),n2,
			     root);
#else
        num = mergeAscending(poly,n1,poly+2,n2,root);
#endif
    }
    e = 1.0/4.0 * a;

    for (i = 0; i < num; i++)
        root[i] -= e;

    return (num);
}

// ===========================================================================

