// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       SimplePolynomial
//  TYPE:       c++ header
//  PROJECT:    VEGA Math Library
//  CONTENT:    manipulation routines for single variable polynomials
//  VERSION:    0.4.0
// ===========================================================================
//  AUTHORS:    rft     Robert F. Tobler
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 14:06:28  rft     last modification
//  28-Aug-93           rft     created
// ===========================================================================
//  MANUAL:
//
//  DESCRIPTION
//      These routines work on SimplePolynomials which are stored in pairs of
//      length and coefficient array. The coefficients are stored in
//      ascending order of terms:
//
//              int      n;
//              double * pol;
//
//      describes the following SimplePolynomial:
//
//              pol[0] + pol[1] * x + pol[2] * x^2 + ... + pol[n] * x^n
//
//      In a number of cases the order n of the SimplePolynomial is fixed for
//      given routine, and therefore not passed as an explicit parameter.
//
//  SEE ALSO
//      VEGA, MathLibrary
// ===========================================================================
#ifndef _MATH_SIMPLEPOLYNOMIAL_H_
#define _MATH_SIMPLEPOLYNOMIAL_H_

#include <VEGA/Misc/ttools.h>

#include "VEGA/Math/globals.h"
#include "VEGA/Math/smplfun.h"
#include "VEGA/Math/vectorn.h"

#ifdef __GNUC__
#pragma interface
#endif

extern "C++"
{

#define ERROR_POLYNOMIAL_ORDER_TOO_HIGH	1

// ========================================================== class SimplePolynomial

class SimplePolynomial : public SimpleFunction
{
public:
    SimplePolynomial(
	int newOrder,
	const double * newCoeffPtr
	);

    virtual ~SimplePolynomial(
	);

    virtual double evaluateAt(
	double x
	);

    int getOrder(
	) const;


    // ----------------------------------------------- static member functions

    //
    //  Unique roots of SimplePolynomials:
    //
    //  These routines only calculate the unique real roots of real
    //  SimplePolynomials; they skip any multiple or complex roots.
    //

    VectorN uniqueRoots(
	);
    
    //
    //  normalized SimplePolynomials:
    //

    static int uniqueRootsN2(           // roots of 2nd order SimplePolynomial
        const double * pol,
        double * root
        );

    static int uniqueRootsN3(           // roots of 3rd order SimplePolynomial
        const double * pol,
        double * root
        );

    static double oneRootN3(            // one root of 3rd order SimplePol.
        const double * pol
        );
        //
        // A normalized 3rd order SimplePolynomial has at least one root.
        //

    //
    //  general SimplePolynomials
    //

    static int uniqueRootsO1(           // root of 1st order SimplePolynomial
        const double * pol,
        double * root
        );

    static int uniqueRootsO2(           // roots of 2nd order SimplePolynomial
        const double * pol,
        double * root
        );

    static int uniqueRootsO3(           // roots of 3rd order SimplePolynomial
        const double * pol,
        double * root
        );

    static int uniqueRootsO4(           // roots of 4th order SimplePolynomial
        const double * pol,
        double * root
        );

protected:
/*
    friend SimplePolynomial operator%(
	const SimplePolynomial & poly1,
	const SimplePolynomial & poly2
	);

*/

    int         order;
    double *    coefficientTable;

};

//
//  Due to a problem with the SGICC we need to explicitly force instantiation
//  of all templates we use in the library:
//

#ifdef SGI

static void instantiateTemplates()
{
    mergeAscending((double *)0,(int)0,(double *)0,(int)0,(double *)0);
    sort3ascending((double *)0);
}

#endif // SGI

} // extern "C++"

#endif // _MATH_SIMPLEPOLYNOMIAL_H_
// ===========================================================================
