// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       VectorN
//  TYPE:       c++ code
//  PROJECT:    VEGA Math Library
//  CONTENT:    vector and matrix of doubles
//  VERSION:    0.6.0
// ===========================================================================
//  AUTHORS:    rft     Robert F. Tobler
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 14:06:28  rft     last modification
//  28-jul-94 15:24:07  rft     created
// ===========================================================================

#include "VEGA/Math/math.h"

#define VECTORN1_NO_INLINE
#define VECTORN2_NO_INLINE

#include "VEGA/Math/vectorn.h"

#define INLINE_GLOBAL
#define INLINE_MEMBER

#include "VEGA/Math/vectorn1.ipp"
#include "VEGA/Math/vectorn2.ipp"

VectorN::~VectorN()
{
    kill();
}

SubVectorN::~SubVectorN()
{
    valuePtr = NIL(double *);
}

MatrixNM::~MatrixNM()
{
    kill();
}

SubMatrixNM::~SubMatrixNM()
{
    valuePtr = NIL(double *);
}

VectorN VectorN::bitReverseIndexedCopy() const
{
    int n = getLength();
    VectorN result(n);

    if (n == 0 || (n & (n-1)))
    {
	EH_ERROR(VectorN_errors, ERROR_VECTOR_NOT_POWER_OF_2,
		 "vector length not a power of 2: " << n);
	return result;
    }

    int highestBit = n >> 1;
    int secondHighestBit = highestBit >> 1;
    int thirdHighestBit = secondHighestBit >> 1;
    int index = 0;
    int reversed = 0;


    const double * vectorPtr = getLowerPtr();
    double * resultPtr = result.getLowerPtr();

    while (index < n)
    {
	resultPtr[reversed] = *vectorPtr++;
    
	index++;
	reversed ^= highestBit;
	if ((index & 1) == 0)
	{
	    reversed ^= secondHighestBit;
	    int reverseBit = thirdHighestBit;
	    for (int bit = 2; (index & bit) == 0; bit <<= 1)
	    {
		reversed ^= reverseBit; reverseBit >>= 1;
	    }
	}
    }
    return result;
}

VectorN & VectorN::bitReverseIndex()
{
    int n = getLength();
    
    if (n == 0 || (n & (n-1)))
    {
	EH_ERROR(VectorN_errors, ERROR_VECTOR_NOT_POWER_OF_2,
		 "vector length not a power of 2: " << n);
	return *this;
    }

    int highestBit = n >> 1;
    int secondHighestBit = highestBit >> 1;
    int thirdHighestBit = secondHighestBit >> 1;
    int index = 0;
    int reversed = 0;

    double * v = getLowerPtr();

    double help;

    while (index < n)
    {
	if (index < reversed)
	{
	    help = v[reversed]; v[reversed] = v[index]; v[index] = help;
	}
	index++;
	reversed ^= highestBit;
	if ((index & 1) == 0)
	{
	    reversed ^= secondHighestBit;
	    int reverseBit = thirdHighestBit;
	    for (int bit = 2; (index & bit) == 0; bit <<= 1)
	    {
		reversed ^= reverseBit; reverseBit >>= 1;
	    }
	}
    }
    return *this;
}

// sfmap: beginExcludeFrom: IVectorN

//
//  The following function assumes that all indeces are already bit-reversed.
//
//
//  For successive evaluation of the trigonometric functions the
//  following trigonometric recurrence is used:
//
//      a = 2 sin(square(d/2))
//      b  = sin d
//
//      cos(t + d) = cos t - [ a * cos t + b sin t ]
//      sin(t + d) = sin t - [ a * sin t - b cos t ]
//

//  The algorithm is based on the following recursion:
//
//	H[f] = Heven[f]
//	       + cos(2 PI f / n) Hodd[f]
//	       + sin(2 PI f / n) Hodd[n-f]		f in [0,n-1]
//
//	Heven[n/2 + g] = Heven[g]
//	Hodd [n/2 + g] = Hodd [g]			g in [0,n/2-1]
//

VectorN & VectorN::fastHartleyTransform()
{
    double * h = getLowerPtr();

    int n = getLength();

    int oldLen = 0;				// initialize to avoid warning
    int curLen;
    int newLen;

    for (curLen = 1; curLen < n; oldLen = curLen, curLen = newLen)
    {
	newLen = 2 * curLen;

	int i, j;
	double hi, hj;

    	for (i = 0; i < n; i += newLen)		// for all blocks
	{
	    j = i + curLen;			// special case:
	    hi = h[i]; hj = h[j];
	    h[i] = hi + hj;			// f = 0
	    h[j] = hi - hj;			// f = PI
	}
	if (curLen < 2) continue;
	
    	for (i = oldLen; i < n; i += newLen)	// for all blocks
	{
	    j = i + curLen;			// special case:
	    hi = h[i]; hj = h[j];
	    h[i] = hi + hj;			// f = PI/2
	    h[j] = hi - hj;			// f = 3 * PI/2
	}
	if (curLen < 4) continue;
 
	double d = MATH_2_MUL_PI / newLen;

	double a = 2.0 * square(sin(d * 0.5));	// initialize trigonometric
	double b = sin(d);			// recurrence

	double cosT = 1.0;
	double sinT = 0.0;

	int f, k, l;
	double one, two;			// temporary variables

	//
	//  To increse the locality of references, the following two loops
	//  could be exchanged, but then the trigonometric functions should
	//  be tabulated.
	//

	for (f = 1; f < oldLen; f++)		// for all frequencies in
	{					//   the first quadrant
	    one = a * cosT + b * sinT;		// trigonometric recurrence
	    two = a * sinT - b * cosT;	
	    cosT -= one;			//   cos (t + d)
	    sinT -= two;			//   sin (t + d)

	    for (i = f, j = curLen - f;		// for all blocks
		 i < n;
		 i += newLen, j += newLen)
	    {
		k = i + curLen; l = j + curLen;

		one = cosT * h[k] + sinT * h[l];
		two = cosT * h[l] - sinT * h[k];
		hi = h[i]; hj = h[j];
		h[i] = hi + one; h[k] = hi - one;	// all four quadrants
		h[j] = hj - two; h[l] = hj + two;	// (i,j,k,l)
	    }
	    
	}
    }
    return *this;
}

//  The improved algorithm for machines with lots of registers uses following
//  recurrence:
//
//	H[f] = H0[f]					
//	       + cos(    2 PI f / n) H1[f]
//	       + sin(    2 PI f / n) H1[n-f]
//	       + cos(2 * 2 PI f / n) H2[f]
//	       + sin(2 * 2 PI f / n) H2[n-f]
//	       + cos(3 * 2 PI f / n) H3[f]
//	       + sin(3 * 2 PI f / n) H3[n-f]		f in [0,n-1]
//
//	H0[n/4 + g] = H0[g]
//	H1[n/4 + g] = H1[g]
//	H2[n/4 + g] = H2[g]
//	H3[n/2 + g] = H3[g]				g in [0,n/2-1]
//

/*

VectorN & VectorN::fastHartleyTransform()
{
    double * h = getLowerPtr();

    int n = getLength();

    int curLen;
    int newLen;

    for (curLen = 1; curLen < n; curLen = newLen)
    {
	newLen = 4 * curLen;

	int i, j;
	double hi, hj;

    	for (i = 0; i < n; i += newLen)		// for all blocks
	{
	    j = i + curLen;
	    k = j + curLen;
	    l = k + curLen;
	    h1 = h[i]; h2 = h[j]; hipj = h1 + h2; himj = h1 - h2;
	    h1 = h[k]; h2 = h[l]; hkpl = h1 + h2; hkml = h1 - h2;
	    
	    h[i] = hipj + hkpl; h[j] = himj + hkml;
	    h[k] = hipj - hkpl; h[l] = himj - hkml;
	}
	if (curLen < 4) continue;
	
    	for (i = curLen/2; i < n; i += newLen)	// for all blocks
	{
	    j = i + curLen; k = j + curLen; l = k + curLen;
	    h1 = h[i]; h2 = h[j]; hipj = h1 + h2; himj = h1 - h2;
	    h1 = MATH_SQRT_2 * h[k]; h2 = MATH_SQRT_2 * h[l];
	    
	    h[i] = hipj + h1;
	    h[j] = himj + h2;
	    h[k] = hipj - h1;
	    h[l] = himj - h2;
	}
	if (curLen < 16) continue;
 
	double d = MATH_2_MUL_PI / newLen;

	double a = 2.0 * square(sin(d * 0.5));	// initialize trigonometric
	double b = sin(d);			// recurrence

	double c1 = 1.0;
	double s1 = 0.0;

	int f, k, l;
	double one, two;			// temporary variables

	//
	//  To increse the locality of references, the following two loops
	//  could be exchanged, but then the trigonometric functions should
	//  be tabulated.
	//

	for (f = 1; f < curLen/2; f++)		// for all frequencies in
	{					//   the first octant
	    one = a * c1 + b * s1;		// trigonometric recurrence
	    two = a * s1 - b * c1;	
	    c1 -= one;				//   cos (t + d)
	    s1 -= two;				//   sin (t + d)

	    c2 = c1 * c1 - s1 * s1;
	    s2 = 2 * c1 * s1;

	    c3 = c2 * c1 - s2 * s1;
	    s3 = c2 * s1 + s2 * c1;

	    for (i = f, j = curLen - f;		// for all blocks
		 i < n;
		 i += newLen, j += newLen)
	    {
		k = i + curLen; l = j + curLen;
		m = k + curLen; n = l + curLen;
		o = m + curLen; p = n + curLen;
		
		h0 = h[i]; h1 = h[j];

		hi = h[k]; hj = h[l];	h2 = c2 * hi + s2 * hj;
					h3 = s2 * hi - c2 * hj;

		h0p2 = h0 + h2;
		h0m2 = h0 - h2;
		h1p3 = h1 + h3;
		h1m3 = h1 - h3;

		hi = h[m]; hj = h[n];   h0 = c1 * hi + s1 * hj;
					h1 = s1 * hi - c1 * hj;
		hi = h[o]; hj = h[p];   h2 = c3 * hi + s3 * hj;
					h3 = s3 * hi + c3 * hj;
		h4p6 = h0 + h2;
		h4m6 = h0 - h2;
		h5p7 = h1 + h3;
		h5m7 = h1 - h3;

		h[i] = h0p2 + h4p6;
		h[j] = h1p3 + h4m6;
		h[k] = h0m2 
	    }
	    
	}
    }
    return *this;
}

*/

// sfmap: endExcludeFrom: IVectorN

// ===========================================================================
