// ===========================================================================
//  Copyright 1994 Technical University of Vienna. All Rights Reserved.
// ===========================================================================
//  NAME:       DynamicArray
//  TYPE:       c++ inline
//  PROJECT:    VEGA Babel Library
//  CONTENT:    implementation module for template class DynamicArray
//  VERSION:    1.0.0
// ===========================================================================
//  AUTHORS:    hl      Helwig L"offelmann
//              tg      Thomas Galla
//              rft     Robert F. Tobler
// ===========================================================================
//  HISTORY:
//
//  30-sep-94 13:46:16  rft     last modification
//  27-jul-94 17:48:34  tg      added method isNotValidIndex
//  27-jul-94 17:47:52  tg      implemented class DynamicArray
//  27-jul-94 15:28:25  hl      created
// ===========================================================================

// ======================================================== class DynamicArray

// -------------------------------------------------------------- constructors

extern "C++"
{

template<class T>
INLINE_TEMPLATE DynamicArray<T>::DynamicArray(
        )
:
    allocatedSize(4),
    size(0)
{
    elements = new T[allocatedSize];
    
    if (elements == NULL)
	EH_FATAL(DynamicArray_errors, "insufficient memory");
}

template<class T>
INLINE_TEMPLATE DynamicArray<T>::DynamicArray(
	int sizeNew
        )
:
    allocatedSize(sizeNew),
    size(sizeNew)
{
    elements = new T[allocatedSize];
    
    if (elements == NULL)
	EH_FATAL(DynamicArray_errors, "insufficient memory");
}

template<class T>
INLINE_TEMPLATE DynamicArray<T>::DynamicArray(
        CONST_TEMPLATE DynamicArray<T> & array
        )
:
    size(array.getSize()),
    allocatedSize(array.getSize())
{	
    elements = new T[allocatedSize];
    
    if (elements == NULL)
	EH_FATAL(DynamicArray_errors, "insufficient memory");

    copy ((CONST_TEMPLATE T *) array.elements, size, elements);
}
    


// ---------------------------------------------------------------- destructor 

template<class T>
INLINE_TEMPLATE DynamicArray<T>::~DynamicArray(
        )
{
    delete[] elements;
} // DynamicArray<T>::~DynamicArray()



// --------------------------------------------------------- member operators

template<class T>
INLINE_TEMPLATE T & DynamicArray<T>::operator[](
        int index
        )
{
    EH_FATAL_CONDITION(DynamicArray_errors, isNotValidIndex(index));
	
    return elements[index];
}



template<class T>
INLINE_TEMPLATE CONST_TEMPLATE T & DynamicArray<T>::operator[](
        int index
        ) const
{
    EH_FATAL_CONDITION(DynamicArray_errors, isNotValidIndex(index));
	
    return elements[index];
}



template<class T>
INLINE_TEMPLATE DynamicArray<T> & DynamicArray<T>::operator=(
        CONST_TEMPLATE DynamicArray<T> & array
        )
{
    if (&array == this)				// ###: rft: 08-sep-94
	return *this;				//	added this

    int arraySize = array.getSize();
    
    if (allocatedSize < arraySize)		// do we need more memory ?
    {
	allocatedSize = arraySize;
	delete[] elements;			// throw away old elements
	
	elements = new T[allocatedSize];	// allocate new memory
	
	if (elements == NULL)
	    EH_FATAL(DynamicArray_errors, "insufficient memory");
    } // if
    
    copy((CONST_TEMPLATE T *) array.elements, size, elements);
    
    if (arraySize < size)
    {
	for (int i = arraySize; i < size; i++)
	    elements[i] = T();			// clear Slots
    }

    size = arraySize;				// set new size
    
    return *this;
}



// --------------------------------------------------------- member functions

template<class T>
INLINE_TEMPLATE T DynamicArray<T>::getElement(
        int index
        ) const
{
    EH_FATAL_CONDITION(DynamicArray_errors, isNotValidIndex(index));
	
    return elements[index];
}

template<class T>
INLINE_TEMPLATE void DynamicArray<T>::setElement(
        int       index,
        CONST_TEMPLATE T & newElement
        )
{
    EH_FATAL_CONDITION(DynamicArray_errors, isNotValidIndex(index));
	
    elements[index] = newElement;
}

template<class T>
INLINE_TEMPLATE void DynamicArray<T>::setSize(
        int sizeNew
        )
{
    if (allocatedSize < sizeNew)		// do we need more memory ?
    {
	while (sizeNew > allocatedSize)
	    allocatedSize *= 2;

	T * pTemporary = new T[allocatedSize];	// allocate new memory
	
	if (pTemporary == NULL)
	    EH_FATAL(DynamicArray_errors, "insufficient memory");
	    
	copy((CONST_TEMPLATE T *) elements, size, pTemporary);
						// save the elements
	
	delete[] elements;			// free the memory
	
	elements = pTemporary;			// restore elements
    }
    
    if (sizeNew < size)
    {
	for (int i = sizeNew; i < size; i++)
	    elements[i] = T();			// clear Slots
    }

    size = sizeNew;				// set new size
}

template<class T>
INLINE_TEMPLATE void DynamicArray<T>::enlargeBy(
        int additionalSlots
        )
{
    setSize(size + additionalSlots);
}

template<class T>
INLINE_TEMPLATE void DynamicArray<T>::appendElement(
        CONST_TEMPLATE T & newElement
        )
{
    setSize(size + 1);
    setElement(size - 1,newElement);
}

template<class T>
INLINE_TEMPLATE int DynamicArray<T>::getSize(
        ) const
{
    return size;
}

template<class T>
INLINE_TEMPLATE int DynamicArray<T>::isNotValidIndex(
	int index
        ) const
{
    return (index < 0 || index >= size);
}
    
} // extern "C++"

// ===================================================== eof DynamicArray.I ==
